# -*- coding: utf-8; mode: perl; srcfile: /adv/perl/A2E/Blabla.pm -*-
# Time-stamp: <2019-06-09 14:35:30 phm>

=head1 NAME

A2E::Blabla

=cut

package A2E::Blabla;

=head1 DESCRIPTION

Reduce any phrase to a symbol, i.e. something conforming to the Perl regex \w*, or to a filename consisting of a symbol and a suffix joined to it by a dot (i.e. a non-symbol character).

You can configure how non-ascii charaters should be converted.

=cut

=head1 PREREQUISITES

=cut

use base 'A2E::Prog';

=head1 OPTIONS

=over

=cut

sub defvars {
    my $m = shift;
    $m->dpuc('A2E::Blabla::defvars', @_);
    $m->SUPER::defvars;
    $m->cache_define(textchars => '%');
    $m->konfayl('blabla.konf');
    $m->dpop('A2E::Blabla::defvars', 1);
    return 1;
}

=back

=head1 IMPLEMENTATION

=head2 Function Charconv

=cut

sub charconv {
    my $m = shift;
    $m->dpuc('charconv', @_);
    my $num = shift;
    my $chr = chr $num;
    my $str = $m->{konfig}->textchars->{$chr} || $chr;
    $m->dpop('charconv', $str);
    return $str;
};

=head2 Function Blabla2symbol

=cut

sub blabla2symbol {
    my $m = shift;
    $m->dpuc('blabla2symbol', @_);
    local $_ = shift;
    s(\A\s*)()g;
    s(\s*\Z)()g;
    s(\s)(_)g;
    my @nums = unpack 'U*', $_;
    $m->dgot(nums => \@nums);
    $_ = join '', map { $m->charconv($_) } @nums;
    y([A-Z])([a-z]);
    s(\W)(_)g;
    $m->dpop('blabla2symbol', $_);
    return $_;
}

=head2 Function Blabla2filename

=cut

sub blabla2filename {
    my $m = shift;
    $m->dpuc('blabla2filename', @_);
    local $_ = shift;
    my ($bas, $sfx) = m(\A(.*)\.(\w+)\Z);
  TRY: {
    SFX: {
	last unless $sfx;
	$_ = join '.', $m->blabla2symbol($bas), $sfx;
	last TRY;
      };
      $_ = $m->blabla2symbol($_);
    };
    $m->dpop('blabla2filename', $_);
    return $_;
}

=head1 TODO

test

=cut

return 1
