#!/usr/bin/perl

=head1 NAME 

       A2E::Dbmvals(3)

=cut

package A2E::Dbmvals;

=head1 SYNOPSIS

	my $p = new_ready A2E::Dbmvals progver => '0.0.1';
	$p->infaylz('j2f_tab.txt');
	my @vals = $p->dbmvals('简', '体');
	print $_, "\n" for @vals;
	簡
	體

=head1 DESCRIPTION

Query translations, as typically found in the DBM files in the A2E document file system.

Give a list of keys, receive a list of values. 

The DBM files contain more than the contents of the corresponding text file, they contain also information from parent text files.   This utility makes it convenient to find out what is in the files.

=head1 PREREQUISITES

    A2E::Prog(3) -- base class from where options are inherited

=cut

use strict;
use base 'A2E::MLDK';

=head1 OPTIONS

inherited from base class named in PREREQUISITES section.

=cut

sub defvars ($) {
    my ($m,$prog,$lihs) = shift->qlpuc('A2E::Dbmvals::defvars') or return;
    $m->SUPER::defvars;

=over
=item --lang | --ll | -l C<lang>

language used for guessing the name of the dbm file if not supplied via infaylz option

=cut
#    $m->konfig_define(['lang','ll','l'],'s');

=back

=head1 FILES

=head2 Configuration

	/etc/opt/a2e/dbmvals.konf
	~/.dbmvals_konf
	./_dbmvals

=cut

    $m->konfayl('dbmvals.konf');
    $m->qpop($prog,1);
}

=head2 Applications

    dbmvals(1) -- in RPM package a2e-bin

=head1 IMPLEMENTATION

=head2 Execution Functions

=head3 Function dbmvals

=cut

sub dbmvals ($) {
    my ($m,$prog,$lihs,@noms) = shift->lpuc('dbmvals',undef,@_);
    $m->mldk('lits');
    my %tups = ();
    foreach my $nom (@noms) { my $val = $lihs->getstr($nom);next unless $val;$tups{$nom} = $val };
    $m->dpop1($prog,\%tups);
};

=head1 THE END

=cut

return 1;

# Local Variables:
# coding: utf-8
# srcfile: /adv/perl/A2E/Dbmvals.pm
# mode: perl
# End:
