#!/usr/bin/perl

=head1 NAME

       Dokdata2db.pm - preprocess a document node, update database and cache

=head1 VERSION

This describes version B<0.1.1> of Dokdata2db.pm

=cut

our $VERSION = '0.1.1';

=head1 SYNOPSIS

    dokdata2db

=head1 DESCRIPTION

dokdata2db looks at the files in a documentation node, collects document
metadata which it can recognize there, and writes these into the
database.

dokdata2db is invokded under the target 'dokdata' in the document generation
makefiles.

=head1 OPTIONS

=item --dok ...

options of upper libraries

=back

=head1 BUGS

Tell adv at a2e de about them.

=head1 PREREQUISITES

    A2E::Dokfs(3)

=head1 SCRIPT CATEGORIES

Web

=head1 SEE ALSO

    dokdata2db(1)
    mlht(1)
    perl(1)

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

Copyright (c) 2006-8 Hartmut PILCH

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

=head2 Initialisation

=head3 Prerequisites

    A2E::Dokfs(3)
    SDBM_File(3)

et al

=cut

package A2E::Dokdata2db;
use base A2E::MLHT;
use Fcntl;
use SDBM_File;
use DBM_Filter;
use Cwd;
use File::Spec::Functions qw(splitpath);
use DateTime;
our $dabarels = {};

=head3 $arg = $m->valid_\w+($lihs,$arg,$vahr): validators
=cut

our %MSGS = ( varnnil => 'variable %n is empty' );

sub valids_dabarel8relspe {
    my ($m,$lihs,$dabarel,$vahr) = @_;
    my $dabarels = $vahr->{dabarels} || $m->dabarels($lihs);
    $vahr->{dabarels} = $dabarels;
    my $relspe = $dabarels->{$dabarel};
    $m->pstop('unknown relation %r', r => $dabarel) if !$relspe;
    $vahr->{dabarel} = $dabarel;
    ($dabarel,$relspe);
};
sub valid_dabarel {
    my ($m,$lihs,$dabarel,$vahr) = @_;
    ($dabarel,my $relspe) = $m->valids_dabarel8relspe($lihs,$dabarel,$vahr);
    $dabarel;
};
sub valid_rel2spe {
    my ($m,$lihs,$dabarel,$vahr) = @_;
    ($dabarel,my $relspe) = $m->valids_dabarel8relspe($lihs,$dabarel,$vahr);
    $relspe;
};

=head3 Function defvars

=cut

sub dokdata2db_defvars {
    my ($m,$prog,$lihs) = shift->qlpuc('dokdata2db_defvars',@_) or return;
=head4 listref idelits

Names of lits whose values must (presumably) be available so that the values of the identifying columns can be collected from $lihs before they are written to the rdbm tables.
This is not fully understood and has been suspected of being kludgy, obsolete or harmful.
It can be turned on and off by the --idelits_ok option of A2E::Dokfs.

=cut
    $m->{idelits} = [];
    $m->konfig_define(['reksekts','rs','r'], '!');
    $m->{var2validp}->{rel2spe} = \&valid_rel2spe;
    $m->{var2validp}->{dabarel} = \&valid_dabarel;
    $m->konfayl('dokdata2db.konf');
    $m->qpop1($prog,1);
};

sub defvars {
    my ($m,$prog,$lihs) = shift->qlpuc('Dokdata2db::defvars',@_) or return;
    $m->SUPER::defvars($lihs);
    $m->dokdata2db_defvars($lihs);
    $m->qpop1($prog,1);
};
sub dabarels {
    my ($m,$prog,$lihs) = shift->lpuc('dabarels',@_);
    %dabarels = $m->ref2hash($m->nom2xhac('__dabarels__',$lihs));
    $m->dpop1($prog,\%dabarels);
}


=head2 Subroutines

=head3 Function vars_get_doklvars

Read-select those text chunks that can be put into the global backend database as metadata.
This depends on the tuples supplied with the special lit2mlid directive and on the document-type-specific extended metadata supplied with the special tpmltoks directive.

=cut

sub lihs_get_doklvars {
    my ($m,$prog,$lihs) = shift->lpuc('lihs_get_doklvars',@_);
    my %doklvars = ();
    $m->lit2mlid_kre_revers_map;
    my $lit2mlid = $m->{cache}->{lit2mlid};
    my @noms = keys %$lit2mlid;
    if ($m->{cache}->{tmpltoks_ok}) {
	my @toks = $m->nom2xlst('_tmpltoks_',$lihs);
	push @noms, @toks if @toks;
    };
    if ($m->{cache}->{idelits_ok}) {
	push @noms, @{$m->{idelits}};
    };
    $doklvars{$_} = $lihs->getval($_) for @noms;
    $m->dpop($prog,%doklvars);
};

=head3 $ok = $m->dok_lang_getput_littups($dok,$tmpl,$lang,%opts{lihs,prelang})

Obtain variables for one language from $lihs and write them to the database.
LAB2LFOL: Find language variant of directory leaf
SYMLINK: Create symlinks to any language variants of the directory leaf

The \%littups variable that is used further downstream originates here.
It transports section/grup tuples that may be suitable for writing to db.
Via littups_put_dabarels? we call on putreldb for each relation to apply if it if applicable.

=cut
sub dok_lang_getput_littups {
    my ($m,$prog,$lihs,$dok,$tmpl,$lang,$prelang,%opts) = shift->opuc('dok_lang_getput_littups',[[qw(dok tmpl lang)],[qw(prelang)],{ prelang => 'lang_let' }],@_);my $vahr = $opts{$vahr};
    return $m->Wpop0($prog, w => funnovar => v => 'lang') if !$lang;
    my $i = 0;
    # $m->littups_kk_dbmopts(\%littups, bd => { l => $lang }) or $m->pstop('cannot retrieve text chunks for document %d language %l', $dok, $lang);
    # my %littups = %{$lihs->{vars}};
    # return $m->Wpop0($prog, w => 'no text chunks found for document %d language %l', d => $dok, l => $lang) if !%littups;
    my %littups = ();
    my $set_littups = sub {
	$littups{__prelang} = $prelang;
	$littups{prelang} = $prelang;
	$littups{__lang} = $lang;
	$littups{__dok} = $dok;
	$littups{$_} => $m->pvar($_) for qw(hodie dok);
	$littups{lang} = $lang;
    };
    &$set_littups;
    if ($prelang) {
	$i += $m->littups_put_dabarel($lihs,$vahr,\%littups,'lang');
    } else {
	$i += $m->littups_put_dabarels($lihs,$vahr,\%littups,$tmpl);
    };
    my $lfol = $littups{lfol};
  LAB2LFOL: {
      last unless $littups{lab2lfol};
      last if $lfol;
      last if $fol =~ m(\A\d+\Z);
      last unless $littups{lab} =~ m(\A[a-z,A-Z]{1,12}\Z);
      $lfol = lc $littups{lab};
    };
  SYMLINK: {
      last unless $lfol;
      my $cwd = $m->dir2dir;
      my $fol = $m->dir2fol($cwd);
      last if $lfol eq $fol;
      last if $m->{topdoks}->{$dok};
      do { $m->uwarn('unallowable symlink dir name %f', f => $lfol) } unless $lfol =~ m(\A[a-z,A-Z]{1,12}\Z);
      my $supdir = $m->supdir;
      my ($dir, $lwd) = $m->chwd($supdir);
    SUPDIR: {
	do { $m->vecho('in %d the needed link from %f to %l already exists', d => $dir, f => $fol, l => $lfol);last } if -l $lfol and $fol eq readlink $lfol;
	$m->ustop('can not link %f to %l!', f => $fol, l => $lfol) if -e $lfol;
	$m->vecho('linking from %f to %l in %d', f => $fol, l => $lfol, d => $dir);
	$m->svn_symlink_add($fol, $lfol);
      };
      $m->chwd($lwd);
      # 150527 substitute '$fol' for '$cwd': we can not cd into an $lfol-based directory if we are working offline
      # 150527 substitute '$fol' for '$lwd'
    };
    my %doklvars = $m->lihs_get_doklvars($lihs);
    $i += $m->littups_put_db(\%doklvars, $lang, $dok);
    $m->dpop1($prog,$i);
};

=head3 Function dir_fol_src

directory leaf name that is not a link for a particular language

=cut

sub dir_fol_src {
    my ($m,$prog,$lihs,$dir) = shift->opuc('dir_fol_src',[qw(dir)],@_);
    my ($vol,$sup,$fol) = splitpath $dir;
    $m->chwd($sup);
    while (1) {
	my $src = readlink $fol;
	last unless $src;
	last if $src eq $fol;
	last unless $src =~ m(\A\w+\Z);
	$fol = $src;
    };
    $m->chwd($dir);
    $m->dpop1($prog,$fol);
}


=head3 $ok = $m->dokprop_put($lihs,$dok,$url): Write out general document properties.
This is also based on some $lihs text chunks and therefore works only if a valid document vocabulary is present.

=cut

sub dokprop_put {
    my ($m,$prog,$lihs,$dok,$url,$dir,%opts) = shift->opuc('dokprop_put',[[qw(dok)],[qw(url_get dir_get)]],@_);
    my %vals = ();
    my $i = 0;
    $i += $m->put(mlhtdok => { dok => $dok }, doktmp => 'now()', dokurl => $url);
    $vals{doknom} = $m->dir_fol_src($dir, lihs => $lihs);
  DOKDATUM: {
      my $dokdatum = $lihs->getstr('datum') || '';
      my $matches = ();
      @matches = $dokdatum =~ m(\A\s*(\d{4}\-\d{2}\-\d{2})\s*\Z);
      do { $vals{dokdatum} = $matches[0];last } if @matches;
    DOKDATUM_VAR: {
	my $dokdatum_var = $lihs->getstr('dokdatum_var') || '';
	last unless $dokdatum_var;
	@matches = $dokdatum_var =~ m(\A\s*(\w+)\s*\Z);
	last unless @matches;
	$dokdatum_var = $matches[0];
	$dokdatum = $lihs->getstr($dokdatum_var) || '';
	@matches = $dokdatum =~ m(\A\s*(\d{4}\-\d{2}\-\d{2})\s*\Z);
	last unless @matches;
	$vals{dokdatum} = $matches[0];
	last DOKDATUM;
      };
      $dokdatum = $lihs->getstr('hodie') || '';
      @matches = $dokdatum =~ m(\A\s*(\d{4}\-\d{2}\-\d{2})\s*\Z);
      do { $vals{dokdatum} = $matches[0];last } if @matches;
      $dokdatum = $m->{konfig}->dokdatum || '';
      @matches = $dokdatum =~ m(\A\s*(\d{4}\-\d{2}\-\d{2})\s*\Z);
      do { $vals{dokdatum} = $matches[0];last } if @matches;
      $m->ustop('document %d has no date, neither dokdatum nor hodie was set', d => $dok);
    };
    $vals{lastmod} = eval { $m->svn_lastmod }; 
  MODDIST: {
      do { $m->pwarn('incomplete data entry in dokprop_put because unable to obtain last modification date from server: %m', m => $@);last } if !$vals{lastmod};
      my $moddist = $m->{konfig}->sitemap_moddist;
      do { $vals{moddist} = $moddist;last } if defined $moddist;
      my ($aaaa, $mm, $dd) = $vals{lastmod} =~ m(\A(\w{4})\-(\w{2})\-(\w{2})\Z);
      do { $m->uwarn('invalid modification date %m', m => $vals{lastmod});last } if !$dd;
      my $last_dt = new DateTime year => $aaaa, month => $mm, day => $dd;
      my $now_dt = now DateTime;
      my $dur = $now_dt - $last_dt;
      $vals{moddist} = $dur->days;
    };
    my $val = $lihs->getstr('smprior') // $m->{konfig}->sitemap_priority;
    $vals{smprior} = $val if defined $val;
    $val = $m->{konfig}->verdatum;
    $vals{verdatum} = $val if $val;
    $val = $lihs->getstr('privat') || $m->{konfig}->privat;
    $vals{privat} = $val if $val;
    $vals{doktyp} = $m->get_tmpl($lihs);
    $m->pstop('document type variable tmpl not defined, edit _pre.%d.txt and retry', d => $dok) if !$vals{doktyp};
    $val = $m->{konfig}->user;
    $vals{dokuid} = $val if $val;
    $vals{dokvers} = $lihs->getstr('dokvers') || '1';
    $i += $m->put(dokprop => { dok => $dok }, %vals);
    # $m->dpop($prog,\%vals);
    $m->dpop($prog,$i);
};

=head3 $ok = $m->littups_put_dabarel($lihs,$vahr,$littups,$dabarel): Write variables to one or more specified database relations

	args := littups, dabarel+
        littups := hashref((nom, val)+)
	dabarel := symbol

=cut
sub littups_put_dabarel {
    my ($m,$prog,$lihs,$vahr,$littups,$relspe) = shift->vlpuc('littups_put_dabarel',[qw(littups rel2spe)],@_);
    my $putrek = $m->rellits2putrek($lihs,$vahr,$relspe,littups => $littups);
    return $m->Wpop0($prog, d => varnnil => n => 'putrek') if !$putrek;
    my $i = $m->put(@$putrek);
    # push @{$m->{idelits}}, @$idelits;
    $m->dpop1($prog,$i);
}

=head3 $ok = $m->tups_put_rels($lihs,$littups,$tmpl,@rels): write template-specific metadata to db

Select template-specific metadata from a vocabulary hash $littups and rearrange it so that it can be written out to the global backend database using &rellits2putrek.
Find out what fields to retrieve, store the concerned values as equivalents of valid database records with non-null index fields or, if that is not possible, discard them.
Construct a reference to a list of records consisting of
 1: tab: string: table name
 2: ide: hashref: key-val pairs for identificatory columns
 3ff: atrs: list of key-val pairs for attributive columns 
each.

Then write these to the database.

=cut

sub tups_put_rels {
    my ($m,$prog,$lihs,$vahr,$littups,$tmpl,@relspes) = shift->vlpuc('tups_put_rels',[[qw(littups tmpl)],'rel2spe'],@_);
    return $m->Wpop0($prog, d => varnnil => n => 'rels') unless @rels;
    my $i = 0;
    foreach my $relspe (@relspes) { 
	my $putrek = $m->rellits2putrek($lihs,$vahr,$relspe,littups => $littups);
	next if !$putrek;
	$i += $m->put(@$putrek);
	if ($m->{cache}->{idelits_ok}){
	    foreach my $idelit (@$idelits) { 
		next if grep { $_ eq $idelit } @{$m->{idelits}};
		push @{$m->{idelits}}, $idelit } } };
    $m->dpop($prog,$i);
};

=head3 $ok = $m->littups_put_dabarels($lihs,$littups,$tmpl): Write data for all doktyp/tmpl-defined relations to db
=cut
sub littups_put_dabarels {
    my ($m,$prog,$lihs,$vahr,$littups,$tmpl) = shift->vlpuc('littups_put_dabarels',[qw(littups tmpl)],@_);
    my @rels = $m->ref2list($m->nom2xlst('_tmplrels_'));
    my $ok = $m->tups_put_rels($lihs,$vahr,$littups,$tmpl,@rels) ? 1 : 0;
    $m->dpop1($prog,$ok);
};

=head3 $ok = $m->dok_langs_put_littups($lihs,$dok,@langs): Write variables for the langugage specified on the commandline or for all languages of the document

=cut

sub dok_langs_put_littups {
    my ($m,$prog,$lihs,$vahr,$dok,$tmpl,@langs) = shift->vlpuc('dok_langs_put_littups',[[qw(dok tmpl)],'lang'],@_);
    my $prelang = '';
    my $i = 0;
    foreach my $lang (@langs) {
	$i += $m->dok_lang_getput_littups($dok,$tmpl,$lang, prelang => $prelang, lihs => $lihs, vahr => $vahr);
	$prelang = $lang;
    };
    $m->dpop1($prog,$i);
};
=head3 $ok = $m->relputdb($lihs,$rel,$fol,$tups): Write record pertaining to one relation to db

=cut
sub relputdb {
    my ($m,$prog,$lihs,$vahr,$relspe,$sem,$tups) = shift->vlpuc('relputdb',[[qw(rel2spe sem littups)],{ sem => 'nom' }],@_);
    my $putrek = $m->rellits2putrek($lihs,$vahr,$relspe, littups => $tups, grup => $sem);
    return $m->Wpop0($prog, d => varnnil => n => 'putrek') if !$putrek;
    my $i = $m->put(@$putrek);
    $m->dpop($prog,$i);
};
=head3 $ok = $m->relputdb($lihs,$fol,$tups,@rels): Write relation records to db
=cut
sub rekputdb {
    my ($m,$prog,$lihs,$vahr,$grup,$tups,@rels) = shift->vlpuc('rekputdb',[[qw(fol littups)],'dabarel',{ fol => 'nom' }],@_);
    my $i = 0;
    foreach my $rel (@rels) { $i += $m->relputdb($lihs,$vahr,$rel,$grup,$tups) }
    $m->dpop1($prog,$i);
};
=head3 $ok = $m->grupputdb($lihs,$grup,@rels): Write doc section records to db

$grup :: section/group name from which data are to be sought
@rels :: data relations associated with sections like $grup

=cut
sub grupputdb {
    my ($m,$prog,$lihs,$vahr,$grup,@rels) = shift->vlpuc('grupputdb',[[qw(grup)],'dabarel',{ grup => 'nom' }],@_);
    # file:/phm/19/09/10/sig/_log07adv.txt::303204
    my %tups = ();
    my $get_atrs = sub {
	# my @atrs = grep { m(\A$grup\_[a-z]+\d*\Z) } keys %{$lihs->{vars}};
	local @_ = $m->text2list($lihs->getstr($m->grupvar($grup,'vars'))); # qw(_nom _lvl _sup mail tit _frm _fol)
	grep { m(\A[a-z]) } @_; # qw(mail tit)
    };
    my @atrs = &$get_atrs;
    my $i = 0;
    my $set_atrtups = sub {
	foreach my $atr (@atrs) {
	    # my ($sfx) = $atr =~  m(\A$grup\_(.*)\Z);
	    # $m->pstop('no sfx found for atr %a under fol %f', a => $atr, f => $grup) unless $sfx;
	    # $tups{$sfx} = $m->{tmplvars}->{$atr};
	    my $nom = join '_', $grup, $atr;
	    $tups{$atr} = $lihs->getstr($nom);
	};
	%tups;
    };
    &$set_atrtups; # => (tit => 'Kern Wien', mail => 'KERN.Wien@e-kern.com') file:/phm/19/09/10/sig/_log07adv.txt::316751
    my $set_lintups = sub {
	my $numkontp = $grup !~ m([0-9]\Z); # whether section name ends in non-numeral, so that numbers can be attached
	foreach my $n (1..3) { # read lines as potential positional fields; usable as 'L1', 'L2', 'L3' in dabarel spec, further use envisageable
	    my $kont = $numkontp ? $n : chr 96 + $n;  # a..c if ends in number, 1..3 otherwise
	    my $nom = $grup . $kont;
	    my $val = $lihs->getstr($nom);
	    last if !$val;
	    my $lin = 'L' . $n;
	    $tups{$lin} = $val;
	};
	%tups;
    };
    &$set_lintups;
    $i += $m->rekputdb($lihs,$vahr,$grup,\%tups,@rels);
    $m->dpop1($prog,$i);
};
=head3 $ok = $m->sektpubdb($lihs,$sekt) 
=cut
sub sektputdb {
    my ($m,$prog,$lihs,$vahr,$sekt) = shift->vlpuc('sektputdb',[[qw(sekt)],{ sekt => 'nom' }],@_);
    my $relsnom = $lihs->getstr('_skt_'.$sekt.'_');
    $m->pstop('no sektrek string found for sekt %t', t => $sekt) if !$relsnom;
    my @rels = $m->ref2list($m->str2arb($relsnom));
    $m->pstop('no rels found for rek %r', r => $rek) if !@rels;
    my @grups = $m->text2list($lihs->getstr($sekt.'__fols'));
    do { $m->pwarn('no fols for sekt %t', t => $sekt);$m->dpop($prog) } if !@grups;
    @grups = map { if (m(\A\~(\w+)\Z)) { $sekt.$1 } elsif (m(\A\_(\w+)\Z)) { $sekt.$_ } else { $_ }  } @grups;
    my $i = 0;
    foreach my $grup (@grups) { $i += $m->grupputdb($lihs,$vahr,$grup,@rels) }
    $m->dpop1($prog,$i);
};
=head3 $ok = $m->tupsputdb($lihs): put records for all sections to db
=cut
sub tupsputdb {
    my ($m,$prog,$lihs,$vahr) = shift->vlpuc('tupsputdb',[],@_);
    my $reksekts = $lihs->getstr('_reksekts_'); 
    return $m->Wpop0($prog, d => varnnil => n => 'reksekts') if !$reksekts;
    my @sekts = $m->ref2list($m->str2arb($reksekts));
    my $i = 0;
    foreach $sekt (@sekts) { 
	$i += $m->sektputdb($lihs,$vahr,$sekt) };
    $m->dpop($prog,$i);
}
=head3 $ok = $m->dokputdb($lihs): put records for current document in all langs to db
=cut
sub dokputdb {
    my ($m,$prog,$lihs,$vahr,$dok,$lang) = shift->vlpuc('dokputdb',[qw(dok lang)],@_);
  DABARELS: {
      last if !$lihs->getstr('__dabarels__'); 
      # $m->dok2tmplvars(l => $lang0);
      $m->set_litsdir($lihs, l => $lang);
      $m->dabarels($lihs);
    };
    my $i = $m->tupsputdb($lihs,$vahr);
    $m->dpop1($prog,$i);
}

=head2 Interface

=head3 $ok = $m->dokdata2db($lihs,%opts)

=cut

sub dokdata2db {
    my ($m,$prog,$lihs,%opts) = shift->lpuc('dokdata2db',@_);
    $m->lihs_get_vars($lihs);
    $m->lihs_var2arb($lihs); # __dabarels__
    my $rs = $opts{rs} || $m->{konfig}->reksekts;
    $m->{progvars}->{cwd} = $m->get_dir($lihs);
    my $dok = $m->get_dok($lihs);
    my $tmpl = $m->get_tmpl($lihs);
    my @langs = $m->set_dok_langs($dok,lihs => $lihs) or $m->pstop('unable to set_dok_langs');
    my %langp = ();$langp{$_} = 1 for @langs;
    my $lang0 = $langs[0];
    # $m->dok2tmplvars(l => $lang0) or $m->ustop('cannot retrieve text chunks for document %d language %l', d => $dok, l => $lang0);
    $m->set_litsdir($lihs, l => $lang0);
    my $i = 0;
    my $vahr = { __flagp__ => { }, langp => \%langp, dok => $dok, tmpl => $tmpl };
    $i += $m->dokprop_put($dok,lihs => $lihs, vahr => $vahr); # write meta data to db
    $i += $m->dok_langs_put_littups($lihs,$vahr,$dok,$tmpl,@langs);
    $i += $m->dokputdb($lihs,$vahr,$dok,$lang0) if $rs; # write data from sections of document to db
    $i += $m->dok_kre_kanal($m->{konfig}->news_kanal);
    $i += $m->dok_put_kanals($m->{konfig}->putkanals);
    if ($tmpl and $dok) {
	$i += $m->put(doktexts => { dok => $dok, varnom => 'tmpl' }, varval => $tmpl, valsymp => 'true');
    } else {
	$m->pstop('need configuration variable tmpl in dokdata2db with document %d', d => $dok);
    };
    $m->vecho('written %i dokdata records to db', i => $i);
    $m->dpop1($prog,$i);
};

=head1 THE END

=cut

1;

# Local Variables: ;
# coding: utf-8 ;
# url: http://a2e.de/adv/perl/A2E/Dokdata2db.pm ;
# mode: perl ;
# End: ;
