#!/usr/bin/perl

=head1 SYNOPSIS

Convert directory from old to new sytax based on configuration using a combination of predefined functions.

Accept a couple of subcommands to achieve this flexibly.

Allow simplest possible specificaton of conversion rules for different ranges of documents specified by regex matched to document name.

When configured fully, conversion of complete directory trees should become possible, e.g.

    subdirs exec:dokkonv:dir /sig/oas/13/01

For Debugging, set

    export DEBUG_BLOKS konv_dir

or similar an invoke with

    perltest dokkonv dir

or similar.

In this case nothing is actually written to file but instead everything is typed out to screen.


=head1 TODO

Move to tree-based parsing and deparsing, possibly using something like Parse::Yapp or Tree::Simple.


=head1 Reference Materials

started 150313

Similar converter projects have been in use before:

    top2dok(1)
    A2E::Privat::Top2dok(3)
    A2E::Eupat

=cut

package A2E::Dokkonv;
use base A2E::Dokfs;
use File::Basename;
use File::Spec::Functions qw(splitdir splitpath catdir catfile);
use POSIX qw(strftime);
use feature 'switch';

sub defvars { 
    my $m = shift;my $prog = 'Dokkonv::defvars';$m->dpuc($prog, @_);
    do { $m->dpop;return } if $m->{ready}->{dokkonv};
    $m->SUPER::defvars;
    $m->svn_defvars;
    $m->dokkonv_defvars;
    $m->dpop($prog, 1);return 1;
};

sub getkonv {
    my $m = shift;my $prog = 'getkonv';$m->dpuc($prog,@_);
    my $dok = shift;
    my @konvs = keys %{$m->{konv2konfig}};
  KONV: {
      last unless @konvs;
      $konv = shift @konvs;
      $m->dpuc('konv', $konv);
    DOKRXPS: {
	my $dokrxps = $m->{konv2konfig}->{$konv}->{dokrxps};
	last unless $dokrxps;
	foreach my $dokrxp (@$dokrxps) { $m->dgot(dokrxp => $dokrxp);do { $m->dpop('konv', $konv);last KONV } if $dok =~ m(\A$dokrxp\Z) };
      };
    DOKGETS: {
	my $dokgets = $m->{konv2konfig}->{$konv}->{dokgets};
	last unless $dokgets;
	foreach my $dokget (@$dokgets) { do { $m->dpop('konv', $konv);last KONV } if &$dokget($dok) };
      };
      $konv = '';
      $m->dpop;
      redo;
    };
    $m->dpop($prog, $konv);return $konv;
};

sub dokkonv_defvars {
    my $m = shift;my $prog = 'dokkonv_defvars';$m->dpuc($prog, @_);
    $m->{konfig}->define('dry_run|n!');
    $m->{konv2konfig} = {
	_ => {
	    srclng => 'de',
	},
	dag => {
	    dokrxps => [ 'phm_pub_sig\w{6}' ],
	    # dokgets => [ $closure_function, sub { ... } ],
	    srcdok => 'phm_pub_sig150324',
	    srcdir => '/phm/15/03/24/sig',
	    fyls =>  {
		nav => { frm => '_nav.%d.txt', srcp => 1 },
		pdr => { frm => 'Rakefile', srcp => 1 },
		tmpl => { frm => '_dok.%d.txt', srcp => 1 },
		dlng => { frm => '_lng.%d.txt', funs => [ 'dokl', 'topsekts' ] },
		llng => { frm => '_lng.%d.%l.txt', funs => [ 'dokl', 'atrs_pre', 'atrs_post' ] },
		pre => { frm => '_pre.%d.txt', funs => [ 'dokl', 'datum' ] },
	    },
	},
	Dag => {
	    dokrxps => [ 'phm_pub\w{6}' ],
	    srcdok => 'phm_pub150324',
	    srcdir => '/phm/15/03/24',
	    fyls =>  {
		nav => { frm => '_nav.%d.txt', srcp => 1 },
		tmpl => { frm => '_dok.%d.txt', srcp => 1 }, 
		dlng => { frm => '_lng.%d.txt', funs => [ 'dokl', 'topsekts' ] },
		llng => { frm => '_lng.%d.%l.txt', funs => [ 'dokl', 'atrs_pre', 'atrs_post' ] }, 
		pre => { frm => '_pre.%d.txt', funs => [ 'dokl', 'dokprogver', 'datum' ] },
	    },
	}
    };
    $m->konfig_define('srcdok', 's%');
    $m->konfig_define('srcdir', 's%');
    $m->konfayl('dokkonv.konf');
    $m->dpop($prog,1);
};

sub postkonfig {
    my $m = shift;
    $m->dpuc('Dokkonv::postkonfig', @_);
    do { $m->dpop;return } if $m->{ready}->{dokfs};
    $m->SUPER::postkonfig;
    $m->dokkonv_postkonfig;
    $m->dpop('Dokkonv::postkonfig', 1);
    return 1;
};

sub dokkonv_postkonfig {
    my $m = shift;my $prog = 'dokkonv_postkonfig';$m->dpuc($prog, @_);
    do { $m->dpop;return } if $m->{ready}->{dokkonv};
    $m->pstop('reserved object variable dag already existed at start') if $m->{dag};
    $m->{ready}->{dokkonv} = 1;
    $m->dpop($prog,1);return 1;
}

=head3 Function konv_dir


convert files of current directory according to configuration.

Invoke from commandline with

    dokkonv dir

Configuration is laid down in object variable konv2konfig.

The configuration to be used is identified by means of a regex specified in object variable konvdokrex

TODO: move part of the configuration to AppConfig variables.


=head4 Subroutine pcpdpe

Set date of today, yesterday and tomorrow as well as weekday for use in _pre.${dok}.txt of Dag-type documents.

=cut

sub getdag {
    my $m = shift;my $prog = 'getdag';$m->dpuc($prog, @_);
    my $dok = shift;
    local $_;
    my $dag = {};
    $dag->{pd} = {};
    ($dag->{pd}->{aa}, $dag->{pd}->{mm}, $dag->{pd}->{dd}) = $dok =~ m([a-z](\d{2})(\d{2})(\d{2}\Z));
    $dag->{pd}->{aaaa} = '20'.$dag->{pd}->{aa};
    $m->ustop('la dokumento %d ne reprezentas dagon', { d => $dok }) unless $dag->{pd}->{aa} and $dag->{pd}->{mm} and $dag->{pd}->{dd};
    my @intdags = ('pc', 'pb');
    my @ontdags = ('pe', 'pf');
    my %mondags = ( 1 => 31, 2 => 28, 3 => 31, 4 => 30, 5 => 31, 6 => 30, 7 => 31, 8 => 31, 9 => 30, 10 => 31, 11 => 30, 12 => 31 );
    my $mondag = sub { my $pd = shift;my $mm = $dag->{$pd}->{mm} % 12;$aa = $dag->{$pd}->{aa} % 100;return $mondags{$mm} unless $mm == 2 and not $aa % 4;return 29 }; 
    my $last = '';
    my $next = 'pd';
  INTDAGS: {
      last unless @intdags;
      $last = $next;
      $next = shift @intdags;
      $dag->{$next} ||= { %{$dag->{$last}} };
      if ($dag->{$last}->{dd} > 1) { 
	  $dag->{$next}->{dd} = sprintf '%02d', $dag->{$last}->{dd} - 1;
      } else {
	  if ($dag->{$last}->{mm} > 1) {
	      $dag->{$next}->{mm} = sprintf '%02d', $dag->{$next}->{mm} - 1;
	  } else {
	      $dag->{$next}->{mm} = '12';
	      $dag->{$next}->{aa} = sprintf '%02d', $dag->{$last}->{aa} - 1;
	  };
	  $dag->{$next}->{dd} = &$mondag($next);
	  $m->dgot($next => { aaaa => $dag->{$next}->{aaaa}, mm => $dag->{$next}->{mm}, dd => $dag->{$next}->{dd} }, $last => { aaaa => $dag->{$last}->{aaaa}, mm => $dag->{$last}->{mm}, dd => $dag->{$last}->{dd} });
      };
      $dag->{$next}->{aaaa} = '20'.$dag->{$next}->{aa};
      redo;
    };
    $next = 'pd';
  ONTDAGS: {
      last unless @ontdags;
      $last = $next;
      $next = shift @ontdags;
      $dag->{$next} ||= { %{$dag->{$last}} };
      if ($dag->{$last}->{dd} >= &$mondag($last)) {
	  $dag->{$next}->{dd} = 1;
	  if ($dag->{$last}->{mm} >= 12) {
	      $dag->{$next}->{mm} = '01';
	      $dag->{$next}->{aa} = sprintf '%02d', $dag->{$next}->{aa} + 1;
	  } else {
	      $dag->{$next}->{mm} = sprintf '%02d', $dag->{$last}->{mm} + 1;
	  };
      } else {
	  $dag->{$next}->{dd} = sprintf '%02d', $dag->{$last}->{dd} + 1;
      }
      $dag->{$next}->{aaaa} = '20'.$dag->{$next}->{aa};
      $m->dgot($next => { aaaa => $dag->{$next}->{aaaa}, mm => $dag->{$next}->{mm}, dd => $dag->{$next}->{dd} }, $last => { aaaa => $dag->{$last}->{aaaa}, mm => $dag->{$last}->{mm}, dd => $dag->{$last}->{dd} });
      redo;
    };
    foreach my $pd (qw(pb pc pd pe pf)) { 
	my $saa = $dag->{$pd}->{aaaa} - 1900;
	my $smm = $dag->{$pd}->{mm} - 1;
	my $sdd = $dag->{$pd}->{dd};
	my @time = (0, 0, 0, $sdd, $smm, $saa);
	$dag->{$pd}->{wd} = strftime '%w', @time;
	$dag->{$pd}->{ww} = strftime '%W', @time;
	$dag->{$pd}->{AMD} = join '-', $dag->{$pd}->{aaaa}, $dag->{$pd}->{mm}, $dag->{$pd}->{dd};
	$dag->{$pd}->{amd} = join '', $dag->{$pd}->{aa}, $dag->{$pd}->{mm}, $dag->{$pd}->{dd};
    };
    $m->dgot(wd => $dag->{wd});
    foreach my $p (qw(pb pc pd pe pf)) { $dag->{$p}->{amd} = join '', map { $dag->{$p}->{$_} } qw(aa mm dd) };
    $m->dpop($prog, $dag);
    # $m->vecho('dag ends up being %d', { d => $dag });
    return $dag;
};

sub konv_dir {
    my $m = shift;my $prog = 'konv_dir';$m->dpuc($prog, @_);
    my $konv = shift;
    $konv = '' if $konv == '.'; 
    my $dok = $m->get_dok or $m->ustop('cant get dok');
    $konv ||= $m->getkonv($dok) || $m->ustop('found no conversion rules for document %d', { d => $dok });
    my $konfig = $m->{konv2konfig}->{$konv} || $m->ustop('no konfig specified for konv %s', $konv);
    $m->dgot(konfig => $konfig);
    my $srcdok = $konfig->{srcdok} || $m->pstop('no srcdok for konv %k', { k => $konv });
    my $srcdir = $konfig->{srcdir};
    my $srclng = $konfig->{srclng} || $m->{konv2konfig}->{_}->{srclng} || $m->pstop('missing srclng');
    my $fyls = $konfig->{fyls} || $m->pstop('no fyls for konv %k', { k => $konv });
    my @fylargs = @_;
    @fylargs = (keys %$fyls) unless @fylargs;
    local $_;
    my $objs = {};
    my $srcs = {};
    my $sym = '';
    my $lang = $srclng;
    my $src = '';
    my $obj = '';
    foreach $fyl (@fylargs) {
	my $frm = $fyls->{$fyl}->{frm} or $m->pstop('no valid info for fyl %f', { f => $fyl });
	$obj = $m->hsprintf($frm, d => $dok, l => $lang) or $m->pstop('no obj for frm %f and dok %d', { f => $frm, d => $dok });
	$m->vecho('obj of fyl %f is %o', { o => $obj, f => $fyl });
	$m->ustop('mi ne trovis konverteblan objekton %o tcitie', { o => $obj }) unless -r $obj and -w $obj;
	$objs->{$fyl} = $obj;
	next unless $fyls->{$fyl}->{srcp};
	my $srcfrm = $fyls->{$fyl}->{srcfrm} || $frm;
	$src = catfile $srcdir, $m->hsprintf($srcfrm, d => $srcdok, l => $lang) or $m->pstop('no src for frm %f and srcdok %d', { f => $frm, d => $srcdok });
	$m->ustop('mi ne trovis legeblan fonton %s tcitie', $obj) unless -r $src;
	$srcs{$fyl} = $src;
    };
    # my $kmd_eksek = sub { if ($m->{debug}) { $m->vecho(join(' ', @_)) } else { system @_ } };
    my $mtime = sub { my $fyl = shift;my @stat = stat $fyl;$stat[9] };
    my $srcnovp = sub { return 1 if &$mtime($src) > &$mtime($obj) };
    my $rtrim_blanks = sub { if (m(\A(.*)\s+\Z)) { $_ = $1 } };
    my $voidlini_ok = 1;
    my $voidlinip = sub { return $voidlini_ok if m(\A\Z) };
    my $komentp = sub { return 1 if m(\A\#) };
    my $ignorkomentp = 0;
    my $delp = 0;
    my $fun = 'dokl';
    my %nbr = ();
    my @funs = qw(dokl atrs_pre atrs_post);
    my $funshift = sub { $fun = shift @funs };
    my $funsubs = { 
	dokl => sub { # -*- dokl: PhmPubSig150324LngDe; mode: fundamental; coding: utf-8; srcfile: /phm/15/03/24/sig/_lng.phm_pub_sig150324.de.txt -*-
	    if (m(\A(.*)\s*dokl:\s*\w+;?(.*)\Z)) { $_ = "$1$2";$nbr{$fun}++;$m->dgot(_ => $_, lok => 'konv1') };
	    if (m(\A(.*)\s*srcfile\:\s*\S+;?(.*)\Z)) { $_ = "$1$2";$nbr{$fun}++;$m->dgot(_ => $_, lok => 'konv2') };
	    if (m(\A(.*\S)\s\s+(\-\*\-)\Z)) { $_ = "$1$2";$nbr{$fun}++;$m->dgot(_ => $_, lok => 'konv3') };
	    $fun = shift @funs;
	    $ignorkomentp = 1;
	},
	atrs_pre =>  sub { 
	    if (m(\A(\w+)\s*\:\=\s*(.*)\Z)) { $_ = "$1 ?= $2";$nbr{$fun}++;$m->dgot(_ => $_, lok => 'konv');$fun = shift @funs };
	},
	atrs_post =>  sub {
	    if (m(\A(\w+)\s*\:\=\s*(.*)\Z)) { $_ = "$1 ?= $2";$nbr{$fun}++;$m->dgot(_ => $_, lok => 'konv') } else { $fun = shift @funs };
	},
	kolonsetvar =>  sub {
	    # $ignorkompentp = 1;
	    if (m(\A(\w+)\:\s*(.*)\Z)) { $_ = "$1 ?= $2";$nbr{$fun}++;$m->dgot(_ => $_, lok => 'kolonsetvar') } else { $fun = shift @funs };
	},
	topsekts => sub {
	    if ('(top' == substr($_,0,4)) { $_ = '(top @topsekts';$nbr{$fun}++;$fun = shift @funs };
	},

	dokprogver => sub {
	    $m->dgot(fun => $fun, lok => 'ini', _ => $_); 
	    if (m(\A(PC|PD|PE|pc|pd|pe|wd)\b)) { $fun = shift @funs } elsif (m(\Adokprogver\b)) { 
		$m->{dag} ||= $m->getdag($dok) unless $m->{dag};
		my @lins = map { my $pd = $_;my $PD = uc $pd;my $amd = $m->{dag}->{$pd}->{amd};("$PD ?= phm_pub$amd", "$pd ?= phm_pub_sig$amd") } qw(pc pd pe);
		push @lins, "wd ?= $m->{dag}->{pd}->{wd}";
		push @lins, $_;
		$_ = join "\n", @lins;
		$nbr{$fun}++;
		$fun = shift @funs;
		$voidlini_ok = 0; # suppress any empty lines between here and next i.e. 'datum'
	    };
	    $m->dgot(fun => $fun, lok => 'fin', _ => $_, funs => \@funs);
	},
	datum => sub {
	    $m->dgot(fun => $fun, _ => $_); 
	    $m->{dag} = $m->getdag($dok) unless $m->{dag}->{pd}->{AMD};
	    my $AMD = $m->{dag}->{pd}->{AMD};
	    $m->pstop('unable to obtain AMD from dag %d', { d => $m->{dag} }) unless $AMD;
	    if (m(\Asmprior\s*\?\=\s*(0\.\d+)\Z)) { 
		my $nf = $1;
		$nf = $nf + 0.2 if $nf < 0.4;
		$_ = sprintf "smprior ?= %.1f", $nf;
		$nbr{$fun}++;
	    };
	    DATUM: {
		last unless m(\A(datum|hodie)\b);
		$_ = "$1 ?= $AMD";
		$nbr{$fun}++;
		last unless $1 eq 'datum';
		$fun = shift @funs;
		$voidlini_ok = 1; # no longer suppress empty lines hereafter
	    };
	    $m->dgot(fun => $fun, lok => 'fin', _ => $_); 
	},
    };
    my $konv_fyl = sub {
	my $fyl = shift;
	@funs = $m->ref2list($fyls->{$fyl}->{funs});
	$fun = shift @funs;
	%nbr = ();
	$ignorkomentp = 0;
	foreach my $f (@funs) { $nbr{$f} = 0 };
	$obj = $objs->{$fyl} || $m->pstop('nothing found in objs %o for fyl %f', { o => $objs, f => $fyl });
	if ($srcs{$fyl}) {
	    $src = $srcs{$fyl};
	    $m->uwarn('object %o is newer than source %s', { o => $obj, s => $src }) unless &$srcnovp;
	} else {
	    $src = join '.', $obj, 'bak';
	    system 'cp', '-vf', $obj, $src;
	};
	$hw = $fyl.'_w';
	$hr = $fyl.'_r';
	$m->vecho('writing from %s to %o', { s => $src, o => $obj });
	$m->open_fayl($src, $hr, 'r');
	$m->open_fayl($obj, $hw, 'w') unless $m->{debug} or $m->{konfig}->dry_run;
	my $prn = ($m->{debug} || $m->{konfig}->dry_run) ? sub { print "PRN $fun: " . shift } : sub { $m->print_tu_fayl($hw, shift) };
      LINI: {
	  $delp = 0; # do not delete line unless so required for current line by converter
	  $_ = $m->getlini($hr);
	  last unless $_;
	  chomp;
	  &$rtrim_blanks;
	KONV: {
	    last unless $fun;
	    last if &$voidlinip;
	    do { $nbr{voidlini}++;redo LINI } unless $_; # suppress any empty line that voidlinip returned as not ok i.e. !$voidlini_ok
	    last if $ignorkomentp and &$komentp;
	    my $funverb = $funsubs->{$fun};
	    $m->pstop('invalida klozurfunktsio %f', { f => $fun }) unless $funverb;
	    $m->decho('using function %f on line %l', { f => $fun, l => $_ });
	    &$funverb;
	  };
	  redo if $delp;
	  &$prn($_."\n");
	  redo;
	};
	$m->kloz_fayl($hr);
	$m->kloz_fayl($hw);
	foreach my $f (keys %nbr) { $m->vecho('%v successful replacements with %k in file %f', { v => $nbr{$f}, k => $f, f => $fyl }) };
    };
    foreach $fyl (@fylargs) { &$konv_fyl($fyl) };
    $m->dpop($prog, 1);return 1;
};

sub help {
    my $m = shift;my $prog = 'help';$m->dpuc($prog, @_);
    my $prog = basename $0;
    my $dok = $m->get_dok;
    my $konv = $m->getkonv($dok);
    my @fyls = keys %{$m->{konv2konfig}->{$konv}->{fyls}};
    my $s_fyls = join ' ', @fyls;
    for ($prog) {
	$prog = $1 when m(\A(.*)\.pl\Z);
    }
    print STDERR "
$prog -- upgrade mlht files and directories from older to newer versions using various subcommands, e.g.

	$prog help
	$prog -nv dir
	$prog -nv dir $konv
	$prog -nv dir $konv $s_fyls
	$prog -n dir . $s_fyls
";
    local $_;print "\t$prog -v dir . $_\n" for @fyls;
    print "\t$prog dir\n";
    $m->dpop($prog, 1);return 1;
}

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# End:
