# -*- coding: utf-8; mode: perl; srcfile: /adv/perl/A2E/Eupat/_dok.perl_a2e_eupat_pub.txt -*-

package A2E::Eupat;

=head1 Name

A2E::Eupat -- Migration Tools for the Eupat Website

=head2 Description

Assuming that we are in a directory of http://eupat.ffii.org, convert that directory to the new multiligual hypertext format currently used on that site.

=head2 Prerequisistes

=cut

use base 'A2E::MLHT', 'A2E::Subdoks';
use strict;
use File::Spec::Functions qw(splitdir catdir catfile);
use HTML::Entities;
use feature 'switch';
# use Data::Dumper;

=head1 Version

This describes version B<0.0.0> of A2E::Eupat

=cut

our $VERSION = '0.0.0';


=head1 Options

=cut

sub eupat_defvars {
    my $m = shift;
    $m->dpuc('eupat_defvars', @_);
=head2 Option eupattmpldir

Directory where eupat migration templates are to be found.

=cut
    $m->cache_define('eupattmpldir', 's', '/adv/perl/A2E/Eupat');
=over

=item Option blablabla
=cut

    $m->cache_define('blablabla', '!');

=back

=head1 Files

    /etc/opt/a2e/eupat.konf

=cut

    $m->konfayl('eupat.konf');
    $m->dpop('eupat_defvars', 1);
    return 1;
}

=head1 Bugs

=head1 To Do

=head1 See Also

    eupatmigr(1)
    http://a2e.de/adv/perl/A2E

=head1 Implementation

=head2 Documentation Conventions

We use a localised form of EBNF to explain the syntax of functions.

An explanation of this and some of our pre-definitions should be found at http://a2e.de/adv/perl/ebnf.

=head2 Initialisation Functions

These are used in the C<new_ready> routine.

=head2 Function defvars

=cut

sub defvars {
    my $m = shift;
    $m->dpuc('A2E::Eupat::defvars', @_);
    do { $m->dpop;return } if $m->{ready}->{eupat_defvars};
    $m->SUPER::defvars;
    $m->subdoks_defvars;
    $m->eupat_defvars;
    $m->konfayl('eupat.konf');
    $m->{doklvars} = {};
    $m->{ready}->{eupat_defvars} = 1;
    $m->dpop('A2E::Eupat::defvars');return 1;
};

=pod

=head3 Function eupat_postkonfig

=cut

sub eupat_postkonfig {
    my $m = shift;
    $m->dpuc('eupat_postkonfig', @_);
    $m->dpop('eupat_postkonfig', 1);
    return 1;
};

=head3 Function postkonfig

=cut

sub postkonfig {
    my $m = shift;
    $m->dpuc('A2E::Eupat::postkonfig', @_);
    do { $m->dpop;return } if $m->{ready}->{eupat_postkonfig};
    $m->SUPER::postkonfig;
    $m->eupat_postkonfig;
    $m->{ready}->{eupat_postkonfig} = 1;
    $m->dpop('A2E::Eupat::postkonfig', 1);
    return 1;
}

=head2 Implementation Functions

=head3 Function mov_svn_faylz_tu_olddir

make space: move svn files to 'OLD' subdirectory

=cut

sub mov_svnfaylz_tu_olddir {
    my $m = shift;
    $m->dpuc('mov_svn_faylz_tu_olddir', @_);
    $m->svn_mkdir('OLD') unless -d 'OLD';
    opendir DIR, '.';
    my @svn_faylz = grep { -f $_ && ! m(\Adeskr\.\w{2}\.html\Z) && $m->svn_status($_) == 3 } readdir DIR;
    closedir DIR;
    foreach my $fi (@svn_faylz) { my $fo = catfile 'OLD', $fi;printf "%s -> %s.", $fi, $fo if $m->{verbose};$m->svn_move($fi, $fo);print '.OK'."\n" if $m->{verbose}; };
    $m->dpop('mov_svn_faylz_tu_olddir', 1);
    return 1;
}

=head3 Function kre_htaccess

create OLD/.htaccess file from template

=cut

sub kre_htaccess {
    my $m = shift;
    $m->dpuc('kre_htaccess', @_);
    my $fi = catfile $m->{cache}->{eupattmpldir}, 'eupat_htaccess_tmpl.txt';
    my $fo = catfile 'OLD', '.htaccess';
    $m->tmpl_skrib_fayl($fi, $fo, coding => 'utf-8', mode => 'tt'); 
    $m->svn_add($fo);
    $m->dpop('kre_htaccess', 1);
    return 1;
}

=head3 Function pars_index

Parse index file of one language, write out the important parts to various filehandles

    args := fi
    fi :: input file e.g. 'index.en.html'
    rets := lits
    lits := hashref
    lits :: metadata text chunks found in index.??.html

The output for the main file _dok.*.txt is written to STDOUT.

=cut

sub pars_index {
    my $m = shift;
    $m->dpuc('pars_index', @_);
    my $fi = shift;
    $m->open_fayl($fi, 'i', 'r');
    my $lits = {};
    my %mlid2lit = ( keywords => 'kws', title => 'tit', description => 'des' );
    my $oldstyl = 0;
  HEAD: while ($m->getlini('i')) {
      given($_) {
	  when (m(\A\<meta\s+name\=\"(\w+)\"\s+content\=\"(.*)\"\>)) { LITS: { my $lit = $mlid2lit{$1};last unless $lit;my $val = HTML::Entities::decode($2);$lits->{$lit} = $val } };
	  when (m(\A\<meta\s+name\=\"review\"\s+content\=\"(\d{4}.\d{2}.\d{2})\"\>)) { $1 =~ s(\/)(-)g;$m->pvar('dokdatum') = $1 };
	  when (m(\A\<title\>(.*)\<\/title\>)) { $lits->{tit} ||= HTML::Entities::decode($1) };
	  when (m(\A\<div\s+id\=\"dokmid\"\>)) { last HEAD };
	  when (m(\A\<div\s+align\=\"center\"\>\<h1\>\<a\s+name\=\"top\">(.*)\</a\>\<\/h1\>\<\/div\>)) { $lits->{tit} ||= HTML::Entities::decode($1);$oldstyl = 1;last HEAD };
      };
  };
    print '#Native fmt=html <<_EOHTML_'."\n";
    my $last = $m->getlini('i');
  DOKMID: while ($m->getlini('i')) {
      given($_) {
	  when (m(\A\<div\s+id=\"dokped\"\>)) { last DOKMID };
	  break unless $oldstyl;
	  when (m(\A\<br\>)) { last DOKMID };
      };
      print $last;
      $last = $_;
  };
    print '_EOHTML_'."\n";
    $m->kloz_fayl('i');
    $m->dpop('pars_index', $lits);
    return $lits;
}

=head3 Function index2doklng

read index.??.html and write _dok.*.txt and _lng.*.??.txt

=cut

sub langrank {
    my $lang = shift;
    my %langrank = ( en => 1, de => 2, fr => 3, ca => 4, ja => 5 ) ;
    my $rank = $langrank{$lang} || 99;
    return $rank;
}

sub index2doklng {
    my $m = shift;
    my $dok = shift || $m->pstop('need dok argument');
    $m->dpuc('index2doklng', @_);
    my $fd = sprintf '_dok.%s.txt', $dok;
    $m->out2sav($fd);
    print join "\n", '% ['.'% emacs_top_local_variables %'.']', '% '.'['.qq{% PROCESS 'deplate_tmpl.txt' %}.']', '['.'%# PROCESS toc %'.']', '['.'%# topvar.litscat %'.']', '['.'% SWITCH lang -%'.']';
    print "\n";
    my @langs = ();
    my $fl_tmpl = catfile $m->{cache}->{eupattmpldir}, 'eupat_lng_xx_tmpl.txt';
    foreach my $fi (glob('OLD/index.??.html')) {
	my ($lang) = $fi =~ m(\AOLD\/index\.(\w{2})\.html\Z);
	$m->ustop('cant identify language of html index file %s', $fi) unless $lang;
	push @langs, $lang;
	print '['.qq{% CASE '$lang' -%}.']'."\n";
	my $lits = $m->pars_index($fi);
	my $fl = sprintf '_lng.%s.%s.txt', $dok, $lang;
	$m->tmpl_skrib_fayl($fl_tmpl, $fl, lits => $lits, mode => 'fundamental', coding => 'utf-8');
	$m->svn_add($fl);
    };
    print '['.'% END %'.']'."\n";
    print $m->emacs_bottom_local_variables('['.'%#', '%'.']', { mode => 'tt', coding => 'utf-8', srcfile => $m->absfayl($fd) });
    $m->sav2out;
    $m->svn_add($fd);
    my @langs_sortit = sort { langrank($a) <=> langrank($b) } @langs;
    $m->fayl_put_str('@langs', join ' ', @langs_sortit);
    $m->svn_add('@langs');
    $m->dpop('index2doklng', 1);
    return 1;
}

=head3 Function kre_pre

Write out _pre.*.txt
	
=cut

sub kre_pre {
    my $m = shift;
    $m->dpuc('kre_pre', @_);
    my $dok = shift || $m->pstop('need dok argument');
    my $fi = catfile $m->{cache}->{eupattmpldir}, 'eupat_pre_tmpl.txt';
    my $fo = sprintf '_pre.%s.txt', $dok;
    $m->tmpl_skrib_fayl($fi, $fo, coding => 'utf-8', mode => 'makefile', fol => $m->dir2fol($m->pvar('cwd')), hodie => $m->pvar('hodie'), timestamp => $m->timestamp, dok => $dok); 
    $m->svn_add($fo);
    $m->dpop('kre_pre', 1);
    return 1;
}

=head3 Function kre_lng

Write out _lng.*.txt but not _lng.*.??.txt
	
=cut

sub kre_lng {
    my $m = shift;
    $m->dpuc('kre_lng', @_);
    my $dok = shift || $m->pstop('need dok argument');
    my $fi = catfile $m->{cache}->{eupattmpldir}, 'eupat_lng_tmpl.txt';
    my $fo = sprintf '_lng.%s.txt', $dok;
    $m->tmpl_skrib_fayl($fi, $fo, coding => 'utf-8', mode => 'fundamental', fol => $m->dir2fol($m->pvar('cwd')), hodie => $m->pvar('hodie'), timestamp => $m->timestamp, dok => $dok); 
    $m->svn_add($fo);
    $m->dpop('kre_lng', 1);
    return 1;
}

=head3 Function kre_makefile

Write the Makefile

=cut

sub kre_makefile {
    my $m = shift;
    $m->dpuc('kre_makefile', @_);
    my $dok = shift || $m->pstop('need dok argument');
    my $fi = catfile $m->{cache}->{eupattmpldir}, 'eupat_makefile_tmpl.txt';
    my $fo = 'Makefile';
    my $dokl = ucfirst $dok;
    $dokl .= 'Make';
    $m->tmpl_skrib_fayl($fi, $fo, coding => 'utf-8', mode => 'makefile', dok => $dok, dokl => $dokl); 
    $m->svn_add($fo);
    $m->dpop('kre_makefile', 1);
    return 1;
};

=head3 Function kre_nav

create navigation file '@nav'.

This is a template which is why in writing it we use TAG_STYLE star

=cut

sub kre_nav {
    my $m = shift;
    $m->dpuc('kre_nav', @_);
    my $dok = shift || $m->pstop('need dok argument');
    my $fi = catfile $m->{cache}->{eupattmpldir}, 'eupat_nav_tmpl.txt';
    my $fo = sprintf '@nav.%s.txt', $dok;
    $m->tmpl_skrib_fayl($fi, $fo, tmplkonf => { TAG_STYLE => 'star' }, coding => 'utf-8', mode => 'tt', dok => $dok); 
    $m->svn_add($fo);
    $m->dpop('kre_nav', 1);
    return 1;

}

=head3 Function kre_dokfs

create configuration file '@dokfs'.

This is a template which is why in writing it we use TAG_STYLE star

=cut

sub kre_dokfs {
    my $m = shift;
    $m->dpuc('kre_dokfs', @_);
    my $dok = shift || $m->pstop('need dok argument');
    my $fi = catfile $m->{cache}->{eupattmpldir}, 'eupat_dokfs_tmpl.txt';
    my $fo = '@dokfs';
    my $dokl = ucfirst $dok;
    $dokl .= 'Dokfs';
    $m->tmpl_skrib_fayl($fi, $fo, dokl => $dokl, coding => 'utf-8', mode => 'conf', dok => $dok); 
    $m->svn_add($fo);
    $m->dpop('kre_dokfs', 1);
    return 1;
}

=head3 Function db_put_dok

    args := dir, dok, supdok

=cut

sub db_put_dok {
    my $m = shift;
    $m->dpuc('db_put_dok', @_);
    my $dir = shift;
    my $dok = shift;
    my $supdok = shift;
    my ($urldir) = $dir =~ m(\A\/eupat(.*)\Z);
    $urldir ||= $dir;
    my $dokurl = 'http://eupat.ffii.org'.$urldir;
    my $ok = $m->put('mlhtdok', { dok => $dok }, doktmp => 'now()', dokurl => $dokurl );
    $m->pstop('cant write url %s of dok %s to table mlhtdok', $dokurl, $dok) unless $ok;
    my ($supurl) = $m->get_record('mlhtdok', { dok => $supdok }, 'dokurl');
    $m->pstop('cant find url of dok %s in mlhtdok', $supdok) unless $supurl;    
    my ($dom, $supurldir) = $supurl =~ m(\Ahttp\:\/\/(eupat\.ffii\.org)(.*)\Z);
    $m->pstop('supurl %s doesnt seem to belong to the website', $supurl) unless $dom;
    my $dokstyp = $urldir =~ m(\A$supurldir\/) ? 'subdoks' : 'grpdoks';
    my $supdir = $m->dokdir($supdok);
    $m->pstop('supdok %s has no directory', $supdok) unless $supdir;
    my @urldirs = splitdir $urldir;
    my $doknom = pop @urldirs;
    $m->put('dokprop', { dok => $dok }, doknom => $doknom, smprior => '0.5', doktyp => 'pub', dokdatum => $m->pvar('dokdatum') || $m->pvar('hodie'));
    $m->{dir_fol} = 
	$ok = $m->add_subdoks($dok, $supdir, $dokstyp) or $m->ustop('kann %s nicht zu subdoks hinzufügen', $m->pvar('dok'));
    $m->dpuc('db_put_dok', $ok);
    return $ok;
}


=head3 Function eupatmigr

migrate one directory.

	args := dir, dok
	rets := ok
        dir => '/pikta'     
	dok => 'swpatpikta'    
	dok := string
        dok :: document identifier under which the generated document is to become known
	ok := boolean
	ok :: executed successfully?
=cut
 
sub eupatmigr {
    my $m = shift;
    $m->dpuc('eupatmigr', @_);
    my $dir = shift;
    $m->chwd($dir);
    my $dok = shift;
    my $supdok = shift;
    $m->mov_svnfaylz_tu_olddir;
    $m->kre_htaccess;
    $m->kre_pre($dok);
    $m->kre_lng($dok);
    $m->kre_nav($dok);
    $m->kre_dokfs($dok);
    $m->kre_makefile($dok);
    $m->index2doklng($dok);
    $m->svn_msg_commit('eupatmigr');
    $m->db_put_dok($dir, $dok, $supdok);
    $m->dpop('eupatmigr');
    return 1;
}

=head1 

=cut

return 1;

