=head1 NAME

A2E::Gettext

=cut

package A2E::Gettext;
use strict;

=head1 VERSION

    This describes version B<0.0.1> of A2E::Gettext.

=cut

    our $VERSION = '0.0.1';

=head1 SYNPOSIS

    use base 'A2E::MLHT', 'A2E::Gettext', 'A2E::Mbox2dok';
    my $p = new_ready A2E::MLHT;
    $p->add_subdoks($dok, $pwd);

=head1 DESCRIPTION

Add Gettext support to A2E::Prog or its children.

Mixin library, does not invoke the constructor of its base library.

This very thin add-on used to reside in A2E::Prog but appeared unused by many invoking higher level programs.

Also Locale::Gettext has recently been causing noisy error messages that call for minimizing its invocation and studying it seperately.
The noise occurred when moving from Perl version 5.12.1 to 5.18.1.

=head1 Prerequisites

    A2E::Prog(3)
    Locale::gettext(3)

=head1 AUTHOR

    Hartmut Pilch <http://a2e.de/phm>

=cut

use base 'A2E::Prog';
use Locale::gettext;
use POSIX;

=head1 Implementation

=head2 Options

=cut

sub gettext_defvars {
    my $m = shift;
    $m->dpuc('gettext_defvars');
=item --localedir /opt/a2e/share/locale

=cut
    $m->{konfig}->define('localedir', { DEFAULT => '/opt/a2e/share/locale' });
    $m->dpop('gettext_defvars', 1);
    return 1;
}

=head2 Subroutines

Superfluous  if used in mix-in mode as intended.



=head2 Routines

=head3 Function new_textdomain

Needed by programs that use gettext for message internationalization, e.g. spez2invs, A2E::Privat::Invs.

See latest difficulties: /phm/18/04/25/sig/_lng.phm_pub_sig180425.txt

=cut

sub new_textdomain {
    my $m = shift;my $prog = 'new_textdomain';$m->dpuc($prog, @_);
    my $domain_name = shift or $m->pstop('%p needs a domain name argument', p => $prog);
    my $locale = $m->pvar('locale');
  SETLOCALE: {
      # last if $m->{locale} or $m->{debug};
      setlocale(LC_MESSAGES, $locale) or $m->ustop('konnte Sprachparameter nicht auf %l setzen', l => $locale);
      my $lcl = setlocale(LC_MESSAGES, '') or die sprintf 'failed to set locale %s in gettext domain %s', $locale, $domain_name;
      $m->ustop('no resulting locale') unless $lcl;
      $m->uwarn('resulting locale %r differs from initial one %i', r => $lcl, i => $locale) unless $lcl eq $locale;
      $m->setpvar(locale => $lcl);
    };
    my $domain = domain Locale::gettext $domain_name;
    my $localedir = $m->{konfig}->localedir;
    $domain->dir($localedir) if $localedir;
    $m->dpop($prog, $domain);return $domain;
};

=head1 Changelog

=over
=item 2015-02-06 Version 0.0.1 phm ref 1100gettext
=back

=head1 TODO

Recently gettext has been causing noisy Perl debugging messages that call for study, cf dok:od_gettext1502

=head1 The End

=cut

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: /adv/perl/A2E/Gettext.pm.tmpl
# End:
