=head1 NAME

A2E::SArb::Hannum - Basic functions like hannum (han numeral) that 
 - do not depend on any A2E library but yet are not basic enough to be part of A2E::Prog
 - underly SArb/Make.pm functions 
 - are useful elswhere, e.g. in VMethods of A2E::MLHT.

=cut 

package A2E::Hannum;
use base 'A2E::Prog';
use POSIX qw(strftime); # TODO: transfer everyting from A2E::SArb::Make that requires this 
use utf8;

=head3 Function hannum_defvars
=cut

our %MSGS = (
    nonintarg => 'argument %n is not an integer',
    aammnoww => 'no week %w found in month %m of year %a'
);
sub hannum_defvars {
    my $prog = 'hannum_defvars';my $m = shift;$m->dpuc($prog,@_);
    do { $m->dpop;return } if $m->ready($prog);
    $m->prog_msgdefs('hannum',%MSGS);
    $m->ready($prog,1);$m->dpop($prog,1);return 1;
};
=head3 Function hannum: han numeral

   $m->hannum(124) => 一百二十四

Used by A2E::SArb::Make and A2E::MLHT as mixin.

=cut

our @HANMODS = ('零','一','二','三','四','五','六','七','八','九');
our @HANDEKS = ('','十','百','千');
our @HANMYRS = ('','萬','億','兆');

sub monint {
    local $_ = shift;
  NOCENT: {
      my ($num) = m(\A(.*)[.,]\d{2}\Z);
      last unless $num;
      $_ = $num;
    };
    s([.,])()g;
    return $_;
};
sub hannum {
    my $prog = 'hannum';my $m = shift;$m->dpuc($prog,@_);
    local $_ = monint shift;
  INTARG: {
      last if m(\A\d+\Z);
      $m->pstop(nonintarg => n => $_);
    };
    my $num = $_;
    my $mod = 0;
    my $hannum = '';
    my $handek = '';
  MYR: foreach my $myr (0..3) {
      my $restp = 0; # avoid trailing zeroes in 三百零零邁
      $hannum = $HANMYRS[$myr] . $hannum;
      foreach my $dek (0..3) {
	  $handek = $HANDEKS[$dek];
	  my $extra = '';my $hanmod = '';
	  $mod = $num % 10;
	  $restp ||= 1 if $mod;
	MOD: {
	    last unless $restp;
	    $extra = $handek if $mod; # avoid 零百零十 
	    $hanmod = $HANMODS[$mod] unless $mod == 1 and $dek == 1 and $num < 10; # avoid initial 一十二
	    $extra = $hanmod . $extra;
	  };
	  $hannum = $extra . $hannum;
	  last MYR if $num < 10;
	  $num = int $num / 10;
      };
  };
    $m->dpop($prog,$hannum);return $hannum;
};

=head3 mmddminmax

first day $ddmin and last day $ddmax of a month $mm of a year $aaaa
first day may be supplied as renderable tree argument
common subroutine of render_mondags and render_monwoks

=cut

sub mmddminmax {
    my $m = shift;$m->dpuc(mmddminmax => @_);
    my ($ddmin, $mm, $aaaa) = @_;
    my $ddmax = '30';
  MONMAX: {
      do { $ddmax = '31';last } if grep { $_ eq $mm } '01', '03', '05', '07', '08', '10', '12';
      last unless '02' eq $mm; # stay with 30 unless february
      $ddmax = '28';
      last if $aaaa % 4; # stay with 28 if *not* divisible by 4
      last unless $aaaa % 100; # stay with 28 nonetheless if divisible by 100
      $ddmax = '29';
    };
    $m->dpop(mmddminmax => [$ddmin, $ddmax]);return $ddmin, $ddmax;
}

=head3 mmwwminmax

first day $ddmin and last day $ddmax of a week (ww) in a month $mm of a year $aaaa
first day may be supplied as renderable tree argument
common subroutine of render_wokdags and render_dddmax

=cut

sub wwddminmax {
    my $m = shift;$m->dpuc(wwddminmax => \@_);
    my ($ddmin,$ww,$mm,$aaaa) = @_;
    my $ddmax = '30';
    ($ddmin, $ddmax) = $m->mmddminmax($ddmin, $mm, $aaaa, @_);
    my $smm = $mm - 1;
    my $saa = $aaaa - 1900;
    my $sw = 0;
    my @dds = ();
    foreach my $dd ($ddmin .. $ddmax) {
	$sw = strftime '%W', 0, 0, 0, $dd, $smm, $saa ;
	last if $ww < $sw;
	next if $ww > $sw;
	push @dds, $dd;
    };
    do { $m->uwarn(aammnoww => w => $ww, m => $mm, a => $aaaa);$m->dpop('wwddminmax');return } unless @dds;
    $ddmin = $dds[0];
    $ddmax = $dds[$#dds];
    $m->dpop('wwddminmax', $ddmin, $ddmax);
    return $ddmin, $ddmax;
}

=head3 Function get_monwoks

Subroutine of monwoks that does most of the work once monwoks has determined how the month starts
E.g. for December 2018 the invocation will be

	get_monwoks('W','48','01',6,'31')

Iterate through all weeks starting from $sdd => '01' up to $ddmax => '31', where $sdd is Saturday ($swd => 6) of week 48.
The result is that of monwoks.
 
=cut

sub get_monwoks { 
    my $prog = 'get_monwoks';my $m = shift;$m->dpuc($prog,@_); 
    my ($ssww_pfx,$sww,$sdd,$swd,$ddmax) = @_; # 'W', '48', '01', 6, '31'
    $swd ||= 7;
    my @monwoks = ();
    my $ssww = $ssww_pfx . $sww; # W48
    my $ddo = $sdd - $swd; # -5: weekday offset for initial week is negative in case of 'W' or zero in case of 'w' weeks
    $ssww_pfx = 'w'; # upcoming weeks all begin with 'w'
    my @wok = ();
    push @wok, $ssww, $ddo, sprintf '%02d', $sdd; # ('W48',-5,'01')
    my $finp = 0;
  DAG: while ($sdd < $ddmax) {
      $sdd = $sdd + 7 - $swd;
      $sdd = $ddmax if $sdd > $ddmax;
      ++$sww;$swd = 0;$ddo = $sdd;
      my $sdz = sprintf '%02d', $ddo;
      push @wok, $sdz; # ('W48',-5,'01','01'); ('w49',-5,'01','02')
      push @monwoks, [ @wok ];
      @wok = ();
      $ssww = $ssww_pfx . sprintf '%02d', $sww; # 'w49'
      my $sda = $sdd + 1;
      last if $sda > $ddmax;
      push @wok, $ssww, $ddo, sprintf '%02d', $sda; # ('w49', 2, '03')
  };
  FINI: {
      last unless @wok;
      push @wok, sprintf '%02d', $sdd;
      push @monwoks, [ @wok ];
    };
    $m->dpop($prog,@monwoks);return \@monwoks;
};

=head3 Function monwoks

	$woks0 = $m->monwoks('2018','12')
        $woks0 => [ ['W48', -5, '01','02'], ['w49',2,'03','09'], ['w50',9,'10','16'], ['w51',16,'17','23'], ['w52',23,'24','30'], ['w53',30,'31','31'] ]
	$woks1 => $m->monwoks('2018','12','01');

=cut

sub monwoks {
    my $m = shift;my $prog = 'monwoks';$m->dpuc($prog,@_);
    my ($aaaa,$mm,$ddmin) = @_;
    my $ddmax = '30';
    ($ddmin, $ddmax) = $m->mmddminmax($ddmin, $mm, $aaaa);
    # now prepare invocation of strftime to obtain the month's first week $sww0 %W and $swd0 weekday %w
    my $smm = $mm - 1;
    my $saa = $aaaa - 1900;
    my $sdd = $ddmin || '01';
    my $sww0 = strftime '%W', 0, 0, 0, 1, $smm, $saa ;
    my $sww = $sww0;
    my $swd0 = strftime '%w', 0, 0, 0, 1, $smm, $saa ;
    my $swd = $swd0;
  DDMIN: { # additionally obtain separate $sww and $swd for ddmin if needed
      last unless $sdd > 1;
      $sww = strftime '%W', 0, 0, 0, $sdd, $smm, $saa ;
      $swd = strftime '%w', 0, 0, 0, $sdd, $smm, $saa ;
    };
    # obtain correct letter for first week: W if it began in previous month, w if it starts on 01
    my $ssww_pfx = 'w';
  SSWW: { # see whether we need 'W' to distinguish from same week at end of last month
      # special case where month starts with monday
      last if $swd0 == 1;
      last if $sww > $sww0;
      $ssww_pfx = 'W';
    }
    my $monwoks = $m->get_monwoks($ssww_pfx,$sww,$sdd,$swd,$ddmax);
    $m->dpop($prog,$monwoks);return $monwoks;
};
sub monwok {
    my $prog = 'monwok';my $m = shift;$m->dpuc($prog,@_);
    my ($aaaa,$mm,$www) = @_;
    my $monwoks = $m->monwoks($aaaa,$mm,'01');
    foreach my $monwok (@$monwoks) {
	my @wok = @$monwok;
	my $xxx = shift @wok;
	next unless $xxx eq $www;
	$m->dpop($prog,\@wok);return \@wok;
    };
    $m->uwarn(aammnoww => w => $www, a => $aaaa, m => $mm);
    $m->dpop($prog);return;
};

=head4 Function wokmon

Derive month from week

    wokmon(2015,'w13') => '03'
    wokmon(18,'W48') => '12'

=cut

sub wokmon {
    my $prog = 'wokmon';my $m = shift;$m->dpuc($prog,@_);
    my ($aaaa,$www,$w,$ww) = @_;
    my $saa = $aaaa - 1900;
    my $smm = int $ww * 12 / 53; 
    my $plus = 0;
    my ($mm, $mm1, $ww0, $ww1) = ();
    my $aamm2ww = {};$aamm2ww->{$aaaa} = {};
    my $aamm2wd = {};$aamm2wd->{$aaaa} = {};
  MOV: {
      $smm = sprintf '%02d', $smm;
      $mm = sprintf '%02d', $smm + 1;
      $aamm2ww->{$aaaa}->{$mm} ||= strftime '%W', 0, 0, 0, 1, $smm, $saa ;
      $ww0 = $aamm2ww->{$aaaa}->{$mm};
      if ($ww0 > $ww) { $m->techo("ww:$ww lt ww0:$ww0 for mm:$mm and smm:$smm");$m->pstop('attempting to move up and down in wokmon at smm %m', m => $smm) unless $plus <= 0;$plus = -1;$smm--;redo MOV };
      $mm1 = sprintf '%02d', $mm + 1;
      $aamm2ww->{$aaaa}->{$mm1} ||= $mm > 11 ? '55' : strftime '%W', 0, 0, 0, 1, $mm, $saa ;
      $ww1 = $aamm2ww->{$aaaa}->{$mm1};
      if ($ww1 < $ww) { $m->techo("ww:$ww gt ww1:$ww1 for mm:$mm an mm1:$mm1");$m->pstop('attempting to move down and up in wokmon at smm %m', m => $smm) unless $plus >= 0;$plus = +1;$smm++;redo MOV };
      last if $ww0 < $ww and $ww < $ww1;
    EQ_WW0: {
	last unless $ww eq $ww0;
	last MOV if $w eq 'W';
	$aamm2wd->{$aaaa}->{$mm} ||= strftime '%w', 0, 0, 0, 1, $smm, $saa ;
	last MOV if $aamm2wd->{$aaaa}->{$mm} eq 1; # we are in first week of month that begins with monday and therefor $w is 'w';
	# we would be in first week of month if it began with 'W' but it doesnt so that we are in last month which is why we have to move down one and then are done
	$m->pstop('attempting to move first up then down in wokmon.eq_ww0 at smm %m', m => $smm) unless $plus <= 0;$smm--;last MOV;
      };
    EQ_WW1: {
	last unless $ww eq $ww1;
	$aamm2wd->{$aaaa}->{$mm1} ||= strftime '%w', 0, 0, 0, 1, $mm, $saa ;
	last MOV unless $w eq 'W' or $aamm2wd->{$aaaa}->{$mm1} eq 1;
	# we are now really in the next month's first week rather than in the current one's incipient last week and therefore must move up one month and then are done
	$m->pstop('attempting to move first down then up in wokmon.eq_ww1 at smm %m', m => $smm) unless $plus >= 0;$smm++;last MOV;
      };
    };
    $mm = sprintf '%02d', $smm + 1;
    $m->dpop($prog,$mm);return $mm;
};

sub wokdags {
    my $prog = 'wokdags';my $m = shift;$m->dpuc($prog,@_);
    my ($aaaa,$mm,$ww,$ddmin) = @_;
    ($ddmin, my $ddmax) = $m->wwddminmax($ddmin, $ww, $mm, $aaaa);
    do { $m->dpop($prog);return } unless $ddmax;
    my @wokdags = ($ddmin .. $ddmax);
    $m->dpop($prog,@wokdags);return @wokdags;
};


return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/Hannum.pm"
# End:
