package A2E::Lihs;
use base A2E::Prog;
use File::Spec::Functions qw(catdir catfile);
use A2E::File;
use strict;use warnings 'all';no warnings 'uninitialized';
use A2E::Vartyps;
our %MSGS = (
    getnonvar => 'trying to read invalid variable %n',
    nrnotyp => 'variable %n with value %v is not of type %t but %r',
    varsvref => 'supposed string variable %n is a %r' );
sub new {
    my ($class,$dir,%opts) = @_;
    my $lihs = {};
    bless $lihs, $class;
    # $lihs->{prog} = shift;
    # printf STDERR "lihs attribute prog is of class %s\n", ref $lihs->{prog}; # OK
    if ($dir) { $lihs->setdir($dir) } else { $lihs->{vars}->{'__'.$_.'__'} = 1 for qw(nolitsdir nonstorp) };
    foreach my $k (keys %opts) { $lihs->{$k} = $opts{$k} };
    $lihs->{$_} ||= {} for qw(vars varp varq lsts lstp hacs hacp vrbs doms lvlmax);
    $lihs->{typs} = new A2E::Vartyps;
    $lihs->{lsts}->{__dirnoms__} = [qw(llitsdir tlitsdir litsdir)];
    $lihs->prog_minidefs;
    $lihs->prog_msgdefs('lihs',%MSGS);
    $lihs;
};
sub setdir {
    my ($lihs,$dir,%tags) = @_;
    return $lihs->{vars}->{__litsdir__} if $dir eq $lihs->{vars}->{__dir__};
    do { mkdir $dir or die "lihs directory $dir does not exist and cannot be created\n" } unless -d $dir;
    $lihs->{vars}->{__dir__} = $dir;
    my $litsdir = catdir $dir, '.lits';
    $lihs->{vars}->{__litsdir__} = $litsdir;
    return unless -w $dir;
  MKDIR: {
      last if -d $litsdir;
      mkdir $litsdir;
    };
    foreach my $k (keys %tags) {
	next unless $tags{$k};
	my $sdir = catdir $litsdir, join '.', $k, $tags{$k};
	mkdir $sdir unless -d $sdir;
	$lihs->{vars}->{'__'.$k.'litsdir__'} = $sdir;
    };
    my ($cwd,$lwd) = $lihs->chwd($litsdir);
    die "not in a lits working directory at $cwd\n" unless $cwd =~ m(\.lits\Z) and $lihs->inod($cwd) == $lihs->inod('.');
    $lihs->{vars}->{'__'.$_.'__'} = 0 for qw(nolitsdir nonstorp);
    $lihs->put_tups('vars',fyl => $lihs->tupsfyl('prae'), k2v => sub { $lihs->getstr($_[0]) } );
    while (my ($k,$v) = each %{$lihs->{vars}}) { # make sure all vars that were stored when no directory existed are now recorded
	$lihs->fayl_put_str($k,$v) if length $v and !-f $k;
    };
    $lihs->chwd($lwd);
    $litsdir;
};
sub addrolitsdirs { # set readonly litsdirs (e.g. for inheriting parent's values in bootstrap phase file:/adv/perl/A2E/MLMK.pm.tmpl::addrolitsdirs)
    my $lihs = shift;
    while (@_) {
	my $nom = shift;my $val = shift;
	$lihs->{vars}->{'__'.$nom.'__'} = $val;
	$lihs->uniqpush($lihs->{lsts}->{__dirnoms__}, [$nom]);
    };
    $lihs->{vars}->{__nolitsdir__} = 0;
    1;
};

=head4 m->tregfyl('vars', '%', 'l') => '.lits/zh/vars.hac'

	m->tregfyl('vars', '%') => '.lits/vars.hac'
	m->tregfyl('bars', '%', '', '.litsbak', 'zh', 'latex') => '.litsbak/zh/bars.hac'
	m->tregfyl('dars', '%', '', '', 'tmp') => '/tmp/dars.hac'

=cut
sub tregfyl {
    my ($lihs,$nom,$typ,$litstag,%opts) = @_;
    my @dfls = $lihs->ref2list($opts{dfls});
    $typ ||= '@';
    if (@dfls) { # dirfols argument given
	if ($dfls[0]) { # relative directory with no arg0 '', supply absolute base
	    unshift @dfls, $lihs->{vars}->{__dir__} || '.';
	}
    } else { # m->tregfyl('vars', '%', 'l') => '.lits/zh/vars.hac'
	push @dfls, $lihs->litstag2litsdir($litstag,$opts{flng});
    };
    $typ = $lihs->typ($typ); # '%' => 'hac'
    my $fyl = catfile @dfls, join '.', $nom, $typ; # .lits/vars.hac
    $fyl;
};
sub tupsfyl {
    my ($lihs,$treg,%opts) = @_;(my $litstag,%opts) = $lihs->shifthash('litstag',%opts);
    $lihs->tregfyl($treg,'%',$litstag,%opts);
};
sub nomsfyl {
    my ($lihs,$treg,%opts) = @_;(my $litstag,%opts) = $lihs->shifthash('litstag',%opts);
    $lihs->tregfyl($treg,'@',$litstag,%opts);
};
sub vrbsfyl {
    my ($lihs,%opts) = @_;
    $lihs->tupsfyl('vrbs',%opts);
};
sub varqfyl {
    my ($lihs,%opts) = @_;
    $lihs->tupsfyl('varq',%opts);
};
sub hacsfyl {
    my ($lihs,%opts) = @_;
    $lihs->nomsfyl('hacs',%opts);
};
sub varsfyl {
    my ($lihs,%opts) = @_;
    $lihs->tupsfyl('vars',%opts);
};
sub opts2nomsfyl {
    my ($lihs,$treg,%opts) = @_;
    $opts{fyl} || $lihs->nomsfyl($treg,%opts);
};
sub opts2tupsfyl {
    my ($lihs,$treg,%opts) = @_;
    $opts{fyl} || $lihs->tupsfyl($treg,%opts);
};
sub put_noms {
    my ($lihs,$treg,$trgr,%opts) = @_;
    $trgr ||= $lihs->{$treg};
    my $fyl = $lihs->opts2nomsfyl($treg,%opts);
    $lihs->fayl_put_lst($fyl,keys %$trgr);
    $fyl;
};
sub get_noms {
    my ($lihs,$treg,%opts) = @_;
    my $fyl = $lihs->opts2nomsfyl($treg,%opts);
    my @noms = $lihs->fayl_get_lst($fyl);
  STOR: {
      my $stor = $opts{stor};
      last unless $stor;
      $lihs->pstop('stor must be hashref') unless 'HASH' eq ref $stor;
      $stor->{$_} = 1 for @noms;
    };
    @noms;
};
sub treg_tups {
    my ($lihs,$treg) = @_;
    $lihs->{$treg};
};
sub treg_noms {
    my ($lihs,$treg) = @_;
    keys %{$lihs->{$treg}};
};
sub vrbs_noms {
    shift->treg_noms('vrbs');
}
sub put_tups { # 
    my ($lihs,$treg,%opts) = @_;local $_;
    die "invoked put_tups with void treg argument\n" unless $treg;
    my $fyl = $lihs->opts2tupsfyl($treg,%opts);
    my $tups = $opts{tups} || $lihs->treg_tups($treg);
    $lihs->fayl_put_tups($fyl,$tups,%opts);
};
sub get_tups { 
    my ($lihs,$treg,%opts) = @_;
    my $fyl = $lihs->opts2tupsfyl($treg,%opts);
    do { warn "no file $fyl to restore $treg from\n";return } unless -f $fyl;
    $opts{stor} ||= $lihs->{$treg}; 
    $lihs->fayl_get_tups($fyl,%opts);
};
sub typl { ('$','@','%') };
sub typ { shift->{typs}->typ(shift) };
sub treg { shift->{typs}->treg(shift) };
sub ttyr { shift->{typs}->ttyr(shift) };
sub typp { shift->{typs}->typp(shift) };
sub tyrp { shift->{typs}->tyrp(shift) };
sub tref { shift->{typs}->tref(shift) };
sub iniarb { shift->{typs}->iniarb(shift) };
sub arbtypp { 
    my ($lihs,$typ) = @_;
    $lihs->{typs}->arbp($typ);
};
sub litstag2litsdir {
    my ($lihs,$litstag) = @_;
    !$litstag ? $lihs->{vars}->{__litsdir__} : 't' eq $litstag ? $lihs->{vars}->{__tlitsdir__} : $lihs->{vars}->{__llitsdir__};
};
sub getnomf {
    my ($lihs,$nom) = @_;
    my @dirs = grep { $_ } map { $lihs->{vars}->{'__'.$_.'__'} } @{$lihs->{lsts}->{__dirnoms__}};
    foreach my $dir (@dirs) {
	my $fi = catfile $dir, $nom;
	next unless -s $fi;
	return $fi;
    };
  LANG0DIR: { # file:/phm/22/03/18/sig/_lng.phm_pub_sig220318.txt::lang0dir
      my $langs = $lihs->{lsts}->{langs};
      last unless $langs;
      my $lang0 = $langs->[0];
      last unless $lang0;
      my $lang = $lihs->{vars}->{lang};
      last unless $lang;
      last if $lang0 eq $lang;
      my $dir = catdir $lihs->{vars}->{__litsdir__}, join '.', 'l', $lang0;
      my $fi = catfile $dir, $nom;
      last unless -s $fi;
      return $fi;
    };
};
sub fgetval {
    my ($lihs,$nom) = @_;local $_ = $nom;
    my $fi = $lihs->getnomf($nom);
    $lihs->fayl_get_str($fi);
};
# <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
sub vgetval { # pre-stored values
    my ($lihs,$nom) = @_;
    $lihs->{lvars}->{$nom} || $lihs->{vars}->{$nom} || $lihs->{tvars}->{$nom};
}; # </nov>
sub getval {
    my ($lihs,$nom,%opts) = @_;local $_ = $nom;
    $lihs->pstop(getnonvar => n => $_) if ref $_ or !m(\A\w+\Z) and not $opts{badlitnom_ok}; 
    $lihs->{varq}->{$nom}++;
    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
    # local $_ = $lihs->{vars}->{$nom};
    local $_ = $lihs->vgetval($nom); # </nov>
    my $ref = ref $_;
    my $val = '';
    my $fh = undef;
    if ('A2E::File' eq $ref) {
	$fh = $_;
    } elsif ($ref) {
	return $_;
    } else {
	$val = $_;
    };
    my $mt0 = $lihs->{vars}->{__mt_minim__}; # allow ignoring of old stored values during variable setting at startup
  IO: {
    NUL: {
	# last unless $lihs->{varp}->{$nom};
	last if $fh;
	last if $lihs->{vars}->{__nolitsdir__}; # file:/adv/perl/A2E/Tmplfil.pm.tmpl::__nolitsdir__
	last if !$mt0 and $val;
	my $fi = $lihs->getnomf($nom);
	my @stat = stat $fi;
	last if !@stat;
	return if $mt0 and $mt0 > $stat[9];
	$val = $lihs->fayl_get_str($fi);
      };
    FH: {
	last unless $fh;
	$val = $fh->gettoplini;
	last IO;
      }; 
    };
    $val;
};
our @varsnoms = ('vars','lvars','tvars');
sub vars_fh_get_strs { # return all values as a hash of strings, resolving any filehandles
    my $lihs = shift;
    my %strs = ();
    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
    foreach my $varsnom (@varsnoms) {
	while (my ($k,$v) = each %{$lihs->{$varsnom}}) {
	    $strs{$k} = 'A2E::File' ne ref $v ? $v : $v->gettoplini;
	};
    }; # </nov>
    %strs;
};
sub getvalr {
    my ($lihs,$nom,$typ) = @_;$typ ||= '$';
    my $treg = $lihs->treg($typ);
    $lihs->{$treg}->{$nom};
};
sub getstr {
    my ($lihs,$nom) = @_;
    local $_ = $lihs->getval($nom);
    my $ref = ref $_;
    $lihs->pstop(varsvref => { n => $nom, r => $ref }) if $ref;
    die "string variable $nom is a denatured $_\n" if m(\AA2E\:\:File\=GLOB\b);
    $_;
};
sub getarb {
    my ($lihs,$nom) = @_;
    $lihs->{lsts}->{$nom} || $lihs->{hacs}->{$nom} || $lihs->{vars}->{$nom};
};
sub gettypr {
    my ($lihs,$nom,$typ) = @_;
    return unless $nom;
    return $lihs->getstr($nom) if !$typ or '$' eq $typ;
    my $treg = $lihs->treg($typ);
    my $tref = $lihs->tref($typ);
    my $val = $lihs->{$treg}->{$nom};
    return unless $val;
    return $val if $tref eq ref $val;
    $lihs->pstop(nrnotyp => n => $nom, v => $val, t => $typ, r => $tref);
};
sub getlstr {
    my ($lihs,$nom) = @_;
    $lihs->gettypr($nom,'@');
};
sub gethacr { 
    my ($lihs,$nom) = @_;
    $lihs->gettypr($nom,'%');
};
sub getlst { 
    my ($lihs,$nom,$typ) = @_;
    my $lstr = $lihs->getlstr($nom);
    return unless $lstr;
    @$lstr;
};
sub gethac { 
    my ($lihs,$nom,$typ) = @_;
    my $hacr = $lihs->gethacr($nom);
    return unless $hacr;
    %$hacr;
};
sub chkvar {
    my ($lihs,$nom,$val) = @_;$val //= $lihs->{vars}->{$nom};local $_ = $val;my $ref = ref $val;
    do { warn "resolving a $ref that should have been resolved already\n";return $val->gettoplini() } if 'A2E::File' eq $ref;
    die "string variable $nom refers to a $ref\n" if $ref;
    die "string variable $nom is a denatured $val\n" if m(\AA2E\:\:File\=GLOB\b);
    $val }
=head4 $ok = $m->setvartyp($nom,$typ,%alt): set a variable to a type and fail if already set to conflicting type

%alt :: hash saying whether type conflicts with $typ; 1 means there is a conflict dont change; 2 means fail.

=cut
sub setvartyp {
    my ($lihs,$nom,$typ,%alt) = @_;
    while (my ($k,$v) = each %alt) {
	next if !$v;
	printf STDERR "variable $nom is of type $k, not changing to $typ\n";
	return 1 if $v == 1;
	return;
    };
    $lihs->{var2typ}->{$nom}->{$typ} = 1;
    $lihs->{typ2var}->{$typ}->{$nom} = 1;
    1;
};

# =head4 $ok = $lihs->chkvartyp($nom,$val,$fol,$typ,$litstag): classify variable and prevent resetting outside type scope
# 
# so far we are only classifying so as to obtain data, the immediate application being somethig like 
# A2E::Lihs::litstag_undef_vars.
# 
# as an initialization routine to be invoked by A2E::MLDK::dk_init_dst etc before vocabulary of new language is read. 
# 
# Other uses include enforcement of type constraints.
# A special verb 
# 
# 	special = |vartyp|d|dok|
# 
# could help correct the implicit classification offered here.
# 
# Also type info could be made insertable in the assignement statement:
# 
# 	dok.d = ...
# 	tit.t:t,l:de = ...
# 	tit.t,de,html = ...
# 
# =cut
# 
# sub chkvartyp {
#     my ($lihs,$nom,$val,$fol,$typ,$litstag) = @_;$val ||= $lihs->{vars}->{$nom};local $_ = $val;my $ref = ref $val;
#     return 1 if $lihs->{var2typ}->{$nom};
#     $lihs->setvartyp($nom,$typ);
#     if ($litstag) {
# 	$lihs->setvartyp($nom,$litstag);
#     };
#     return 1 if $lihs->tref($typ);
#     local $_ = $val;
#     if (m(\A\d+\Z)) {
# 	$lihs->setvartyp($nom,'i', t => 1);
#     } elsif (m(\A\d+[.,]\d+\Z)) {
# 	$lihs->setvartyp($nom,'r', t => 1);
#     } elsif (m(\A[a-z,|&+]+\Z)) {
# 	$lihs->setvartyp($nom,'d', t => 1);
#     } else {
# 	$lihs->setvartyp($nom,$litstag, d => 1); # 201226 s($litstag)('t')
#     };
#     1 };
sub litstag_undef_vars {
    my ($lihs,$litstag) = @_;
    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
    # $lihs->{vars}->{$_} = undef for keys %{$lihs->{typ2var}->{$litstag}};
    # foreach $k (keys %{$lihs->{vars}}) { my %typ = %{$lihs->{var2typ}->{$k}};next if $typ{d} or $typ{i} or $typ{r};$lihs->{vars}->{$k} = undef } 
    $litstag ||= 'l' ;
    my $varsnom = $litstag . 'vars';
    %{$lihs->{$varsnom}} = (); # </nov>
};

sub putval {
    my ($lihs,$prog,$nom,$val,%opts) = shift->ppuc('putval',@_);
    $lihs->{vars}->{$nom} = $val;
    $lihs->dpop1($prog,$val);
};
sub putstr {
    my ($lihs,$prog,$nom,$val,%opts) = shift->ppuc('putstr',@_);
    my $linp = $opts{linp} ? 1 : 0; # lines that come from body i.e. file:/adv/perl/A2E/Tmplfil.pm.tmpl::gruplini_nosplit and are written to their respective files only
    $lihs->pstop('trying to set invalid variable %n to value %v', n => $nom, v => $val) unless $nom =~ m(\A\w+\Z) or $opts{badnom_ok};
    $val = $lihs->chkvar($nom,$val);
    my $len = length $val;
    $lihs->{varp}->{$nom} = $len unless $linp;
    my $nonstorp = $opts{nonstorp} || $lihs->{vars}->{__nonstorp__};
    $lihs->decho('not writing %n to disk because nonstorp is on', n => $nom) if $nonstorp; 
    TYP: {
	last unless $opts{typ};
	$lihs->{var2typ}->{$nom}->{$opts{typ}} = 1;
    };
    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
    # my $litstag = $lihs->{var2typ}->{$nom}->{d} && !$opts{flng} ? '' : $opts{litstag};
    my $litstag = $opts{litstag}; # set_litstag
  SET_LITSTAG: {
      last unless $litstag;
      last if $litstag eq 't';
      last if $opts{flng} and $litstag eq 'l';
      $litstag = '';
    };
    my $varsnom = $litstag . 'vars';
    # </nov>
    my $ripetp = !$litstag && $lihs->{vars}->{__ripetp__} ? 1 : 0; # 避免寫第二語版以後重寫無語變量，但又允許某語獨有資料 file:/adv/perl/A2E/Tmplfil.pm.tmpl::__ripetp__
    my $litsdir = $nonstorp ? '' : $lihs->litstag2litsdir($litstag);
    my $fh = undef;
  FAYL: {
      last if !$litsdir;
      my $fayl = catfile $litsdir, $nom;
      last if -r $fayl and $ripetp;
      # die "lits file path $fayl lacks .lits component\n" unless $fayl =~ m(\.lits\b);
      if ($len) { $lihs->fayl_put_str($fayl,$val) } else { unlink $fayl };
      last if $linp;
      my $freq = $lihs->{freq}->{$nom} || $lihs->{varq}->{$nom}; # query frequency either from last time per lihs_var2arb or from this time, cf file:/adv/perl/A2E/MLHT.pm.tmpl::freq
      if ($freq) { $len = int $len / $freq };
      if ($lihs->{lsts}->{$nom}) {
	  $len = int $len / 2;
      } elsif ($lihs->{hacs}->{$nom}) {
	  $len = int $len / 3;
      };
      last if 64 > $len;
      $fh = new A2E::File $fayl, 'r';
    };
  SETVAR: {
      last if $linp;
      # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
      # $lihs->{vars}->{$nom} = $fh || $val;
      $lihs->{$varsnom}->{$nom} = $fh || $val; # </nov>
    };
  LVLMAX: {
      last if $linp;
      my $lvlmax = $opts{lvlmax};
      last unless $lvlmax;
      $lihs->{lvlmax}->{$nom} = $lvlmax;
    };
  ECHO: {
      last unless $opts{echo};
      print STDERR $litstag . ':' .  $nom . "\t" . $val . "\n";
    };
  PUTVARS: {
      my $putvars = $opts{putvars};
      last if !$putvars;
      $putvars->{$nom} = $val;
    };
    $lihs->dpop1($prog,$val);
};
sub defstr { # ||=
    my ($lihs,$nom,$val,%opts) = @_;
    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
    # my $val0 = $lihs->{vars}->{$nom};
    my $val0 = $lihs->vgetval($nom); # </nov>
    return $val0 if $val0;
    $lihs->putstr($nom => $val, %opts);
};
sub del {
    my ($lihs,$nom,%opts) = @_;
    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
    # my $val = $lihs->{vars}->{$nom};
    my $val = $lihs->vgetval($nom);
    # </nov>
    if ($val) {
	$val->close if 'A2E::File' eq ref $val;
	# <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
	$lihs->{lvars}->{$nom} = undef;
	$lihs->{tvars}->{$nom} = undef;
	# </nov>
	$lihs->{vars}->{$nom} = undef;
    };
    foreach my $dirnom (@{$lihs->{lsts}->{__dirnoms__}}) {
	my $dir = $lihs->{vars}->{$dirnom};
	next unless $dir and -d $dir;
	my $fayl = catfile $dir, $nom;
	next unless -r $fayl;
	unlink $fayl;
    };
    foreach my $treg (qw(lsts hacs)) {
	next unless defined $lihs->{$treg}->{$nom};
	$lihs->{$treg}->{$nom} = undef;	    
    };
    $nom;
};

=head4 DESTROY
=cut    
sub DESTROY {
    my $lihs = shift;
    # undef $lihs->{prog};
    # undef $lihs->{typs};
    1;
};

1;
# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/Lihs.pm"
# End:
