=head1 NAME

A2E::MLDK

=cut

package A2E::MLDK;

=head1 DESCRIPTION

Generate files in several languages from one multilingual template, as done by A2E::MLHT, but in addition generate a system of dependency rules as  attempted by A2E::MakefileCache and execute targets based thereon, using the Daiku system.
Each document $m->pvar('dok') gets its own Daiku workspace, cf Daiku
In the long run this is likely to completely replace A2E::MakefileCache and the current system of external makefiles, and to become the base of A2E::Mktdir (which currently calls `mldk lits`).
Started file:/phm/18/09/26/sig/_lng.phm_pub_sig180926.txt::mldk

=head1 PREREQUISITES

    A2E::MakefileCache(3)
    A2E::MLHT(3)
    Daiku(3)	

=cut

use base A2E::MLHT, A2E::Dokdata2db;
use strict;use warnings 'all';no warnings 'uninitialized';
# use Data::Dumper;
use Daiku;
use File::Spec::Functions qw(:ALL);
use File::Touch;
use POSIX;
use List::MoreUtils qw(uniq);
our $old_sdbm_p = 0;
our @lihs_progvars = qw(tmpl tmls dok); # progvars that are overridden by what is found in _pre.*.txt et al
our %MSGS = (
    baddkr => 'invalid daiku registry %r',
    dkbt => 'mldk building target %t',
    dkbnt => 'mldk built %n jobs en route to %D',
    notmls => 'no tmls found in %p, assuming html'
    );
sub mldk_defvars {
    my ($m,$prog,$lihs) = shift->qlpuc('mldk_defvars',@_) or return;
    $m->dokdata2db_defvars($lihs);
=head4 (--chdir|--chwd) I<dir>    
Change to the specified working directory before doing anything.  Similar to 'make -C I<dir>'.
=cut    
    $m->konfig_define(['chdir','chwd','C'],'s');
    $m->konfig_define('postproc','%', { latex => 'daiku' });
    $m->konfig_define('no_postproc','!');
=head4 (--cdok|-D) I<dok>    
Change to the specified working directory before doing anything.  Similar to 'make -C I<dir>'.
=cut    
    $m->konfig_define(['cdok','D'],'s');
=head4 --svn_no_deps
Whenever a file is in the Subversion repository, it shall be treated as source text and not be built from other files.
This can be overridden by setting the svn:keywords deps=1
=cut
    $m->cache_define('svn_no_deps','!',0);
    $m->konfig_define('ssh_host','s', 'a2e.de');
    $m->konfig_define('langlands','%', { zh => '|tw|cn|' }, 1);
    $m->konfayl('mldk.konf');
    $m->prog_msgdefs('mldk',%MSGS);
    $m->setpvar(svn_add_ok => 1, lihs => $lihs);
    $m->qpop1($prog,1);
};
=head3 defvars
=cut
sub defvars {
    my ($m,$prog,$lihs) = shift->qlpuc('A2E::MLDK::defvars',@_) or return;
    $m->SUPER::defvars($lihs);
    $m->mldk_defvars($lihs);
    $m->dpop1($prog,1);
};

=head3 init_daiku_klas($class) -- initiate a class of daiku workspaces
A class can e.g. be 'prog' for "program" and the member could be 'mldk'.
Choose names freely in range m(\A\w+\Z) so as to avoid special classes _dok and _tmpl.
=cut
sub init_daiku_klas {
    my ($m,$prog,$k) = shift->ppuc('init_daiku_klas',@_);
    $m->pstop('class %k already exists', k => $k) if $m->{daiku}->{$k};
    $m->{daiku}->{$k} = {};
    $m->dpop1($prog,$m->{daiku}->{$k});
};

=head3 mldk_postkonfig
=cut
sub mldk_postkonfig {
    my ($m,$prog,$lihs) = shift->qlpuc('mldk_postkonfig',@_) or return;
    # $m->mlht_postkonfig($lihs);
    $m->{daiku} = {};
    foreach my $k (qw(_dok _tmpl)) { $m->init_daiku_klas($k) };
    my %langlands = ();
    foreach my $k (keys %{$m->{konfig}->langlands}) { $langlands{$k} = $m->textsep2listref($m->{konfig}->langlands->{$k}) };
    $m->setphac(langlands => \%langlands, lihs => $lihs);
    $m->qpop1($prog,1);
};
=head3 postkonfig
=cut
sub postkonfig {
    my ($m,$prog,$lihs) = shift->qlpuc('A2E::MLDK::postkonfig',@_) or return;
    $m->SUPER::postkonfig($lihs);
    $m->mldk_postkonfig($lihs);
    $m->qpop1($prog,1);
};

=head2 Daiku Document Workspace instructions

TODO: Integrate Daikufile
Maybe the C<namespace> concept can be used to connect a particular Daikufile to a document workspace.

=head3 $dkr:Daiku::Registry = $m->new_dkr

Setup a namespace for a document/directory/function.

=cut
sub new_dkr {
    my $m = shift;
    mkdir '.lits' unless -d '.lits';
    my $dir = catdir '.lits','.mldk';
    mkdir $dir unless -d $dir;
    new Daiku::Registry;
};
=head3 $dkr:Daiku::Registry = $m->get_dkr($dkr?): check a Daiku::Registry object or create one
=cut
sub get_dkr {
    my ($m,$dkr) = @_;
  NEW: {
      last if $dkr;
      $dkr = $m->new_dkr;
      $m->{svn_add_ok} = {};
    };
    $m->pstop(baddkr => r => ref($dkr)) unless $dkr->can('find_task');
    $dkr;
};
=head3 dkpuc: allow simpler and stricter invocation of methods of the $m+$dkr object
=cut
sub dkpuc {
    my ($m,$prog,$dkr,@args) = shift->ppuc(@_);
    ($m,$prog,$m->get_dkr($dkr),@args);
};

=head $m->dkreg($dkr,$dst,$dto1): register dto1 with destination dst in registry $dkr 

Check whether something is in the way and report.

=cut
sub dkreg {
    my ($m,$dkr,$dst,$dto1,$init) = @_;
    my $dto0 = $dkr->find_task($dst);
    my $err = ref($dto1) ne ref($dto0);
    $m->pstop_or_warn(!$err, 'daiku destination %D is already occupied by a %r with deptree %A', D => $dst, r => ref($dto0), A => $m->dkt_chkdeps($dto0)) if $dto0;
    $dkr->register($dto1);
  INIT: {
      last if !$init;
      $m->pstop('bad init for dst %D', D => $dst) if 'CODE' ne ref $init;
      $m->{dkr}->{$dst}->{init} = $init;
    };
    $m->vecho('destination %D is now registered as a %f at %r', f => $dto1, r => $dkr, D => $dst);
    $dst;
};

=head3 $dst = $m->dkrf($dkr,$dst,$depsr,$desc,$code,%opts): make a file like Daiku->file but in the document workspace

Files that are in the SVN repository are not made but human-edited and therefor do not have dependencies.
An exception can be made for non-editable files $fi which we share via svn, provided that we mark them with
	svn propset svn:keywords format("deps = %s",join(',',@$deps)) $fi
cf 
 file:/phm/18/ont/tdir/phm_pub_sig181005/_lng.phm_pub_sig181005.txt::ekso
 file:/phm/18/ont/tdir/phm_pub_sig181005/_lng.phm_pub_sig181005.txt::svn_get_deps
We are not actually checking the deps value for the time being. Anything non-null will do.
=cut
sub dkrf {
    my ($m,$prog,$dkr,$dst,$depsr,$desc,$code,%opts) = shift->dkpuc('dkrf',@_);
    $m->pstop('empty daiku file destination') unless $dst;
  DEPS0: {
      my $dt = $dkr->find_task($dst);
      last unless $dt;
      my $deps0 = $dt->deps;
      last unless @$deps0;
      $m->pwarn('destination %d already registered with dependencies %D, attempting to reregister with dependencies %E', d => $dst, D => $deps0, E => $depsr);
    };
    undef $depsr if $m->pvar('svn_no_deps') && !$m->svn_status_ekso($dst) && !$m->svn_get_deps($dst); # do not rebuild anything that is already in svn unless it has svn:keywords deps set
    my @args = (dst => $dst);
    my %depp = ($dst => 1);
    my @deps = ();
    my $chk_deps = sub {
	foreach my $dep (@$depsr) {
	    $m->pstop('dkf: non-string prerequisite %d in list %D of target %t', d => $dep, D => $depsr, t => $dst) if ref $dep;
	    do { $m->uwarn('empty prerequisite in list %D of target %t', t => $dst, D => $depsr);next } unless length $dep;
	    do { $m->uwarn('duplicate prerequisite %e in list %D of target %t', e => $dep, t => $dst, D => $depsr );next } if $depp{$dep};
	    push @deps, $dep;
	    $depp{$dep} = 1;
	};
    };
    &$chk_deps;
    my $fun = $code ? sub { 
	my $f = shift;my $dst = $f->dst;
	my $ok = &$code($f,@_);
	$m->uwarn('make function for target %d ended without success', d => $dst) unless $ok;
	$m->ustop('target %d not made', d => $dst) unless -e $dst;
      TSUKAISUTE: {
	  last if !$opts{tsukaisute};
	  my $dt = $dkr->find_task($dst);
	  last if !$dt;
	  @{$dt->deps} = ();
	};
      SVN_ADD: {
	  last if !$opts{svn_add_ok};
	  $m->svn_add($dst);
	  $dst = $m->absfayl($dst);
	  $m->{svn_add_ok}->{$dst} = 1
	};
	$dst;
    } : sub { 
	my $f = shift;my $dst = $f->dst;my @deps = @{$f->deps};
	$m->touch_dst_unless_newest($dst,@deps);
	$dst };
    my $push_args = sub {
	push @args, deps => \@deps;
	$desc ||= "make $dst from @deps";
	push @args, desc => $desc if $desc;
	push @args, code => $fun;
    };
    &$push_args;
    $m->dkreg($dkr,$dst,Daiku::File->new(@args),$opts{init});
    $m->dpop1($prog,$dst);
};

# sub dkf { my ($m,$prog,$dst,$depsr,$desc,$code,%opts) = shift->ppuc('dkf',@_);my $dst = $m->dkrf($m->{daiku}->{set},$dst,$depsr,$desc,$code,%opts);$m->dpop1($prog,$dst) };

=head4 $m->dkrfh($dst,\@depnoms,$desc,$code,%depvars): Allow simpler and stabler expression of how targets are generated from subtargets and parameters

The function supplied in code has more information at its disposal than dkf functions and is free from having to keep track of positions in a @deps sequence.
@depnoms is a list of files or their %depvars indices on which making depends.
%depvars could be
 (tml_txt => $tml_txt, tpag => $tpag, tvars => 't.html/vars.hac', lvars => 'de/vars.hac', lang => 'de', tml => 'html', svn_add_ok => 1)
%depvars can be used to pass subtarget values in random order, while @depnoms specifies the order.
%depvars can also be used to pass non-target parameters and even dkf %opts like svn_add_ok => 1. 
Parameters ($dst,%depvars) are passed to the making function so that rich information is at its disposal and referencable by name, whereas the enforcing of dependencies is by means of @depnoms is entirely Daiku's job.
PROBLEMS: use %depvars requires discipline on the part of the user (programmer), some sanity checks could help.

=cut
sub dkrfh {
    my ($m,$prog,$dkr,$dst,$depnomsr,$desc,$code,%depvars) = shift->dkpuc('dkrfh',@_);
    $dst = $depvars{$dst} || $dst;
    my @deps = map { $depvars{$_} || $_ } $m->ref2list($depnomsr); # qw(tml_txt tpag lvars tvars hacsfyl);
    push @deps, $m->fdeps($dst);
    my $fun = sub {
	my $f = shift;my $dst = $f->dst;
	&$code($dst,%depvars);
    };
    $m->dkrf($dkr,$dst,\@deps,$desc,$fun,%depvars); # %depvars may contain some options such as svn_add_ok
    $m->dpop1($prog,$dst);
};
# sub dkfh { my ($m,$prog,$dst,$depnomsr,$desc,$code,%depvars) = shift->ppuc('dkfh',@_);$dst = $m->dkrfh($m->{daiku}->{set},$dst,$depnomsr,$desc,$code,%depvars);$m->dpop1($prog,$dst) };

sub ls_lt  {
    my %f2t = ();do { local @_ = stat $_;$f2t{$_} = strftime '%Y-%m-%d %H:%M:%S', localtime $_[9] } for @_;
    my @fis = sort { $f2t{$b} cmp $f2t{$a} } keys %f2t;
    foreach my $fi (@fis) { my @stat = stat $fi;print "$f2t{$fi}\t$fi\t$stat[7]\n" };
    @fis;
};
sub mt { @_ = stat shift;$_[9] };
sub insist_dst_newest { # useful function to execute within code passed to dkf
    my $m = shift;
    my ($dst,@deps) = @_;
    my @fis = ls_lt $dst, @deps;
    my $mt1 = mt $dst;
    $m->ustop('failed to make target file %f', f => $dst) unless $mt1;
    foreach my $dep (@deps) {
	my $mt0 = mt $dep;
	$m->ustop('failed to provide source file %f', f => $dep) unless $mt0;
	next unless $mt0 > $mt1;
	$m->ustop('dk target file %f is older than source file %e', e => $dep, f => $dst);
    };
    $dst;
};
sub renew_dst_unless_newest { # useful function to execute within code passed to dkf; e.g. use when effect of deps on dst is unsure
    my ($m,$prog,$dst,$depsr,%opts) = shift->ppuc('renew_dst_unless_newest',@_);
    my @deps = $m->ref2list($depsr);
    my $renew = $opts{renew} || sub { touch $dst };
  TOUCH: {
      my $dep = '';
      my $dst_mt = mt $dst;
    TRY: {
	last if !$dst_mt;
	for (@deps) {
	    my $dep_mt = mt $_;
	    # next unless $dep_mt and $dep_mt > $dst_mt;
	    next if !$dep_mt or $dep_mt < $dst_mt;
	    $dep = $_;
	    last TRY;
	};
	last TOUCH;
      }; # TRY
      $m->vecho('%(SW|creating|renewing) destination %f %(SW|without prerequisite|dependent on prerequisite %e)', SW => $dst_mt, f => $dst, e => $dep);
      &$renew;
    }; # TOUCH
    my ($fi0) = ls_lt $dst, @deps;
    $m->pstop('even after build-or-touch, destination file %f is still older than prerequisite %e', f => $dst, e => $fi0) if $fi0 ne $dst and mt($fi0) > mt($dst);
    $m->dpop1($prog,$dst);
};
sub touch_dst_unless_newest {
    my ($m,$prog,$dst,@deps) = shift->ppuc('put_deps_unless_newest',@_);
    $m->dpop1($prog,$m->renew_dst_unless_newest($dst,\@deps));
};
sub putdeps_dst_unless_newest {
    my ($m,$prog,$dst,@deps) = shift->ppuc('put_deps_unless_newest',@_);
    $m->dpop1($prog,$m->renew_dst_unless_newest($dst,\@deps,renew => sub { $m->fayl_put_lst($dst,@deps) }));
};

=head3 dkt -- make a dask like Daiku->task but in the document workspace
=cut

sub dkrt {
    my ($m,$prog,$dkr,$dst,$deps,$desc,$code,%opts) = shift->dkpuc('dkrt',@_);
    my @args = (dst => $dst);
    push @args, deps => $deps if $deps;
    push @args, desc => $desc if $desc;
    push @args, code => $code if $code; 
    $m->dkreg($dkr,$dst,Daiku::Task->new(@args),$opts{init});
    $m->dpop1($prog,$dst);
};
# sub dkt { my ($m,$prog,$dst,$deps,$desc,$code) = shift->ppuc('dkt',@_);$dst = $m->dkrt($m->{daiku}->{set},$dst,$deps,$desc,$code);$m->dpop1($prog,$dst) };

=head3 dkft -- make a phony task file like Daiku->file + Daiku->task but in the document workspace
=cut
sub mldk_taskf { mkdir '.lits' if !-d '.lits';my $dir = catfile '.lits', '.mldk';mkdir $dir if !-d $dir;catfile $dir, $_[0] };
sub mldk_phony { local $_ = shift;s(\W)(_)g;mldk_taskf $_ };
sub dkrft {
    my ($m,$prog,$dkr,$tsk,$deps,$desc,$code) = shift->dkpuc('dkrft',@_);
    my $dst = mldk_taskf $tsk;
    $deps ||= [];
    $desc ||= $m->hsprintf('dkft phony task %t uses file %f', t => $tsk, f => $dst);
    $m->dkrf($dkr,$dst,$deps,$desc, sub {
	my $f = shift;my $dst = $f->dst;my @deps = @{$f->deps};
	my $s = $code && 'CODE' eq ref $code && &$code($f);
	$s ||= join "\n", 'OK', $tsk, $dst, $desc, @$deps; 
	$m->fayl_put_str($dst, $s);
	$dst } );
    $m->{dkr}->{$tsk}->{ref} = $dst;
  TSK: {
      do { $m->decho('cannot register %t as taiku task because file of same name exists; never mind, it is registered as a mldk task anyway');last } if -e $tsk;
      $m->dkrt($dkr, $tsk => [ $dst ], $m->hsprintf('task %f depends on phony task file %f', t => $tsk, f => $dst));
    };
    $m->dpop1($prog,$dst);
};
# sub dkft { my ($m,$prog,$tsk,$deps,$desc,$code) = shift->ppuc('dkft',@_);$dst = $m->dkrft($m->{daiku}->{set},$tsk,$deps,$desc,$code);$m->dpop1($prog,$dst) };

=head3 $m->dstdeps($dtf,\%depp,$lvl) => $arb: build and debug dependency tree before execution

$dtf :: Daiku::Task or Daiku::File
%depp :: targets that should not reoccur, default {}
$lvl :: hierarchy level, default 0
  
=cut
our $errmark = '!';
sub dstdeps {
    my ($m,$prog,$dtf,$deppr,$lvl) = shift->ppuc('destdeps',@_);$lvl ||= 0;my %depp = $m->ref2hash($deppr);
    $m->pstop('invalid daiku task/file object') unless $dtf->can('registry');
    my $dst = $dtf->dst;
    my $spc = '';$spc .= ' ' for 0..$lvl;
    $m->decho($spc.'[ '.$dst);
    $depp{$dst} = 1;
    my @dd = ($dst);
    my @deps = @{$dtf->deps};
    $lvl++;
    my $err = '';
    for my $dep (@deps) {
	my $err0 = $depp{$dep} ? $errmark : '';
	do { $m->pwarn('circular dependency chain: upper target %D reoccuring below', D => $dep);$err = $errmark } if $err0;
	my $sdt = $err ? '' : $dtf->registry->find_task($dep); # Daiku::File or Daiku::Task
	push @dd, $sdt ? do { my $arb = $m->dstdeps($sdt,\%depp,$lvl);my $errsub = shift @$arb;$err ||= $errsub;$arb } : do { $dep = $err0.$dep;$m->decho($spc.'  '.$dep);$dep };
    };
    $m->decho($spc.']');
    unshift @dd, $err;
    $m->dpop1($prog,\@dd);
};
=head3 dkb -- build a target like Daiku->build but in the document workspace

First build the dependency tree, check against circularity and execute initialization routines registered in $m->{dkr}->{$dst}->{init}

=head4 $dst = dk_init_dst($dst): run constructor $m->{dkr}->{$dst}->{init} if available and return $dst when done
=cut
sub dk_init_dst { # TODO: integrate $m->{dkr}, $m->{svn_add_ok} et al into A2E::DKR object
    my ($m,$dst) = @_;
    my $init = $m->{dkr}->{$dst}->{init};
    return if !$init;
    $m->pstop('dst %D has invalid initializer', D => $dst) if 'CODE' ne ref $init;
    &$init;
    $dst;
};
=head4 @dsts = $m->dk_init_dsts($dst,@deps): run constructors associated with targets of a dependency tree $arb, return destinations for which this was done
=cut
sub dk_init_dsts {
    my ($m,$dst,@deps) = @_;
    my @dsts = ();
    foreach my $dep (@deps) {
	if (ref $dep) {
	    $m->pstop('bad dependency list %D', d => $dep) if 'ARRAY' ne ref $dep;
	    push @dsts, $m->dk_init_dsts(@$dep)
	} else {
	    push @dsts, $m->dk_init_dst($dep);
	};
    };
    push @dsts, $m->dk_init_dst($dst);
    @dsts;
};

=head4 dkt_chkdeps: check dependencies and represent them as a hierarchy tree
=cut
sub dkt_chkdeps {
    my ($m,$dt) = @_;
    my $arb = $m->dstdeps($dt);
    my $err = shift @$arb;
    $m->pstop('dependency tree %A contains circular dependencies marked with %E', E => $errmark, A => $arb) if $err;
    $arb;
};
=head4 dkr_dst2arb: do more checks and obtain a cleaner tree than dkt_chkdeps
=cut
sub dkr_dst2arb {
    my ($m,$dkr,$dst) = @_;
    my $dt = $dkr->find_task($dst);
    $m->ustop('invalid daiku destination %D', D => $dst) if !$dt;
    $m->pstop('destination %D is of type %r rather than Daiku::Task nor Daiku::File', r => $dt, D => $dst) unless $dt->can('deps');
    my $arb = $m->dkt_chkdeps($dt);
    $m->{dkr}->{$dst}->{dt} = $dt;
    $arb;
};
=head4 dk_dst_init_arb: execute all constructors along the dependency tree in the right sequence

To be done with care: do not do this on trees for top-level targets like 'main' that encompass versions for various languages.
 
=cut
sub dkr_dst_init_arb {
    my ($m,$dkr,$dst) = @_;
    my $arb = $m->dkr_dst2arb($dkr,$dst);
    $m->dk_init_dsts(@$arb);
};
=head4 $dst = $m->dkrb($dkr,$dst,%opts): execute a dependency tree
$opts{initp} :: execute constructor, of one level only if value is 1, of whole tree if > 1
TODO: A2E::DKR class with nj field et al
=cut
sub dkrb {
    my ($m,$prog,$dkr,$dst,%opts) = shift->dkpuc('dkrb',@_);my $initp = $opts{initp} || 0;
    $m->pstop('asking daiku to build empty target') if !$dst;
    $dst = $m->{dkr}->{$dst}->{ref} || $dst;
    my $arb = $m->dkr_dst2arb($dkr,$dst);
    $m->decho('running daiku on dependency tree %A', A => $arb);
    ($initp > 1) ? $m->dk_init_dsts(@$arb) : $initp ? $m->dk_init_dst($dst) : $m->decho('executing no constructor en route to %D', D => $dst);
    $m->{dkr}->{$dst}->{nj} = eval { $dkr->build($dst) };
    $m->pstop('Daiku unable to build dependency tree %A: %m', A => $arb, m => $@) if $@;
    $m->vecho(dkbnt => n => $m->{dkr}->{$dst}->{nj}, D => $dst);
    $m->dpop1($prog,$dst);
};

=head1 set_templates

set $m->{template} and $m->{star_template}, used here and in A2E::MLMK::transkopi

=cut
sub set_templates {
    my ($m,$prog,$lihs) = shift->lpuc('set_templates',@_);
    my $tmpl = $m->pvar(tmpl => 1);
    my @tmls = $m->get_tmls;
    $m->setpvar(tmplang => $tmls[0], lihs => $lihs);
    foreach my $tml (@tmls) { $m->set_template($lihs,$tml,$tmpl) };
    $m->dpop1($prog,1);
}

=head3 Function tmplsrcobjfayls

Find the appropriate template file, as specified by fnom2ffol, by seeking it in a series of directories specified in A2E::Tmplkonf(3), from the lowest (most specific) to the highest (most generic) possible template.
Even this may be no longer needed because we will just invoke all members of fnom2ffol to generate the object files from the templates in one run.

=cut

sub tmplsrcobjfayls {
    my ($m,$prog,$lihs,$fnom,%opts) = shift->opuc('tmplsrcobjfayls',1,@_);
    my ($fsrc,$fobj,%fops) = $m->fnom2ffol($fnom,%opts);
    $fsrc = $m->hsprintf($fsrc);
    $m->pstop('source file base name %f still contains unresolved percent prefix expression, something wrong with hsprintf_repl %H?', f => $fsrc, H => $m->{hsprintf_repl}) if $fsrc =~ m(\%);
    $fsrc = $m->tmplsrcfayl($fsrc, %fops);
    $fobj = $m->hsprintf($fobj);
    # $m->{tmplsrcfayls}->{$fnom} = $fsrc;
    # $m->{tmplobjfayls}->{$fnom} = $fobj;
    $m->dpop($prog,$fsrc,$fobj,%fops);
};

=head3 $m->ffol_tmplsrcfayl('dok_%f_tmpl.txt') => '/adv/tmpl/a2e/oas_akt_aammdd/html/dok_html_tmpl.txt'

Return name of template generation source file, such as e.g.

    /adv/tmpl/a2e/oas_adv_aamm/lng_tmpl.txt
    /adv/tmpl/a2e/oas_adv_aamm/lng_tmpl.de.txt

if such is found based on an expression $fnom e.g. posttmplfnom0 or posttmplfnom such as

    %d/lng_tmpl.txt
    %d/lng_tmpl.%l.txt

or others such as

    %ppre.%d.%l.txt

etc.  Return nil if nothing is found

=cut

sub ffol_tmplsrcfayl {
    my ($m,$prog,$ffol,%tups) = shift->ppuc('ffol_tmplsrcfayl',@_);
    my $tmpl = $m->pvar(tmpl => 1);
    $tups{d} ||= $tmpl;
    $tups{p} ||= $m->{konfig}->srcfpfx; 
    my $fib = $m->hsprintf($ffol, %tups);
    my $tmpldir = $m->pvar('tmpldir');
    my $tmplang = $m->pvar('tmplang');
    my $fi = catfile $tmpldir, $tmplang, $fib;
    return $m->dpop1($prog,$fi) if -r $fi;
    $fi = catfile $tmpldir, $fib;
    return $m->dpop1($prog,$fi) if -r $fi;
    $m->pwarn('template %i not found in tmpldir %d with tmplang %l', i => $fib, d => $tmpldir, l => $tmplang);
    my $tmplappdir = $m->pvar(tmplappdir => 1);
    $fi = catfile $tmplappdir, $fib;
    return $m->dpop1($prog,$fi) if -r $fi;
    $m->dpop($prog);
};

=head3 $m->skrib_fnom($fnom,$lihs,$dkr,%opts) => $target

Make a make-rule for Writing out one member of fnom2ffol hash , return the target file.

=cut 

sub skrib_fnom { # 修改中 file:/phm/19/06/02/sig/_lng.phm_pub_sig190602.txt::kdok
    my ($m,$prog,$lihs,$dkr,$fnom,%opts) = shift->lpuc('skrib_fnom',@_);
    $opts{lihs} = $lihs;
    my ($fi,$fo,%fops) = $m->tmplsrcobjfayls($fnom,%opts);
    $fo = catfile $opts{dir}, $fo if $opts{dir};
    my $proc = sub { 
	$m->template_process($lihs,$fi,$fo,%fops);
	$fo };
    $m->dkrf($dkr,$fo,[$fi],'',$proc,svn_add_ok => $opts{svn_add_ok});
    $m->dpop1($prog,$fo);
};

=head1 my $dst = $m->dkmk($dstvar,$deps,%opts): Generate make dependencies for vocabulary files

Make sure vocabulary sources are reread whenever they are renewed.
Make sure the vocabulary is always fully available even if some sources were reread and others weren't.
Generate a vocabulary cache file $dst on the basis of $dstvar and $opts{litstagl} that serves as a target and as a cache.
Make use of %riidvar_opts in %ropts and especially of $ropts{putvars} which allows collecting a separate vocabulary hash which is stored as cache.
Apart from the main target we have subtargets which are phony vocabula/lit files or phony target files. 

To be able to generate a phony vocabula file the source/dep needs in the first 3 lines to contain a unique and definite variable assignment of the form

	tra_demcW5_pre = Time-stamp: <2019-07-30 09:20:49 phm>

or 

	tra_demcW5_pre = file:/adv/tmpl/lang_pre.txt::2

where the assigned value must conform to regex $dstvarval i.e. contain no spaces and begin with at 'Time' or 'file' or 'https?' or 'ftp' followed by a punctuation mark character.
$opts{oldnomp} prevents duplicate target vocabula variables.  By default the namespace is the function but it can be enlarged by supplying an oldnomp hashref from above.

Otherwise a phony target file is generated in the '.mldk' rather than '.lits' subdirectory.

During startup treat values as known only if they are in memory.
If values older than $mt_minim are found on disk, treat these as void during startup.
When reading a second time ($ripetp), do not rewrite .lits but only variant subdirectories ('zh', 'latex').
TODO variantp: sort out what to do with special grup structures defined in *.zh.txt.  These belong into .lits/zh/, contrary to the established rule.  llng (and ttml?) files have a distinct status. file:/adv/perl/A2E/Tmplfil.pm.tmpl::variantp
190823 由 faylz_src_dk 更名

=cut

our %riid_opts = (msglini => 1, insistp => 1);
our $dstvarval = '\A(\w+)\s\=\s(?:file|https?|ftp|Time)\W';
sub dkmk {
    my ($m,$prog,$lihs,$dstvar,$faylzr,%opts) = shift->opuc('dkmk',2,@_);my @faylz = $m->ref2list($faylzr);my $oldnomp = $opts{oldnomp} || {};my @edeps = $m->ref2list($opts{edeps});my @litstagl = $m->ref2list($opts{litstagl});my $remk = $opts{remk} ? 1 : 0;my $oldfaylp = $opts{oldfaylp} || {};my $dkr = $m->get_dkr($opts{dkr});my $tsukaisute = $opts{tsukaisute};my %faylp;my $putvars = {};my %ropts = %riid_opts;$ropts{putvars} = $putvars;$ropts{lihs} = $lihs;($ropts{ripetp},$ropts{litstag},$ropts{litstagarg}) = @litstagl;
    do { $m->pwarn('trying to invoke dkmk with dstvar %d on lihs without litsdir, correcting', d => $dstvar);$lihs->putstr(__nolitsdir__ => 0);$lihs->{vars}->{__nonstorp__} = 0 } if $lihs->{vars}->{__nolitsdir__};
    my $litsdir = $m->litstag_opts_setvars(%ropts);
    $m->pstop('no point in reading without litsdir') if !$litsdir;
    $m->pstop('no point in reading into bad litsdir %D', D => $litsdir) if !-d $litsdir;
    my $lastdst = '';
    my $novnomp = sub {
	my $nom = shift;
	return if !$nom;
	$m->uwarn('landmark variable %n appearing a second time', n => $nom) if $oldnomp->{$nom};
	$nom;
    };
    my $dst = catfile $litsdir, join '.', $dstvar, 'hac';
    my @deps = ();
    my %depp = ($dst => 1);
    my %no_remk = ();
    my $riidfaylx = sub {
	my ($faylx,$fayl) = @_; # $faylx is $fayl with additional parameters appended after a '?' for parsing by riid_tmplvars_fayl
	$m->pstop('no point in reading fayl %f without litsdir write permission', f => $fayl) if $lihs->{vars}->{__nolitsdir__};
	my @fyls = ($faylx);
      DIRBASLL: { # additionally read neutral file if we are a lang file; _lng.*.txt comes before, lang.txt after; neutral files are fixed appendices to lang files which alone are subject to makefile arithmetics
	  my ($vol,$dir,$basl) = splitpath $fayl;
	  my ($bas,$ll) = $basl =~ m(\A(lang|\_lng\.\w+)\.(\w{2}).txt\Z); 
	  last unless $ll;
	  my $fyl = catfile $dir, join '.', $bas, 'txt';
	  if ($bas eq 'lang') {
	      push @fyls, $fyl;
	  } else {
	      unshift @fyls, $fyl;
	  };
	};
	foreach my $fyl (@fyls) { 
	    next if $faylp{$fyl}; # dirbasp
	    $m->riid_tmplvars_fayl($fyl, %ropts);$faylp{$fyl} = 1 };
	1 };
    my $varnomdst = sub {
	my ($varnom,$faylx,$fayl,@subdeps) = @_;
	$oldnomp->{$varnom} = 1;
	my $get_ldir = sub { # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">kludge to address change of target to .lits/l.$ll/, duplicating fib2flng from riid_tmplvars_fayl, assuming that $ll never wanders outside the current $ll
	    return $litsdir if $ropts{litstag} eq 't';
	    my $flng = $m->fib2flng($fayl);
	    $lihs->litstag2litsdir($flng ? 'l' : '');
	}; # </nov>
	my $dep = catfile &$get_ldir, $varnom;
	$m->dkrf($dkr,$dep, \@subdeps, '', sub { 
	    my $f = shift;my $dep = $f->dst;my @deps = @{$f->deps};
	    my $ok = eval { &$riidfaylx($faylx,$fayl) };
	    do { unlink $dep;$m->ustop('failure en route to targets %a and %b in vocabulary %f: %m', a => $dep, b => $dst, f => $fayl, m => $@) } if $@ or !$ok;
	    $m->insist_dst_newest($dep,@deps);
	    $m->fayl_put_lst($dep.'.deps',@deps);
	    $dep } );
	$dep };
    my $phonydst = sub {
	my ($faylx,$fayl,@subdeps) = @_;
	my $dep = mldk_phony $subdeps[0];
	$m->dkrf($dkr,$dep, \@subdeps, '', sub {
	    my $f = shift;my $dep = $f->dst;my @deps = @{$f->deps};
	    &$riidfaylx($faylx,$fayl);
	    $m->putdeps_dst_unless_newest($dep,@deps);
	    $dep } );
	$dep };
  FAYL: foreach my $faylx (@faylz) {
      my ($fayl) = $faylx =~ m(\A(.*)\?);
      next if $fayl and $faylx =~ m(\-\Z); # leave out low-rank files because control variable isnt even read from there; BUG
      $fayl ||= $faylx;
      next unless -f $fayl;
      do { $m->uwarn('ignoring duplicate daiku target file %f en route to %D, ignore', f => $fayl, D => $dst);next } if $oldfaylp->{$fayl};
      $oldfaylp->{$fayl} = 1;
      my @subdeps = eval { $m->lng_fayl_subfaylz($fayl, lihs => $lihs, oldfalp => $oldfaylp) }; # 190826 could elimination of oldfaylp subdeps conflicts with remk?
      $m->ustop('failed to obtain included files from %f: %m', f => $fayl, m => $@) if $@;
      $m->pstop('nothing to read for file %f', f => $fayl) unless @subdeps;
      my ($varnom) = $m->fayl_get_lst($fayl,1,$dstvarval,$novnomp,1,5);
      my $dep = $varnom ? &$varnomdst($varnom,$faylx,$fayl,@subdeps) : &$phonydst($faylx,$fayl,@subdeps);
      unlink $dep if $remk and !$no_remk{$fayl};
      $m->pstop('duplicate prerequisite %e in list %E of dst %d', e => $dep, d => $dst, D => \@faylz ) if $depp{$dep};
      push @deps, $dep;$depp{$dep} = 1;
  };
    @deps = $m->uniqpush(\@deps,\@edeps);
    my $riidtups = sub {
	$m->{grupvars} = {};
	$m->{nom2gvars} = {};
	return if !-r $dst;
      PUTVARS: {
	  last if %$putvars;
	  # my $dt = ('l' eq $opts{litstag}) ? 'd' : 't'; # exclude data if lang, exclude text otherwise
	  # my $v2k = sub { 
	  # my ($stor,$k,$v) = @_;
	  # # return if $lihs->{var2typ}->{$k}->{$dt};
	  # $stor->{$k} = $v;
	  # };
	  $putvars = $m->fayl_get_tups($dst); # v2k => $v2k;
	};
	$m->hashpush(ae => $lihs->{vars}, $putvars);
	1;
    };
    &$riidtups;
    $m->dkrf($dkr,$dst, \@deps,"produce local vocabulary", sub {
	my $f = shift;my $dst = $f->dst;my @deps = @{$f->deps};
      NO_PUTVARS: {
	  last if %$putvars;
	  foreach my $dep (@deps) { next if -e $dep;unlink $dep };
	  $m->pstop('generation of target %D yielded no putvars', D => $dst)
	};
	my $dt = ('l' eq $opts{litstag}) ? 'd' : 't'; # exclude data if lang, exclude text otherwise
	# my $k2v = sub { 
	   # my ($k,$stor) = @_;
	   # return if $lihs->{var2typ}->{$k}->{$dt};
	   # $stor->{$k};
	# };
	my $get_putvars = sub { # kludge; $putvars is buggy, loses newly read values of __fols/__vars in favour of some old value
	    # file:/phm/19/09/16/sig/_log13adv.phm_pub_sig190916.txt::74482
	    my %putvars2 = ();$putvars2{$_} = $lihs->getstr($_) for keys %$putvars;
	    \%putvars2;
	};
	$m->fayl_put_tups($dst,&$get_putvars); # k2v => $k2v
	$m->fayl_put_lst($dst.'.deps',@deps);
	$dst }, init => $riidtups, tsukaisute => $tsukaisute );
    $m->dpop1($prog,$dst);
};
=head4 $dst = $m->dkrmk($dkr, ....): like dkmk but with initial $dkr argument

=cut
sub dkrmk {
    my $m = shift;
    my $dkr = shift;
    $m->dkmk(@_, dkr => $dkr);
};
=head5
=cut
sub lits_dk_trg {
    my ($m,$prog,$lihs) = shift->lpuc('lits_dk_trg',@_);
    my $fyl = $lihs->hacsfyl;
    $m->dpop1($prog,$fyl);
};

=head1 Subroutines of mldk

Make targets @trg with daiku
Replacement for C<make>

=cut

sub fdeps {
    my ($m,$prog,$fo) = shift->ppuc('fdeps',@_);
    my $fd = join '.', $fo, 'deps';
    return $m->dpop($prog) unless -r $fd;
    $m->dpop($prog,$m->fayl_get_lst($fd,1));
};
sub set_lihs_dok_dir_langs {
    my ($m,$prog,%opts) = shift->ppuc('set_lihs_dok_dir_langs',@_);
    my $dok = $opts{dok} || '';
    my $dir = $m->{konfig}->chdir || $opts{dir} || '';
    my $lihs = $opts{lihs};
  DIR: {
    DOK: {
	last if $dir;
	$dok = $m->{konfig}->cdok;
	last unless $dok;
	$m->setpvar(dok => $dok, lihs => $lihs);
	$dir = $m->dokdir($dok);
      };
      last unless $dir;
      $m->chwd($dir);
      $dok ||= $m->dir2dok($dir);
    };
    ($lihs,$dok,$dir) = $m->mlht_set_dokvars(lihs => $lihs, dok => $dok, dir => $dir);
    $lihs->setdir($dir);
    my @langs = $m->plst(langs => 1);
    my $lang0 = $langs[0];
    my $lang = $m->{konfig}->lang;
    if ($lang) { $lang0 = $lang;@langs = ($lang);$m->setplst(langs => \@langs, lihs => $lihs) } else { $lang = $lang0 };
    $m->setpvar(lang => $lang, lihs => $lihs);
    $m->set_litsdir($lihs, l => $lang);
    $m->dpop($prog,$lihs,$dok,$dir,@langs);
};

=head4 ($lihs,$dok,$dir,$dvr) = $m->mldk($trg,lihs=>$lihs,dok=>$dok,dir=>$dir,edeps=>$srcs,dvr=>\%dv)
$trg => (?:lits|main|dok|all)
$srcs :: specify additional read-only sources which must be there as prerequisites
$lihs|$dok|$dir|$dvr :: receive document identifiers as return values or pass them e.g. in recursive document creation in file:/adv/perl/A2E/MLMK.pm.tmpl::setsupvars

=cut
sub mldk {
    my ($m,$prog,$lihs,$trg,%opts) = shift->opuc('mldk',1,@_);my $dok = $opts{dok};my $dir = $opts{dir};my @langs = ();my @trgs = 'ARRAY' eq ref $trg ? @$trg : ($trg);my @edeps = $m->ref2list($opts{edeps});my %dv = $m->ref2hash($opts{dvr});my $remk = $opts{remk} ? 1 : 0;
    my @svndeps = ();
    $m->{svn_add_ok}->{$_} = 1 for map { $m->absfayl($_) } glob $m->hsprintf('%p*s');
    $m->{svn_add_ok}->{$_} = 1 for map { $m->absfayl($_) } glob $m->hsprintf('*file'); # editable and svn-able files of the current directory
    $m->uniqpush(\@edeps, [ keys %{$m->{svn_add_ok}} ]);
    ($lihs,$dok,$dir,@langs) = $m->set_lihs_dok_dir_langs(lihs => $lihs, dok => $dok, dir => $dir);
    $m->pstop('no litsdir to write tlihs object unfit for writing') if !$lihs->{vars}->{__litsdir__} or $lihs->{vars}->{__nolitsdir__};
    $m->setpvar(dok => $dok, lihs => $lihs);
    $m->setplst(langs => \@langs);
    my $lang = $langs[0];
    $m->setpvar(lang => $lang, lihs => $lihs);
    my $lang0 = $lang;
    my $ll = $lang;
    my $pre_txt = $m->fnom2fobj('pre');
    $m->riid_tmplvars_fayl($pre_txt); # $deps[0] => "./_pre.$dok.txt?+", current antefayl0, needs to be read initially no matter what
    my $tmplappdir = $m->pvar(tmplappdir => 1);
    my @lits = ();
    my %lihs_opts = (lihs => $lihs, dir => $dir, dok => $dok);
    $m->riid_tmplvars_init;
    $m->special_set_sems;
    my $svn_add_ok = $m->pvar(svn_add_ok => 1);
    my $dkr = $m->get_dkr($opts{dkr});
    my $oldnomp = { special => 1, tit => 1, lab => 1, sut => 1, des => 1, kws => 1, langs => 1, tmls => 1, tmplang => 1 };
    my $oldfaylp = {};
    my $dkrdst = sub {
	my ($dstvar,$deps,%opts) = @_;
	$opts{lihs} = $lihs;
	$opts{oldnomp} = $oldnomp;
	$opts{remk} = $remk;
	$opts{oldfaylp} = $oldfaylp;
	my $dst = $m->dkrmk($dkr,$dstvar,$deps,%opts); # mldkrmk
	@edeps = () if $opts{edeps};
	$dv{$dstvar} ||= $dst;
	$dv{$opts{dstvar}} = $dst if $opts{dstvar};
	if (-r $dst) { 
	    push @lits, $dst;
	} else { 
	    # <nov dag=200106 id=oak7dbg>原來重複嘗試難謂長久之計也
	    # eval { $m->dkrb($dkr,$dst,initp => 1) };
	    $m->dkrb($dkr,$dst,initp => 1);
	    # $m->dkrb($dkr,$dst,initp => 1) if $@;
	    # </nov>		
	};
	$m->pstop('destination %D still not there', D => $dst) unless -r $dst;
	$dst };
    $lihs->get_tups($_) for qw(vrbs varp varq data);
    $lihs->get_noms($_) for qw(lstp hacp);
    # build or get basis without tmpl-related stuff, suitable for transfer to child doc in file:/adv/perl/A2E/MLMK.pm.tmpl::sup7lihs
    TOKS: { 
	my @deps = $m->pre_faylz($dok, typ_min => 2);
      PRE1: {
	  last if $deps[0] =~ m(\b$pre_txt\b);
	  $m->pstop('pre_faylz is not putting %p but %D in first place', p => $pre_txt, D => $deps[0]); 
	};
	&$dkrdst(toks => \@deps, edeps => \@edeps); # dkr_toks
    };
    my $tmpl = $lihs->getstr('tmpl') || $m->pvar(tmpl => 1);$m->setpvar(tmpl => $tmpl, lihs => $lihs);
    my @tmls = $m->get_tmls($lihs);
    $m->ustop(notmls => p => $pre_txt) unless @tmls;
    my $tmplang = $tmls[0] || $m->pvar('tmplang') || 'html';
    $m->setpvar(tmplang => $tmplang, lihs => $lihs);
    $m->set_templates;
    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
    $dv{vars0} ||= $m->lihs_varsfyl($lihs); # </nov>
    $dv{lvars0} ||= $m->lihs_varsfyl($lihs, l => $lang);
    $dv{tvars0} ||= $m->lihs_varsfyl($lihs, t => $tmplang);
    $dv{hacsfyl} ||= $lihs->hacsfyl;
    my @litsdeps = map { $dv{$_} } qw(vars0 lvars0 tvars0 hacsfyl);
    # foreach my $nom (@lihs_progvars) { my $val = $lihs->getarb($_);$m->setpvar($nom => $val, lihs => $lihs) if defined $val };
    $m->uwarn(funnovar => v => 'pub_find') unless $lihs->{vars}->{pub_find};
    my $push_fnoms = sub {
	my $tmpl_fnoms = $lihs->getstr('tmpl_fnoms') || $m->{konfig}->tmpl_fnoms;
	my @fnoms = $m->textsep2list($tmpl_fnoms);
	foreach my $fnom (@fnoms) { 
	    next if $fnom eq 'nav'; # handled separately in HTML_NAV block below
	    push @edeps, $m->skrib_fnom($lihs,$dkr,$fnom, svn_add_ok => $svn_add_ok);
	};
    };
    &$push_fnoms;
    # my %lpag;
    $m->uniqpush(\@edeps,[ map { $m->absfayl($_) } $m->fayl_fil_lst('_deps') ]);
    my $i = 0;
    $ll = $lang0;
    $m->setpvar(lang => $ll, lihs => $lihs);
    TMPLFYL: {
	my @deps = $m->pre_tmk_faylz(typ_min => 2, tmpl => $tmpl);
	&$dkrdst(tmpl => [ @deps ], edeps => [ @edeps ]);
    };
    my $tml = $tmls[0];
    $i = 0;
    my @vardsts = ();
    $m->lihs_get_vars($lihs, fyls => [$dv{vars0}]) unless %{$lihs->{vars}};
    foreach my $tml (@tmls) {
	$m->setpvar(tmplang => $tml, lihs => $lihs);
	$m->set_litsdir($lihs, t => $tml);
	my @deps = ();
	push @deps, $m->tml_tmk_faylz(typ_min => 2, tmpl => $tmpl, tml => $tml);
	my $tvars = 'tvars_'.$tml;
	%{$lihs->{tvars}} = ();
	&$dkrdst(vars => \@deps, litstagl => [$i, t => $tml], dstvar => $tvars);
	$m->lihs_get_vars($lihs, fyls => ['tvars',$dv{$tvars}]) unless %{$lihs->{tvars}};
	$m->pstop('tvars %v differs from tvars0 %u although tml %t is same', v => $dv{$tvars}, u => $dv{tvars0}, t => $tml) if $tml eq $tmls[0] and not $dv{tvars0} eq $dv{$tvars};
	$dv{tml_txt} = $m->skrib_fnom($lihs,$dkr,'tml', tml => $tml);
	my $j = 0;
	# <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
	# $dv{$tvars} = $m->lihs_varsfyl($lihs,'t',$tml); 
	# $m->dkrf($dkr,$dv{$tvars},\@deps, '', sub {
	# my $f = shift;my $dst = $f->dst;my @deps = @{$f->deps};
	# $m->set_template_riidvars($lihs,$tml);
	# $m->lihs_put_vars($lihs, treg => 'tvars', t => $tml, fyl => $dst);
	# } );
	# </nov>
	foreach $ll (@langs) {
	    $m->setpvar(lang => $ll, lihs => $lihs); # mldk_for_langs
	    my @deps = $m->lng_faylz($ll, $dok, typ_min => 2, tmpl => $tmpl); # file:/phm/20/12/28/sig/_log07adv.phm_pub_sig201228.txt::98311
	    push @deps, $m->fayl_fil_lst('_deps_'.$ll);
	    my $topvar = 'topvar_'.$ll;
	    my $lvars = 'lvars_'.$ll;
	    $dv{$lvars} = $m->lihs_varsfyl($lihs, l => $ll); # made as side effect of topvar, kludgy
	    %{$lihs->{lvars}} = ();
	    &$dkrdst(topvar => \@deps, litstagl => [$i, l => $ll], dstvar => $topvar); # lldkrdst
	    $m->lihs_get_vars($lihs, fyls => ['lvars',$dv{$lvars}]) unless %{$lihs->{lvars}};
	    $m->set_litsdir($lihs, l => $ll);
	    # $m->dk_init_dst($dv{$topvar});
	    $lihs->{vars}->{__mt_minim__} = 0;
	    $m->lihs_put_vars($lihs, treg => 'lvars', l => $ll, fyl => $dv{$lvars}); # dkr_lihs_put_lvars
	    if ($langs[0] eq $ll) {
		$m->lihs_put_vars($lihs, treg => 'vars', fyl => $dv{vars0});
		$m->lihs_arb2var($lihs);
	    };
	    $m->pstop('even lihs_arb2var failed to make %f', f => $dv{hacsfyl}) unless -r $dv{hacsfyl};
	    my $lpag2out = sub {
		my ($dst,%dv) = @_;
		# $lihs->litstag_undef_vars($_) forq qw(l t);
		# <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
		$m->lihs_get_vars($lihs, l => $ll, fyls => ['lvars',$dv{$lvars},'tvars',$dv{$tvars}]); # lpag2out_get_lvars
		# </nov>
		$m->lihs_var2arb if !$lihs->{lsts}->{langs};
		$m->pstop('even lihs_var2arb failed to set list langs') if !$lihs->{lsts}->{langs};
		$m->template_process($lihs,$dv{tml_txt},$dst,l => $ll,f => $tml) or $m->pstop('failed to generate %o from template %i', i => $dv{tml_txt}, o => $dst);
 		# kludge file:/phm/19/05/13/sig/_org.phm_pub_sig190513.txt::tml2outfayl
		my $system_postproc = sub {
		    return $dst if $m->{konfig}->no_postproc;
		    my $postproc = $m->{konfig}->postproc->{$tml};
		    return $dst unless $postproc;
		    $ENV{MLTML} = $tml;$ENV{MLANG} = $lang;
		    my $err = system $postproc;
		    return $dst unless $err;
		    $m->ustop('failed to prostprocess with %p', p => $postproc);
		};
		&$system_postproc;
	    };
	    my $outdep = join '_', 'out', $tml, $ll;
	    $dv{$outdep} = $m->hsprintf($m->{konfig}->tml2outfayl->{$tml}); # html => 'index.de.html', latex => 'olgmV9.de.tex', multemp => '_top-olgmV9.de.txt'
	    push @vardsts, $m->dkrfh($dkr,$outdep => ['tml_txt', $lvars, $tvars, 'topvar', 'hacsfyl'],"generate final output after which external postprocessor may be invoked",$lpag2out,%dv);
	} continue { $j++ };
      HTML_NAV: {
	  my $nonav = $m->pvar('nonav');      
	  last if $nonav;
	  last unless 'html' eq $tml;
	  # file:/adv/perl/A2E/dokfs.konf::/fnom2ffol.nav/
	  # fnom2ffol nav = nav_tmpl.txt %pnav.%d.txt insist:1 mode:tt tag:* oobj:%d.nav.%l.html
	  foreach my $fnom (qw(nav row news)) {
	      my ($bf, $fi_r, %ffol_opts) = $m->fnom2ffol($fnom);
	      my $fnomtmpl = $fnom . '_tmpl';
	      $dv{$fnomtmpl} = $m->tmplsrcfayl($bf); # '/adv/tmpl/a2e/nav_tmpl.txt'
	      $dv{$fnom} = $m->hsprintf($fi_r); # '_nav.phm_pub1811.txt'
	      my @doksfiles = $m->ref2list($m->doksfiles($dok));
	      $m->dkrfh($dkr,$fnom => [$fnomtmpl,@doksfiles],'',sub {
		  my ($dst,%dv) = @_;
		  $m->template_process($lihs,$dv{nav_tmpl},$dst, f => $tml);
		  $dst }, svn_add_ok => $svn_add_ok, %dv);
	      foreach my $ll (@langs) {
		  $m->setpvar(lang => $ll, lihs => $lihs);
		  my $navl = join '_', 'nav',$ll;
		  $dv{$navl} = $m->hsprintf($ffol_opts{oobj}); # 'phm_pub1811.nav.de.html'
		  push @vardsts, $m->dkrfh($dkr,$navl => ['nav','lvars_'.$ll],'',sub {
		      my ($dst,%dv) = @_;
		      $m->pstop('before template process lihs is unfit for writing') if $lihs->{vars}->{__nolitsdir__} or !$lihs->{vars}->{__litsdir__};
		      $m->template_process($lihs,$dv{nav_tmpl},$dst,l=>$ll,f=>$tml);
		      $m->pstop('after template process lihs has become unfit for writing') if $lihs->{vars}->{__nolitsdir__} or !$lihs->{vars}->{__litsdir__};
		      $dst },%dv);
	      };
	  };
	}; # HTML_NAV
    } continue { $i++ }; # tmls
    $m->pstop('lihs object unfit for writing') if $lihs->{vars}->{__nolitsdir__} or !$lihs->{vars}->{__litsdir__};
    $m->setpvar(lang => $lang0, lihs => $lihs);
	$ll = $lang0;
	$tml = $tmls[0];
    $m->dkrt($dkr,svn => [ keys %{$m->{svn_add_ok}} ], 'put external depended-on files into svn repository',sub { 
	my $f = shift;my $dst = $f->dst;my @deps = @{$f->deps};
	system 'svn', 'add', '--parents', glob('_*s'), glob('_???.*.txt'), glob('*file');
	$m->svn_add(\@deps,0); # non recursive add
	$dst } );
    $m->dkrt($dkr,dokdata => \@litsdeps, 'put document metadata into db', sub { 
	my $f = shift;my $dst = $f->dst;my @deps = @{$f->deps};
	my %orig = ();
	my @flags = qw(sqlecho verbose);
	foreach my $flag (@flags) {
	    $orig{$flag} = $m->{konfig}->get($flag);
	    $m->{konfig}->set($flag,1);
	};
        # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
	$m->lihs_get_vars($lihs, fyls => ['vars',$dv{vars0},'lvars',$dv{lvars0},'tvars',$dv{tvars0}]);
	# </nov>
	$m->lihs_var2arb; # __dabarels__
	$m->dokdata2db($lihs, rs => 1);
	foreach my $flag (@flags) {
	    $m->{konfig}->set($flag,$orig{$flag});
	}; 
	$dst } );
    $m->dkrt($dkr, main => [ @vardsts ], 'create target documents');
    my @dok = ('main','dokdata');
    $m->dkrt($dkr, rdokdata => [ @dok ], 'make here and there', sub {
	my $f = shift;my $dst = $f->dst;my @deps = @{$f->deps};
	$m->exek('ssh', 'a2e.de', 'mldk', '-v', 'dok');
	$dst } );
    $m->dkrt($dkr, ssh => [], 'remote targets like ssh[main dokdata]', sub {
	my ($f,@args) = @_;my $dst = $f->dst;my @deps = @{$f->deps};
	$m->dkrb($dkr,@args);
	my $host = $m->{konfig}->{ssh_host} || $m->pvar('ssh_host') || $m->ustop('dont know where to remote-execute, specify with variable ssh_host in mldk.konf or _dokfs');
	$m->exek('ssh', $host, 'mldk', '-v', '--cdok', $m->pvar('dok'), @args);
	$dst } );
    $m->dkrt($dkr, lits => \@litsdeps);
    foreach my $trg (@trgs) {
	$trg = $dv{$trg} || $trg;
	$m->pstop('lihs object unfit for building') if $lihs->{vars}->{__nolitsdir__} or !$lihs->{vars}->{__litsdir__};
	$m->dkrb($dkr,$trg);
    };
    $lihs->{vars}->{__mt_minim__} = 0;
    $dir = $m->dir2dir;
    my $cwd = $m->pvar(cwd => 0) || $m->dir2dir;
    $m->vecho('mldk ends in directory %D and cwd %C', D => $dir, C => $cwd);
    $m->dpop($prog,$lihs,$dok,$dir,\%dv);
};
1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/MLDK.pm.tmpl"
# End:
