# -*- coding: utf-8; mode: perl; srcfile: /adv/perl/A2E/MLHT.pm.tmpl -*-

=head1 NAME

A2E::MLHT

=cut

package A2E::MLHT;

=head1 DESCRIPTION

Generate files in several languages from one multilingual template, using
=over
=item static variable lists generated by the (dynamic) means of the maketext syntax described in A2E::Tmplfil and A2E::SArb::Make
=item reactivate dynamic variables and fill the with values using the Textbau filter implemented in A2E::Template::Plugin::Textbau
=back

=head1 PREREQUISITES

    A2E::Dokfs(3)
    A2E::Valid(3)
    SDBM_File

et al

=cut

BEGIN { unshift @INC, $ENV{A2E_MLHT_VOK_LIBDIR} || '/sig/oas/19' }; # file:/sig/oas/19/A2E/MLHT/Vok.pm
use base 'A2E::Dokfs', 'A2E::Valid', 'A2E::Texchars', 'A2E::Hannum', 'A2E::MLHT::Vok'; # A2E::Tmplfil via A2E::Valid
use strict;use warnings 'all';no warnings 'uninitialized';
use IO::File;
use File::Spec::Functions;
use File::Basename;
use Fcntl;
our $old_sdbm_p = 0; # avoid deprecated SDBM_File stuff
# use SDBM_File
# use DBM_Filter;
use Data::Dumper;
use utf8;
our $novitas = $ENV{NOVITAS} || 0;
our %MSGS = (
    sdbmko => 'cannot store value %v of length %l at key %k in sdbm vars %V: %m',
    nokopilits => 'failed to transfer empty variable %n to %o',
    progargsnovar => 'var %v not set at beginning of %P',
    fldlitnul => 'field %f i.e. lit %l has no content',
    nodnojoin => 'failed to join node %n at level %l number %i with joiner %j: %m',
    kvnondbm => 'dbm-incompatible value (ref) found for key %k',
    klsplitff => 'value of variable %k has length of %l bytes, maximum being %m we must split it into sublines %k0_ ff: %v',
    kviparts => 'value of string variable %k was split into %l parts: %P',
    nulstem => 'stem %s at level %l number %n has neither content nor subnodes',
    blokfail => 'failed to process block %b with grupvars %g: %m',
    template_ko => 'invalid template object',
    fi_insist => 'source file %f does not exist',
    tmplvoid => 'template %i generated empty result %o',
    pvarlstp => 'program variable %n should refer to a list',
    pvarhacp => 'program variable %n should refer to a hash',
    nodok => 'failed to identify document',
    nodoktyp => 'failed to identify document type for dok %d',
    nodokset => 'cant set properties for unknown document',
    tnodok => 'tmplvars dok not known yet, setting to %d',
    notsrcf => 'unable to find a source file for template %t',
    norft => 'required template file source %t does not exist',
    nordt => 'required template directory tmpldir is not known',
    nonomfyl => 'found no vocabule file %f', 
    dirvarnodir => 'no directory found at %v %d',
    dokpnodir => 'need to restore dir for dok %d',
    );

=head3 $arg = $m->valid_\w+($lihs,$arg,$vahr): validators
for use with vlpuc, opuc, vopuc
=cut

sub get_tml {
    my ($m,$prog,$lihs) = shift->lpuc('get_tml',@_);
    my $tml = $lihs->{vars}->{__tml__} || $lihs->{vars}->{tml};
  TML: {
      last if $tml;
      $tml = $m->{konfig}->tmplang;
      $m->setpvar(tmplang => $tml);
    };
    $m->pwarn(funnovar => v => 'tml') if !$tml;
    $m->dpop1($prog,$tml);
};
sub valid_tml {
    my ($m,$lihs,$tml,$vahr) = @_;
  NOTML: {
      last if $tml;
      return '' if $vahr->{__flagp__}->{let};
      $tml = $m->get_tml($lihs) if $vahr->{__flagp__}->{get};
    }
    $m->ustop('bad tml %T', T => $tml) if !$m->{tmlp}->{$tml};
    $tml };

=head1 OPTIONS

=cut

sub mlht_defvars {
    my ($m,$prog,$lihs) = shift->qlpuc('mlht_defvars',@_) or return;
    $m->{tmlp} = { html => 1, latex => 1 }; # primitive version, more elaborate see file:/adv/perl/A2E/Dokfs.pm.tmpl::set_langs

=over

=item --sdbm_lenmax 768
limit on length of records in DBM files.
SDBM_File has a limit of 1008 according to perldoc.
We try to overcome the limit by encoding and decoding.
We set 768 as a limit for the value field so as to be sure to keep clear.
960 didnt seem to suffice.
=cut    
    $m->cache_define('sdbm_lenmax', 's', '256');
=item --sdbm_linbrk: break long lines of length beyond --sdbm_lenmax, so that text containing them can be handled
=cut
    $m->cache_define('sdbm_linbrk', '!');
    # $m->cache_define('dbm_overwrites_txt', 's');
=item --doktyp

whether we are working on a 'dok' or a 'nav' document.
deprecated, for internal use only.

=cut

    $m->prog_define('doktyp', 's', 'dok');

=item --tmplkonf

Configuration parameters for the templates used for creation of makefile, navigation file, main document etc.

=cut

    $m->cache_define('tmplkonf', 's%');
    # file:/phm/18/10/14/sig/_lng.phm_pub_sig181014.txt::rm_dbmopts2fibr
    $m->texchars_defvars;
    $m->tmplfil_defvars($lihs);
    $m->konfayl('mlht.konf');
    $m->{doklvars} = {};
    $m->prog_msgdefs('mlht',%MSGS);
    $m->{var2validp}->{tml} = \&valid_tml;
    $m->qpop1($prog,1);
}

=head1 FILES

general configuration

    /etc/opt/a2e/mlht.konf

and site-specific configuration, loaded depending on --daba value

    /etc/opt/a2e/mlht_a2e.konf
    /etc/opt/a2e/mlht_ffii.konf

as well as user-specific additions

    ~/mlht.konf
    ~/mlht_ffii.konf
    ~/mlht_a2e.konf

and document specific configuration files

    ./mlht.konf
    @mlht

=head1 BUGS, CAVEATS

=head1 TODO

=head1 SEE ALSO

    A2E::Template::Plugin::Textbau
    tmplfil(1)

=head1 IMPLEMENTATION

=cut

=head2 Header

=cut

=head2 Overwrite Functions

=head3 Function defvars

=cut

sub defvars {
    my ($m,$prog,$lihs) = shift->qlpuc('A2E::MLHT::defvars',@_) or return;
    $m->SUPER::defvars($lihs);
    $m->dokfs_defvars($lihs);
    $m->mlht_defvars($lihs);
    $m->hannum_defvars;
    $m->qpop1($prog,1);
};

=head3 Function doks2dokurllabs

Lightweight version of doks2reks, possibly now obsolete.

=cut

sub doks2dokurllabs {
    my $prog = 'doks2dokurllabs';my $m = shift;$m->dpuc($prog, @_);
    my @reks = map { my $dok = $_;my ($url, $lab) = $m->dok2urllabs($dok);{ dok => $dok, url => $url, lab => $lab } } @_;
    $m->dpop($prog,@reks);
}

=head3 Function doks2reks

Return a list of hashrefs, each having fields dok, url and lab, which contain the info needed for a navigation menu.
Subroutine of vmethods predokreks, postdokreks, subdokreks

=cut

sub doks2reks {
    my $prog = 'doks2reks';my $m = shift;$m->dpuc($prog,@_);
    my @reks = grep {$_} map { eval { $m->dok2data($_) } } @_; # dont stop just because a document info in some obscure menu branch is broken
    $m->dpop($prog,@reks);
}

=head3 Function dok2text

Get value of text chunk nom from a different dok and lang

=cut

sub dok2text {
    my ($m,$prog,$dok,$nom,$lihs,@langs) = shift->ppuc('dok2text',@_);$lihs = $m->get_lihs($lihs);
    my $var = join '_', $dok, $nom; 
    my $text = $lihs->getstr($var);
    do { $m->dpop1($prog,$text);return $text } if $text;
    my $url = $m->db_dok2absurl($dok);
    do { $m->uwarn('unknown document %d', d => $dok);$m->dpop($prog);return } unless $url;
    my $dir = $m->dokdir($dok);
    my $data = {};
    @langs = $m->doklangs($dok, @langs);
    foreach my $lang (@langs) {
	$data = $m->dok2data($dok, $lang, dir => $dir, url => $url);
	last if $data->{$nom};
    };
    $text = $data->{$nom};
    $m->dpop1($prog,$text);return $text;
}
=head3 $m->nomlangtext($nom,$lang) 

Get value of text chunk $nom from a different $lang, supposing that it is found in ".lits/l.$lang/$nom" or ".lits/$nom" or ultimately $topdir/l.$lang/$nom
Useful in templates cf file:/adv/tmpl/a2e/oas_ont_aa/lng_tmpl.txt::nomlangtext or file:/adv/tmpl/a2e/pub_ont_aa/lng_tmpl.txt e.g. for multilingual month variables mm01 etc to be used there even when values are not yet available or not available in the current document.

=cut
sub langnomtext {
    my ($m,$prog,$lihs,$lang,$nom) = shift->lpuc('langnomtext',@_);
    my $litsdir = $lihs->{vars}->{__litsdir__};
    my @litsdirs = ($litsdir);
    my $dok = $lihs->getval('dok');
    my $fol = $lihs->getval('fol');
    my $dir = '';
    my $ldir = $m->dirtagarg2llitsdir($litsdir, 'l', $lang);
    my $fyl = catfile $ldir, $nom;
  NOMLANGFYL: { # obtain the right file, up to sth like $topdir/.lits/l.$lang/$nom, or fail
      last if -r $fyl;
      $m->uwarn(nonomfyl=> f => $fyl);
      my @predoks = $m->predoks($dok, typ_min => 2, nonrekurs => 1);
      $dok = pop @predoks;
      my $dokdirp = sub {
	  $dir = $m->{dokprop}->{$dok}->{dir};
	  unless ($dir) { $m->pwarn(dokpnodir => d => $dok);$dir = $m->dokdir_pur($dok) };
	  $m->pstop(doknodir => $dok) unless $dir;
	  $m->ustop(doknedir => d => $dir, k => $dok) unless -d $dir;
	  $litsdir = catdir $dir, '.lits';
	  -d $litsdir;
      };
    SUPDOK: { # parent is special case, look up $ldir/${fol}_${nom} file:/phm/21/04/20/sig/_lng.phm_pub_sig210420.txt::dokdirp
	last unless &$dokdirp($dok);
	push @litsdirs, $litsdir;
	my $ldir = $m->dirtagarg2llitsdir($litsdir, 'l', $lang);
	$m->ustop(doknedir => n => $ldir) unless -d $ldir; # only parent/supdok always has $lang version
	$fyl = catfile $ldir, join '_', $fol, $nom;
	last NOMLANGFYL if -r $fyl;
	$fyl = catfile $ldir, $nom;      
	last NOMLANGFYL if -r $fyl;
      };
      foreach $dok (reverse @predoks) { # all nodes from grandparent to topdir
	  next unless -d &$dokdirp($dok);
	  push @litsdirs, $litsdir;
	  my $ldir = $m->dirtagarg2llitsdir($litsdir, 'l', $lang);
	  next unless -d $ldir;
	  $fyl = catfile $ldir, $nom;
	  last NOMLANGFYL if -r $fyl;
      };
      foreach $litsdir (@litsdirs) { # nonlingual file from current to top dir
	  $fyl = catfile $litsdir, $fyl;
	  last NOMLANGFYL if -r $fyl;
      };
      $m->ustop('unable to find any value for %n in lit hierarchy', n => $nom);
    };
    my $text = $m->fayl_get_str($fyl);
    # $text //= $lihs->getstr($nom);
    $m->dpop1($prog,$text);
};
sub nomlangtext {
    my ($m,$prog,$lihs,$nom,$lang) = shift->lpuc('nomlangtext',@_);
    my $text = $m->langnomtext($lihs,$lang,$nom);
    $m->dpop1($prog,$text);
};

sub spanlst { # file:/adv/tmpl/a2e/latex/tabel_tmpl.txt::spanlst
    # &_(['a','b','c','d','e'], 'b','d') => ['b','c','e']
    # 有待與[[skiplst]]集成
    my $prog = 'spanlst';my $m = shift;$m->dpuc($prog,@_);
    my ($lstref,$first,$last) = @_;
    local @_ = @$lstref;
    my $next = '';
    until ($next eq $first) {
	return unless @_;
	$next = shift @_;
    };
    my @lst = ($next);
    until ($next eq $last) {
	return unless @_;
	$next = shift @_;
	push @lst, $next;
    };
    $m->dpop1($prog,\@lst);
};
sub spanlen { # file:/adv/tmpl/a2e/latex/tabel_tmpl.txt::spanlen
    # &_(['a','b','c','d','e'], 'b','d') => 3
    my $prog = 'spanlen';my $m = shift;$m->dpuc($prog,@_);
    my @lst = $m->ref2list($m->spanlst(@_));
    my $len = $#lst + 1;
    $m->dpop1($prog,$len);
};
sub skiplst { # file:/adv/tmpl/a2e/latex/tabel_tmpl.txt::skiplst
    # &_(['a','b','c','d','e'], 'a','e') => ['b','c','d']
    # 有待與[[spanlst]]集成
    my $prog = 'skiplst';my $m = shift;$m->dpuc($prog,@_);
    my ($lstref,$prae,$post) = @_;
    local @_ = @$lstref;
    my $next = '';
    my $iprae = -1;
    my $ipost = $#_ + 1;
    my $i = $iprae;
  PRAE: {
      last unless $prae;
      while ($i++ < $ipost) {
	  next unless $_[$i] eq $prae;
	  $iprae = $i;
	  last PRAE;
      };
      $m->ustop('no such preceding element %e in sequence %c?', e => $prae, c => $lstref);
    };
  POST: {
      last unless $post;
      while ($i++ < $ipost) {
	  next unless $_[$i] eq $post;
	  $ipost = $i;
	  last POST;
      };
      $m->ustop('no such postceding element %e in sequence (e.g. column-specifying leaves) %c', e => $post, c => $lstref);
    };
    my @lst = @_[$iprae+1..$ipost-1];
    $m->dpop1($prog,\@lst);
};
sub skiplen { # file:/adv/tmpl/a2e/latex/tabel_tmpl.txt::skiplen
    # &_(['a','b','c','d','e'], 'a','e') => 3
    my $prog = 'skiplen';my $m = shift;$m->dpuc($prog,@_);
    my @lst = $m->ref2list($m->skiplst(@_));
    my $len = $#lst + 1;
    $m->dpop1($prog,$len);
};
sub nth { # file:/sig/oas/19/04/jkinW4/tra/jkinW4_latex_tmpl.txt::nth
    my $lstref = shift;
    my $i = shift;
    $i--;
    $lstref->[$i];
};

sub lst2hac {
    my ($m,$prog,$nom,$lihs) = shift->ppuc('lst2hac',@_);
    $lihs = $m->get_lihs(shift);
    @_ = $lihs->getlst($nom,$lihs);
    $lihs->{hacs}->{$nom}  = {};
    while (my ($k,$v) = (shift, shift)) { $lihs->{hacs}->{$nom}->{$k} = $v };
    $m->dpop1($prog,1);
};

sub procjung { # use in file:/sig/oas/19/05/demcW5/tra/absv.txt::pl_proc, connect with special |hacputq|sfx2fun|%sfx2fun|proc|+procjung+|
    my ($m,$prog,$lihs,$jungils,$varsref,@litvals) = shift->lpuc('procjung',@_);
    $m->pstop('invoked procjung without a joiner') unless $jungils;
    $m->pstop('invoked procjung without anything to join') unless @litvals;
    my @subjuns = $m->ref2list($jungils);
    my $jungil = shift @subjuns;
    local $_;
    do { $_ = join ' ', @litvals;$m->dpop1($prog,$_);return $_ } unless $jungil;
    my %vars = $m->ref2hash($varsref);
    $m->pstop('invoked procjung without any variables') unless %vars;
    my $litlvl = $vars{litlvl} // -1;
    my $indlvl = $vars{indlvl} // 0;
    my $litnum = $vars{litnum} || 0;
    $litlvl++;
    $vars{litlvl} = $litlvl;
    $vars{indlvl} = $indlvl;
  SUBJUNS: {
      last unless @subjuns;
      @litvals = map {
	  my @subvals = m(\A(\W).*\1\Z) ? $m->text2list($_) : ($_);
	  $m->procjung($lihs,\@subjuns,\%vars,@subvals);
      } @litvals;
    };
    my @jungil_seq = ref $jungil ? @$jungil : $m->text2list($jungil);
    my $jungil_typ = shift @jungil_seq;
    my $jungil_nom = shift @jungil_seq;
    while (@jungil_seq) { # extra args such as => ('jungil', 'alineas')
	my $k = shift @jungil_seq;
	my $v = shift @jungil_seq;
	$vars{$k} = $v;
    };
    $vars{litvals} = \@litvals;
    my $tml = $vars{tml} || $m->pvar(tml => 1);
    local $_ = $m->jungblok($lihs, $jungil_nom,\%vars,$tml,$jungil_typ);
    $m->dpop1($prog,$_);
};

=head3 Function defin_mlht_renfuns

Define some extensions of the mlht maketext (A2E::SArb::Make) language to be used in text chunk files, such as C<tit> in

    text1 := The title of oas_varb0904 is $(tit:oas_varb0904)
    (
    (_kolors @ol
    yellow
    green
    blue
    )
    The colors are $(lit|kolors)
    )

=cut

sub ksum2sizs {
    my $prog = 'ksum2sizs';my $m = shift;$m->dpuc($prog,@_);
    my ($kols,$sum,%rest) = @_;
    my %sizs = ();
    my $nkols = scalar @$kols;
    my ($k,$v) = ('','');
    foreach $k (@$kols) { 
	$v = $rest{$k};
	next unless $v;
	$sizs{$k} = $v;
	$sum -= $v;
	$nkols--;
    };
    do { $m->dpop($prog,%sizs);return %sizs } unless $nkols > 0;
    $v = int $sum / $nkols;
    $m->pstop('size of remainig columns is %i', i => $sum) unless $sum > 0;
    foreach $k (@$kols) {
	next if $sizs{$k};
	$sizs{$k} = $v;
    };
    $sizs{'_keys_'} = $kols;
    $m->dpop($prog,%sizs);return %sizs;
};

=head4 $m->defin_mlht_renfuns($lihs)

=cut
sub defin_mlht_renfuns {
    my ($m,$prog,$suplihs) = shift->qlpuc('defin_mlht_renfuns',@_) or return;
    my $c = $m->template_context;
    my $url_fun = sub { 
	my $sarb = shift;my $vars = shift;
	my ($dok,$lang,@junk) = map { $sarb->render($vars, $_) } @_;
	$m->warn('url function has invalid trailing arguments %J', \@junk) if @junk;
	$m->ustop('url function has no document argument') if !$dok;
	my $url = $m->dok2lokurl($dok, lang => $lang);
	$url };
    $c->{sarb}->add_renfun(tit => sub { my $sarb = shift;$m->dpuc('tit', @_);my $vars = shift;my ($dok, @args) = map { $sarb->render($vars, $_) } @_;my $body = $m->dok2text($dok, 'tit', $sarb->{lihs}, @args);$m->dpop1('tit', $body);$body });
    $c->{sarb}->add_renfun(lab => sub { my $sarb = shift;$m->dpuc('lab', @_);my $vars = shift;my ($dok, @args) = map { $sarb->render($vars, $_) } @_;my $body = $m->dok2text($dok, 'lab', $sarb->{lihs}, @args);$m->dpop1('lab', $body);$body });
    $c->{sarb}->add_renfun(sut => sub { my $sarb = shift;
					my $vars = shift;my ($dok, @args) = map { $sarb->render($vars, $_) } @_;my $body = $m->dok2text($dok, 'sut', $sarb->{lihs}, @args);
					return $body });
    $c->{sarb}->add_renfun(des => sub { my $sarb = shift;$m->dpuc('des', @_);my $vars = shift;my ($dok, @args) = map { $sarb->render($vars, $_) } @_;my $body = $m->dok2text($dok, 'des', $sarb->{lihs}, @args);$m->dpop1('des', $body);$body });
    $c->{sarb}->add_renfun(url => $url_fun);
    # draft adv1801.renlit, sub litscat_simpl
    $c->{sarb}->add_renfun(lit => sub { my $sarb = shift;$m->dpuc('lit', @_);my $vars = shift;my ($litstem, @junk) = map { $sarb->render($vars, $_) } @_;$m->ustop('url function has invalid trailing arguments %J', \@junk) if @junk;my $text = $m->litscat_simpl($litstem);$m->dpop1('lit', $text);$text });
    $c->{sarb}->add_renfun(vok => sub { my $sarb = shift;my %vars = $m->ref2hash(shift);local @_ = map { $sarb->render(\%vars, $_) } @_;local $_ = $m->vok($sarb->{lihs},\%vars, @_);return $_ });
    $c->{sarb}->add_renfun(lv => sub { my $sarb = shift;my %vars = $m->ref2hash(shift);my $lang = $m->pvar(lang =>1);local @_ = map { $sarb->render(\%vars, $_) } @_;
				       local $_ = $m->lvals2val($sarb->{lihs},$lang,\@_,$vars{slngs},$vars{_sup});
				       return $_ });
    $c->{sarb}->add_renfun(ksum2sizs => sub { my $sarb = shift;my %vars = $m->ref2hash(shift);
					      local $_ = $sarb->render(\%vars, shift);
					      my $kols = m(\A\w+\Z) ? $vars{$_} : $m->textsep2listref($_);
					      my $sum = $sarb->render(\%vars,shift) || 100;
					      my @rest = map { $sarb->render(\%vars, $_) } @_;
					      my %sizs = $m->ksum2sizs($kols,$sum,@rest);
					      $_ = $m->arb2str(\%sizs);
					      return $_ } );
    $c->{sarb}->add_renfun(procjung => sub { 
	my $sarb = shift;
	my %vars = $m->ref2hash(shift);
	my $nom = $sarb->render(\%vars, shift); # 'pc_prc'
	local $_; # file:/adv/perl/A2E/Tmplfil.pm.tmpl::globsfxvars
	$_ = $vars{$nom} || $sarb->{lihs}->getstr($nom) or $m->pstop(funnovar => v => 'jungils');
	my $jungils = ref $_ ? $_ : $m->str2arb($_,2);
	my @litvals = $sarb->render(\%vars, @_);
	local $_ = $m->procjung($sarb->{lihs}, $jungils,\%vars,@litvals);
	return $_ } );
    $m->qpop1($prog,1);
};

=head3 Function defin_mlht_tmplfuns

See also &A2E::Template::Context::defin_tmplfuns
More generic methods/filters/blocks that are independent of the A2E::Dokfs application layer should go there.

=head4 define_vmethod

Make define_vmethod from Template::Context readily available. 

=cut

sub define_vmethod { return shift->{cache}->{tmplopts}->{CONTEXT}->define_vmethod(@_) };

=head4 define_filter

Make define_filter from Template::Context readily available.
Make define_filter from there available.
arg0 is context, arg1 filter name, arg2 function body, arg3 $is_dynamic.
$is_dynamic must be 1 it seems.  At least even a simple filter like 'u2tex' does not work without it; however 'dok2text' has $is_dynamic nil, maybe just untested?

=cut
sub define_filter { return shift->{cache}->{tmplopts}->{CONTEXT}->define_filter(@_) };
sub defin_miniverb {
    my ($m,$prog,$mini,$verb,$lihs) = shift->ppuc('defin_miniverb',@_);$lihs = $m->get_lihs($lihs);
    my $c = $m->template_context;
    return $m->dpop($prog) unless $mini and $verb;
    $m->{sarb}->miniverb_put($mini,$verb,$lihs);
    $c->define_vmethod(SCALAR => $mini => sub { return $m->{sarb}->nom_args_call($verb, undef, @_) });
    $m->dpop1($prog,1);
};
our $litvaldir = '';
sub setlitvaldir {
    my ($m,$prog) = shift->ppuc('setlitvaldir',@_);
    my $tml = $m->pvar(tml => 1);
    my $lang = $m->pvar(lang => 1);
    my @dirs = ();my $dir = '';
    push @dirs, '.lits';$dir = catdir @dirs;mkdir $dir unless -d $dir;
    push @dirs, join '.', $lang, $tml;$dir = catdir @dirs;mkdir $dir unless -d $dir;
    $m->dpop1($prog,$dir);
};
sub fayl_get_str_expand {
    my ($m,$prog,$s) = shift->ppuc('fayl_get_str_expand',@_);
    local $_ = $s;
    s(\$[(](\w+)[)])($m->fayl_get_str(catfile($litvaldir,$1)))eg;
    $m->dpop1($prog,$_);
};
sub urlfile {
    my ($m,$prog,$s) = shift->ppuc('urlfile',@_);
    local $_ = $s;
    $_ = m(\A\s*(\S+)\s*\Z) ? $1 : $m->ustop('urlfile %u seems to contain blank characters', u => $_);
    $s = m(\Afile\:(.*)\Z) ? $1 : $_;
    $m->ustop('no such urlfile %f', f => $s) unless -e $s;
    $m->dpop1($prog,$s);
};
sub defin_mlht_tmplfuns {
    my ($m,$prog,$lihs) = shift->qlpuc('defin_mlht_tmplfuns',@_) or return;
    my $gvars = $m->{grupvars};
    my $c = $m->template_context;
    $c->define_vmethod(SCALAR => cgistr => sub { local $_ = shift;s(\:)(%3A)g;s(\/)(%2F)g;$_ });
    $c->define_vmethod(SCALAR => urlfile => sub { $m->urlfile(@_) });
    $c->define_vmethod(SCALAR => dokrek => sub { my $dok = shift;my $data = $m->dok2data($dok) || { };$data });
    $c->define_vmethod(SCALAR => tpaf => sub { my $bas = shift;my ($paf) = $m->tmplvars_fayl($bas);$paf });
    $c->define_vmethod(SCALAR => lv => sub { my $text = shift;my $lang = $m->pvar(lang => 1);return $text unless $text =~ m(\A\|.*\|\s*\Z);my ($sep,@vals) = $m->textsep2seplist($text,'|', noblanksepp => 1);return $text unless $sep;return $text unless $#vals > 0;my $lihs => $m->get_lihs;$m->lvals2val($lihs,$lang,\@vals) });
    $c->define_vmethod(SCALAR => tmplpath => sub { $m->{template}->service->context->config->{INCLUDE_PATH} });
    $c->define_vmethod(LIST => dokreks => sub { my $list = shift;return [ map { $m->dok2data(@_) } @$list ] });
    $c->define_vmethod(LIST => spanlst => sub { $m->spanlst(@_) });
    $c->define_vmethod(LIST => spanlen => sub { $m->spanlen(@_) });
    $c->define_vmethod(LIST => skiplst => sub { $m->skiplst(@_) });
    $c->define_vmethod(LIST => skiplen => sub { $m->skiplen(@_) });
    $c->define_vmethod(LIST => nth => sub { nth(@_) });
    $c->define_vmethod(LIST => safelist => sub { local $_ = shift; # file:/phm/19/07/22/sig/_lng.phm_pub_sig190722.txt::safelist
						 my @l = 'ARRAY' eq ref $_ ? @$_ : ($_);
						 push @l, @_;return \@l } );
    $c->define_vmethod(LIST => mapjoin => sub { my $list = shift;my $vrb = shift;my $sep = shift;join $sep, map { $m->{sarb}->nom_args_call($vrb, undef, $_) } $list });
    $c->define_vmethod(SCALAR => predokreks => sub { my $dok = shift;my @doks = $m->predoks($dok);[ $m->doks2reks(@doks) ] });
    $c->define_vmethod(SCALAR => litfols => sub { my ($lit) = @_;[ $m->litfols(undef,$lit) ] }); # file:/phm/18/09/sig/_org.phm_pub_sig1809.txt::var_put_fols
    $c->define_vmethod(SCALAR => litvars => sub { my ($lit) = @_;[ $m->litvars(undef,$lit) ] });
    $c->define_vmethod(SCALAR => predoks => sub { my ($dok) = @_;[ $m->predoks($dok) ] });
    $c->define_vmethod(SCALAR => predokreks => sub { my ($dok) = @_;my @doks = $m->predoks($dok);[ $m->doks2reks(@doks) ] });
    $c->define_vmethod(SCALAR => postdoks => sub { my ($dok) = @_;[ $m->postdoks($dok) ] });
    $c->define_vmethod(SCALAR => postdokreks => sub { my ($dok) = @_;my @doks = $m->postdoks($dok);[ $m->doks2reks(@doks) ] });
    $c->define_vmethod(SCALAR => subdoks => sub { my ($dok) = @_;[ $m->menu_subdoks($dok) ] });
    $c->define_vmethod(SCALAR => subdokreks => sub { my ($dok) = @_;my @doks = $m->menu_subdoks($dok);[ $m->doks2reks(@doks) ] });
    $c->define_vmethod(SCALAR => topdokreks => sub { my ($dok) = @_;[ $m->get_tuple_records(qq{select dok, dokurl, lastmod from topdok join mlhtdok on ( topdok.topdok = mlhtdok.dok ) where inddok = '$dok'}) ] });
=head4 sqltups

Works as follows

  [sql = 'select speznum, spez from spez where speznum > 169999 order by speznum desc']
  [FOREACH tup = sql.sqltups]
  [tup.speznum] :: [tup.spez]
  [END]

and has been tested in this capacity.

=cut
    $c->define_vmethod(SCALAR => sqltups => sub { my $sql = shift;[ $m->get_tuple_records($sql) ] });
=head4 sqlreks

Same but returns reference to array rather than hash
Used to build tables with BLOCK sqlreks, whereas BLOCK sqltups has to be defined individually
  <% BLOCK sqltups -%>
  || <% sepjoin(' || ') %> ||
  <% FOREACH t = sql.sqltups -%>
  | <% t.speznum %> | <% t.spez %> | <% t.spezmon %> | <% t.spzfdat %> |
  <% END -%>
  <% END %><%# sqltups -%>
  <% BLOCK sqlreks -%>
  || <% sepjoin(' || ') %> ||
  <% FOREACH r = sql.sqlreks -%>
  | <% r.join(' | ') %> |
  <% END -%>
  <% END %><%# sqlreks -%>


=cut

    $c->define_vmethod(SCALAR => sqlreks => sub { my $sql = shift;[ $m->get_records($sql) ] });
    $c->define_vmethod(SCALAR => url => sub { my ($dok,$lang) = @_;$m->dok2lokurl($dok, lang => $lang) });
    $c->define_filter('url', sub { $m->dok2text(shift, 'url') });
    $c->define_vmethod('SCALAR', 'lab', sub { my $lihs = $m->get_lihs;$m->dok2text(shift, 'lab', $lihs, @_) });
    $c->define_filter('lab', sub { my $c = shift;my @args = @_;my $lihs = $m->get_lihs;sub { my $dok = shift;$m->dok2text($dok, 'lab', $lihs, @args) } }, 1);
    $c->define_vmethod('SCALAR', 'tit', sub { return $m->dok2text(shift, 'tit', @_) });
    $c->define_filter('tit', sub { my $c = shift;my @args = @_;my $lihs = $m->get_lihs;sub { my $dok = shift;$m->dok2text($dok, 'tit', $lihs, @args) } }, 1);
    $c->define_vmethod('SCALAR', 'sut', sub { return $m->dok2text(shift, 'sut', @_) });
    $c->define_filter('sut', sub { my $c = shift;my @args = @_;my $lihs = $m->get_lihs;sub { my $dok = shift;$m->dok2text($dok, 'sut', $lihs, @args) } }, 1);
    $c->define_vmethod('SCALAR', 'des', sub { return $m->dok2text(shift, 'des', @_) });
    $c->define_filter('des', sub { my $c = shift;my @args = @_;my $lihs = $m->get_lihs;sub { my $dok = shift;$m->dok2text($dok, 'des', $lihs, @args) } }, 1);
    $c->define_vmethod('SCALAR', 'meta', sub { $m->dok2text(@_) });
    $c->define_filter('meta', sub { my $c = shift;my $tag = shift;my @args = @_;my $lihs = $m->get_lihs;sub { my $dok = shift;$m->dok2text($dok, $tag, $lihs, @args) } }, 1);
    $c->define_vmethod('SCALAR', 'nomlangtext', sub { my $lihs = $m->get_lihs;$m->nomlangtext($lihs, @_) });
    $c->define_vmethod('SCALAR', 'langnomtext', sub { my $lihs = $m->get_lihs;$m->langnomtext($lihs, @_) });
    $c->define_vmethod('SCALAR', 'cit', sub { my $nom = shift;my $dok = shift;my $lang = shift;my $lihs = $m->get_lihs;return $m->dbmeta($dok,$nom,$lihs,$lang) });
    $c->define_vmethod(SCALAR => vok => sub { my ($text,$vok,@rest) = @_;
					      return unless $text;
					      return $text if $gvars->{novok};
					      my $lihs = $m->get_lihs;
					      $text = $m->vok($lihs,$gvars,$vok,$text,@rest);
					      $text });
    $c->define_vmethod(SCALAR => dbmeta => sub { return $m->dbmeta(@_) });

=head4 vmethod nom2xlst
    
Turn a list variable from the maketext language into a list variable that can be used by the Template language.

Example:

If antevar is defined in the maketext file

	special = deflst chunks
	special = lstpucq chunks mm
	special = lstpucq chunks aamm
	special = lstpucq chunks tem

as defined by A2E::Tmplfil, then a template text

       [* FOREACH var = chunks.nom2xlst -*]
       - [* var *]
       [* END -*]

leads to an output of

	- mm
	- aamm
	- tem

The list variable can be quite complex.
Lower layers need not be lists but can be hashes or strings or references to variables of such types.

=cut
    $m->define_vmethod(SCALAR => nom2xlst => sub { my $nom = shift;
						 LST: {
						     my $lst = $m->{grupvars}->{$nom};
						     last unless $lst;
						     return $lst if 'ARRAY' eq ref $lst;
						     $m->pstop('lst is %v', v =>  $lst) if ref $lst;
						     $lst = $m->strl2arb($lst,1);
						     return $lst if 'ARRAY' eq ref $lst;
						     $m->pstop('bad list reference %v', v =>  $lst);
						   };
						   $m->nom2xlst($nom) });
    $c->define_vmethod(SCALAR => nom2xhac => sub { my $nom = shift;
						 HAC: {
						     my $hac = $m->{grupvars}->{$nom};
						     last unless $hac;
						     return $hac if 'HASH' eq ref $hac;
						     $m->pstop('hac is %v', v =>  $hac) if ref $hac;
						     $hac = $m->strl2arb($hac,1);
						     return $hac if 'HASH' eq ref $hac;
						     $m->pstop('bad hash reference %v', v =>  $hac);						     
						   };
						   $m->nom2xhac($nom) });
    $c->{jungkont} = { };
    my $litscat = sub { 
	my ($lit) = shift =~ m(\A\s*(\w+)\s*\Z);$m->ustop('invoking litscat without a value to start from') unless $lit;
	$litvaldir = $m->setlitvaldir;
	my $lihs = $m->get_lihs;
	local $_ = $m->litscat($lihs,$lit,{ });
	$_ = $m->fayl_get_str_expand($_);
	$_;
    };
    my $litreks = sub { my ($lit) = shift =~ m(\A\s*(\w+)\s*\Z);$lihs = $m->get_lihs;$m->ustop('invoking litreks without a value to start from') unless $lit;$m->litreks($lihs,$lit, @_) };
    my $suplit = sub { my $grup = shift;return unless $grup;my $nom = $grup . '__sup';$lihs = $m->get_lihs;my $val = $lihs->getstr($nom);$m->techo('suplit of %n is %v', n => $grup, v => $val);$val }; 
    my $supfols = sub { my $grup = shift;return unless $grup;my $nom = $grup . '__sup';$lihs = $m->get_lihs;my $val = $lihs->getstr($nom);$m->techo('suplit of %n is %v', n => $grup, v => $val);my @fols = $m->getlst($val . '__fols');@fols }; 
    my $rekputfayl = sub { my $rek = shift;my $fo = shift;$m->rekputfayl($rek,$fo);'' };
    my $monwoks = sub { return $m->monwoks(@_) };
    my $monwok = sub { return $m->monwok(@_) };
    $c->define_vmethod(SCALAR => litscat => $litscat);
    $c->define_vmethod(SCALAR => litreks => $litreks);
    $c->define_filter(litscat => $litscat);
    $c->define_vmethod(HASH => rekputfayl => $rekputfayl);
    $c->define_vmethod(HASH => klavs => sub { my $hac = shift;$m->klavs($hac) }); # 未試
    $c->define_vmethod(LIST => seqpucq => sub { my ($m,$lst,@vals) = @_;$m->seqpucq($lst,\@vals) }); # 未試
    $c->define_vmethod(SCALAR => monwoks => $monwoks);
    $c->define_vmethod(SCALAR => monwok => $monwok);
    $c->define_vmethod(SCALAR => expand => sub { my $text = shift;my $grup = shift;$m->grupfill($text, grup => $grup) });
    $c->define_filter(expand => sub { my $context = shift;my $grup = shift;return sub { my $text = shift;return $m->grupfill($text, grup => $grup) }, 1 });
    $c->define_vmethod(SCALAR => u2tex => sub { $m->str_u2tex(shift) });
    $c->define_filter(u2tex => sub { my $context = shift;return sub { my $text = shift;return $m->str_u2tex($text) } }, 1);
    $c->define_vmethod(SCALAR => grupval => sub { $m->grupval(@_) } );
    $c->define_vmethod(SCALAR => sf2nom => sub { my ($sup,$fol) = @_;
						 my $lihs = $m->get_lihs;
						 my $text = $m->sup8fol2nom($lihs,$sup,$fol);
						 $text });
    $c->define_vmethod(SCALAR => supdok => $suplit); # misnomer, 應更名 suplit, 而 supdok 應是祖先文件而非文塊 file:/phm/20/09/28/sig/_lng.phm_pub_sig200928.txt::vargrup
    $c->define_vmethod(SCALAR => suplit => $suplit); 
    $c->define_vmethod(SCALAR => supgrup => $suplit); 
    $c->define_vmethod(LIST => supfols => $supfols); 
    $c->define_vmethod(SCALAR => vargrup => sub { my $grup = shift;return if ! $grup;my $var = shift;$m->vargrup($var,$grup) });
    $c->define_vmethod(SCALAR => num2alpha => sub { $m->num2alpha(shift) });
    $c->define_vmethod(SCALAR => num2Alpha => sub { $m->num2Alpha(shift) });
    $c->define_vmethod(SCALAR => romnum => sub { $m->romnum(shift) });
    $c->define_vmethod(SCALAR => hannum => sub { $m->hannum(shift) });
    $c->define_vmethod(SCALAR => varkont => sub { my ($nom) = @_;my $lihs = $m->get_lihs;my $val = $lihs->getval($nom);$val++;$lihs->putstr($nom,$val);$val });
    $c->define_vmethod(SCALAR => numpad => sub { my ($text,$len,$pad) = @_;$m->numpad($text, $len, $pad) });
    $c->define_vmethod(SCALAR => modnum => sub { my( $text,@args) = @_;$m->modnum($text, @args) });
    $c->define_vmethod(SCALAR => varnom => sub { my $nom = shift;my $lihs = $m->get_lihs;return '@'.$nom if $lihs->{lsts}->{$nom};return '%'.$nom if $lihs->{hacs}->{$nom};$nom });
    $c->define_vmethod(SCALAR => varval => sub { my $nom = shift;my $lihs = $m->get_lihs;my $val = $lihs->getval($nom);$val = $m->arb2str($val) if ref $val;$val });
    $c->define_vmethod(SCALAR => text2list => sub { my $text = shift;$m->{sarb}->text2list($text) });
    $c->define_vmethod(SCALAR => call => sub { my ($text,@args) = @_;
					       $m->{sarb}->val_args_call($text,undef,@args) } );
    $c->define_vmethod(SCALAR => argscall => sub { my ($args,$text) = @_;
						   return $text unless 'ARRAY' eq ref $args;
						   $m->{sarb}->val_args_call($text,undef,@$args) } );
    $m->qpop1($prog,1);
};


=head3 Function postkonfig

=cut

=head3 Encoding and Decoding from structure info to tmplvars

Formerly in SDBM: Store a long list in the form of __tmpllsts1, __tmpllsts2, .... __tmpllsts1543

=head4 &lihs_put_fols($lihs, 'lsts', @fols), &lihs_get_fols($lihs, 'lsts'), tregfyl et al
moved to
	file:/adv/perl/A2E/Lihs.pm::put_noms
	file:/adv/perl/A2E/Lihs.pm::get_noms
	file:/adv/perl/A2E/Lihs.pm::put_tups
	file:/adv/perl/A2E/Lihs.pm::get_noms
=cut

=head4 &lihs_put_lvlmax
encode lvlmax info in tmplvars, decoder lihs_get_lvlmax 
=cut

sub lihs_put_lvlmax {
    my ($m,$prog,$lihs) = shift->lpuc('lihs_put_lvlmax',@_);
    my $fyl = $lihs->put_tups('lvlmax');
    return $m->dpop($prog) unless $fyl;
    $m->dpop1($prog,$fyl);
    # file:/adv/perl/A2E/OLD/MLHT.old::put_lvlmax_old_sdbm
};
sub lihs_get_lvlmax {
    my ($m,$prog,$lihs) = shift->lpuc('lihs_get_lvlmax',@_);
    my $stor = $lihs->get_tups('lvlmax');
    return $m->dpop($prog) unless $stor;
    $m->dpop1($prog,$stor);
    # file:/adv/perl/A2E/OLD/MLHT.old::put_lvlmax_old_sdbm
};

=head5 $m->progvars_set_lihs_vars($m,$lihs,%opts)

仿效 setptvar 效果, 使記憶中 lihs->{vars} 服從 progvars
曽惹禍 file:/phm/18/10/15/sig/_lng.phm_pub_sig181015.txt::179, 但似已充份限制

=cut
sub progvars_set_lihs_vars {
    my ($m,$prog,$lihs,%opts) = shift->lpuc('progvars_set_lihs_vars',@_);
    return $m->dpop($prog) unless $lihs eq $m->{lihs};
    $opts{lihs} = $lihs;
    $opts{nonstorp} = 1;
    foreach my $nom (keys %{$m->{progvars}}) { 
	my $val = $m->{progvars}->{$nom};
	next if ref $val;
	next unless $lihs->getval($nom);
	$m->lihs_putarb($lihs,$nom,$val,%opts);
	$m->lihs_putarb($lihs,'__'.$nom,$val,%opts) if $m->{progvars2hsprintf}->{$nom};
    };
    $m->dpop1($prog,1);
};    
sub lihs_put_vrbs { #
    my ($m,$prog,$lihs) = shift->lpuc('lihs_put_vrbs',@_);
    $m->dpop1($prog,$lihs->put_tups('vrbs')) # 已無需 sub { $m->{sarb}->miniverb_get(shift,$lihs) }
}
sub lihs_get_vrbs { 
    my ($m,$prog,$lihs) = shift->lpuc('lihs_get_vrbs',@_);
    $m->dpop1($prog,$lihs->get_tups('vrbs', v2k => sub { my ($k,$v) = @_;$m->defin_miniverb($k,$v,$lihs) }));
};
=head4 $m->lihs_varsfyl(l => 'zh'), (t => 'html')
=cut
sub lihs_varsfyl {
    my ($m,$prog,$lihs,%opts) = shift->lpuc('lihs_varsfyl',@_);
    $m->set_litsdir($lihs, %opts);
    my $litstag = $opts{l} ? 'l' : $opts{t} ? 't' : '';
    my $fyl = $lihs->varsfyl(litstag => $litstag);
    $m->dpop1($prog,$fyl);
};
sub opts2treg {
    my %opts = @_;
    $opts{treg} || $opts{l} ? 'lvars' : $opts{t} ? 'tvars' : 'vars';
}
sub lihs_put_vars { # dump vars from memory into cache for use in mldk
    my ($m,$prog,$lihs,%opts) = shift->lpuc('lihs_put_vars',@_);local $_;
    $opts{k2v} = sub { my ($k,$stor) = @_;$lihs->getstr($k) };
    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
    # $lihs->put_tups('vars', %opts);
    my $treg = opts2treg %opts;
    $lihs->put_tups($treg, %opts);
#   DATA: {
#       my $fdat = $lihs->tupsfyl('data');
#       my %tups = ();
#       foreach my $k (keys %{$lihs->{typ2var}->{d}}) { $tups{$k} = $lihs->getstr($k) };
#       $lihs->fayl_put_tups($fdat,\%tups);
#     };
# </nov>
    $m->dpop1($prog,$lihs);
};

=head4 $m->lihs_get_vars($lihs, %opts): Read string vars into memory from cache
opts are either of variants 
	fyl => '.lits/l.zh/vars.hac'
 	l => 'zh', t => 'latex'
# <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
# 	fyls => ['.lits/l.zh/vars.hac', '.lits/t.html/vars.hac']
	fyls => ['vars', '.lits/vars.hac', 'lvars', '.lits/l.zh/vars.hac', 'tvars', '.lits/t.html/vars.hac']
# </nov>

This is done before template processing because everything needs to put into a static hash that becomes Template::Stash.
Read $varsfyl if not done yet, mark it as read by setting $last_varsfyl, return $varsfyl.
Return nil if $varsfyl does not exist (e.g. during A2E::MLMK::submktdir), in which case we have alternatives, as decided in lihs_get_arbs.

=cut
# <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
# our $last_varsfyl = ''; # </nov>
sub lihs_get_vars {
    my ($m,$prog,$lihs,%opts) = shift->lpuc('lihs_get_vars',@_);my $fyl = $opts{fyl};my $treg = opts2treg %opts;
    my $lang = $opts{l} || $m->pvar(lang => 1);
    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
    my %tag2opt;$tag2opt{t} = 'f'; # </nov>
    my @fyls = ();
    if ($opts{fyls}) {
	push @fyls, $m->ref2list($opts{fyls});
    } elsif ($fyl) {
	push @fyls, $treg, $fyl if -e $fyl;
    } else {
	push @fyls, vars => $m->lihs_varsfyl($lihs);
	foreach my $litstag (qw(l t)) {
	    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC">
	    my $opt = $tag2opt{$litstag} || $litstag;
	    next unless $opts{$opt};
	    $fyl = $m->lihs_varsfyl($lihs, $litstag => $opts{$opt});
	    next unless -e $fyl;
	    my $varsnom = $litstag . 'vars';
	    push @fyls, $varsnom, $fyl;
	    # </nov>
	};
    };
    while (@fyls) {
	$treg = shift @fyls;$fyl = shift @fyls;
        # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC"><old>next if $fyl eq $last_varsfyl;</old></nov>
	$lihs->get_tups($treg => fyl => $fyl);
    };
    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC"><old>$last_varsfyl = $fyl;</old></nov>
    $m->dpop1($prog,$fyl);
};
=head4 $m->lihs_get_arbs($lihs, %opts): Read vars/lsts/hacs into memory from cache
opts are those of lihs_get_vars.
This is done before template processing because everything needs to put into a static hash that becomes Template::Stash.
Resolve any file handle variables where necessary with var_fh_get_strs.
Overwrite any string vars with lsts/hacs where available.
=cut
sub lihs_get_arbs { # read vars,lsts,hacs from cache and return a hash suitable for template processing
    my ($m,$prog,$lihs,%opts) = shift->lpuc('lihs_get_arbs',@_);
    my %arbs = $lihs->vars_fh_get_strs;
    foreach my $treg (qw(lsts hacs)) { foreach my $k (keys %{$lihs->{$treg}}) { $arbs{$k} = $lihs->{$treg}->{$k} } };
    %arbs;
};

=head4 Function arb2var

Store info that is needed to reconstruct complex variables from their string representations, as attempted by &A2E::MLHT::lihs_var2arb
file:/adv/perl/A2E/MLHT.pm.tmpl::/lihs_var2arb/
=cut
our @tuptregs = qw(vrbs varp varq lvlmax);

sub lihs_arb2var { # file:/phm/18/09/sig/_org.phm_pub_sig1809.txt::var2arb
    my ($m,$prog,$lihs) = shift->lpuc('lihs_arb2var',@_);
    foreach my $typ ('@','%') {
	my $treg =$lihs->treg($typ);
	$lihs->put_noms($treg);
    };
    foreach my $treg (@tuptregs) {
	$m->pstop('void element in list tuptregs %T', T => \@tuptregs) unless $treg;
	$lihs->put_tups($treg);
    };
    $m->dpop1($prog,$lihs);
};

=head3 Special Directives of MLHT level

=head4 Special Directive litvalid

Validate a C<lit> (template variable) here and now, using the validation information given through some special lits.

TODO: migrate the specification of those special lits to more robust structured variables such as __valids__, __errmsg__ etc which are internally created and appended using defxhac, hacxput etc.
=cut
sub special_litvalid { 
    my ($m,$lit,$dialog) = @_;
    # $dialog //= !$m->pvar('batch');
    $m->litvalid($lit,dialog => $dialog);
};

=head4 Special directive set_sems

When used in maketext

	special upon_setvar sems set_sems

or in A2E::Txtdbm, it ensures that the information from textchunk @sems, as soon as it set, is analysed and registered internally, so that __sem1, __sem2, __sems etc are set and the sems itself will not be changed again.

	args := 
	rets := ok
	ok :: true if operation was achieved i.e. with sems was non-null and usable

As of 2020-01-01 becoming a regular part of function mldk.
	
=cut

sub special_set_sems { 
    my ($m,$prog,$lihs) = shift->lpuc('special_set_sems',@_);
    my @sems = $lihs->getlst('sems');
    return $m->dpop($prog) if !@sems;
    my $sem = $m->arb2str(@sems);
    @sems = $m->lihs_set_sems($lihs,$sem);
    $m->dpop($prog,@sems);
};

=head4 defperson
    define a person for use in a2e templates
  TODO: move to a localisation module
=cut
sub special_defperson {
    my ($m,$prog,$nom,@rest) = shift->ppuc('',@_);
    my $lihs = $m->get_lihs;
    my ($lab,$url,@args) = map { $m->grupfill($_, lihs => $lihs) } @rest;
    $lihs->putstr($nom => $lab);
    $lihs->putstr('person_'.$nom.'_lab' => $lab);
    $lihs->putstr('person_'.$nom.'_url' => $url) if $url;
    $m->dpop1($prog,$nom);
};
=head4 mlht_postkonfig
=cut

sub defblok {
    my $m = shift;
    my ($tml,$nom,$kod) = @_;
    $m->{tmlbloks}->{$tml}->{$nom} = $kod;
    return 1;
};
    
sub grupval {
    my ($m,$prog,$grup,$fol,$lihs) = shift->ppuc('grupval',@_);
    return $m->dpop($prog) unless $grup and $fol;
    $lihs = $m->get_lihs($lihs);
    my $nom = join '_', $grup, $fol;
    my $val = $lihs->getstr($nom);
    $m->dpop1($prog,$val);
};
sub vargrup { # 未試、用意見 file:/phm/20/09/28/sig/_lng.phm_pub_sig200928.txt::vargrup
    # 可簡化、應將文塊名存於 %{$gvars->{_vargrup}} file:/adv/perl/A2E/Tmplfil.pm.tmpl::_vargrup
    my ($m,$prog,$grup,$var,$lihs) = shift->ppuc('vargrup',@_);
    return $m->dpop($prog) unless $grup and $var;
    $lihs = $m->get_lihs($lihs);
    my $nom = join '_', $grup, $var, '_grup';
    $m->dpop1($prog,$nom);
};
sub konv_linioi {
    my ($m,$gvars) = @_;
    join "\n", @{$gvars->{litvals}};
};
our %latex_paraln = (l => 'raggedright', r => 'raggedleft', c => 'centering');
our %latex_fsz2size = (ti => 'tiny', si => 'scriptsize', no => 'normalsize', la => 'large', La => 'Large', LA => 'LARGE', hu => 'huge', Hu => 'Huge', HU => 'HUGE');
sub ltxkmd {
    my ($kmd,$arg) = @_;
    return $arg unless $kmd;
    $kmd = "\\".$kmd;
    return $kmd unless $arg;
    $kmd.'{'.$arg.'}';
};
sub latex_tabhjung {
    my ($m,$gvars) = @_;
    join ' & ', @{$gvars->{litvals}};
};
sub latex_tabvjung {
    my ($m,$gvars) = @_;
    my @litvals = @{$gvars->{litvals}};
    return '' unless @litvals;
    my $cline = '';
    my $hline = '';
  CLINE: {
      my $border = $gvars->{border};
      last unless $border;
      $hline = ltxkmd 'hline';
      my $tkols = $gvars->{tkols}; # file:/sig/oas/20/07/ecodX7/tra/wart/_lng.ecodX7_wart.txt::tkols
      unless ($tkols) { $cline = $hline;last };
      my @kols = $m->ref2list($gvars->{kols});
      $m->ustop('tkols %t given but no kols', t => $tkols) unless @kols;
      my $nkols = scalar @kols;
      $m->ustop('too many title columns %t among a total of %n', t => $tkols, n => $nkols) unless $tkols < $nkols;
      $cline = ltxkmd 'cline', join '-', ++$tkols, $nkols; # 待査要用 \cline 乎 \cmidrule 乎
    };
    my $lnbrk = sub { "\\\\". shift .  "\n" };
    my $hlnbrk = &$lnbrk($hline);
    my $clnbrk = &$lnbrk($cline);
    local $_ = join $clnbrk, @litvals;
    $_ = $hline . "\n" . $_ . $hlnbrk;
    if ($gvars->{longtable} && $gvars->{tabtit}) {
	$_ .=  '\endhead';
    }
    $_;
};
sub latex_tabvgjung {
    my ($m,$gvars) = @_;
    my @litvals = @{$gvars->{litvals}};
    my $hline = $gvars->{border} ? '\hline' : '';
    $hline .= "\n";
    join $hline, @litvals;
};

sub latex_par { 
    my ($m,$gvars) = @_;
    my $litstem = $gvars->{litstem};
    my $fsz = $m->grupval($litstem,'fsz');
    my $lfsz = ltxkmd $latex_fsz2size{$fsz};
    my $aln = $gvars->{aln};
    my $laln = ltxkmd $latex_paraln{$aln};
    my $tabtit = $gvars->{tabtit} || 0;
    my $ftf = $tabtit ? 'bf' : $gvars->{ftf};
    my $textftf = $ftf ? 'text' . $ftf : '';
    my @litvals = @{$gvars->{litvals}};
    local $_ .= $laln;
    $_ .= $lfsz;
    $_ .= ' ' if $_;
    my $litvals_size = scalar @litvals;
    my $litval = '';
    while ($litval = shift @litvals) {
	next unless $litval;
	$_ .= ltxkmd $textftf, $litval;
	last unless @litvals;
	$_ .= "\\par "; 
    };
    $_ = '{'.$_.'}' if 1 < $litvals_size or $laln or $lfsz;
    $_;
};
sub repeat { my $s = shift;my $i = shift;local $_ = '';while ($i--) { $_ .= $s };$_ };
sub latex_tabfld {
    my ($m,$gvars) = @_;
    my @kols = $m->ref2list($gvars->{kols});
    $m->ustop('tabfld used without columns') unless @kols;
    my @lastspan = $m->ref2list($gvars->{lastspan});
    my @litspan = $m->ref2list($gvars->{litspan});
    my @nextspan = $m->ref2list($gvars->{nextspan});
    my $lastkol = $lastspan[$#lastspan];
    my $nbr = $m->skiplen(\@kols,$lastkol,$litspan[0]);
    my $litstem = $gvars->{litstem};
    my $kol = $gvars->{kol} || $litspan[0];
  KOL: {
      last if $kol;
      $kol = $litspan[0];
    LITFOL: {
	last if $kol;
	local $_ = $gvars->{litfol};
	last KOL unless m(\A\_(\w+)\Z);
	($kol) = split '_', $1;
      };
    };
    $m->pstop('unable to recognize column in litstem %l', l => $litstem) unless $kol;
    $gvars->{kol} = $kol;
    my %alns = $m->ref2hash($gvars->{alns});
    $m->ustop('tabfld used without alns') unless %alns;
    my %sizs = $m->ksum2sizs(\@kols,100,$m->ref2hash($gvars->{sizs}));
    $m->ustop('tabfld used without sizes') unless %sizs;
    my $siz = $sizs{$kol} || $m->pstop('unable to derive size from column %k', k => $kol);
    my $aln = $gvars->{aln} || $alns{$kol};
    $m->ustop('unable to recognize alignement of column %k in litstem %l based on alns %a', k => $kol, l => $litstem, a => \%alns) unless $aln;
    $gvars->{aln} = $gvars->{tabtit} ? $aln : $m->grupval($litstem,'aln');
    $_ = repeat '& ', $nbr;
    my $nbrkols = scalar @litspan;
    $m->ustop('bad number of spanned columns %d', d => $nbrkols) unless $nbrkols > 0;
    my @subkols = ();
    my $arg = $m->latex_par($gvars);
  MULTICOL: {
      last unless $nbrkols > 1;
      @subkols = @{$m->spanlst(\@kols, $litspan[0], $litspan[$#litspan])};
      $siz = 0;foreach my $k (@subkols) { $siz += $sizs{$k} };
      last unless @subkols;
      my $rlim = $gvars->{border} ? '|' : '';
      my $llim = $rlim if $gvars->{litnum} <= 1;
      $arg = "\\multicolumn{$nbrkols}{${llim}${aln}${rlim}}{$arg}";
    };
  MULTIROW: {
      my $vspan = $gvars->{vspan};
      last unless $vspan;
      $m->ustop('bad vspan value %v, must be numeric', v => $vspan) unless $vspan =~ m(\A\d+\Z);
      $m->ustop('size %d out of range', d => $siz) if $siz <= 0 or $siz > 100;
    SUP: { # when vspan is 1, compute it from number of rows in supsup node (which is probably joined by tabvjung)
	last unless $vspan == 1;
	my $sup =  $gvars->{_sup};
	$m->ustop('no sup for vspan in nom %n', n => $gvars->{_nom}) unless $sup;
	my $lihs = $m->get_lihs;
	$sup = $lihs->getstr($sup . '__sup');
	$m->ustop('no supsup for vspan in nom %n', n => $gvars->{_nom}) unless $sup;
	my $fols = $lihs->getstr($sup . '__fols');
	$m->ustop('no supsupfols for vspan in nom %n', n => $gvars->{_nom}) unless $fols;
	$vspan = scalar $m->text2list($fols);
      };
      $siz = sprintf '%02d', $siz;
      $arg = "\\multirow{$vspan}{.$siz\\textwidth}{$arg}";
    };
    $_ .= $arg;
  KOLSFIN: { 
      last if @nextspan;
      $nbr = $m->skiplen(\@kols,$litspan[$#litspan]);
      $_ .= repeat ' &', $nbr;
    };
    $_;
}; # latex_tabfld
sub latex_imgtxt {
    my ($m,$gvars) = @_;
    my $nom = $gvars->{imgnom};
    my $tex = join '.', $nom, 'tex';
    my $tpl = join '.', $tex, 'tmpl', 'txt';
    $m->template_process($tpl, $tex, %$gvars);
    "\\input{$nom}";
};
sub html_imgtxt {
    my ($m,$gvars) = @_;
    my $nom = $gvars->{imgnom};
    "<img src=\"$nom.jpg\" alt=\"$nom\"/>";
};
sub jungil_litvals {
    my ($m,$prog,$gvars,$lihs) = shift->ppuc('jungil_litvals',@_);$lihs = $m->get_lihs($lihs);
    my $tml = $gvars->{tml} || $m->pvar(tml => 1); 
    my $jungil_nom = $gvars->{jungil};
    my $litval = $m->jungblok($lihs,$jungil_nom,$gvars,$tml);
    $m->dpop1($prog,$litval);
};
sub latex_tabrowx {
    my ($m,$prog,$gvars,$lihs) = shift->ppuc('latex_tabrowx',@_);local $_;$lihs = $m->get_lihs($lihs);
    local $_ = '';
  HSKIP: { # file:/sig/oas/19/04/jkinW4/tra/kijun/_lng.jkinW4_kijun.txt::hskip, 舊式、新近使用 tkols  file:/sig/oas/20/07/ecodX7/tra/wart/_lng.ecodX7_wart.txt::tkols
      my $hskip = $gvars->{hskip} || 0;
      last unless $hskip;
      my @kols = $m->ref2list($gvars->{kols});
      $m->ustop('hskip specified but no kols found') unless @kols;
      my $nkols = scalar @kols;
      $m->ustop('hskip %h must be smaller than nkols %n', h => $hskip, n => $nkols) unless $hskip < $nkols;
      $_ .= ltxkmd 'cmidrule', join '-', ++$hskip, $nkols;
      $_ .= ' ';
    };
    $_ .= $m->jungil_litvals($gvars,$lihs);
    $m->dpop1($prog,$_);
};
sub mlht_postkonfig {
    my ($m,$prog,$lihs) = shift->qlpuc('mlht_postkonfig',@_) or return;
    $m->tmplfil_postkonfig($lihs);
    $m->dokfs_postkonfig($lihs);
    $m->valid_postkonfig;
    $m->set_template($lihs); # cf file:/adv/perl/A2E/MLDK.pm.tmpl::set_templates
    $m->defin_mlht_renfuns($lihs);
    $m->defin_mlht_tmplfuns($lihs);
    $m->tmplfil_special_put(litvalid => \&special_litvalid);
    $m->tmplfil_special_put(set_sems => \&special_set_sems);
    $m->tmplfil_special_put(defperson => \&special_defperson);
    $m->tmplfil_special_put(miniverb => sub { shift->defin_miniverb(shift,shift,$lihs) });
    $m->defblok(latex => linioi => \&konv_linioi);
    $m->defblok(html => linioi => \&konv_linioi);
    $m->defblok(latex => tabhjung => \&latex_tabhjung);
    $m->defblok(latex => tabrowx => \&latex_tabrowx);
    $m->defblok(latex => tabvjung => \&latex_tabvjung);
    $m->defblok(latex => tabvgjung => \&latex_tabvgjung);
    $m->defblok(latex => tabfld => \&latex_tabfld);
    $m->defblok(latex => txtimg => \&latex_txtimg);
    $m->defblok(latex => par => \&latex_par);
    $m->qpop1($prog,1);
};
sub postkonfig {
    my ($m,$prog) = shift->qpuc('A2E::MLHT::postkonfig',@_) or return;
    $m->SUPER::postkonfig;
    $m->mlht_postkonfig;
    # $m->set_langs unless $novitas;
    $m->dpop1($prog,1);
}


=head2 Implementation Functions

=head3 ($lihs,$dir,$dok) = $m->lihs_dir_dok_init(%opts): setup document variable space

=cut

sub lihs_dir_dok_init {
    my ($m,$prog,%opts) = shift->ppuc('lihs_dir_dok_init',@_);
    my ($lihs,$dir,$dok);
    ($lihs,%opts) = $m->shifthash('lihs',%opts);
    ($dir,%opts) = $m->shifthash('dir',%opts);
    ($dok,%opts) = $m->shifthash('dok',%opts);
    my $pwd = $m->pvar(pwd => 1);
  DIR: {
      last if $dir;
      $dir = $lihs->{vars}->{__dir__} if $lihs;
      last if $dir;
      $dir = $m->dokdir($dok) if $dok;
      last if $dir;
      $m->pstop(funnovar => v => 'dok');
    };
    $dir = $m->absfayl($dir);
  DOK: {
      last if $dok;
      $dok = $lihs->{vars}->{__dok__} if $lihs;
      last if $dok;
      $dok = $m->dirdok($dir) if $dir;
      last if $dok;
      $dok = $m->pvar(dok => 1);
    };
  LIHS: {
      last if $lihs;;
      $lihs = $m->{dok2lihs}->{$dok};
      last if $lihs;
      $lihs = $m->{lihs} if $dir eq $pwd;
      last if $lihs;
      $lihs = $m->get_lihs;
    };
    $lihs->setdir($dir);
    $lihs->{vars}->{__dok__} = $dok;
    $m->{dok2lihs}->{$dok} = $lihs;
    $m->dpop($prog,$lihs,$dir,$dok,%opts);
};

=head3 Function mlht_set_dokvars

Read in document configuration
pvar dok must have been set before.

=cut

sub mlht_set_dokvars {
    my ($m,$prog,$lihs,$dok,$dir,%opts) = shift->opuc('mlht_set_dokvars',[[],[qw(nom_let dir_let)]],@_);
    ($dok,$dir) = $m->setdokdir($dok, $dir);
    $lihs->setdir($dir);
    $m->setpvar(dok => $dok, lihs => $lihs);
    my $doktyp = $m->pvar('doktyp');
    my $pfx = $m->pvar(srcfpfx => 1);
    my $mp = quotemeta $pfx;
    my $rdok = quotemeta $dok;
    unless ($dok && $doktyp) {
	($doktyp,$dok) = $m->{konfig}->infaylz->[0] =~ m(\A$mp(dok|nav)\.($rdok)\.txt\Z);
	$m->setpvar(doktyp => $doktyp, lihs => $lihs);
	$m->setpvar(dok => $dok, lihs => $lihs);
    };
    $m->ustop('nodok') unless $dok;
    $m->ustop('nodoktyp', d => $dok) unless $doktyp;
    $m->set_pwd($lihs);
    ($lihs,$dir,$dok) = $m->lihs_dir_dok_init(dir => $dir, dok => $dok, lihs => $lihs);
    $m->set_topdoks($lihs);
    $m->set_topdir($dok, lihs => $lihs);
    $m->set_langs($lihs);
    $m->dpop($prog,$lihs,$dok,$dir);
};

=head3 Function mlht_set_restvars

=head4 Subroutine mlht_get_dokl

=cut

sub mlht_get_dokl {
    my $prog = 'mlht_get_dokl';my $m = shift;$m->dpuc('mlht_get_dokl', @_);
    my $Dok = $m->CamelCase($m->pvar('dok'));
    my $Typ = $m->{doktyp} eq 'dok' ? '' : $m->{doktyp};
    $Typ = ucfirst $Typ;
    my $Lng = ucfirst($m->pvar('lang'));
    my $dokl = $Dok . $Typ . $Lng;
    $m->dpop1($prog,$dokl);
};

=head4 mlht_set_restvars

set some special, language-independent variables 

=cut

sub mlht_set_restvars {
    my ($m,$prog,$lihs,$lang) = shift->lpuc('mlht_set_restvars',@_);
    $m->pstop(funnovar => v => 'lang') unless $lang;
    $m->set_litsdir($lihs, l => $lang);
    $lihs->putstr(dokl => $m->mlht_get_dokl, lihs => $lihs);
    my $srcfile = $m->{konfig}->infaylz->[0];
    $srcfile = catfile $m->pvar('pwd'), $srcfile unless file_name_is_absolute $srcfile;
    $m->set_emacs_lokvars($lihs, srcfile => $srcfile, dokl => $lihs->getstr('dokl'));
    my $val;
    foreach my $nom (qw(dok pwd topdir topdok emacs_timestamp lang daba)) {
	$val = $m->pvar($nom);
	do { $m->pwarn('no value to set progvar %n to', n => $nom);next } unless $val;
	$lihs->putstr($nom => $val, lihs => $lihs);
    };
    $m->setpvar(lang => $lang, lihs => $lihs);
    $lihs->putstr(lang => $lang, lihs => $lihs);
    $lihs->putstr(__lang => $lang, lihs => $lihs);
    $m->{hsprintf_repl}->{l} = $lang;
  DATUM: {
      my $val = $m->pvar('datum') || $m->timestamp('%Y-%m-%d');
      do { $m->pwarn('no value to set progvar datum to');next } unless $val;
      $lihs->putstr(datum => $val, lihs => $lihs);
    };
    my $DIR = $m->pvar('DIR') || $m->pvar('pwd');
  DIR: { # obsolete?
      do { $m->pwarn('no value to set progvar DIR to');next } unless $DIR;
      $lihs->putstr(DIR => $DIR, lihs => $lihs);
    };
  URL: { # obsolete?
      $val = $m->pvar('url') || $m->dir2url($DIR);
      do { $m->pwarn('no value to set progvar DIR to');next } unless $val;
      $lihs->putstr(URL => $val, lihs => $lihs);
    };
    foreach my $daba (keys %{$m->{konfig}->dabas}) {
	next unless $m->{konfig}->dabas->{$daba};
	my $var = join '_', $daba, 'dbh';
	$lihs->putval($var => $m->{dbi}->{$daba}, lihs => $lihs);
=head5

Define database handles that can be invoked in templates using

     [USE DBI(dbh = a2e_dbh)]

as explained in

        perldoc Template::Plugin::DBI
	/phm/17/06/22/sig/_lng.phm_pub_sig170622.txt
	/adv/packages/perl-Template-DBI/_dok.perl_template_dbi_pub.txt

=cut
    };
=pod

Allow handover of parsed tree variables via the stash to external systems like A2E::Template::Plugin::Call:

    m->{SArb}->{vars} = context->{STASH};
    m->{SArb}->{arbs} = context->{STASH}->{__ARBS__};

=cut
    $lihs->putval(__ARBS__ => $m->{tmplarbs}, lihs => $lihs); # obsolete?
    $m->dpop1($prog,1);
};

=head3 Function vars_kk_vars

Concatenate vocabulary hashrefs $avars and $bvars

	$m->vars_kk_vars($m->{avars}, bd, { dok => ewicV9 });

Arg1 is a name of a concatenation function that may be used, the currently defined ones being
	bd	:	latter wins when it is defined
	be	:	latter wins when it is true/nonnull
	ae	:	former wins when it is true/nonnull
	ad	:	former wins when it is defined
	an	:	latter wins when former is defined, otherwise nothing
Args 4ff are options, wherein keys => [ ... ]  is of interest.  It limits the scope that is to be reset, e.g.

	$m->vars_kk_vars({ dok => ewicV9 }, [ bd => ak => 1 ], $m->{avars})


will result in only the keys of the first hashref $avars being updated from the second $bvars and those elements of $bvars whose hashref is not found in $avars are ignored.  Moreover updating can be limited to specific keys:

	$m->vars_kk_vars({ dok => ewicV9 }, [ bd => keys => [ 'dok', 'tmpl' ] ], $m->{tmplvars}, '|', $lang_tmplvars)

As seen here, concatenation can go on infinitely with pairs of concat-operator and hashref. 
The operators have alternative names defined in %konkatnoms.  Here '|' is used, which is reminiscent of logical or, meaning that $lang_tmplvars take precedence only where they fill blanks.

Another operator option 

        kv2kv => sub { my ($k1,$bv) = @_;my ($k2) = m(\A$dok\_(\w+)\Z);$k = $k2 || $k1;return $k, $bv->{$k1} } 

can be used to transform records on the fly as in file:/adv/perl/A2E/Mktdir.pm.tmpl::/lang_set_tmplvars/

It is to be recommended to use litsdirs2lihs and d2dbmvars in order to generate hashref arguments from dbm files.

Yet another option

	put => 'tpa'

allows to choose a writing routine other than the normal plain entry of a string value, e.g. one where the value is encoded as a string, which requires specific bookkeeping side-effects outside a plain hashref parameter.
The routine of greatest interest here ist $lihs->putarb, dubbed 'tpa'.

The option
	nonvars => \@tratoks
	nonvars => $m->{progvars}
causes any variables from the given name range to be ignored.

=cut

our %konkatnoms = ( dbm_ovr_txt => 'be', dbm_del_txt => 'bd', txt_ovr_dbm => 'ae', txt_del_dbm => 'ad', '||' => 'bd', '??' => 'ad', '?' => 'ae', '|' => 'ae', '&&' => 'bd', '!!' => 'bd', '&' => 'be', '!' => 'be', ad => 'ad', ae => 'ae', bd => 'bd', be => 'be' );
sub get_knom { my $m = shift;my $knom0 = shift || 'bd';my $knom = $konkatnoms{$knom0} || $m->pstop('unknown hash concatenation function %k', k => $knom0);return $knom };

=head3 $lihs = $m->litsdirs2lihs(dir => '/sig/oas') || $m->litsdirs2lihs(lihs => $lihs)

Read string variables from store, formerly sdbm, now .lits directory tree.
Argument indicates directory to read from.

Can be eliminated in favor or storage of all memorizable vars in a tups file, since now the body lines all stay on disk only.
See work on lihs_put_vars, lihs_get_vars, invoked from lihs_var2arb, lihs_arb2var

=cut
    
sub vars_set_proc {
    my ($m,$prog,$vnom,$vars,$klav,$knil) = shift->ppuc('vars_set_proc',@_);
    my($val,$ref);
    $ref = ref $klav;
    if ($ref) {
	$val = $klav;
    } else {
	$klav ||= $knil;
	$val = $vars->{$klav};
	$m->pstop('unknown %n object %k, known ones are %v', n => $vnom, k => $klav, v => $m->koniug_et(keys(%$vars))) unless $val;
	$ref = ref $val;
    };
    $m->pstop('%n object %k must not be %r but CODE', n => $vnom, k => $klav, r => $ref) unless 'CODE' eq $ref;
    $m->dpop1($prog,$val);
};

sub vars_kk_vars {
    my ($m,$prog,$avars,$lihs) = shift->ppuc('vars_kk_vars',@_);$lihs = $m->get_lihs($lihs);
    $m->pstop('avars is of invalid variable type, should be hashref') unless 'HASH' eq ref $avars;
    my $knom = '';
    my $pnom = '';
    my ($kk,$bvars,%opts);
    my %puts = (
	put => sub { my ($k,$v) = @_;$avars->{$k} = $v },
	dbm => sub { my ($k,$v) = @_;$m->dbmvars_put($avars, $k => $v) },	
	tps => sub { my ($k,$v) = @_;$lihs->putstr($k,$v) },
	tpa => sub { my ($k,$v) = @_;$lihs->putarb($k,$v) } );
    my %kv2kvs = (
	bv => sub { my ($k,$bv) = @_;return $k, $bv->{$k} },
	tgs => sub { my ($k,$bv) = @_;return $k, $lihs->getstr($k) },
	tga => sub { my ($k,$bv) = @_;return $k, $lihs->getarb($k) } );
    my $set_kk_opts = sub {
	my $ref = ref $kk;
	%opts = ();
	if (!$ref) { 
	    $knom = $kk 
	} elsif ('ARRAY' eq $ref) {
	    my @args = @$kk;
	    $knom = shift @args;
	    do { my $k = shift @args;$opts{$k} = shift @args } while @args;
	} else { $m->pstop('unknown concatenation operator %k', k => $kk) };
    };
    while (@_) {
	$kk = shift; # concatenation options, e.g. 'bd', { kk => 'bd', k => 1 }, [ 'bd', keys => [ 'dok', 'tmpl' ], k => 1 ]
	$bvars = shift; # vocabulary hashref, can be generated from $m->litsdirs2vars or $m->d2dbmvars etc
	&$set_kk_opts($kk);
	my $vars = $opts{ak} ? $avars : $bvars;
	my @keys = $opts{keys} ? @{$opts{keys}} : keys %$vars;
	my $gethac = sub {
	    my $arb = shift;
	    my %hac = ();
	    return %hac unless $arb;
	    my $ref = ref $arb;
	    return %$arb if 'HASH' eq $ref;
	    if ('ARRAY' eq $ref) {
		$hac{$_} = 1 for @$arb;
	    } else {
		$hac{$arb} = 1;
	    }; 
	    return %hac;
	};
	my %nonvars = &$gethac($opts{nonvars});
	my $put = $m->vars_set_proc('put', \%puts, $opts{put}, 'put');
	my %konkats = ( 
	    # &$put 一成不變 file:/phm/18/10/14/sig/_lng.phm_pub_sig181014.txt::put_konkats; 或許是 'my' 造成、可以試々放回上面
	    bd => sub { my ($k,$v) = @_;&$put($k,$v) },
	    be => sub { my ($k,$v) = @_;return unless $v;&$put($k,$v) },
	    ae => sub { my ($k,$v) = @_;return if $avars->{$k};&$put($k,$v) },
	    ad => sub { my ($k,$v) = @_;return if defined $avars->{$k};&$put($k,$v) } );
	my $konkat = $m->vars_set_proc('konkat',\%konkats, $m->get_knom($knom), 'bd');
	my $kv2kv = $m->vars_set_proc('kv2kv',\%kv2kvs, $opts{kv2kv}, 'bv');
	foreach my $k (@keys) {
	    next if $nonvars{$k};
	    my ($k,$v) = &$kv2kv($k,$bvars,$avars);
	    &$konkat($k,$v);
	};
    };
    $m->dpop1($prog,$avars);
};
sub tmplvars_kk_vars {
    my ($m,$prog,$lihs,@rest) = shift->lpuc('tmplvars_kk_vars',@_);
    while (@rest) {
	my $kk = shift @rest;
	$kk = $m->kk_set($kk, put => 'tpa');
	my $vars = shift @rest;
	$m->vars_kk_vars($lihs->{vars},$kk,$vars);
    };
};

=head3 Function lihs_var2arb

Reconstructs complex variables and thus reconstructs lihs->{lsts} and lihs->{hacs} after lihs->{vars} were built either by constructing them from scratch with $m->riid_tmplvars_faylz / $m->faylz_src_dk or by reading in what was previously constructed with m->lihs_get_vars.
#OLD file:/phm/18/09/sig/_org.phm_pub_sig1809.txt::var2arb
#OLD TODO: integrate with vars_get_sdbm, apply not to tmplvars but to any newly read dbmvars. file:/phm/18/10/28/sig/_org.phm_pub_sig181028.txt::var2arb
#OLD TODO2: what about miniverb reconstruction?  If nothing is done here, using miniverb definitions unless riid_tmplvars_faylz was done

=cut

sub lihs_var2arb {
    my ($m,$prog,$lihs,%opts) = shift->lpuc('lihs_var2arb',@_);
    $m->pstop(funnovar => v => 'lihs') unless 'A2E::Lihs' eq ref $lihs;
    # my %arbs = $m->lihs_get_arbs($lihs,fyl => $opts{fyl});
    # 正好不要 file:/phm/18/10/15/sig/_lng.phm_pub_sig181015.txt::179, 但似已充份限制
    $m->progvars_set_lihs_vars;
    foreach my $treg (@tuptregs) { $lihs->get_tups($treg) };
    $lihs->{freq}->{$_} = $lihs->{varq}->{$_} for keys %{$lihs->{varq}}; # keep last count separate so that we can weight importance of variables so as to determine storage stragety in file:/adv/perl/A2E/Tmplvars.pm.tmpl::
    $m->lihs_get_lvlmax($lihs);
    my $proc_fols = sub {
	my ($typ,@fols) = @_;
	foreach my $nom (@fols) {
	    my $val = $lihs->getstr($nom);
	    $m->pstop('variable %n has implausible value %v', n => $nom, v => $val) if $val =~ m(\A(?:ARRAY|HASH)\b);
	    do { $m->vecho('omit empty %t%k', t => $typ, k => $nom);next } unless $val;
	    eval { $m->strval_defarb($lihs,$typ,$nom,$val) }; # cf file:/adv/perl/A2E/Tmplfil.pm.tmpl::/var2arb/
	    $m->pstop('failed to set %t%n to %v: %s', t => $typ, n => $nom, v => $val, s =>  $@) if $@;
	};
    };
    foreach my $typ ('@', '%') {
	my $treg = $lihs->treg($typ);	
	my $typp = $lihs->typp($typ);
	my @fols = $lihs->get_noms($treg, stor => $lihs->{$typp});
	# file:/phm/18/10/15/sig/_lng.phm_pub_sig181015.txt::arb2var
	# @fols = grep { !$m->{progvars}->{$_} } @fols;
	next if $opts{noarbs}; 
	&$proc_fols($typ,@fols);
    };
    $m->dpop1($prog,$lihs);
};
=head3 kk_set($kk, kv2kv => sub { ... }, ak => 1, ...) => $kk
=cut
sub kk_set {
    my ($m,$prog,$kk,%opts) = shift->ppuc('kk_set',@_);
    my $ref = ref $kk;
    my (@lkk, $kk0,%hkk,$k,$v);
    if ($ref) {
	unless ('ARRAY' eq $ref) {
	    $m->pstop('unallowable reftype %r in concatenation operator %k', r => $ref, k => $kk);
	};
	@lkk = @$kk;
	$kk0 = shift @lkk;
	%hkk = @lkk;
    } else {
	$kk0 = $kk;
	%hkk = ();
    }
    $hkk{$k} = $v while ($k,$v) = each %opts;
    $kk = [ $kk0, %hkk ];
    $m->dpop1($prog,$kk);
};


=head3 $ok = &tmplarbs_kk_litsdirs2vars (($kk:konkat_op, $dbmvars:hashref)+)
This is becoming the basic routine for fully restoring vocabulary including complex variables from dbm/litsdir files.
=cut
sub tmplvars_kk_arbvars {
    my ($m,$prog,$lihs,@rest) = shift->lpuc('tmplvars_kk_arbvars',@_);
    while (@rest) {
	my $kk = shift @rest;
	my $arbvars = shift @rest;
	$m->tmplvars_kk_vars($lihs, $kk, $arbvars);
    };
    $m->dpop1($prog,1);
}
sub tmplvars_kk_progvars {
    my ($m,$prog,$lihs,@rest) = shift->lpuc('tmplvars_kk_progvars',@_);
    my $kk = $m->kk_set(@rest);
    my $res = $m->tmplvars_kk_arbvars($lihs,$kk,$m->{progvars});
    $m->dpop1($prog,$res);
};

=head3 $m->arbvars_kk_tmplvars($arbvars, $kk) => %vars: import the real tree values of complex tmplvars into arbvars 

$arbvars => $m->{cache}
$kk => [ ae => ak => 1 ]

=cut
sub arbvars_kk_tmplvars {
    my ($m,$vars,$kk) = @_;
    $kk = $m->kk_set($kk, kv2kv => 'tga');
    $m->vars_kk_vars($vars, $kk, $m->{tmplvars});
};
sub progvars_kk_tmplvars {
    my ($m,$kk) = @_;
    return $m->arbvars_kk_tmplvars($m->{progvars},$kk);
}
sub progvars_kk_arbvars {
    my $m = shift;
    while (@_) {
	my $kk = shift;
	my $vars = shift;
	return $m->vars_kk_vars($m->{progvars},$kk,$vars);
    };
};
sub dbmvars_kk_tmplvars { # 由於所叫程序複雜故不甚需要 Txt2dbm::txt2dbm 未尾可考慮用上
    my ($m,$vars,$kk) = @_;
    $kk = $m->kk_set($kk, kv2kv => 'tgs');
    return $m->vars_kk_vars($vars,$kk,$m->{tmplvars});
};

=head4 $m->dbmopts2paf(l => 'de', d => 'oas_akt1809', m => 'lng') => '/sig/oas/18/09/akt/_lng.oas_akt1809.de.txt'

Select a particular dbmfayl based on parametric info supplied as %opts.
Option m is the file name template configured via fnom2ffol

=cut

sub dbmopts2paf {
    my ($m,$prog,%opts) = shift->ppuc('dbmopts2paf',@_);
    my $fnom = 'llng';
    if ($opts{m}) { 
	$fnom = $opts{m};delete $opts{m};
    }
    my $dok = $opts{d} || $m->{hsprintf_repl}->{d};
    my $lang = $opts{l} || $m->{hsprintf_repl}->{l};
    $m->set_dok_lang($dok,$lang);
    my $fibr = $m->fnom2fobj_r($fnom);
    my $fi = $m->hsprintf($fibr,%opts);
    my $dir = '';
  DIR: {
      last if $dok eq $m->pvar('dok');
      $dir = $m->dokdir($dok);
      $m->ustop('no directory found for document %d', d => $dok) unless $dir;
      $m->uwarn('directory %d inaccessible', d => $dir) unless -r $dir and -d $dir;
      $fi = catfile $dir, $fi;
    };
    $m->dpop1($prog,$fi);
};

=head3 dok2tmplvars(d => 'oas1809', l => 'de') => %vars -- read vocabulary from a .lits directory and return $lihs object
=cut
sub dok2tmplvars { 
    my $m = shift;
    $m->litsdirs2lihs($m->dbmopts2paf(@_));
};

=head3 $m->tmplarbs_kk_dbmopts( bd => { d => $m->pvar('supdok'), l => 'de', m => 'l' })

Find the dbm file based on its parameters (opts), where 'm' is the dmb

former
	$m->dok2tmplvars($dok, ll => $lang, kk => 'dbm_del_txt', tv => $m->{tmplvars}, kv2kv => sub { ... } };
now
	$m->tmplvars_kk_dokfayl([ bd => kv2kv => sub { ... }] => { x => l, d => $dok, l => $lang });

SET_LANG: If no language argument is given, we check whether the quoted document is available in our current language.  If this is the case we choose the current language, otherwise the source language of the quoted document.

=cut

sub tmplarbs_kk_dbmopts {
    my $m = shift;
    $m->tmplarbs_kk_litsdir(shift,$m->dbmopts2paf(%{shift()})) while @_;
    1;
};
sub vars_kk_dbmopts {
    my $m = shift;
    my $vars = shift;
    $m->vars_kk_litsdir($vars, shift,$m->dbmopts2paf(%{shift()})) while @_;
    1;
};
sub vars_kk_litsdir {
    my $m = shift;
    my $vars = shift;
    while (@_) {
	my $kk = shift;
	my $dirvars = $m->litsdirs2vars(shift);
	$m->vars_kk_vars($vars, $kk, $dirvars);
    };
    1;
};

=head3 Function template_arbvars

orl file:/phm/18/09/sig/_org.phm_pub_sig1809.txt::template_arbvars
Make sure we have the right set of template variables when we invoke a template

Invocation Example:

	template_arbvars(l => 'de', f => 'html', __tf => ['_var.'.$dok.'.txt'], __tv => { dok => 'ewicV9' }, m => 'llng') => $tmplvars

such that the content of $m->{tmplvars} is changed when needed, and further one-time additions can be made with __tf and __tv

This becomes necessary as we introduce make-style rules in file:/adv/perl/A2E/MLDK.pm.tmpl::/template_process/

Since we are no longer directly using dbmvars for our templates, we might as well directly use list and hash variables therein.

A simpler version that may serve as hashref source for vars_kk_vars is d2dbmvars:

	$m->d2dbmvars('l', l => 'de', d => 'oas1809')
returns hashref read from file
	'_lng.oas1809.de.txt'

TODO see file:/phm/18/09/sig/_org.phm_pub_sig1809.txt::/dbm_konkat/

=cut
sub __opt2fnom { my $k = shift;return 'tml' if $k eq 'f';return 'llng' };
sub d2dbmvars {
    my $prog = 'd2dbmvars';my $m = shift;$m->dpuc($prog,@_);
    my $k = shift;
    my %opts = @_;
    $opts{m} ||= __opt2fnom{$k};
    my $fayl = $m->dbmopts2paf(%opts);
    my $vars = $m->litsdirs2vars($fayl);
    $m->dpop1($prog,$vars);
};
our %TMPLARBS_NONOPTS = (oobj => 1);
sub template_arbvars {
    my ($m,$prog,$lihs,%opts) = shift->lpuc('template_arbvars',@_);
    $m->decho('set arbvars for opts %O', O => \%opts);
    # <nov orl="file:/phm/20/12/27/sig/_lng.phm_pub_sig201227.txt::varsXC"> rm_old file:/adv/perl/A2E/OLD/MLHT.old::PROC_OPT </nov>
    my %arbs = $m->lihs_get_arbs($lihs, %opts);
    $m->dpop1($prog,\%arbs);
};

=head3 Function template_process

Simplify template invocation based on template_arbvars:

$m->template_process("_dok.$dok.html.txt", 'index.de.html', l => 'de', f => 'html', __tf -> [ "_var.$dok.mk" ], __tv => { dok => 'ewicV9' }, __tt => $m->{lang_template});

=cut
sub tml2outmode { # file:/phm/18/12/06/sig/_lng.phm_pub_sig181206.txt::tml2outmode
    my ($m,$prog,$tml) = shift->ppuc('tml2outmode',@_);
    do { $m->pwarn('invoked tml2outmode without tml argument');$m->dpop1($prog) } unless $tml;
    my $mode = $m->{konfig}->tml2outmode->{$tml};
    $m->dpop1($prog,$mode);
};
sub fi_opts_get_template_arbvars {
    my ($m,$prog,$lihs,$fi,$fo,%opts) = shift->lpuc('fi_opts_get_template_arbvars',@_);
  INSIST: {
      my $insist = $opts{insist};
      last unless $insist;
      delete $opts{insist};
      last if -f $fi and -r $fi;
      $m->ustop(fi_insist => f => $fi);
    };
    my $template = $opts{__tt};
  TT: {
      last unless $template;
      delete $opts{__tt};
    };
    my $tmpl = $opts{t} || $m->pvar(tmpl => 1);
    my $tml = $opts{f} || $m->pvar(tmplang => 1);
    $template ||= $m->{template};
    $m->pstop(template_ko => undef) unless 'Template' eq ref $template;
    # $m->mlht_set_langvars($lihs,$opts{l});
    my $set_lokvars = sub {
	my $mode = $opts{mode} || $m->tml2outmode($tml); # 'latex'
	if ($mode) { delete $opts{mode} } else { $mode = 'fundamental' };
	my $lokvars = $m->emacs_lokvars(mode => $mode);
	$lokvars->{fsrc} = $fi;
	$lokvars->{fobj} = $fo;
	my ($dok) = $fi =~ m(\b\_(?:lng|pre|dok)\.(\w+)\.);
	$dok ||= $m->pvar(dok => 1);
	$lokvars->{dok} = $dok;
	my $olddok = $lihs->getstr('dok');
	$m->pwarn('correcting old tmplvars dok %o to new %D', o => $olddok, d => $dok) if $dok ne $olddok;
	while (my ($k,$v) = each %$lokvars) {
	    $lihs->{vars}->{$k} = $v;
	};
    };
    &$set_lokvars;
    my $arbvars = $m->template_arbvars($lihs,%opts);
  TESTS: {
      last unless $m->{debug};
    VAR2ARBTEST: {
	my $hacnom = 'latex' eq $tml ? 'latex_class2sektplus' : 'langlands';
	last if ref $arbvars->{$hacnom};
	$m->pstop('failed to reconstruct complex test variable %v', v => $hacnom);
      };
    KDOKTEST: { # file:/phm/19/06/02/sig/_lng.phm_pub_sig190602.txt::kdok
	my $tdok = $arbvars->{dok};
	my $pdok = $m->{progvars}->{dok};
	last if $tdok eq $pdok;
	$m->pstop('tmplvar dok %t unequal to progvar dok %p', t => $tdok, p => $pdok);
      };
    };
    $m->dpop($prog,$template,$arbvars);
};

our @litsdir_io_opts = qw(nonstorp nolitsdir);
sub template_process {
    my ($m,$prog,$lihs,$fi,$fo,%opts) = shift->lpuc('template_process',@_);
    $m->vecho('In %(PL) to write from %i to %o with opts %O', i => $fi, o => $fo, O => \%opts);
    #220318 try $m->mlht_set_lihs($lihs,$opts{l});
    my $tml = $m->pvar(tml => 1);
    my ($template,$arbvars) = $m->fi_opts_get_template_arbvars($lihs,$fi,$fo,%opts);
    my $ipath = $template->service->context->config->{INCLUDE_PATH};
    $m->vecho('processing template %i with include path %I and opts %O to produce %o', i => $fi, I => $ipath, o => $fo, O => \%opts);
    $m->vecho('writing to outfile %o with tml %f and lang %l and undl_verb %u with include path %I', o => $fo, u => $arbvars->{undl_verb}, I => $ipath);
  NOLITSDIR_INCLUDE: { # we have stored everything in memory, dont need disk i/o
      my %io_opts = ();
      my $sarbvars = $m->{sarb}->{lihs}->{vars};$m->{sarb}->{lihs}->{vars} = $arbvars; # vmethods like litscat have only this variable to look to; not even $arbvars nor the Template::Stash can be used to pass anything to them
      $io_opts{$_} = $lihs->{vars}->{'__'.$_.'__'} ? 1 : 0 for @litsdir_io_opts;
      my $varqfyl = $lihs->tupsfyl('varq');
      $lihs->get_tups('varq',fyl => $varqfyl);
      $lihs->{vars}->{__nostorp__} = 1;
      $lihs->{vars}->{__nolitsdir__} = 0;      
      my $res = $m->template_include($fi, $arbvars, $fo, %opts) ? 1 : 0; # file:/phm/19/05/12/sig/_lng.phm_pub_sig190512.txt::riidvars 改用include
      $m->vecho('written to %o %(SW|KO|OK).', o => $fo, SW => $res);
      $m->ustop('Attempt to write template %i to %o reported unsuccessful by TT', i => $fi, o => $fo) unless $res;
      $m->ustop(tmplvoid => i => $fi, o => $fo) unless -s $fo;
      $lihs->{vars}->{'__'.$_.'__'} = $io_opts{$_} for @litsdir_io_opts;
      $lihs->put_tups('varq', fyl => $varqfyl);
      $m->{sarb}->{lihs}->{vars} = $sarbvars;
    };
    $m->dpop1($prog,$fo);
};

=head3 Function mlht_set_langvars

    set tmplvars for one language for postprocessing.

=cut

sub mlht_set_langvars {
    my ($m,$prog,$lihs,$lang) = shift->lpuc('mlht_set_langvars',@_);
    my @langs = $m->set_langs($lihs);
    my $lang0 = $langs[0];
    $m->pstop("no lang0 found in mlht_set_langvars among langs @langs in directory %d", d => $m->dir2dir) unless $lang0;
    $lang ||= $lang0;
    # my $fi = $m->antefayl;
    $m->lihs_get_vars($lihs, l => $lang);
    # $m->dbm2tmplvars($m->{tmplvars}, bd => $fi) if -r $fi;
    # $m->dbm2tmplvars($m->{tmplvars}, bd => $m->lng_fayl(l => $lang)) if -r $fi;
    # $m->dbm2tmplvars($m->{tmplvars}, ae => $m->lng_fayl(l => $lang0)) unless $lang0 eq $lang;
    $m->dpop1($prog,1);
};

=head3 Function mlht_set_lihs

    set template variables

=cut

sub mlht_set_lihs {
    my ($m,$prog,$lihs,$lang) = shift->lpuc('mlht_set_lihs',@_);
    $m->mlht_set_langvars($lihs,$lang);
    $m->mlht_set_restvars($lihs,$lang);
    $m->riid_tmplvars_post;
    $m->dpop1($prog,1);
};

=head3 Function proc_faylz

file:/phm/18/09/sig/_org.phm_pub_sig1809.txt::mlhts

=cut
sub get_tmls {
    my ($m,$prog,$lihs) = shift->lpuc('get_tmls',@_);
    my @tmls = $lihs->getlst('tmls');
  KONFIG_TML: { # 允許選擇性輸出 file:/phm/21/07/16/sig/_lng.phm_pub_sig210716.txt::tml
      my $tml = $m->{konfig}->tml;
      last unless $tml;
      last unless $lihs->getval(konfig_tml => 1);
      @tmls = ($tml);
    }; 
    return $m->dpop($prog,@tmls) if @tmls;
    @tmls = $m->uniq_plst('tmls');
    my @tmls2 = ();
    my %tmlp2 = ();
    foreach my $tml (@tmls) {
	push @tmls2, $tml if $m->{tmlp}->{$tml} and !$tmlp2{$tml};
	$tmlp2{$tml} = 1;
    };
    $m->setptvar(tmls => \@tmls2, lihs => $lihs);
    $m->dpop($prog,@tmls2);
}
sub proc_faylz {
    my ($m,$prog,$lihs) = shift->lpuc('proc_faylz',@_);
    my @faylz = @{$m->{konfig}->infaylz};
    my @langs = $m->plst('langs');
    my $lang0 = $langs[0];
    $m->mlht_set_lihs($lihs,$lang0);
    my @tmls = $m->get_tmls;
    my $tmpl = $m->pvar(tmpl => 1);
    my $toptmpl = $lihs->getstr('toptmpl') || 'konv_tmpl.txt';
    my $tml_tmpl_r = $lihs->getstr('tml_tmpl') || '%f_tmpl.txt';
    my $infayl_repl = $lihs->getstr('infayl') || '_dok.%d.txt';
    my $tml_infayl = $lihs->getstr('tml_infayl') || '_dok.%d.%f.txt';
    $m->{hsprintf_repl}->{t} = $tmpl;
    my $dok = $m->pvar(dok => 1);
    $m->{hsprintf_repl}->{d} = $dok;
    my $tmplproc = sub { my $hi = shift;my $ho = shift;$m->{template}->process($hi, $m->{tmplvars}, $ho) or die $m->{template}->error() };
    # file:/phm/18/09/20/sig/_lng.phm_pub_sig180920.txt::proctext 
	my $topprocr = $lihs->getstr('topprocr') || 'a2e_dok_%f';
    my $topproc = '';
    # my $proc = undef;
    my $pre_txt = $m->hsprintf('%ppre.%d.txt');
    my $pre_pag = join '.', $pre_txt, 'pag';
    my $topvar = $lihs->getstr('topvar');
    $m->ustop('topvar undefined, need to rebuild vocabulary with mldk dbms?') unless $topvar;
    if (@faylz) { 
	$m->template_process($lihs, $faylz[0], $m->{konfig}->outfayl, f => $m->{konfig}->tmplang || 'html', l => $m->{konfig}->lang || $lang0, t => $tmpl);
    } elsif (@tmls) {
	my $infayl = $m->hsprintf($infayl_repl);
	my $fi = -r $infayl ? $infayl : $m->hsprintf($toptmpl);
	&$tmplproc($fi);
	my $litscat = $Template::VMethods::TEXT_VMETHODS->{litscat};
	$m->pstop('litscat undefined') unless $litscat;
	foreach my $tml (@tmls) {
	    $m->pstop('empty tml in %T', T => \@tmls) unless $tml;
	    my $mode = $m->tml2outmode($tml);
	    my $tml_tmpl = $m->hsprintf($tml_tmpl_r);
	    $m->{hsprintf_repl}->{f} = $tml;
	    my $tmplang = $tml;
	    $m->setpvar(tmplang => $tmplang, lihs => $lihs);
	    $m->set_template_riidvars($lihs, $tml, $tmpl); # file:/phm/19/05/12/sig/_lng.phm_pub_sig190512.txt::riidvars
	    my $tml_txt = $m->hsprintf($tml_infayl);
	    my $tml_pag = join '.', $tml_txt, 'pag';
	    $m->tmplarbs_kk_litsdir(bd => $tml_txt) if -r $tml_pag;
	    $topvar = $lihs->getstr('topvar');
	    $m->ustop('topvar still undefined even after %f was read', f => $tml_pag) unless $topvar;
	    my $pi = '';
	    if (-r $tml_txt) {
		$pi = $tml_txt;
	    } else {
		$m->{template}->process($tml_tmpl, $m->{tmplvars});
	      PROC: {
		  $topproc = '';
		  do { $m->uwarn('top process name %n not defined', n => 'topprocr');last } unless $topprocr;
		  $topproc = $m->hsprintf($topprocr);
		  $pi = eval { $m->{template}->context->template($topproc) };
		  $m->uwarn('no template command called %c found: %m', c => $topproc, m => $@) if $@;
		};
	    };
	    foreach my $lang (@{$m->{langs}}) {
		$m->mlht_set_lihs($lihs,$lang);
		my $fo = $m->hsprintf($m->{konfig}->tml2outfayl->{$tml});
		$m->set_emacs_lokvars($lihs, srcfile => $fo, mode => $mode);
		if ($pi) {
		    &$tmplproc($pi, $fo);
		} else {
		    $m->vecho('Calling litscat on topvar %t.', t => $topvar, c => $topproc);
		    $m->fayl_put_str($fo,&$litscat($lihs,$topvar));
		};
	    };
	};
    };
    $m->dpop($prog,@faylz);
};

=head3 Function mlht

file:/phm/18/09/sig/_org.phm_pub_sig1809.txt:mlhts

=cut

sub mlht {
    my $prog = 'mlht';my $m = shift;$m->dpuc($prog, @_);
    $m->{konfig}->infaylz($_) for @_;
    my ($lihs,$dok,$dir) = $m->mlht_set_dokvars if -r '_dokfs';
    $m->proc_faylz;
    $m->dpop1($prog,1);
};

=head2 Listcat & Co

=head3 Function revargs

=cut

sub revargs {
    my $prog = 'revargs';my $m = shift;$m->dpuc($prog,@_);
    my @args = map { 'ARRAY' eq ref $_ ? @$_ : $_ } @_;
    my $body = $m->{sarb}->render_revargs(@args);
    $m->dpop1($prog,$body);
}

=head3 Function setlistarg

subroutine for processing arguments of litscat, but with more potential use.

    args := arg, litstem, varnom, listlevel, priv?
    rets := arg
    arg => [ '1', [ '&', 'q', 'a' ], '_', 'a' ]
    arg => [ [ 'proc', 'alineas' ], [ 'call', 'ln' ] ], 'top'
    arg := listref or string, depending on listlevel 	
    arg :: value of the argument, produced in litscat by 'shift' while parsing the commandline and enhanced by us
    litstem => 'top'
    litstem :: stem of stash variable
    varnom => 'jungils'
    varnom => 'priv'
    varnom :: name of variable to be underscore-joined to stem and searched in stash
    listlevel :: level of listification (use of textsep2listref) leading to a more or less complex return value
    listlevel 0 :: string
    listlevel 1 :: simple listref
    listlevel 2 :: listref of elements of which splittable ones become listrefs, others scalar
    listlevel 3 :: listref of listrefs

=cut

sub setlistarg {
    my ($m,$prog,$lihs,$arg,$litstem,$varnom,$listlevel) = shift->lpuc('setlistarg',@_);
    $arg ||= $listlevel ? [] : '';
    my $nom = $m->dokprogverpre(400) ? join('_', $litstem, $varnom) : $m->grupvar($litstem, $varnom);
    my $val = $lihs->getstr($nom);
    do { $m->dpop1($prog,$arg);return $arg } unless $val;
    if ($listlevel == 3) { 
	@$arg = map { $m->textsep2listref($_) } @{$m->textsep2listref($val)};
    } elsif ($listlevel == 2) {
	@$arg = map { my ($sep, @lst) = $m->textsep2seplist($_);$sep ? [ @lst ] : $_ } @{$m->textsep2listref($val)};
    } elsif ($listlevel == 1) {
	@$arg = @{$m->textsep2listref($val)};
    } else {
	$arg = $val;
    };
    $m->dpop1($prog, $arg);return $arg;
};

=head3 Function litsjung

Process blocks for litscat

=cut
sub jungblok {
    my ($m,$prog,$lihs,$jungil_nom,$gvars,$tml,$jungil_typ) = shift->lpuc('jungblok',@_);
    local $_;
  VOK: {
      last unless 'vok' eq $jungil_typ;
      $_ = $m->vok($lihs,$gvars,$jungil_nom);
      return $m->dpop1($prog,$_);
    };
    my $jung = $m->{tmlbloks}->{$tml}->{$jungil_nom};
  JUNG: {
      last unless $jung;
      $_ = &$jung($m,$gvars);
      return $m->dpop1($prog,$_);
    };
    $jungil_typ ||= 'proc';
  PROC: {
      last if 'proc' eq $jungil_typ;
      $m->ustop('bad section joiner %j operating on block %b, must be proc or vok', j => $jungil_typ, b => $jungil_nom);
    };
  EVAL: {
      $_ = eval { $m->template_include($jungil_nom,$gvars) };
      last unless $@;
      $m->ustop(blokfail => b => $jungil_nom, g => $gvars, m => $@);
    };
    $m->dpop1($prog,$_);
};

sub litsjung {
    my ($m,$prog,$lihs,$tml,$jungil,$grupvarsr,@litvals) = shift->lpuc('litsjung',@_);
    my %grupvars = $m->ref2hash($grupvarsr);
    my @jungil_seq = @$jungil; # [ 'call', 'ln' ]
    my $jungil_typ = shift @jungil_seq;
  REVARGS: {
      last unless $jungil_typ eq 'revargs';
      my $argnum = shift @jungil_seq;
      $m->ustop('revargs requires integer argument, not %n', n => $argnum) unless $argnum =~ m(\A\d+\Z);
      $jungil_typ = shift @jungil_seq;
      my @prevals = splice @litvals, 0, $argnum;
      @litvals = reverse @litvals;
      unshift @litvals, @prevals;
    };
    my $litval = '';
    my $jungil_nom = shift @jungil_seq;
    $grupvars{jungnom} = $jungil_nom; # 'linioi_spi'
    $m->pwarn('%n operating on empty block', n => $jungil_nom) unless @litvals;
    $grupvars{litvals} = \@litvals;
    while (@jungil_seq) { # extra args such as => ('jungil', 'alineas')
	my $k = shift @jungil_seq;
	my $v = shift @jungil_seq;
	$grupvars{$k} = $v;
    };
    my $jungkont_nom = $grupvars{kontnom} || $jungil_nom; # 'linioi_spi'
    if ($grupvars{kontval}) {
	$m->{jungkont}->{$jungkont_nom} = $grupvars{kontval};
    } else {
	$m->{jungkont}->{$jungkont_nom}++;      
    };
    $grupvars{jungkont} = $m->{jungkont}->{$jungkont_nom};
    $litval = $m->jungblok($lihs, $jungil_nom, \%grupvars, $tml, $jungil_typ);
    $m->dpop1($prog,$litval);
};

=head3 Function litscat

Construct hierarchy and typesetting rules of a document based on naming of text chunks found in database (lits) and typeset (cat) the document.

    args := litstem, kontils, jungils, litlvl, litnum
    litstem :: stem of top level text chunk names
    litstem => 'top'
    kontils := listref of strings or listrefs
    kontils :: counters to be used
    kontils => [ '1', [ 'q', 'a' ], '1' ]
    jungils := listref of listrefs
    jungils ::  joiners to be used
    jungils => [ [ 'proc', 'alineas' ], [ 'proc', 'qeta' ], [ 'call', 'al' ], [ 'call', 'ln' ] ]
    litlvl := integer
    litlvl := sectioning level of any subsectional chunks (sublits), default 0
    litlvl => 0
    litnum := integer
    litnum :: sequential number of current lit
    litnum => 0
 

Of these, kontils is no longer used in the new implicit naming syntax described in dok:adv_pub_litscat1004.
Also, in the case of implicit naming, jungils is preferably read from the internal variables such as '_top_jungils' in every group by &A2E::Tmplfil::tmplvars_opengrup.

A call like

    $m->litscat($lihs, 'top', [ '1', [ 'q', 'a' ], '1' ], [ [ 'proc', 'al' ], [ 'proc', 'qeta' ], [ 'call', 'al' ], [ 'call', 'ln' ] ], 0)

should ensure that text chunks implicitely named with

    (q8a /proc+al/proc+qeta/call+al/call+ln/
    (intro
    (q8a_q
    ...
    )
    (q8a_a
    (
    this is line a of paragraph 1 of A part of the intro section of our Q&A
    this is line b of paragraph 1 of A part of the intro section of our Q&A 
    )
    )
    )
    )

are typeset based on the definitions in template language of

    block alineas
    block qeta

and definitions in maketext language of

    chunk al
    chunk ln
 
where block qeta is a user-defined template language block named 'qeta', and 'al' and 'ln' are user-defined 'call'able verbs, and all of these elements are fed with the text chunks that belong to their level.

The joiners (e.g. 'proc+qeta', 'call+al') can be reset for any element at any level by saying, in the textchunk file, e.g. 

    (q8a
    ...
    (
    ( |call+lnol|call+ln|
    )
    )
    )

meaning that within the question part of block 2 we use 'lnol' for concatenation at the first level and ln for concatenation at the next level.

Joiners used most are proc and call, also in conjunction with revargs, e.g.

    des5_jungils := |revargs:1:proc:desclist&tagdesc_regex&\A(\w+)\s\:\:\s(.*)\Z|call:al|call:ln|

where revargs (reverse arguments) assures that textchunks are counted downwards, which might be useful in a chronological list, see example below.

Note that parameters passed to proc (equivalent template process statement) must in this case be all contained in a single argument together with the command verb itself so as not to interfere with the argument count of C<revargs>.  In the above example this single argument uses '&' as its internal separator, because most others are blocked by the letters that appear in the regexp.  It might be more elegant to write

    desclist_regexp = \A(\w+)\s\:\:\s(.*)\Z
    des5_jungils := |revargs:0:proc:desclist+tagdesc_regex+${desclist_regexp}|call:al|call:ln|

The numeric argument or revargs indicates how many initial list elements are exempted from the reversal.
This can be useful in case of sections like

    log0 = Chronology
    log1 = 2009-03-01 first event
    log2 = 2009-03-02 second event

where we want a reverse chronology of events but a subject line ("Chronology") coming first in any case.
The 'revargs' function is similar but unrealted to the 'revargs' function of text textchunk language.

The 'revargs' function has become less important in the new system of implicit naming.

Allowable counters in the old syntax were:

    '1' : /[0-9]+/ : any arabic number, decimal system, starting from from number indicated here
    'a' : /[a-z]+/ : any alphabetic number, eicosahexesimal ($ALPHA_BASIS-based) system with 'z' as 0, 'a' as 1 and 'az' as 26, starting from the alphabetic number indicated here 
    'nom+des+rev+img' : enumerated set of new stems
    '++1+2+2a+2b' : enumerated set of suffixes attached to existing stem
    '++a..c+e..h' : enumerated as above, with ranges (e.g. here leaving out 'd' even if it exists and going up to 'h' even if not every one of these exists)
    '++1..3+5..8' : enumerated as above, with ranges (e.g. here leaving out '4' even if it exists and going up to '8' even if not every one of these exists); TODO: ensure that this is implemented
    '_' : underscore separator between levels, needed when counters take forms like '+1+2+2a+2b'

In the new syntax counting is always done implicitely has no choice regarding the numbers and formats used.
Where this is a problem, the solution is to be sought at the level of Template programming using mechanisms such as the one sketched by 'tagdesc_regex' above, which is supported in our 'desclist' block definition in 'bautext_tmpl.txt'.
    
Any template language blocks invoked in the jungils argument have the following template variables available for their use:

    litvals := list of strings
    litvals :: evaluated chunks of text to be concatenated by the block
    litstem := symbol
    litstem :: stem name of the block, usable e.g. as an anchor name
    litlvl := integer
    litlvl :: hierarchy level of the block, usable e.g. for generating a section heading
    indlvl := integer
    indlvl :: indentation level of the block, whis is incremented when the block is one of those set true in a hash variable such as %indproc = ||enumlist2|1||itemlist2|1||ilinioi2|1|, as demonstrated in a2e document nitaO2_pat
    litnum := integer
    litnum :: ordinal number of the block within the sequence of peer blocks at litlvl, usable e.g. for generating numbered section titles
    litlen :: number of leaves in node

Moreover, for debugging purposes, at the moment we hand the following info to each template process:
 
    jungnom => desclist
    jungnom := symbol
    jungnom :: name of current joining function 
    jungkont => 133
    jungkont := integer
    jungkont :: frequency of invocation of jungnom so far

Complex information such as jungils is passed by means of a structured represenation such as

    proc+alineas,call+ln # wrong syntax
    ,proc+alineas,call+ln
    /proc+alineas/call+ln/

in which the separators can be chosen freely but the higher-level separator must come before the lower-level separator.
The second and third of the above three examples will work, the first not.

=cut

sub litscat_simpl {
    my ($m,$prog,$litstem) = shift->ppuc('litscat_simpl',@_);
    # $m->uwarn('wrapper litscat_simpl around litscat unfinished.  need to reconstruct arguments from existing tree structure');
    $m->{lits} = $m->{nom2lits}->{$litstem} || $m->nom2lits($litstem);
    $m->set_grupvars($litstem);
    my $litval = $m->litscat($m->{lihs}, $litstem, $m->{grupvars});
    $m->dpop1($prog,$litval);
}
sub val2lst { # TODO: make this extra safety unnecessary so that we can use text2list: @kopilits etc should be properly nested from the start
    my ($m,$prog,$val) = shift->ppuc('val2lst',@_);
    my @lst = ref $val ? $m->ref2list($val) : $m->text2list($val);
    $m->dpop($prog,@lst);
};

sub fayl_put_str_expand {
    my ($m,$prog,$litstem,$s) = shift->ppuc('fayl_put_str_expand',@_);
    local $_ = $s;
    $_ = $m->fayl_get_str_expand($_);
    my $litvalfyl = catfile $litvaldir,$litstem;
    $m->fayl_put_str($litvalfyl,$_);
    $_ = '$('. $litstem . ')';
    $m->vecho($_);
    $m->dpop1($prog,$_);
};

sub litscat {
    my ($m,$prog,$lihs,$litstem,$grupvarsr,$jungils) = shift->lpuc('litscat',@_);
    $m->pwarn(funnovar => v => 'litstem') unless $litstem;
    my %grupvars = $m->ref2hash($grupvarsr);
    my $tml = $grupvars{tml};
    my $optsr = { lihs => $lihs, gvars => \%grupvars };
  TML: {
      last if $tml;
      $tml = $m->pvar(tmplang => 1);
      $grupvars{tml} = $tml;
    };
    my $lang = $m->pvar(lang => 1);
    my $litval = '';
  PRE_LITVAL: { # see if we are already in a value rather than a structure node and if so process and return
      my $get_non_litval_p = sub {
	  # to be recognized as a value, there must be no structure; simply having a value under $litstem name is not enough
	  my $litstem_nom = $m->grupvar($litstem,'nom');
	  my $litstem_val = $lihs->getstr($litstem_nom) || '';
	  return unless $litstem_val; # 或名或値規則 file:/phm/20/11/19/sig/_lng.phm_pub_sig201119.txt::nomnom_xor_nomval
	  $m->pstop('litstem_nom %n must point to litstem %l not %v', a => $litstem_nom, l => $litstem, v => $litstem_val) unless $litstem_val eq $litstem;
	  1;
      };
      my $get_litval = sub { # see whether $litstem is a node or a value, protect node namespace
	  # local $_ = $lihs->getstr($litstem);
	  # my ($litval) = m(\A\s*(.*\S)\s*\Z)s;
	  # $litval;
	  $lihs->getstr($litstem);
      };
      my $non_litval_p = &$get_non_litval_p;
      last if $non_litval_p;
      $litval = &$get_litval;
      $m->ustop('empty litval for litstem %l', l => $litstem) unless length $litval;
      last if ref $litval; # |a|b|c|d|e|
      my $statvar = $m->grupvar($litstem, 'stat'); # top__stat
    STAT: {
	my $stat = $lihs->getstr($statvar); # true if body started with '!:' or '?:'
	last if $stat; # if so, expansion has already been done
	$litval = $m->grupfill($litval, lihs => $lihs, gvars => \%grupvars);
      };
    VOKS: { # apply vok function to a field, either literally based on vok scalar or dependent on lit name and voks hash; useful for tables where a special transformation is applied to each column (%voks) but this again may be overridden locally (vok).  This must come before TMPLANG.  Each field has potentially several enumerated litvals which are transformed separately.  This used to be done in file:/adv/tmpl/a2e/latex/tabel_tmpl.txt::voks and similar operations are still done to give fields special font typefaces textftf file:/adv/tmpl/a2e/latex/tabel_tmpl.txt::textftf
	my $vok = $grupvars{vok};
      VOK: {
	  last if $vok;
	  my $voks = $grupvars{voks};
	  last unless $voks;
	  my $ref = ref $voks;
	  $m->ustop('special variable voks must be HASH not %r', r => $ref) unless 'HASH' eq $ref;
	  my ($kol) = $litstem =~ m(\_([a-z]+)\d+\Z); # Each field has potentially several enumerated litvals which are transformed separately.
	  last unless $kol;
	  $vok = $grupvars{voks}->{$kol};
	};
	last unless $vok;
	$litval = $m->vok($lihs,\%grupvars,$vok,$litval);
      };
    TMPLANG: {
	do { $litval = $m->str_u2tex($litval);last } if $tml eq 'latex';
	do { $litval = $m->str_u2htm($litval);last } if $tml eq 'html';
      };
      return $m->dpop1($prog,$litval);
    };
    $jungils = $m->setlistarg($lihs, $jungils, $litstem, 'frm', 3); # [ [ 'call', 'al' ], [ 'call', 'ln' ] ]
    $m->kopilits($optsr,$litstem);
    my @subfols = $m->litfols($lihs,$litstem);
    my @subjuns = @$jungils;
    my $jungil = shift @subjuns;
    return $m->dpop1($prog,$litval) unless 'ARRAY' eq ref $jungil and @$jungil;
    my $litlvl = $grupvars{litlvl} // -1;
    my $indlvl = $grupvars{indlvl} // 0;
    my $litnum = $grupvars{litnum} || 0;
    $litlvl++;
    $grupvars{litlvl} = $litlvl;
    $grupvars{indlvl} = $indlvl;
    $grupvars{litlen} = scalar @subfols;
  SUBFOLS2LITVAL: { # 有連接器故必有分枝
    GVARS: {
	# %{$m->{grupvars}} = %grupvars; # 2019-06-02為簡化偵錯而設
	# $m->grup_add_grupvars($litstem, %$optsr);
	my $gvars2 = $m->set_grupvars($litstem, lihs => $lihs);
	$m->hashpush(ff => \%grupvars, $gvars2);
      };
    TOKVOKS: { # file:/phm/19/07/19/sig/_lng.phm_pub_sig190719.txt::tokvoks
	my $tokvoks_s = $m->grupval($litstem,'tokvoks');
	last unless $tokvoks_s;
	my %tokvoks = $m->ref2hash($m->str2arb($tokvoks_s));
	my @keys = @{$tokvoks{'_keys_'}};
	@keys = keys %tokvoks unless @keys;
	foreach my $k (@keys) {
	    my $v = $tokvoks{$k};
	    next unless $v;
	    my $gv = $grupvars{$k};
	    next unless $gv;
	    $gv = 1 eq $v ? $m->grupfill($gv, lihs => $lihs, gvars => \%grupvars) : $m->vok($lihs,\%grupvars,$v,$gv);
	    $grupvars{$k} = $gv;
	    my $lk = join '_', $litstem, $k;
	    $lihs->putstr($lk => $gv);
	};
      };
      my @litvals = ();
    GET_LITVALS: {
	last unless @subfols;
	my %subvars = %grupvars;
	$subvars{subfols} = \@subfols;
	$subvars{lastfol} = '';
	$subvars{lastspan} = [];
	my $setnext = sub { my $i = shift;my $fol = $subfols[$i] || '';$subvars{nextfol} = $fol;my @span = grep { $_ } split '_', $fol;$subvars{nextspan} = \@span };
	my $i = 0;
	&$setnext($i);
	my $fol2litval = sub {
	    my $fol = $_;
	    $i++;
	    $m->pstop('next fol unequal to nextfol') unless $fol eq $subvars{nextfol};
	    $subvars{litfol} = $fol;
	    $subvars{litspan} = $subvars{nextspan};
	    $subvars{litnum} = $i;
	    &$setnext($i);
	    my $nom = $m->sup8fol2nom($lihs,$litstem,$fol);
	    local $_ = $lihs->getstr($nom) || '';
	    my $val = m(\A([|+/%]).*\1\s*\Z)s ? $m->textsep2listref($_) : $m->litscat($lihs, $nom, {%subvars}, [ @subjuns ]);
	    $subvars{lastfol} = $subvars{litfol};
	    $subvars{lastspan} = $subvars{litspan};
	    return $val;
	};
	@litvals = map { &$fol2litval($_) } @subfols;
	# 似為安全起見、待査是否真有必要重設
      };
      $grupvars{litnum} = $litnum; 
      $grupvars{litstem} = $litstem;
    VOKANT: { # application see file:/sig/oas/19/07/jnwpW7/tra/publ.txt::/^|/ file:/sig/oas/19/07/jnwpW7/tra/papl.txt::/^|/
	last unless @litvals;
	local $_ = $litvals[0];
	last unless 'ARRAY' eq ref $_;
	if (@subjuns and $grupvars{linisplit}) { # allow turning lines into records, e.g. table rows of form |a|b|c|d| 
	    @litvals = map { ref $_ ? $m->procjung($lihs, \@subjuns, \%grupvars, @$_) : $_ } @litvals;
	    last VOKANT;
	};
	my $vokant = shift @litvals;
	@litvals = map { $m->{sarb}->val_args_call($_,{%grupvars},@litvals) } @$vokant;
      };
      $litval = $m->litsjung($lihs,$tml,$jungil,\%grupvars,@litvals);
      $litval = $m->fayl_put_str_expand($litstem,$litval);
      $m->pstop(nodnojoin => n => $litstem, l => $litlvl, i => $litnum, j => $jungil, m => $@) if $@;
    };
    delete $m->{nom2gvars}->{$litstem} if $m->{nom2gvars}->{$litstem};
    delete $m->{nom2lits}->{$litstem} if $m->{nom2lits}->{$litstem} and $novitas; # hesitating, maybe still needed for kopilits in some fragile case?
    $m->dpop1($prog,$litval);
};
=head4 Function litfols, litvars etc

	$m->litfols($lihs,'top') => ('ant','ont','int')
	$m->litfols($lihs,'top',$lihs) => ('ant','ont','int')
	$m->lihs_get_typfols($lihs,'top','fols') => ('ant','ont','int')
	
=cut

sub lihs_get_typfols {
    my ($m,$prog,$lihs,$stem,$ftyp) = shift->lpuc('lihs_get_typfols',@_);
    $m->pstop(funnovar => v => 'fol') unless $ftyp;
    my $nom = $m->grupvar($stem,$ftyp); # 'top__fols'
    my @fols = $m->text2list($lihs->getstr($nom)); # ('adr', 'akt', 'aqt')
    $m->dpop($prog,@fols);
};
sub litfols {
    my ($m,$prog,$lihs,$stem) = shift->lpuc('litfols',@_);
    my @fols = $m->lihs_get_typfols($lihs,$stem,'fols');
    $m->dpop($prog,@fols);
};
sub litvars {
    my ($m,$prog,$lihs,$stem) = shift->lpuc('litvars',@_);
    my @fols = $m->lihs_get_typfols($lihs,$stem,'vars');
    $m->dpop($prog,@fols);
};
sub litlsts {
    my ($m,$prog,$lihs,$stem) = shift->lpuc('litlsts',@_);
    my @fols = $m->lihs_get_typfols($lihs,$stem,'lsts');
    $m->dpop($prog,@fols);
};
sub lithacs {
    my ($m,$prog,$lihs,$stem) = shift->lpuc('lithacs',@_);
    my @fols = $m->lihs_get_typfols($lihs,$stem,'hacs');
    $m->dpop($prog,@fols);
};

=head4 Function litreks -- mlht replacement for Template::Plugin::Datafile

/sig/oas/17/06/gobcU6/gobcU6.sig_tmp2.txt
should be simplified with help of a MLHT vmethod 'litreks' so that instead of this
we can write
FOREACH r = LIT.litreks('fyl','tit')

See newer use in
/sig/oas/17/07/satoU7/satoU7.sig_tmp5.txt 
Similar conventions for field mapping L1 used in dabarels, possible problems in /phm/18/09/01/sig/_lng.phm_pub_sig180901.txt

=cut

sub litreks {
    my ($m,$prog,$lihs,$litstem,$idnom,@lnoms) = shift->lpuc('litreks', @_);
    $idnom ||= 'L0';
    local $_;
    my $litfols = $lihs->getstr($litstem.'__fols') or do { $m->pwarn('no litfols for %l', l => $litstem);$m->dpop;return };
    my @fols = $m->textsep2list($litfols);
  LITFOLS: {
      last if @fols;
      @fols = $m->litfols($lihs,$litstem);
    }
    $m->pwarn('%n litfols for %l', n => scalar(@fols), l => $litfols);
    my @reks = ();my $fld = '';
    my ($nom, $ln, @flds) = ();
    my $i = 0;
    foreach my $fol (@fols) {
	$nom = $m->sup8fol2nom($lihs,$litstem,$fol);
	my $idval = $lihs->getstr($nom.'__nom');
	do { $m->uwarn('no record identifier found for leaf %f', f => $fol);next } unless $idval;
	do { $m->uwarn('identifier %n misregistered as %v', n => $nom, v => $idval);next } unless $idval eq $nom;	
	my $rek = { $idnom => $idval };
	$m->dgot($idnom => $rek->{$idnom});	
	my @flds = $m->litvars($lihs,$nom);
	foreach my $fld (@flds) {
	    next unless $fld =~ m(\A[a-z]);
	    my $lit = join('_', $nom, $fld); 
	    my $val = $lihs->getstr($lit);
	    do { $m->pwarn(fldlitnul => f => $fld, l => $lit);next } unless $val;
	    $rek->{$fld} = $val;
	    $m->dgot($fld => $rek->{$fld});	
	};
	my @subfols = $m->litfols($lihs,$nom);
	my @subnoms = map { $m->sup8fol2nom($lihs,$nom,$_) } @subfols;
	$i = 0;
	$ln = '';
	for my $subnom (@subnoms) {
	    my $subval = $lihs->getstr($subnom);
	    last unless $subval;
	    if (@lnoms) {
		$ln = $lnoms[$i];
		$i++;
		last unless $ln;
	    } else {
		$i++;
		$ln =  'L' . $i;
	    };
	    $rek->{$ln} = $subval;
	};
	push @reks, $rek;
    };
    $m->dpop1($prog,\@reks);
}

=head4 rekputfayl

Write all the record's value pairs into a textfile; e.g.

    
means while write everything into the file of the given name in a form like

    nom = equi
    lng = en
    pgs = 0,1,2,3,4,5
    nums = 000,001,002,003,004,005
    amd = 2016-05-25
    tit = Directory of Recognition of Equivalence

=cut

sub rekputfayl {
    my $m = shift;
    $m->dpuc('rekputfayl', @_);
    my $rek = shift;
    my $fo = shift;
    my $s = join "\n", map { join ' = ', $_, $rek->{$_} } keys %$rek;
    $m->fayl_put_str($fo, $s);
    $m->vecho('written string %t to file %f', t => $s, f => $fo);
    $m->dpop1('rekputfayl', 1);
    return 1;
};
sub dbmeta {
    my $prog = 'dbmeta';my $m = shift;$m->dpuc($prog,@_);
    my $dok = shift;
    my $nom = shift;
    my ($lang) = $m->doklangs($dok, shift);
    my $tmplvars = $m->{doklvars}->{$dok}->{$lang};
  DBMETA_TMPLVARS: {
      last if $tmplvars and 'HASH' eq ref $tmplvars;
      $tmplvars = $m->dok2tmplvars(d => $dok, l => $lang);
      $m->ustop('failed to read tmplvars for dok %d and lang %l', d => $dok, l => $lang) unless $tmplvars and 'HASH' eq ref $tmplvars;
      $m->{doklvars}->{$dok}->{$lang} = $tmplvars;
    };
    my $txt = $tmplvars->{$nom};
    $m->dpop1($prog,$txt);return $txt
};

=head1 Function tmplsrcfayl

Return full pathname for basename arg1, obtained by searching through the template directories.
The returned pathname has been found to exist.
arg2ff is a list of options containing 
    C<insist> ends with error if nothing is found
    C<tmpldir>  limits search to template directory tree, requires that something must be specifically defined for this template (this option is becoming obsolete, was used for 'dok_tmpl.txt' until 2015-11-09)

=cut

sub tmplsrcfayl {
    my ($m,$prog,$ffol,%opts) = shift->ppuc('tmplsrcfayl',@_);
    my $fayl = $m->template_include_file($ffol);
    return $m->dpop1($prog,$fayl) if $fayl;
    my $insist = $opts{insist};
    $m->ustop_or_warn(!$insist, notsrcf => t => $ffol);
    $m->dpop($prog);
};

=head4 setpvar( $nom => $val ) => 1: make a value unchangeable so that reading out dbmvars files will not alter it
Used for values that belong the the control structure of our process.
To the extent that they are scalar, they also overwrite $lihs->{vars} and do so in case of new subdocuments via mlmk/mldk/lihs_arbvar/progvars_set_lihs_vars.
setpvar should be used only for variables that are managed by the program regardless of what vocabulary files say.
This overwrites the sipmler version of A2E::Prog that A2E::Dokfs uses.
With option typ => '@' or typ => '%', we get a better replacement for setplst and setphac because lihs_put makes use of this argument too.

=cut

sub progvar_invalid_p {
    my ($m,$lihs,$nom,$val,$vahr) = @_;
    my $validp = $m->{var2validp}->{$nom};
    return if !$validp;
    $vahr ||= {};
    eval { &$validp($m,$lihs,$val,$vahr) };
    $@ };
sub setpvar {
    my ($m,$prog,$nom,$val,%opts) = shift->ppuc('setpvar',@_);
    (my $lihs,%opts) = $m->shifthash('lihs',%opts);
    (my $vahr,%opts) = $m->shifthash('vahr',%opts);
    $lihs = $m->get_lihs($lihs);
    my $typ = $opts{typ};
    $nom = $m->ali2nom($nom);
    $val = $m->ali2val($nom,$val);
  KO: {
      last if $opts{nochk};
      my $KO = $m->progvar_invalid_p($lihs,$nom,$val,$vahr);
      last unless $KO;
      $m->pstop('attempting to set %n to wrong value %v: %m', n => $nom, v => $val, m => $KO);
    };
    $m->{progvars}->{$nom} = $val;
    $m->lihs_put($lihs,$nom,$val,%opts);
  HSPRVARP: {
      my $k = $m->{progvars2hsprintf}->{$nom};
      last unless $k;
      # $m->{$nom} = $val; # file:/phm/19/05/30/sig/_lng.phm_pub_sig190530.txt::pvars2hsprf
      $lihs->{vars}->{'__'.$nom} = $val unless ref $val;
      $m->pstop('refusing to set hsprintf_repl key %k variable %n to null value', k => $k, n => $nom) unless $val;
      $m->{hsprintf_repl}->{$k} = $val;
    };
    if ($typ) {
	return $m->dpop($prog,$m->ref2list($val)) if '@' eq $typ;
	return $m->dpop($prog,$m->ref2hash($val)) if '%' eq $typ;
    };
    $m->dpop1($prog,$val);
};
sub setpvar_mini { 
    my ($m,$prog,$lihs,$k,$v) = shift->lpuc('setpvar_mini',@_);
    $m->dpop($prog) unless defined $v;
    $m->{progvars}->{$k} = $v;
    $m->lihs_put($lihs,$k,$v);
    $m->dpop1($prog,$v);
}
sub setplst {
    my ($m,$nom,$val,%opts) = @_;
    $m->setpvar($nom,$val, typ => '@', %opts);
};
sub setphac {
    my ($m,$nom,$val,%opts) = @_;
    $m->setpvar($nom,$val, typ => '%', %opts);
};
sub pvar {
    my ($m,$nom,$insist,$lihs) = @_;$lihs = $m->get_lihs($lihs);
    $nom = $m->ali2nom($nom);
    my $val = $m->{progvars}->{$nom};
  VAL: {
      last if defined $val;
      $val = $lihs->getarb($nom);
      last if defined $val;
      $val = $m->{cache}->{$nom};
      last if defined $val;
      $val = $m->{$nom};
    };
  ALI: {
      last if ref $val;
      $val = $m->ali2val($nom,$val);
    };
    $m->pstop(nomisref => n => $nom) if ref $nom;
    $m->pstop(badnom => n => $nom) unless $nom =~ m(\A\w+\Z);
    $m->pstop(pvardef => n => $nom) if $insist and not defined $val;
    $val;
};
sub plst {
    my ($m,$nom,$insist,$lihs) = @_;$lihs = $m->get_lihs($lihs);
    $nom = $m->ali2nom($nom);
    my $lstr = $m->{progvars}->{$nom} || $lihs->getlstr($nom) || $m->{cache}->{$nom} || $m->{$nom};
    do { $m->ustop_or_warn(!$insist, pvardef => n => $nom);return } unless defined $lstr;
    $m->ustop(pvarlstp => n => $nom) unless 'ARRAY' eq ref $lstr;
    @$lstr;
};
sub phac {
    my ($m,$nom,$insist,$lihs) = @_;$lihs = $m->get_lihs($lihs);
    $nom = $m->ali2nom($nom);
    my $hacr = $m->{progvars}->{$nom} || $lihs->gethacr($nom) || $m->{cache}->{$nom} || $m->{$nom};
    do { $m->ustop_or_warn(!$insist, pvardef => n => $nom);return } unless defined $hacr;
    $m->ustop(pvarhacp => n => $nom) unless 'HASH' eq ref $hacr;
    %$hacr;
};

=head3 Function lihs_set_sems

Split sems (first mktdir argument e.g. w30,27,31) and transfer w30 to __sem1, 27 to __sem2, 31 to __sem3, delete any existing __sem4 and following.

=cut

sub lihs_set_sems {
    my ($m,$prog,$lihs,$sem) = shift->lpuc('lihs_set_sems',@_);
    $m->pstop(funnovar => v => 'sems') unless $sem;
    my @sems = grep { $_ } split m([\_\W]), $sem;
    $m->setplst(sems => \@sems, lihs => $lihs);
    $m->lihs_putarb($lihs, sems => \@sems);
    my $nom = '';
    my $i = 0;
    my $val = '';
  SETSEMS: { # set tmplvars __sem? from @sems
      $val = $sems[$i];
      last unless defined $val;
      $i++;
      $nom = 'sem' . $i;
      $m->setpvar($nom => $val, lihs => $lihs);
      $nom = '__'.$nom;
      $m->setpvar($nom => $val, lihs => $lihs);
      redo;
    };
  DELARGS: { # remove possible trailing arguments from previous invocations
      $i++;
      foreach my $nom (('sem'.$i, '__sem'.$i)) {
	  $lihs->del($nom);
      };
    };
    $m->dpop($prog,@sems);
};

=head3 $m->fnom2ffol('tml') => ('dok_%f_tmpl.txt', '%pdok.%d.%f.txt', insist => 1, mode => tt, tag => '*')

Read out info from konfig->fnom2ffol as in file:/adv/perl/A2E/dokfs.konf::fnom2ffol

<nov id=grupfnom>file:/phm/21/04/17/sig/_lng.phm_pub_sig210417.txt::grupfnom</nov>

=cut
sub tmpl_fnom2ffol {
    my ($m, $prog, $fnom) = shift->ppuc('fnom2ffol',@_);
    my $ffolrek = $m->{konfig}->fnom2ffol->{$fnom};
    my ($fsrc, $fobj, @rest) = split m(\s+), $ffolrek;
    return unless $fobj;
    my ($nom, $val, $opt, %fops);
    foreach $opt (@rest) { 
	 ($nom, $val) = split ':', $opt;
	 $m->ustop('empty option %o for fnom %f in fnom2ffol', f => $fnom, o => $opt) unless length $val;
	 $fops{$nom} = $val;
    };
    $m->dpop($prog,$fsrc,$fobj,%fops);
};
sub grup_fnom2ffol {
    my ($m,$prog,$lihs,$fnom,%opts) = shift->opuc('fnom2ffol',1,@_);
    my $fobj = $lihs->getstr($fnom . '_fobj');
    return $m->dpop($prog) if !$fobj;
    $fobj = $m->grupfill($fobj);
    my $fsrc = $lihs->getstr($fnom . '_fsrc') || sprintf '%s_tmpl.txt', $fnom;
    my %fops = %{$m->lihs_gettypr($lihs, $fnom . '_fops', '%')};
    $m->dpop($prog,$fsrc,$fobj,%fops);
}
sub fnom2ffol {
    my ($m,$prog,$lihs,$fnom,%opts) = shift->opuc('fnom2ffol',1,@_);
    my ($fsrc,$fobj,%fops) = $m->grup_fnom2ffol($fnom, lihs => $lihs, %opts);
    return $m->dpop($prog,$fsrc,$fobj,%fops) if $fobj;
    ($fsrc,$fobj,%fops) = $m->tmpl_fnom2ffol($fnom); 
    $m->pstop('leaf not found for fnom %f', f => $fnom) if !$fsrc;
    $m->dpop($prog,$fsrc,$fobj,%fops);
};
sub __fnom2fnon_r {
    my $prog = '__fnom2fnon_r';my $m = shift;$m->dpuc($prog,@_);
    my ($num,$fnom) = @_;
    my @ffoo = split m(\s+), $m->{konfig}->fnom2ffol->{$fnom};
    my $fnon_r = $ffoo[$num];
    $m->dpop1($prog,$fnon_r);return $fnon_r;
};
sub __fnom2fnon {
    my $prog = '__fnom2fnon';my $m = shift;$m->dpuc($prog,@_);
    my ($num,$fnom,%opts) = @_;
    my $fnon_r = $m->__fnom2fnon_r($num,$fnom);
    my $fnon = $m->hsprintf($fnon_r,%opts);
    $m->dpop1($prog,$fnon);return $fnon;
};
=head4 $m->fnom2fsrc('tml') => 'dok_%f_tmpl.txt', fnom2fobj => %pdok.%d.%f.txt
=cut

sub fnom2fsrc_r { return shift->__fnom2fnon_r(0,@_) };
sub fnom2fobj_r { return shift->__fnom2fnon_r(1,@_) };
sub fnom2fsrc { return shift->__fnom2fnon(0,@_) };
sub fnom2fobj { return shift->__fnom2fnon(1,@_) };

=head3 $m->fnom_tmplsrcfayl('tml') => '/adv/tmpl/a2e/oas_akt_aammdd/html/dok_html_tmpl.txt'
=cut
sub fnom_tmplsrcfayl {
    my $prog = 'fnom_tmplsrcfayl';my $m = shift;$m->dpuc($prog,@_);
    my ($fnom,%tups) = @_;
    my $fsrc = $m->fnom2fsrc($fnom);
    my $fi = $m->ffol_tmplsrcfayl($fsrc,%tups);
    $m->dpop1($prog,$fi);return $fi;
};    

sub doklfil {
    my $prog = 'doklfil';my $m = shift;$m->dpuc($prog,@_);
    $m->{konfig}->infaylz($_) for @_;
    $m->mlht_set_dokvars;
    $m->proc_faylz_nur_tmpl;
    $m->dpop1($prog,1);
};

sub lihs_initvars {
    my ($m,$prog,$lihs) = shift->lpuc('A2E::MLHT::lihs_initvars',@_);
    $m->tmplfil_lihs_initvars($lihs);
    $m->dpop1($prog,1);
};

=head1 The End
=cut
1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/MLHT.pm.tmpl"
# End:
