package A2E::MLMK;
use utf8;

=head1 NAME

A2E::MLMK

=head1 DESCRIPTION

Make multilingual hypertext systems, especially subdocument trees, according to the specifications, step by step with make-style targets, concretizing and checking blueprints before taking action on them.

This shall replace A2E::Mktdir(3).

=head1 PREREQUISITES

    A2E::MLDK(3)

=cut

use base 'A2E::MLDK', 'A2E::Subdoks', 'A2E::Mbox2dok';
use strict;use warnings 'all';no warnings 'uninitialized';
use Data::Dumper;
use File::Copy;
use File::Touch;
use utf8;
use List::MoreUtils qw(uniq);
use Encode qw(_utf8_on decode encode);
use English -no_match_vars;
use POSIX qw(strftime setgid getgrnam);
use File::Spec::Functions qw(splitdir splitpath catdir catfile);
use File::MkTemp;
$Template::BINMODE = ':utf8';use Template;
use Term::ReadLine;
use Cwd qw(getcwd);
our $novitas = $ENV{NOVITAS};
sub valid_trg {
    my ($m,$lihs,$trg,$vahr) = @_;
    unless ($trg) { if ($vahr->{__flagp__}->{let}) { return 'prob' } else { $m->ustop('empty target') } };
    $m->pstop('no targets list defined') unless 'ARRAY' eq ref $m->{targets}; 
    $m->ustop('need a valid target i.e. not %t but one of %L', t => $trg, L => $m->{targets}) unless grep { $_ eq $trg } @{$m->{targets}};
    $trg;
};
sub defvars {
    my ($m,$prog,$lihs) = shift->qlpuc('A2E::MLMK::defvars',@_) or return;
    $lihs->{vars}->{progver} ||= '0.0.1';
=head1 VERSION

This describes version B<0.0.1> of A2E::MLMK.

=cut
    $m->SUPER::defvars;
    $m->subdoks_defvars;
    # $m->tmplfil_defvars; included in A2E::MLHT
    $m->mbox2dok_defvars;
    $m->{lngtags} = {};
    $m->{progvars}->{$_} = [] for qw(suplangs);
    $m->setpvar(timestamp => $m->timestamp, lihs => $lihs);
=head1 OPTIONS

=over

=item --basurl I<url>

=cut
    $m->{konfig}->define('basurl=s');
=item --http I<http>

=cut

    $m->{konfig}->define('http=s');

=item --langnoms de=Deutsch

=cut

    $m->cache_define('langnoms', '%');

=item --lang0 de

=cut

    $m->{konfig}->define('lang0=s');

=item --minimal

=cut

    $m->{konfig}->define('minimal|minim!');

=item --topsty tra

Document style.  By setting this you imply that the document is a top node.  Otherwise the style is inherited from the topmost preceding node (topdok).

=cut

    $m->konfig_define(['topsty', 'top'], 's');

=item --mailuser adv

=cut

    $m->konfig_define('mailuser', 's');

=item --conv deplate

=cut

    $m->konfig_define('conv', 's', 'deplate');

=item --menuitem_template

Obsolete, maybe still used for legacy directories at a2e?

=cut

    $m->{konfig}->define('menuitem_template=s%');

=item --mailhost a2e.de

=cut

    $m->{konfig}->define('mailhost=s');

=item --mailadr adv@a2e.de

=cut

    $m->{konfig}->define('mailadr=s');

# =item --makefile_tmpl makefile_tmpl.txt
# 
# =cut
#
#    $m->{konfig}->define('makefile_tmpl=s', { DEFAULT => 'makefile_tmpl.txt' });

=item --skrib_antefaylz

OBSOLETE
write language-specific parameter files, named as specified in antefnom

=cut

    $m->{konfig}->define('skrib_antefaylz!');

=item --tmplvars|--tv|-V mm=01 -V aa=09 -V thema=Italia,de:Italien,en:Italy,fr:Italie

Extra template variables to be set at the beginning and written into the local --antefnom0 (@pre.mydok.txt) or --antefnom (pre.mydok.en.txt) file. 
These variables are also written into the language-neutral --antefnom0 file, so that they precede any vocabulary definitions at compile-time.
The values of these variables can, even though this is not recommended, be multilingual, using the syntax indicated above.  In this case, by implication, skrib_antefaylz is true.
We recommend monolingual --tmplvars values because these variables are, in contrast to the langlexvars, intended to be more closely related to process control than to the textual content of the documents.

=cut

    $m->{konfig}->define('tmplvars|tv|V=s%'); 

=item --langlexvars|llv can=";canned definition;abgepackte Definition" --llv can2='|another ${can}|noch ne ${can}'

extra multilingual textchunk groups to be written into vocubulary files postfayl such as _lng.mydok.en.txt.

These are evaluated to become 

	$m->{langlexvars}->{can}->{en} = 'canned definition'
	$m->{langlexvars}->{can}->{de} = 'abgepackte Definition'

, provided that the languages of the document to be created are en and de.

The value is retrieved&cached using the function C<langlexvar>.

As of 2009-12-12, it seems that we may never want to use this, given that these variables are read late and belong to the text level rather than the program operation level.  For the latter purpose, we already have --tmplvars|-V.

=cut
    
    $m->{konfig}->define('langlexvars|llv=s%');

=item --lexvars|--lv aa=09 --lv mm=01

Extra variables to be written into language-neutral vocabulary file (unless that already exists).
This option is not very useful because the files written to the antefnom (parameter element) file is more important.

=cut

    $m->{konfig}->define('lexvars|lv=s%');

=item --tmplang_outfmts multemp="html pdf" mlwebdb="html"

=cut

    $m->{konfig}->define('tmplang_outfmts|outfmts=s%');

=item --tratoks_extra 'kws'

=cut

    $m->{konfig}->define('tratoks_extra=s@');
    $m->konfig_define('logourl', 's', '/a2e.jpg');
    $m->{konfig}->define('subsub!');
    my $testdir = catfile $ENV{HOME}, 'tdir';
    $m->konfig_define('testdir','s', $testdir);
    $m->konfayl('mlmk.konf');
    $m->{subsems_funs} = {};
    $m->dpop1('A2E::MLMK::defvars', 1);
    $m->{term} = new Term::ReadLine $m->{prognom};
    $m->{targets} = [ qw(prob help dir prob subs scap) ];
    $m->{var2validp}->{trg} = \&valid_trg;
    $m->qpop1($prog,1);
};

=head1 BUGS

None open at the moment

=head1 LIMITATIONS

=over

=item one document template set for everything

=back

=head1 SEE ALSO

    mktdir0(1)
    mktdir5(1)
    A2E::Dokfs(3)

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

    Copyright (c) 2008-9 Hartmut Pilch (phm)

    This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

Names in this program are based on English with International Latin spelling, e.g. 'kopi' for 'copy' and 'meikfayl' for 'makefile', as (to be) explained in

  http://a2e.de/alfabet/interlatin

with with occasional Esperanto (such as replacing words such as 'write'/'rait', that become unwieldy in interlatin spelling, with 'skrib')

  http://a2e.de/lang/eo

for which the same spelling is used.

=cut

=head2 Global switches and variables for internal configuration

=head3 @*_tmplvars et al

Template variables that are not reset or treated specially when template variables are read in from files.

=cut

our @tratoks = ('fol', 'lab', 'tit', 'sut', 'des');

# DOKTYP
# TODO, cleanup, allow more configuration by user, as formerly done with konfig_define(tmplvars_ro => 's@')
# variables that are not taken from the supdok variable store but need to be treated specially.
# these are typically specified in lang_pre.txt or _pre.%d.txt files with '?=' assignment statements and will therefore retain their obsolete values if not wiped out before reading
# variables that are no longer read after the first generation-run of one document (e.g. in a second language)
our @extra_progvars = qw(aa aaaa mm dd ddmax KA KAM KAMD monat);
# remove these from progvars when changing to a new document
our @doktypsupvars = qw(sems2dok sems2fol); # tmpldir variables for the parent document
our @doktypsubvars = qw(tmpl_fnoms sems2subflat sems2subsems sems2subvars sems2subtmpl semslit fol); # tmpldir variables for subdocuments, worth editing in _pre.%d.txt
our %doktypprevars = (grpdokslit => 'grpdoks', subdokslit => 'subdoks', no_subsems => '0', sems2subflat => '0');
our @doktypintlsts = qw(_tmplrels_ _reksekts_ _tmpltoks_ _antevars_ slngs); # internal list variables that are set in a tmpldir or for other reasons should be deleted before a document begins
our @subdoklits_vars = qw(tmls langs); # reserved subdoklits variables that must not be overwritten from progvars
our @starthacs = qw(nom2lits); # hashes that start empty for each subdocument file:/phm/19/06/10/sig/_lng.phm_pub_sig190610.txt::subdoklits; 190831 removed 'subdoklits'
our @heredhacs = qw(progvars tmplarbs); # hashes that are localized but inherit initival values from parent doc; 190819 removed tmplvars,tmpllsts,tmplhacs; 190831 removed 'grupvars'
# variables that are deleted before any further run (in another language), so that they are freshly read once even if specified with ?=
# our @doktypdelvars = (@tratoks, 'thema', 'AA', 'MM', 'DD', 'lab_valids', 'tit_valids', 'sut_valids', 'des_valids');
our @doktypdelvars = (@tratoks, map { join '_', $_, 'valids' } @tratoks);
our @doktyplokvars = (qw(tmls tmplang tem dokstem supstem aa mm dd AA MM DD ww wd));
# variables that need to be set in any document type definition /tmpl/a2e/*/lang_pre.txt: if any of these is null the document type is incomplete
our @nonnull_tmplvars = qw(sems2dok sems2fol);
# string variables that are copied from the object $m into the template variables store $lihs->{vars}
our @konfig_tmplvars = qw(topsty logourl tmpl);
# END DOKTYP

=head2 Finishup and postprocessing

Add accumulated files and some further ones to Subversion repository.
Write metadata to the database. 
Set up a link to an mbox if found.

=head2 Top level functions

=head3 Function lastkonfig

=cut

sub lastkonfig {
    my ($m,$prog,$lihs) = shift->qlpuc('A2E::MLMK::lastkonfig',@_) or return;
    $m->SUPER::lastkonfig($lihs);
    $m->konfsetif(skrib_antefaylz => $m->{konfig}->topsty); # always write language-specific parameter files if we are at top document level.
  LANG_TMPLVARS: { # transfer multilingual values of variable settings such as -V thema=Italia,de:Italien,en:Italy,fr:Italie from $m->{konfig}->tmplvars into $m->{lang_tmplvars} 
      while (my ($nom, $val) = each %{$m->{konfig}->tmplvars}) {
	  my ($val0, @tups) = split ',', $val;
	  next unless @tups;
	  $m->{konfig}->set(skrib_antefaylz => 1);
	  $m->{lang_tmplvars} ||= {};
	  # $m->{konfig}->tmplvars->{$nom} = $val0;
	  $lihs->putstr($nom => $val0);
	  foreach my $tup (@tups) {
	      my ($ll, $vall) = $tup =~ m(\A(\w+)\:(.+)\Z);
	      next unless $vall;
	      $m->{lang_tmplvars}->{$ll} ||= {};
	      $m->{lang_tmplvars}->{$ll}->{$nom} = $vall;
	  };
      };
    };
    $m->qpop1($prog,1);
};

=head3 Function postkonfig

=cut
our %MSGS = (
    funnoopt => 'no such option %o',
    dirnedir => 'no such directory %d',
    sems2fol => q{cannot obtain fol by 'call'ing sems2fol %v on %S},
    novarinprc => 'no %v found in %p', 
    nolangs => 'no languages to write into file %f',
    no_lng_tmpl => 'mi ne trovis na lng_tmpl por produki na %t',
    renfol => 'kann %f nicht zu %t umbenennen',
    foldekz => 'Verzeichnis %f existiert bereits.  Gut so.',
    nochdst => 'failed to chdir to dirstem %d',
    oldfol => 'subdocument %d of type %t already exists at leaf %f',
    novfol => 'subdocument %d of type %t to be created at leaf %f',
    dir2url => 'unable to set url from dir %d',
    mkfol => '%(P|Made|Make) leaf %f i.e. dok %d of template type %t %(Y|identical|subsequent|paraordinate|subordinate) to %u at dir %D',
    nofol => 'No subdocuments to make for document %d',
    tmpldirvarnil => 'variable %n undefined after reading tmpldir %t' );

sub defin_mlmk_tmplfuns { # toplevel template functions for generating makefiles, did not turn out to be as useful as hoped
    # 創建於 file:/phm/19/05/19/sig/_lng.phm_pub_sig190519.txt::96
    my ($m,$prog,$lihs) = shift->lpuc('defin_mlmk_tmplfuns',@_);
    my $c = $m->template_context;
    $c->define_vmethod('LIST', 'subsems', sub { my $semsr = shift;my $lihs = $m->get_lihs;my @sems = 'ARRAY' eq ref $semsr ? @$semsr : $lihs->getlst('sems');$m->subsems($lihs,@sems) }); 
    $m->dpop1($prog,1);
};

sub postkonfig {
    my ($m,$prog,$lihs) = shift->qlpuc('A2E::MLMK::postkonfig',@_) or return;
    $m->SUPER::postkonfig($lihs);
    $m->mldk_postkonfig($lihs);
    # $m->tcek_mailadr_varz; # file:/adv/perl/A2E/OLD/MLMK.old::tcek_mailadr_varz
  KONFPUT: { # OBSOLETE?
      last if $novitas;
      $lihs->putstr(emacs_timestamp => $m->emacs_timestamp, nonstorp => 1);
      foreach my $nom (@konfig_tmplvars) {
	  my $val = $m->{konfig}->get($nom);
	  next unless $val;
	  $lihs->putstr($nom => $val, nonstorp => 1) 
      };
    };
    $m->prog_msgdefs('mlmk',%MSGS);
    $m->defin_mlmk_tmplfuns($lihs);
    $m->setpvar(batch => 1, lihs => $lihs); # 臨時
    # $m->{sarb}->{lihs} = $lihs;
    $m->set_topdoks($lihs);
    $m->qpop1($prog,1);
};

=head2 General Basic Routines

=head3 &set_dirstem(subflat) => dirstem: 

Set the parent directory dirstem to which the leaf is to be appended.

=cut

sub set_dirstem {
    my ($m,$prog,$lihs,$subflat,$pwd) = shift->lpuc('set_dirstem',@_);
    $m->setpvar(pwd => $pwd, lihs => $lihs);
    my $dirstem = $pwd;
    $m->setpvar(supdir => $pwd, lihs => $lihs);
    my @dirlist = splitdir $pwd;
    if ($subflat) {
	pop @dirlist;
	$dirstem = catdir @dirlist;
    };
    $m->setpvar(dirstem => $dirstem, lihs => $lihs);
    $m->dpop1($prog,$dirstem);
};
=head3 Function sems_call

Evaluate an expression as stored in special template variables sems2dok, sems2fol, sems2subsems, sems2subvars, sems2subtmpl.
See examples below.
TODO: debug A2E::SArb and its use by A2E::Tmplfil et al up to A2E::MLMK, so as to find out why $m->{sarb}->{vars} at some point diverges from $lihs->{vars}, retaining old values, resulting in wrong result for for a sems2dok expression '${dokstem}${aamm}'.
See e.g. report in file:/phm/19/05/28/sig/_lng.phm_pub_sig190528.txt::sarbvars
Currently we need to clumsily check for this divergence and correct it at every invocation.

=cut

sub sems_call {
    my ($m,$nom,$lihs,@sems) = @_;$lihs = $m->get_lihs($lihs);$m->dpuc($nom,@sems);
    do { $m->pwarn('no sems arguments given');@sems = $m->plst(sems => 1) } unless @sems;
    my $exp = $lihs->getval($nom);
    return $m->wpop($nom,[d => 'no tmplvar %v known', v => $nom]) if not defined $exp;
    my $val = $m->{sarb}->val_args_call($exp, undef, @sems);
    $m->pstop('var %n value %e called to naught with sems %S', n => $nom, e => $exp, S => \@sems) if not defined $val;
    $m->dpop1($nom,$val);
};

=head3 Function sems2dok

A typical use might be

    $m->sems2dok('victoria', '31')  ==> victoria090731

where two semantic features, given as arguments, are translated into a
document symbol according to a rule specified by the user via template
variables like sems2dok or other mechanisms to be implemented here.

=cut

sub sems2dok { 
    shift->sems_call('sems2dok', @_) };

=head3 Function sems2fol

Like sems2dok, but generate current directory leaf name rather than document symbol.

=cut

sub sems2fol { 
    shift->sems_call('sems2fol', @_) };

=head3 $tmpl = $m->riid_tmpldir_lits($suplihs,$lihs,$tmpltups,$supdok,@sems): Set $lihs of child document based on its tmpl/doktyp and $suplihs

$suplihs :: parent vocabulary
$lihs :: vocabulary to be set
$subflat :: whether document is in sister directory
@sems :: signifiers from which leaf name is constructed
$tmpl :: document type (tmpl/doktyp) of configured $lihs

Most info is returned by setting $lihs.
Much of the info is a temporary kludgy solution.
TODO:
 - reduce, 
   * potentially eliminate reading of lits (language-dependent stuff) from lang.??.txt and lang.txt
 - transfer to userspace, 
   * variables like 'KA' which are expected to have been preset must not be hardwired here

=cut

sub riid_tmpldir_lits {
    my ($m,$prog,$suplihs,$lihs,$tmpltups,$supdok,@sems) = shift->ppuc('riid_tmpldir_lits',@_);$suplihs = $m->get_lihs($suplihs);$lihs = $m->get_lihs($lihs);
    my $subflat = $suplihs->{vars}->{subflat};
    my $subtmpl = $m->sems2subtmpl($suplihs,@sems);
    my $subvars = $m->subvars($suplihs,@sems); # { dokstem => 'phm_pub_moral1805', no_subsems => 1 }
    my $tmpl = $suplihs->getstr('tmpl');
    return $m->dpop($prog) if $subtmpl eq $tmpl;
    $tmpl = $subtmpl;
    $m->setptvar(tmpl => $tmpl, lihs => $lihs); # 待査: file:/sig/oas/19/07/akt/.lits/prae.hac::69 無 __tmpl
    my $tmpldir = $m->tmpl_set_tmpldir($lihs,$tmpl);
    my @faylz = $m->pre_faylz($supdok, typ_min => 2, tmpl => $tmpl, norank => 1);
    $m->set_templates($lihs);
    # $m->rekurs_del_tmplvars(minim => 1);
    # $m->lang_set_tmplvars($m->plst('suplangs')->[0], $m->pvar('supdok'), tmpl => $m->pvar('tmpl'));
    # file:/adv/perl/A2E/OLD/Mktdir.old::lang_set_tmplvars
    my $initvars = sub { # partially obsolete
	do { delete $m->{progvars}->{$_};delete $lihs->{vars}->{$_} } for @doktypsupvars, @doktypsubvars;
	do { delete $m->{progvars}->{$_};$lihs->{lsts}->{$_} = [] } for @doktypintlsts;
	$lihs->del($_) for @doktypdelvars; # OLD
	while (my ($k,$v) = each %doktypprevars) { $m->lihs_put($lihs,$k,$v) if defined $v };
	while (my ($k,$v) = each %$subvars) { next unless defined $v;$lihs->putstr($k,$v);$m->{progvars}->{$k} = $v }; # protect against overwriting
	$lihs->{lsts}->{_antevars_} = [ keys %$subvars ];
	foreach my $nom (@doktyplokvars) { my $val = $m->{progvars}->{$nom};next unless $val;$m->pwarn('doktyplokvar %n must remain readable, deleting it with value %v from progvars', n => $nom, v => $val);delete $m->{progvars}->{$nom} }; 
    };
    &$initvars;
    my $kopilits = sub { # OLD: stuff that may have to be transferred from sublihs to lihs before tmpl-related tmkl files are read
	my @kopivars = ('KA');
	push @kopivars, map { 'mm' . sprintf '%02d', $_ } 01..12;
	push @kopivars, map { 'dd' . $_ } 0..7;
	foreach my $k (@kopivars) { my $v = $suplihs->getval($k);next unless $v;$m->lihs_put($lihs,$k,$v) };
    };
    # &$kopilits;
    foreach my $fayl (@faylz) { # file:/phm/19/12/23/sig/_log17adv.phm_pub_sig191223.txt::63303
	my $redefp = $fayl =~ m(\b$tmpl\b);
	$m->riid_tmplvars_fayl($fayl, msglini => 1, lihs => $lihs, redefp => $redefp, putvars => $tmpltups);
	# last if $redefp; # redefp_quit:可疑、暫除20191224
    };
    my $setpvars = sub { # transfer document creation related key variables to progvars, may be largely obsolete
	$lihs->{vars}->{tmpl_fnoms} ||= $m->{konfig}->tmpl_fnoms;
	# my $semslit = $lihs->{vars}->{semslit} || ($subflat ? $lihs->{vars}->{grpdokslit} : $lihs->{vars}->{subdokslit});
	# $m->setptvar(semslit => $semslit, lihs => $lihs) if $semslit;
	foreach my $nom (@doktypsupvars,@doktypsubvars) { # obsolete?
	    my $val = $lihs->{lsts}->{$nom} || $lihs->getval($nom);
	    do { $m->decho(tmpldirvarnil => n => $nom, t => $tmpldir);next } unless defined $val;
	    # $m->{progvars}->{$nom} = $val;
	};
	$m->setptvar(doktypsubvars => \@doktypsubvars, lihs => $lihs); # needed for generating _pre.*.txt files
    };
    &$setpvars;
    $m->dpop1($prog,$tmpl);
};
=head3 $pretyp = $m->last_supgrp_pretyp($dok): last subordinating node in chain
find last $m->{dokprop}->{$dok}->{typ} that was 2 or 3.
=cut

sub last_supgrp_pretyp {
    my ($m,$dok,$predokp) = @_;
    $predokp->{$dok} = 1; # 循環計數器 file:/phm/21/04/14/sig/_lng.phm_pub_sig210414.txt::predokp
    my $pretyp = $m->{dokprop}->{$dok}->{typ};
    my $predok = $m->{dokprop}->{$dok}->{dok};
    $m->uwarn('predok %d is already in ancestry %A', d => $predok, A => $predokp) if $predokp->{$predok};
    return $pretyp if $pretyp > 1 or !$predok or $predokp->{$predok};
    $m->last_supgrp_pretyp($predok,$predokp);
};
=head3 ($predok,$pretyp) = $m->getpredoktyp($subflat,$supdok,$spredok): basic info for inserting next dok into dokprop chain

The $m->{dokprop}->{$dok}->{(?:dok|typ)} info thus generated by sup7lihs is still tentative and subject to modification once the document is really created and registered in $dirstem/_(?:grp|sub)doks etc.  This is because we are not aware of any existing documents yet.

=cut
sub getpredoktyp {
    my ($m,$subflat,$supdok,$predok) = @_;
    my $pretyp = $subflat ? 2 : 3;
    return $supdok, $pretyp unless $predok;
    my $last_pretyp = $m->last_supgrp_pretyp($predok);
    $pretyp = !$predok ? $last_pretyp : ($last_pretyp == $pretyp) ? 1 : $pretyp;
    $predok = ($pretyp > 1) ? $supdok : $predok;
    ($predok,$pretyp);
};

=head3 ($dirstem,$dokstyp,$fol,$dok,$tmpl,$dir) = $m->sup7lihs($suplihs,$lihs,$predok,@sems)

Spawn a supdirectory/vocabulary $lihs from parent $suplihs.
Derive from @sems and info of $suplihs.
Read tmpl related vocabulary of doktyp.
Etymology: '7' is sept/sette i.e. 'set', similar to '2' but assigns the value that it generates.

=cut
sub sup7lihs { # was 'set_dir_varz'
    my ($m,$prog,$suplihs,$lihs,$supdok,$predok,$probp,@sems) = shift->lpuc('sup7lihs',@_);
    my $subflat = $suplihs->{vars}->{subflat};$lihs->{vars}->{subflat} = $subflat;
    my $dokstyp = $subflat ? 'grpdoks' : 'subdoks';
    ($predok,my $pretyp) = $m->getpredoktyp($subflat,$supdok,$predok);
    my $kopi_progvars = sub { # copy variables that reflect the state of current processing and thus are similar to those set below in set_dir_varz below
	my @vars = (qw(hodie lang lang0 langlands langs privat progver srcfpfx supdir supdok suplangs svn_add_ok timestamp test_modus tmls tmplang topdir topdok topurl verbose prog_time prog_name prog_user));
	foreach my $nom (@vars) {
	    my $val = $m->{progvars}->{$nom};
	    $m->lihs_put($lihs, $nom, $val) if defined $val;
	};
	1;
    };
    # <nov id=kopiarbs>file:/phm/21/04/19/sig/_lng.phm_pub_sig210419.txt::kopiarbs
    foreach my $treg (qw(vrbs lsts hacs)) { %{$lihs->{$treg}} = %{$suplihs->{$treg}} }; # </nov>
    &$kopi_progvars;
    my @tupsregs = qw(toks); # 'prae' contains toxic stuff like __dir__, 'toks' contains 'aaaa' etc
    $suplihs->get_tups($_, stor => $lihs->{vars}) for @tupsregs;
    @tupsregs = qw(); #
    my @nomsregs = qw(); # qw(lstp hacs);
    my $kopi_arbvars = sub { # copy tokens and trunks from above
	# foreach my $treg (@tupsregs) { $suplihs->get_tups($treg, stor => $lihs->{$treg}) };
	# foreach my $treg (@nomsregs) { $suplihs->get_noms($treg, stor => $lihs->{$treg}) };
	foreach my $var (values %{$lihs->{vrbs}}) {
	    next if $lihs->{vars}->{$var};
	    my $val = $suplihs->getstr($var);
	    $lihs->putstr($var,$val) if $val;
	};
	1;
    };
    &$kopi_arbvars;
    my $tmpltups = {};
    $lihs->addrolitsdirs(litsdir => $suplihs->{vars}->{__litsdir__});
    my $tmpl = $m->riid_tmpldir_lits($suplihs,$lihs,$tmpltups,$supdok,@sems);
    $lihs->{__tmpl__} = $tmpl;
    my $dirstem = $m->set_dirstem($lihs,$subflat,$suplihs->{vars}->{__dir__});
    my $fol = $m->sems2fol($lihs,@sems) or $m->pstop('no fol derivable from sems'); # 供給 lang_pre.txt等使用
    my $dok = $m->sems2dok($lihs,@sems);
    my $dir = catdir $dirstem, $fol;
    my %prop = (dir => $dir, tmpl => $tmpl, dok => $predok, typ => $pretyp, fol => $fol); # enrich for reporting in report_predoks
    $m->{dokprop}->{$dok}->{$_} = $prop{$_} for keys %prop;
    STORP:{ # now allow disk storage and write initial files
	last if $probp;  # not if target is 'prob' i.e. we are only probing
	$lihs->setdir($dir);
	$lihs->addrolitsdirs(slitsdir => $suplihs->{vars}->{__litsdir__});
	$lihs->put_tups($_) for @tupsregs;
	$lihs->put_tups(tmpl => tups => $tmpltups);
	$lihs->put_noms($_) for @nomsregs;
    };
    my $set_dir_varz = sub { # set specific variables that distinguish this node from its ancestors
	$m->ustop('unable to set dok based on sems %S', S => $m->koniug_et(@sems)) unless $dok;
	# ($lihs,$dok,$dir) = $m->set_lihs_dok_dir_langs(dok => $dok, dir => $dir, lihs => $lihs); # 過早因尚無文件
	my $topsty = $m->{konfig}->topsty;
	my %dir_varz = (fol => $fol, dok => $dok, dir => $dir, tmpl => $tmpl, topsty => $topsty, sems => \@sems);
	while (my ($k, $v) = each %dir_varz) { $m->setpvar_mini($lihs,$k,$v) if defined $v };
	1;
    };
    &$set_dir_varz;
    $m->dpop($prog,$dirstem,$dokstyp,$fol,$dok,$tmpl,$dir); # file:/phm/22/01/02/sig/_lng.phm_pub_sig220102.txt::foldoktypdir
};

sub supld2ld {
    my ($m,$prog,$sem,$suplihs,$supdir,$supdok,$predok,$probp) = shift->ppuc('supld2ld',@_);
    my ($lihs,$oldlihs) = $m->lihs_new($m->{lihs});
    my @sems = $m->lihs_set_sems($lihs,$sem);
    $suplihs->{vars}->{subflat} = $m->subflat($suplihs,@sems);
    my ($dirstem,$dokstyp,$fol,$dok,$tmpl,$dir) = $m->sup7lihs($suplihs,$lihs,$supdok,$predok,$probp,@sems);
    $m->pstop('lihs object unfit for writing') if $lihs->{vars}->{__nolitsdir__} or !$lihs->{vars}->{__litsdir__};
    $m->dpop($prog,$lihs,$oldlihs,$dirstem,$dokstyp,$fol,$dok,$tmpl,$dir); # file:/phm/22/01/02/sig/_lng.phm_pub_sig220102.txt::foldoktypdir
}

=head4 $m->lihs_import_grupvars($lihs1,$dir,$grup,$lang)

Import subnode info from parent vocabulary $sublihs to present vocabulary $lihs.
The subnode name is found in $grup.
前身為 C<tmplvars_kk_dbmsup> 及 C<tmplvars_kk_subdoklits> file:/adv/perl/A2E/OLD/MLMK.old::tmplvars_kk_subdoklits
lvalnomp 幇助區别二種變量
TODO: dinstinguish litvars $tit,$sut,$des from momolingual datvars slng1 num etc which point to /\A\w+\Z/, divide __vars into __lits vs __dats
=cut
sub lihs_import_grupvars {
    my ($m,$prog,$lihs,$suplihs,$grup,$lang) = shift->lpuc('lihs_import_grupvars',@_);
    my %opts = (echo => 1);
    if ($lang) {
	$m->set_litsdir($suplihs, l => $lang);
	$opts{litstag} = 'l';
	$opts{flng} = 1; # <nov id=grupflng>file:/phm/21/04/17/sig/_lng.phm_pub_sig210417.txt::grupflng</a>
    };
    $m->{grupvars} = {};
    my $gvars = $m->grup_add_grupvars($grup, lihs => $suplihs, non_lits => 1, lang => $lang);
    return $m->wpop($prog,[d => 'parent document has no section called %g, so we fall back to the tmpl values', g => $grup]) if !$gvars;
    my @noms = ();
    foreach my $nom (keys %$gvars) {
	next if $nom =~ m(\A\_); # do not import _sup, _lvl, _frm etc
	# write lab tit sut des only if we are in _lng.$dok.$lang.txt, not _pre.$dok.txt
	next if $m->{konfig}->lvalnomp->{$nom} ? !$lang : $lang;
	my $val = $gvars->{$nom};
	next unless length $val;
	$m->lihs_put($lihs, $nom, $val, %opts);
	push @noms, $nom;
    };
    $m->dpop($prog,@noms);
};

=head2 set_dok_varz

Find out what kind of a location our leaf is (whether we can write there, whether it is a tree trunk), store some basic configuration values into memory, write others to the database.

The aliases and topdoks must be read into memory in any case.

The database must be edited only if we are creating a new document, i.e. not SAMDOK.

=cut

sub set_dok_varz {
    my ($m,$prog,$dok,$lihs) = shift->ppuc('set_dok_varz',@_);$lihs = $m->get_lihs($lihs);
    # M-x -op-p file:/phm/18/09/21/sig/_lng.phm_pub_sig180921.txt::kdok
    # 150211 $m->pstop('document %s has wrong name', $dok) if $ENV{DEBUG_SEMS2DOK} eq '150211' and not $dok eq 'phm_pub_adv_sig15';
    $m->setpvar(dok => $dok, lihs => $lihs);
    $m->set_alidoks(lihs => $lihs);
    $m->set_topdoks($lihs);
    # my $db_absurl = $m->db_dok2absurl($dok);
    # my $url = $m->pvar('url');
    # 去除與依存系統不符之文 file:/phm/18/10/15/sig/_lng.phm_pub_sig181015.txt::rm_oldurl
    $m->dpop1($prog,1);
};
sub set_dok_varz_db {
    my ($m,$prog,$lihs,%dir_varz) = shift->lpuc('set_dok_varz_db',@_);
    my $dpvar = sub { my ($k) = @_;$dir_varz{$k} || $m->pvar($k => 1) };   
    my $dok = &$dpvar('dok');
    my $url = $m->pvar(url => 1);
    my $fol = &$dpvar('fol');
    my $hodie = $m->pvar(hodie => 1);
    my $user = $m->pvar('user') || $m->{konfig}->user;
    my $tmpl = $m->pvar(tmpl => 1);
    my $inddok = $m->pvar('inddok');
    $m->put(mlhtdok => { dok => $dok }, dokurl => $url, doktmp=> 'now()');
    my $put_tmpl = sub { $m->put(dokprop => { dok => $dok }, doknom => $fol, dokvers => 1, dokdatum => $hodie, dokuid => $user, doktyp => $tmpl, lastmod => $hodie) };
  PUT: {
      eval { &$put_tmpl };
      last unless $@;
      $m->uwarn('registering unknown doktyp %t', t => $tmpl);
      $m->put(doktyp => { doktyp => $tmpl }, doktyprem => $dok, doktyptmp => 'now()');
      &$put_tmpl;
    };
  TOPSTY: {
      last unless $m->{konfig}->topsty;
      $m->put(dokflags => { dok => $dok }, varnom => 'topdok', varval => 'true'); # obsolete 2009-04-19
      $m->put(topdok => { topdok => $dok }, lastmod => $hodie, inddok=> $inddok); # dubious: inddok has no real value unless we build an index 2019-01-01, 2022-01-04 
    };
    $m->dpop1($prog,$dok);
};

=head2 Set further variables

Once we have identified the directory, document and position in the hierarchy,
we are ready to set some more variables and perform some checks.

=head2 Install the target directory with some basic files

=head3 Function mkchfol

Safely create a subdirectory after moving blocking data out of the way and go there but return the old directory.
Refrain from creating if the data indicates that the document node shouldn't be here.
Make sure we are svn-wise up to date.
Retrieve-store info about languages of any pre-existing dok (oldlangs).
This comes before pars_args.

=cut

sub mkchfol {
    my ($m,$prog,$lihs,$fol,$dirstem,$dokstyp,$dok,$supdir) = shift->lpuc('mkchfol',@_);
    $m->pstop(funnovar => v => 'fol') unless $fol;
    $m->pstop(funnovar => v => 'dok') unless $dok;
    $m->pstop(funnovar => v => 'supdir') unless $supdir;
    $m->pstop(funnovar => v => 'dirstem') unless $dirstem;
    $m->pstop(dirnedir => d => $dirstem) unless -d $dirstem;
    $m->pstop(funnovar => v => 'dokstyp') unless $dokstyp;
    my %dir_varz = (dok => $dok, fol => $fol, dirstem => $dirstem, supdir => $supdir); # file:/phm/22/01/02/sig/_lng.phm_pub_sig220102.txt::chdok
    while (my ($k, $v) = each %dir_varz) { $m->pstop() unless $v eq $m->pvar($k) };
    $m->set_dok_varz_db($lihs, %dir_varz);
  SVN_UPDATE: {
      last if $m->{svn_repos}->{offline};
      eval { $m->svn_update };
      last unless $@;
      $m->topdir_svn_cleanup($lihs);
      $m->svn_update;
    };
    my ($cwd,$lwd) = $m->chwd($dirstem) or $m->ustop(nochdst => d => $dirstem);
  MKDIR: {
    DIR_P: {
	last unless -d $fol;
	$m->vecho(foldekz => f => $fol);
	($cwd,$lwd) = $m->chwd($fol);
	last MKDIR;
      }
    EXISTS: {
	last unless -f $fol;
	my $ftmp = mktemp($fol.'XXXXXXX');chomp($ftmp);
	rename $fol, $ftmp or $m->ustop(renfol => f => $fol, t => $ftmp);
      };
      mkdir $fol;
      $m->svn_add_if($fol);
      ($cwd,$lwd) = $m->chwd($fol);
    };
    $m->dpop1($prog,$cwd);
};

=head2 Write the files

Write stubs for the main document (@dok_*) and its auxiliary files.

First evaluate all the commandline parameters, start writing only if they are OK.

=head4 ($val,$old) = $m->mlmk_input_lit($lihs,$nom,$dialog): Read a lingual text from the command line and validate it.

Return input value and whether it is old, i.e. the input just confirmed what was already in the template.

=cut

sub mlmk_input_lit {
    my ($m,$prog,$lihs,$nom,$dialog) = shift->lpuc('mlmk_input_lit',@_);
    my $batch = $m->{progvars}->{batch};
    $dialog ||= !$batch;
    # <nov id=inplit orl="file:/phm/21/04/17/sig/_lng.phm_pub_sig210417.txt::inplit">
    my ($val, $old) = $m->litvalid($nom, dialog => $dialog, lihs => $lihs, flng => 1, litstag => 'l', echo => 1); # </nov> 
    $m->dpop($prog,$val,$old);
};

=head3 $m->skrib_postfayl($lihs,$lang,$fi,$fo,%fopts) => $fo: Write language dependent tags from object variable lngtags into a lng_fayl.

=cut


sub skrib_postfayl {
    my ($m,$prog,$lihs,$lang,$fi,$fo,%fopts) = shift->lpuc('skrib_postfayl',@_);
    # $lihs->chkvar('des');
    my @langlits = $m->plst('langlits');
    $m->lihs_putarb($lihs, langlits => \@langlits);
    $m->riid_tmplvars_fayl($fo, msglini => 1) if -r $fo and $fopts{lits};
    # $lihs->chkvar('des');
    my $tmplappdir = $m->pvar('tmplappdir');
    my $fnulvalid = catfile $tmplappdir, 'valid.txt';
    my $flngvalid = catfile $tmplappdir, "valid.$lang.txt";
    my $fvalid = -r $flngvalid ? $flngvalid : $fnulvalid;
    $m->riid_tmplvars_fayl($fvalid, msglini => 1);
    my %langlits_oldvals = (_keys_ => \@langlits);
    $m->set_litsdir($lihs, l => $lang);
    foreach my $lit (@langlits) { 
      FOL2LAB: { # 臨時補洞 file:/phm/20/02/17/sig/_lng.phm_pub_sig200217.txt::/\bFOL2LAB\b/
	  last if $lit ne 'lab';
	  last if $lihs->{vars}->{lab};
	  last if !$lihs->{vars}->{fol};
	  $lihs->{vars}->{lab} = $lihs->{vars}->{fol};
	};
	my ($val, $oldval) = $m->mlmk_input_lit($lihs,$lit);
	next if $oldval eq $val;
	$langlits_oldvals{$lit} = $oldval;
    };
    $m->lihs_putarb($lihs, langlits_oldvals => \%langlits_oldvals);
    $m->template_process($lihs, $fi, $fo, mode => 'fundamental');
    $m->svn_add_if($fo);
    $m->dpop1($prog,$fo);
};    

sub skrib_antefayl0 {
    my ($m,$prog,$lihs,$fi,$fo,$mode) = shift->lpuc('skrib_antefayl0',@_);
    my @dabalits = $m->plst('dabalits');
    my $set_dabalits = sub {
	foreach my $lit (@dabalits) { 
	    my ($val, $old) = $m->mlmk_input_lit($lihs,$lit);
	    next unless $val;
	    $lihs->putstr($lit => $val);
	};
        $m->lihs_putarb($lihs, dabalits => \@dabalits);
    };
    &$set_dabalits;
    my @antevars = $lihs->getlst('_antevars_');
    my $set_antevars = sub {
	foreach my $k (qw(tmpl tmls langs sems)) { next if grep { $_ eq $k } @antevars;unshift @antevars, $k };
    };
    &$set_antevars;
    $m->lihs_putarb($lihs, antevars => \@antevars);
    my $dokprogver = $m->{konfig}->progver;
    $lihs->putstr(dokprogver => $dokprogver);
    $m->template_process($lihs, $fi, $fo, mode => $mode || 'conf');
    $m->dpop1($prog,$fo);
};

=head3 Function tmpl2langlits

=cut

sub tmpl2langlits {
    my ($m,$prog,$lihs,$tmpl) = shift->lpuc('tmpl2langlits',@_);
    my @lits = (@tratoks, @{$m->{konfig}->tratoks_extra});
    if ($m->{progvars}->{tmpltoks_ok}) {
	my @tmpltoks = @{$m->nom2xlst('_tmpltoks_',$lihs)};
	push @lits, @tmpltoks;
    };
    $m->{konfig}->lvalnomp->{$_} = 1 for @lits;
    $m->dpop($prog,@lits);return @lits;
};

=head2 $m->skrib_langs_fayl('_langs','de en zh ja'): Write out the _langs file.

The list of languages of the document is a basic meta info for many possible applications, including web templates.  Therefore it has its own file.

=cut

sub skrib_langs_fayl ($;$$) {
    my ($m,$prog,$fayl,$str) = shift->ppuc('skrib_langs_fayl',@_);
    $fayl ||= $m->hsprintf('%plangs');
    $str ||= join ' ', $m->plst(langs => 1);
    $m->fayl_put_str($fayl,$str);
    $m->svn_add_if($fayl);
    $m->dpop1($prog,$fayl);
};

=head3 Function subtmpl

Identify the subtmpl (subordinate document type) belonging to the current tmpl (document type), based on what is found in the values of the dynamic textchunk variable C<sems2subtmpl>.
Optionally @sems (semantic units passed to C<mlmk>, often only the directory leaf) arguments are given.  These are passed and often required by the C<sems2subtmpl> variable.
If either @sems or the C<sems2subtmpl> variable are missing, we terminate (raise an exception).

=cut

sub sems2subtmpl { 
    shift->sems_call('sems2subtmpl',@_) };
sub sems2subflat { 
    shift->sems_call('sems2subflat',@_) };
sub sems2list {
    my ($m,@args) = @_;
    $m->text2list($m->sems_call(@args)) };
sub sems2subvars { 
    shift->sems2list('sems2subvars',@_) };
sub sems2subsems { 
    shift->sems2list('sems2subsems',@_) };

=head3 Function subflat

Resolve the boolean C<subflat> (whether to create an adjacent directory rather than a subdirectory) value belonging to the current tmpl (document type), based on what is found in the values of the dynamic textchunk variable C<sems2subflat> and the C<sems> (semantic units supplied to mlmk as arguments, often consisting of the directory leaf only).
The sems are passed to the C<sems2subflat> variable as ${1}, ${2} etc.
If nothing is given we assume 0, i.e. opt for the default procedure of creating subdirectory.

=cut

sub subflat {
    my ($m,$prog,$lihs,@sems) = shift->lpuc('subflat',@_);
    @sems = $m->plst('sems') unless @sems;
  SUPLIT: { # incorporate finer mechanism using subdokslit,grpdokslit as outlined at file:/phm/19/05/19/sig/_org.phm_pub_sig190519.txt::semslit
      my $doklit = join '_', @sems;
      my $suplit = $doklit . '__sup';
      my $sup = $lihs->getstr($suplit);
      last unless $sup;
      return $m->dpop1($prog,0) if $sup eq $lihs->getstr('subdokslit');
      return $m->dpop1($prog,1) if $sup eq $lihs->getstr('grpdokslit');
    };
    my $subflat = $m->sems2subflat($lihs,@sems) || '0';
    $m->dpop1($prog,$subflat);
};

=head3 Function subvars

Set variables for subtemplate as specified in sems2subvars and return them.
E.g.

    sems2subvars ?= $(case ${1},,sig,|privat:1|no_subsems:1|)

leads to

    { privat => 1, no_subsems => 1 }

in case $sems[0] eq 'sig'.

2017-06-24 note: the switch no_subsems seems to mean that subdirectories are not made automatically.

=cut

sub subvars {
    my ($m,$prog,$lihs,@sems) = shift->lpuc('subvars',@_);
    my @vars = $m->sems2subvars($lihs,@sems);
    my $subvars = {};
    foreach my $var (@vars) {
	my ($k, $v) = $var =~ m(\A(\w+)\:(.+)\Z);
	next unless $k;
	$subvars->{$k} = $v;
    };
    $m->dpop1($prog,$subvars);
};


=head3 Function subsems

Return a list of argument list strings each of which is to be passed to one subdocument generation ('submktdir') call.

In the frequent and simplest case this argument list string consists of one item that is then immediately used as subdirectory leaf name.

This function optionally takes @sems (semantic units passable to C<mlmk>, usually directory leaf) arguments.  It evaluates the textchunk variable C<sems2subsems> of the current template/doctype, which can but in practise rarely does use @sems as arguments.

If the C<sems2subsems> variable does not exist, we are at the end of directory tree and return nil without warning.

A complex argument like

    sems2subsems ?= $(monwoks ${aaaa},${mm},${ddmin})

produces something like

    |W26,01,01|w27,02,08|w28,09,15|w29,16,22|w30,23,29|w31,30,31|

as a result.

Apart from subdocuments indicated via sems2subsems there are those specific to the particular document indicated via subsections of section ${semslit} of the document or those deliberately specified in variable @{extra_subsems}.

E.g.
	semslit = akt
means that subsections of 'akt' are proto-subdocuments.
However as a precaution we postulate that each would-be subdocument must have an attribute '@langs' before it is considered ready for creation.
OTOH
	@extra_subsems = |olgmV9|frazV9|
means that the listed two are to be created no matter what.

Changelog
 file:/phm/18/10/28/sig/_lng.phm_pub_sig181028.txt::semslit
=cut
sub lihs_get_sems {
    my ($m,$lihs,@sems) = @_;
    return @sems if @sems;
    @sems = $lihs->getlst('sems') unless @sems;
    return @sems if @sems;
    $m->pwarn('failed to get sems from lihs of dir %D, trying progvar sems', D => $lihs->{vars}->{__dir__});
    @sems = $m->plst('sems');
    return @sems if @sems;
    $m->pstop('unable to get sems from anywhere');
};

sub subsems {
    my ($m,$prog,$lihs,@sems) = shift->lpuc('subsems',@_);
    my $no_subsems = $lihs->getstr('no_subsems');
    return $m->dpop($prog) if $no_subsems;
    @sems = $m->lihs_get_sems($lihs,@sems);
    my @subsems = $m->sems2subsems($lihs,@sems);
  SEMSLITVARS: { 
      # file:/adv/tmpl/a2e/oas_akt_aamm/lang_pre.txt::/\bsemslit\b/
      # file:/phm/19/05/19/sig/_lng.phm_pub_sig190519.txt::semslit
      my %lits = ();
      foreach my $litvar (qw(semslit subdokslit grpdokslit)) {
	  my $lit = $lihs->getstr($litvar); # 'akt'
	  next if !$lit;  
	  next if $lits{$lit};
	  $lits{$lit} = 1;
	  my @litfols = $m->litfols($lihs,$lit);
	  next if !@litfols;
	  my @litnoms = map { $m->sup8fol2nom($lihs,$lit,$_) } @litfols;
	  @litnoms = grep { $lihs->getstr($_.'_langs') } @litnoms;
	  next if !@litnoms;
	  @litnoms = map { $_ = $lihs->{vars}->{$lit.'_sems'} || $_;$_ } @litnoms; 
	  push @subsems, @litnoms;
      };
    };
    $m->uniqpush(\@subsems, $lihs->getlstr('extra_subsems'));
    $m->dpop($prog,@subsems);
};

=head3 @predoks = $m->submktdir($suplihs,$trg,$sem,$supdir,$supdok,$predok): Generate single subdirectory tree

$sem => 'm08' || '+ant+m08+'
$suplihs :: vocabulary of our current parent

Make a particular $subdok directory tree either recursively or just up to the named Daiku target $opts{trg}.
Heart of the package.
Here we spawn a child vocabulary $lihs from parent vocabulary $suplihs and use it to create a new directory with all its files that can then, if $trg is at least 'dir', be compiled into a full vocabulary (mldk 'lits') and grandchildren etc. 
All parameters are obligatory.
The return values are the same as in submktdirs.

=cut

our %non_rekurs_trg_p = ( dir => 1, tdir => 1 );
sub submktdir {
    my ($m,$prog,$suplihs,$trg,$sem,$supdir,$supdok,$predok) = shift->lpuc('submktdir',@_);my $destrukts = [];
    $m->pstop(funnovar => v => 'sem') if !$sem;
    $m->pstop(funnovar => v => 'supdir') if !$supdir;
    $m->pstop(dirnodir => v => 'supdir') if !-d $supdir;
    my $dkr = $m->new_dkr;
    my $probp = $trg eq 'prob';
    # $suplihs->{vars}->{dir} ||= $suplihs->{vars}->{cwd};
    my ($lihs,$oldlihs,$dirstem,$dokstyp,$fol,$dok,$tmpl,$dir) = $m->supld2ld($sem,$suplihs,$supdir,$supdok,$predok,$probp); # file:/phm/22/01/02/sig/_lng.phm_pub_sig220102.txt::foldoktypdir
    my %dir_varz = (fol => $fol, dok => $dok, tmpl => $tmpl, dir => $dir);
    my $set_dir_varz = sub { while (my ($k,$v) = each %dir_varz) { $m->setpvar_mini($lihs,$k,$v) } };
    &$set_dir_varz;
    my $re_lihs = sub { $m->lihs_old($oldlihs);foreach my $k (qw(fol dok tmpl dir)) { my $v = $oldlihs->{vars}->{$k};next unless $v;$m->{progvars}->{$k} = $v } };
    push @$destrukts, $re_lihs;
    my $cwd = $m->lihs_get_cwd($suplihs);
    my $re_cwd = sub { $m->chwd($cwd) }; # file:/phm/22/01/02/sig/_lng.phm_pub_sig220102.txt::chdok
    $m->pstop('cwd %C unequal to oldlihs %D', C => $cwd, D => $oldlihs->{vars}->{dir}) if $cwd ne $oldlihs->{vars}->{dir};
    my $semlangs = sub {
	my $lit = join '_', $sem, 'langs';
	my $sl = $suplihs->{vars}->{$lit};
	$sl ? $m->text2list($sl) : $m->lihs_get_langs($suplihs);
    };
    my @langs =  &$semlangs;
    my $lang0 = $langs[0];
    $m->setpvar(lang => $lang0, lihs => $lihs);
    my $verbose = $m->pvar('verbose');
    my $oldp = 0;
    my $set_oldp = sub {
	my $fpre = catfile $dirstem, $fol, $m->fnom2fobj('pre', d => $dok);
	my $flng = catfile $dirstem, $fol, $m->fnom2fobj('lng', d => $dok);
	$oldp = (-r $fpre and -r $flng);
	my $msg = $oldp ? 'oldfol' : 'novfol';
	$m->{dokprop}->{$dok}->{oldp} = $oldp;
	$m->echo($probp || $verbose, $msg, d => $dok, f => $fol, t => $tmpl);
    };
    &$set_oldp;
    return $m->ddpop1($prog,$destrukts,$dok) if $probp;
    $m->set_dok_varz($dok,$lihs);
    my $url = $m->dir2url($dir) or $m->ustop(dir2url => d => $dir);
    $m->setpvar(url => $url, lihs => $lihs);
    my $testdir = $m->{konfig}->testdir || 'tdir';
    my $frms = { pre => '%ppre.%d.txt', lng => '%plng.%d.txt', llng => '%plng.%d.%l.txt' };
    my %dv = (); # depvars as in mldk
    # my $fyls = { };
    foreach my $fyl (qw(pre lng)) {
	# $fyls->{$fyl} = catfile $dir, $m->hsprintf($frms->{$fyl}, d => $dok);
	$dv{$fyl} = catfile $dir, $m->hsprintf($frms->{$fyl}, d => $dok);
    };
    my $oldnodp = -e $dv{pre} and -e $dv{lng};
  OLDNODP: { # file:/phm/22/01/02/sig/_lng.phm_pub_sig220102.txt::oldnodp
      last if !$oldnodp;
      # touch $_ for $dv{pre}, $dv{lng};
      last if $m->{konfig}->subsub;
      # file:/phm/21/02/01/sig/_lng.phm_pub_sig210201.txt::subs_ddpop1
      last if $trg ne 'subs';
      return $m->ddpop1($prog,$destrukts,$dok);
    };
    foreach my $ll (@langs) {
	# $fyls->{llng}->{$ll} = catfile $dir, $m->hsprintf($frms->{llng}, d => $dok, l => $ll);
	$dv{'llng_'.$ll} = catfile $dir, $m->hsprintf($frms->{llng}, d => $dok, l => $ll);
    };
    # $dv{pre} = $fyls->{pre};
    my @langlits = $m->tmpl2langlits($lihs,$tmpl); #
    my @dabalits = $m->tmpl2dabalits($suplihs,$tmpl,@langlits);
    $m->setplst(langlits => \@langlits, lihs => $lihs);
    $m->setplst(dabalits => \@dabalits, lihs => $lihs);
    my $sem1 = $m->pvar(sem1 => 1);
    my $svn_add_ok = $m->pvar(svn_add_ok => 1);
    my @dsts = ();my $push_dst = sub { push @dsts, $_[0] };
    $dv{sup_pag} = $m->lits_dk_trg($suplihs);
    my $set_pre_tmpl = sub {
	($dv{pre_tmpl},$dv{pre_bas},my %fopts) = $m->tmplsrcobjfayls('pre');
	my $pre = catfile $dir, $dv{pre_bas};$m->pstop('%f and %g differ', f => $pre, g => $dv{pre}) unless $dv{pre} eq $pre;
	&$push_dst($m->dkrfh($dkr,pre => ['sup_pag','pre_tmpl'], "make language independent variable file", sub {
	    my ($dst,%dv) = @_;
	    my @noms = $m->lihs_import_grupvars($lihs,$suplihs,$sem1);
	    $m->lihs_put($lihs, __supdats__ => \@noms);
	    $m->skrib_antefayl0($lihs,$dv{pre_tmpl},$dst,$fopts{mode});
	    $dst }, svn_add_ok => $svn_add_ok, tsukaisute => 1, %dv));
    };
    &$set_pre_tmpl;
    my $set_lng_tmpl = sub {
	my $lng_ffol = $m->fnom2fsrc('lng');
	$dv{lng_tmpl} = $m->ffol_tmplsrcfayl($lng_ffol);
	$m->ustop(no_lng_tmpl => t => $dv{lng}) unless $dv{lng_tmpl};
    };
    &$set_lng_tmpl;
    my @tmls = $m->get_tmls($lihs);
    $m->pstop('no tmls found') unless @tmls;
    my $tml = $tmls[0];
    my $topdok = $lihs->getstr('topdok') || $m->pvar(topdok => 1);
  SUPDATS: {
      my @noms = $m->lihs_import_grupvars($lihs,$suplihs,$sem1);
      $m->lihs_put($lihs, __supdats__ => \@noms);
    };
    # @langs = $m->plst(langs => 1);
    $dv{langs} = catfile $dir, $m->hsprintf('%plangs');
    &$push_dst($m->dkrfh($dkr,langs => ['sup_pag'],"make langs file", sub {
	my ($dst,%dv) = @_;
	my $slangs = join ' ', @langs;
	$m->skrib_langs_fayl($dst,$slangs);
	$dst }, svn_add_ok => $svn_add_ok, tsukaisute => 1, %dv ));
    my $dkf_langs = sub {
	my $i = 0;
	my $dkf_lang = sub {
	    my ($lang,$i) = @_;
	    $m->pstop(funnovar => v => 'lang') unless $lang;
	    $m->setpvar(lang => $lang, lihs => $lihs);
	    my $llng = 'llng_'.$lang;
	    my $lvarsdep = 'lvars_'.$lang;
	  LVARSDEP: { # mldk will later read and remake this based on further input files, same for toks and tmpl
	      my ($prez,$tmlz,$lngz) = $m->pre_tml_lng_tmk_faylz($lang, tmpl => $tmpl, tml => $tml);
	      my ($pre,$lng) = $m->dok_pre_lng_faylz($lang,$topdok, lihs => $lihs, topdok => $topdok);
	      unshift @$prez, @$pre;
	      $dv{$lvarsdep} = $m->dkrmk($dkr, vars => [@$prez,@$tmlz,@$lngz], lihs => $lihs, litstagl => [$i, l => $lang], tsukaisute => 1, remk => 1);
	    };
	    ($dv{llng_tmpl},$dv{llng_bas},my %fopts) = $m->tmplsrcobjfayls('llng');
	    $dv{llng} = catfile $dir, $dv{llng_bas};$m->pstop('%f and %g differ', f => $dv{$llng}, g => $dv{llng}) unless $dv{$llng} eq $dv{llng};
	    my $dst = $dv{llng};	    
	    $m->dkrfh($dkr, $dst => [$lvarsdep,'llng_tmpl'], "make lng $lang txt in $dir", sub { 
		my ($dst,%dv) = @_;
		# $lihs->del($_) for @doktypdelvars;
		# $m->riid_tmplvars_fayl($_, msglini => 1, lihs => $lihs) for @fis;
		$m->dk_init_dst($dv{$lvarsdep});
		my @noms = $m->lihs_import_grupvars($lihs,$suplihs,$sem1,$lang);
		$m->lihs_put($lihs, __suplits__ => \@noms);
		# $m->tmplvars_kk_progvars($lihs, be => nonvars => [ keys %{$m->{subdoklits}} ] );
		$m->skrib_postfayl($lihs,$lang,$dv{llng_tmpl},$dst,%fopts);
		$dst }, svn_add_ok => $svn_add_ok, %dv);
	    $dst };
	foreach my $lang (@langs) {
	    &$push_dst(&$dkf_lang($lang,$i++));
	};
    };
    &$dkf_langs;
    &$push_dst($m->dkrfh($dkr, lng => ['lng_tmpl'], "make $dv{lng} from $dv{lng_tmpl}", sub {
	my ($dst,%dv) = @_;
	my @noms = $m->lihs_import_grupvars($lihs,$suplihs,$sem1,$langs[0]);
	$m->lihs_put($lihs, __suplits__ => \@noms);
	# kludge 190205: originally without %fopts, should rather be configurable via file:/adv/perl/A2E/dokfs.konf::insist
	$m->template_process($lihs,$dv{lng_tmpl},$dst, f => $tml, t => $tmpl, mode => 'org', insist => 1);
	$dst }, svn_add_ok => $svn_add_ok, %dv ), tsukaisute => 1 );
    my $push_fnoms = sub {
	my $s = $lihs->getstr('tmpl_fnoms');
	unless ($s) { $m->uwarn('no litvar @tmpl_fnoms found, try konfvar');$s = $m->{konfig}->tmpl_fnoms };
	$m->ustop('no konfvar tmpl_fnoms found either') unless $s;
	my @tmpl_fnoms = $m->textsep2list($s,',');
	my @grup_fnoms = $m->grupvars_getlst($lihs,'fnoms',$fol); # 前題 $nom eq $fol 在此成立
	$m->uwarn('no grup_fnoms found for %g', g => $fol) unless @grup_fnoms;
	my @fnoms = $m->uniqpush(\@tmpl_fnoms,\@grup_fnoms);
	$m->ustop('no @fnoms found at all') unless @fnoms;
	$m->svn_add_if($dir);
	foreach my $fnom (@fnoms) {
	    &$push_dst($m->skrib_fnom($lihs,$dkr,$fnom,dir=>$dir,grup=>$fol,svn_add_ok=>$svn_add_ok));
	};
    };
    &$push_fnoms;
    $m->setpvar(lang => $lang0, lihs => $lihs);
    my $subcwd = '';
    my @predoks = ();
    my $dkf_subs = sub {
	my @tmls = $m->plst('tmls');
	my $tml = $tmls[0];
	my $ftml = catdir $dir, $m->hsprintf($m->{konfig}->tml2outfayl->{$tml}); # html => 'index.de.html', multemp => '_top-olgmV9.de.txt'
	$m->add_subdoks($dok,$supdir,$dokstyp); # file:/adv/perl/A2E/Subdoks.pm.tmpl::add_subdoks
	my $chdok = sub {
	    push @$destrukts, $re_cwd;
	    $subcwd = $m->mkchfol($lihs,$fol,$dirstem,$dokstyp,$dok,$supdir);
	    $m->pstop('changed to %S rather than %D', S => $subcwd, D => $dir) if $subcwd ne $dir;
	    $m->svn_add_if($dir);
	    my $svn_add = sub {
		my @faylz = grep { m(\A(?:\_.*(?:txt|s)|\w+file)\Z) } glob '*';
		$m->svn_add_if(@faylz);
	    };
	    &$svn_add;
	    $m->{dokprop}->{$dok}->{oldp} = 1;
	    %{$m->{nom2lits}} = (); # semi-redundant since already in @starthacs 
	    @{$lihs->{lsts}->{__dirnoms__}} = qw(llitsdir tlitsdir litsdir);
	    $subcwd };
	$m->dkrf($dkr, $ftml,[@dsts],'enter the directory and set the new document', sub {
	    my $f = shift;my $dst = $f->dst;my @deps = @{$f->deps};
	    &$chdok;
	    $m->mldk('main', lihs => $lihs, remk => 1);
	    $dst } );
	$m->dkrt($dkr, 'tdir',[@dsts],"make files @dsts");
	$m->dkrt($dkr, 'dir',[$ftml],"make one subdok $ftml and chdir there");
	$m->dkrt($dkr, 'subs',[@dsts],"make full subdocuments", sub {
	    &$chdok;
	    push @predoks, $m->submktdirs($trg, lihs => $lihs, dok => $dok, dir => $dir, mldk => $m->{konfig}->subsub ? 'lits' : 'main', remk => 1);
	    @predoks } );
	$m->dkrt($dkr, 'scap',[@dsts],"make subdocument sources only, outside svn repository, but including full subtrees even of existing subdocuments", sub {
	    &$chdok;
	    push @predoks, $m->submktdirs($trg, lihs => $lihs, dok => $dok, dir => $dir, mldk => 'lits', remk => 1);
	    @predoks } );
    };
    &$dkf_subs;
    $trg ||= 'prob';
    $m->dkrb($dkr, $trg);
    # my %dirrek = (dirstem => $dirstem, fol => $fol, tmpl => $tmpl, depvars => \%dv);
    # my %dirrek = (fol => $fol, tmpl => $tmpl);
    # $m->{dokprop}->{$dok}->{$_} = $dirrek{$_} for keys %dirrek;
    push @predoks, $dok;
    $m->ddpop($prog,$destrukts,@predoks);
};

=head3 ($lihs,$dok,$dir) = $m->setsupvars($lihs,%opts): Generate vocabulary and basic info for a parent directory 

Generates basic vocabulary and returns info of the document under which we will be creating subdocuments.
Invokes mldk('lits').
Invoked from submktdirs.
$lihs :: our vocabulary which now becomes $suplihs in order to cede the namespace to children but remains the same
$dok :: dok were we are, becomes $supdok
$dir :: absolute directory where we are, becomes $supdir

=head5 further potential options and return values

$dvr => { varsfyl_t0 => '.lits/l.html/vars.hac', varsfyl_l0 => '.lits/l.de/vars.hac', .... } :: files on which mldk depended, could become useful, e.g. for checking state with $m->mldk_litsp(%$dvr)
%opts dok,dir and dvr could be passed to the core of this function, $m->mldk('lits',%opts), in which case these values would be preset.
Especially option $dvr could have been useful for checking whether we have already run mldk('main') before, but we now solve this problem by passing mlht a target mldk => 'main'.

=cut

sub lihs_get_langs {
    my ($m,$prog,$lihs) = shift->lpuc('lihs_get_langs',@_);
    my @langs = $lihs->getlst('langs');
    return $m->dpop($prog,@langs) if @langs;
    $m->pwarn('failed to obtain langs from lihs');
    @langs = $m->plst('langs');
    return $m->dpop($prog,@langs) if @langs;
    $m->pwarn('failed to obtain langs from progvars');
    @langs = $m->get_langs(lihs => $lihs);
    return $m->dpop($prog,@langs) if @langs;
    $m->pstop('unable to obtain langs from anywhere');
};
sub lihs_get_cwd {
    my ($m,$prog,$lihs) = shift->lpuc('lihs_get_cwd',@_);
    my $cwd = $lihs->getstr('cwd') || $m->pvar(cwd => 1);    
    $m->dpop1($prog,$cwd);
};
sub setsupvars {
    my ($m,$prog,$lihs,%opts) = shift->lpuc('setsupvars',@_);
    my $tmpl = $m->pvar(tmpl => 1);
    my @langs = $m->lihs_get_langs($lihs);
    my $lang0 = $langs[0];
    my $lang = $m->{konfig}->lang;
    $lang0 = $langs[0];
    $lang ||= $lang0;
    my ($suplihs,$supdok,$supdir) = $opts{mldk} ? do {
	$m->lihs_var2arb;
	$m->mldk($opts{mldk}, lihs => $lihs, %opts);
    } : do {
	my $supdir = $opts{dir} || $lihs->{vars}->{__dir__} || $m->pstop(funnoopt => o => 'dir');
	my $supdok = $opts{dok} || $lihs->{vars}->{dok} || $m->pstop(funnoopt => o => 'dok');	
	$m->set_lihs_dok_dir_langs(lihs => $lihs, dok => $supdok, dir => $supdir);
    };
    $m->pstop('suplihs and lihs are not the same although both pertain to the new parent') if $lihs ne $suplihs;
    $m->pstop('suplihs unfit for writing') if $suplihs->{vars}->{__nolitsdir__} or !$suplihs->{vars}->{__litsdir__};
    $suplihs->putstr(tmpl => $tmpl);
    $m->setplst(langs => \@langs, lihs => $suplihs);
    $m->setpvar(lang0 => $langs[0], lihs => $suplihs);
    $m->setpvar(lang => $lang, lihs => $suplihs);
    $m->setplst(suplangs => \@langs, lihs => $suplihs);
    $m->setpvar(lang0 => $lang0, lihs => $suplihs);
    $m->setpvar(supdir => $supdir, lihs => $suplihs);
    $m->setpvar(supdok => $supdok, lihs => $suplihs);
    # $m->tmplvars_kk_progvars(be => nonvars => [ keys %{$m->{subdoklits}} ]);
    $suplihs->putstr($_ => $m->pvar($_)) for qw(svn_add_ok); # kludge 190205
    # $m->progvars_kk_tmplvars([ae => ak => 1]);
    # $m->tmplarbs_kk_dmopts(bd => { d => $supdok, l => $lang0, nonvars => [ keys %{$m->{subdoklits}} ] });
    $m->lihs_get_vars($suplihs, l => $lang0);
    $m->lihs_var2arb($suplihs);
    $m->dpop($prog,$suplihs,$supdok,$supdir);
};

=head4 @predoks = $m->submktdirs($trg,sem=>$sem,lihs=>$lihs,mldk=>$mldk): recursively generate subdirectory tree and return one info record for each generated subnode

$trg :: target => (?:prob|tdir|dir|subs|subsub|scap)
$sem :: optional subdocument signifier which narrows the range of subdirectories to one, needed in connection with targets (?:tdir|dir)
$lihs :: optional A2E::Lihs vocabulary object of the parent node, needed in recursion
$mldk :: optional target for mldk invoked in setsupvars, allowing distinction between full generation 'main' and minimal version 'lits' in submktdir.
@predoks :: documents that were generated up to now, in sequence of document tree as in file:/adv/perl/A2E/Dokfs.pm.tmpl::/\bpredoks\b/ with associated $m->{dokprop} registry.

=cut

sub submktdirs {
    my ($m,$prog,$lihs,$trg,%opts) = shift->opuc('submktdirs',[qw(trg)],@_);my $sem = $opts{sem};my $restrus = [];
    my ($suplihs,$supdok,$supdir) = $m->setsupvars($lihs,%opts);
    $m->setpvar(cwd => $supdir, lihs => $lihs);$m->setpvar(dir => $supdir, lihs => $lihs);
    my @predoks = ();
    my @subsems = $sem ? ($sem) : $m->subsems($lihs);
    # my $chwd_supdir = sub { $m->chwd($supdir) };
    # my $sarb7suplihs = sub { $m->{sarb}->{lihs} = $suplihs };
    # push @$restrus, $chwd_supdir, $sarb7suplihs;
  SUBSEMS: {
      last if !@subsems;
      $m->vecho('subsems: %S', S => \@subsems);
      my $sup = {};
      my $starthacs7m2nul = sub { $m->{$_} ||= {};%{$sup->{$_}} = %{$m->{$_}};%{$m->{$_}} = () for @starthacs };
      my $starthacs7m2sup = sub { %{$m->{$_}} = %{$sup->{$_}} for @starthacs };
      my $heredhacs7m2sup = sub { foreach my $hac (@heredhacs) { %{$m->{$hac}} = %{$sup->{$hac}} } };
      my $hacs7sup2m = sub { foreach my $hac (@heredhacs,@starthacs) { $sup->{$hac} ||= {};$m->{$hac} ||= {};%{$sup->{$hac}} = %{$m->{$hac}} };$sup };
      &$hacs7sup2m;
      # push @$restrus, $heredhacs7m2sup, $starthacs7m2sup;
      my ($owd,$lwd) = (); # 僅用於偵錯
      my $fol = $m->pvar('fol');
    FOL: {
	last if $fol;
	my @fols = splitdir $supdir;
	$fol = $fols[$#fols];
	$m->setpvar(fol => $fol);
	$m->pwarn('deriving unknown fol %f from dir %d', f => $fol, d => $supdir);
      };
      my %dir_varz = ( dir => $supdir, dok => $supdok, cwd => $supdir, fol => $fol );
      my $chsup = sub {
	  ($owd,$lwd) = $m->chwd($supdir);
	  while (my ($k,$v) = each %dir_varz) { $m->setpvar($k => $v, lihs => $lihs) }; # 僅為正確性為偵錯
      };
      foreach my $sem (@subsems) {
	SUBVARSMKTDIR: { # file:/phm/19/05/21/sig/_lng.phm_pub_sig190521.txt::subvarsmktdir, file:/phm/19/05/22/sig/_org.phm_pub_sig190522.txt::newsub
	    &$starthacs7m2nul;
	    my $predok = $predoks[$#predoks];
	    push @predoks, $m->submktdir($suplihs,$trg,$sem,$supdir,$supdok,$predok);
	    $m->pstop('sarb lihs was not restored') if $m->{sarb}->{lihs} ne $suplihs;
	    &$heredhacs7m2sup;
	    &$chsup; # 疑是大部可置於循環末尾段中
	  };
      };
      &$starthacs7m2sup;
      # &$chsup; #20220104
    };
    $m->ddpop($prog,$restrus,@predoks);
};

=head4 $ok = $m->report_predoks($dok,$trg,@predoks): Report results of subdirectory creation

$trg := (?:prob|tdir|dir|subs|subsub|scap) :: target
@predokss := dok+ :: list or documents returned by submktdirs and of which properties were registered in $m->{dokprop}->{$dok}

=cut
sub report_predoks {
    my ($m,$prog,$supdok,$trg,@predoks) = shift->ppuc('report_predoks',@_);
    my $verbose = $m->pvar('verbose');
    my $probp = 'prob' eq $trg;
    my $verprob = $verbose || $probp;
    if (@predoks) {
	foreach my $dok (@predoks) {
	    $m->pstop('null dok in returned predoks %P', P => \@predoks) if !$dok;
	    my $dirrek = $m->{dokprop}->{$dok};
	    $m->pstop('no dirrek available for dok %D', dok => $dok) unless $dirrek;
	    $m->echo($verprob, mkfol => { P => !$dirrek->{oldp}, f => $dirrek->{fol}, d => $dok, Y => $dirrek->{typ}, u => $dirrek->{dok}, D => $dirrek->{dir}, t => $dirrek->{tmpl}, T => $trg });
	} 
    } else {
	$m->echo($verprob, nofol => d => $supdok);
    };
    $m->dpop($prog,@predoks);
};

=head2 @predoks = $m->mlmk($trg,$sem): Make planned subdirectories recursively according target and optional choice of single child dir

$trg =: (?:prob|tdir|dir|subs|subsub|scap)
$sem =>  'm08' || '+ant+m08+'  :: child dok signifier as used by sems2dok,sems2fol

Make a mlht document tree in an incremental way, useful for debugging templates and other sources. 
The following $trg targets are available:

	prob:	obtain a list of potential subdocument signifiers ($sem) and resulting subdirectory info
	tdir:	make subdocument sources of first level for inspection, without chdir; to be removed with rmtdir later
	dir:	make full subdocument of first level, registered and compiled
	subs:	make full tree and register everything in svn, break as soon as any document can not be compiled (with mldk main).
	scap:	make minimal tree (barren trunk) for later repair, uncompiled and outside svn but still registered in _grpdoks/_subdoks

C<prob> is default
C<scap> is C<subs> for the impatient, faster and more likely to run to the end.
Etymology: scapus is latin for trunk and scap provides an escape.
To override the svn_related behaviour, you can force SVN by --svn_add_ok and disable it by --svn_off or corresponding environment variables SVN_ADD_OK and SVN_OFF respectively.  Setting SVN_ADD_OK to 0 will not override anything, and SVN_OFFLINE has no effect since adding is only a scheduling operation that can be done offline.

$sem is a list of semiotic elements or single element from which the subdirectory choice is by serving as arguments to sems2dok/sems2fol.
If it is a list, it is joined with the usual punctuation marks as in '+ant+m08+'.

=cut
our %trg2svnp = ( subs => 1, dir => 1);
sub mlmk {
    my ($m,$prog,$trg,$sem) = shift->ppuc('mlmk',@_);
    my $inod = $m->inod('.');
    $trg ||= 'prob';
    my $dok = $m->get_dok;
    $m->{progvars}->{svn_add_ok} = $m->{konfig}->svn_off ? 0 : $m->{konfig}->svn_add_ok || $trg2svnp{$trg} ? 1 : 0;
    $m->{progvars}->{predoks} = [ $m->fayl_predoks($dok, kmdfun => 'putdok') ]; # repair database because we rely on it during creation process
    my @predoks = $m->submktdirs($trg, sem => $sem, mldk => 'lits', remk => 1);
    $m->report_predoks($dok,$trg,@predoks);
    $m->pstop('we have not returned to the initial directory but are in %D', D => $m->get_cwd) unless $inod == $m->inod('.');
    $m->dpop($prog,@predoks);
};
1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/MLMK.pm.tmpl"
# End:
