=head1 NAME

A2E::Makedoksrow - generate a menu from navigation info found in a text file and in a database

=cut

package A2E::Makedoksrow;

=head1 VERSION

This describes version B<0.0.4> of package A2E::Makedoksrow

=cut

our $VERSION = '0.0.4';

=head1 SYNOPSIS

makedoksrow -o <outfile> <dok> <lang>

=head1 DESCRIPTION

Generate pointers to related documents used at bottom of a HTML document, used in documents which were generated at a2e.de before 2009-03-09.  For newer documents, this is done with mlht(1), using template vmethods such as C<predokreks> which are defined in A2E::MLHT(3).


=head1 OPTIONS

listing may be stale

=over

=item --help | -h

Display short help and exit.

=item --manpage | -m

Display full documentation and exit.

=back

General options:

=over

=item --daba I<daba>

Name of database to write to.  Either 'a2e' (default) or 'ffii'.

=item --verbose | -v

verbose reporting for the user

=item --urllab | -u

In the input file the URL comes first, the associated title second
(traditionally it has been the other way around)

=item --dok <dok>

document identifier, e.g. 'makedoksrow_pod0701'

Preferrably found in the name of the input file.

=item --lang <lang>

two-letter ISO language code, e.g. 'en'. 

Preferrably found in the name of the input file.

=item --infiles|i <infile>

File used for input, can be specified repetitively.

Should preferrably be one file with a name like

    makedoksrow_pod0701.nav.en.html

where 'makedoksrow_pod0701' is the document identifier (dok) and 'en' the language (lang).

But it's also possible to specify several files.

Moreover, the dok and lang elements can also be specified by separate commandline options, see above.

=back

=head1 PREREQUISITES

    A2E::Dokfs(3)

=head1 SCRIPT CATEGORIES

Web / mktdir

=head1 SEE ALSO

    A2E::Nav2html(3)
    A2E::MLHT(3)

=head1 ENVIRONMENT

=over

=item INFILE FORMAT

The infile could typically read as follows:

	This project

	 href:makedoksrow.pl
	 href:sqltab.pl		SQLTAB

	@Environment

	Participants

	 dok:phm	Hartmut Pilch
	 dok:adv	Data Processing
	# dok:i2p	I2P


This creates a menu, in which the indented items are links.
The href macro's argument is a relative link 
The dok macro generates a relative link based on a document identifier.
The second column may contain a label.
If it isn't pressent, a default value is used for the label.
This default is either a label found in the database or the link itself.

To compile the above form, the '--urllab' option needs to be specified.
Without this option, the sequence of url and label is inversed. 
In a later version, the 'urllab' sequence will probably become the default.

Unindented lines represent titles.

If an unindented line starts with a '@', it is followed by a menu, as far as
this can be generated from the document hierarchy that is registered in the
database.

Lines that begin with hash marks (#) are treated as comments, i.e. ignored.

=back

=head1 PREREQUISITES

       A2E::Dokfs(3)

=cut

use base 'A2E::Dokfs';
use strict;

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

Copyright (c) 2006-8 Hartmut Pilch

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

=head2 Overwritable Functions

=head3 Function A2E::Makedoksrow::defvars

=cut

sub defvars {
    my $m = shift;
    $m->dpuc('A2E::Makedoksrow::defvars', @_);
    $m->SUPER::defvars;
    $m->{konfig}->htmlesc(1);
    $m->{konfig}->lifmt('%s');
    $m->{konfig}->define('doksrow_separ|separ=s', { DEFAULT => ' | ' });
    $m->{konfig}->define('doksrow_start|start=s', { DEFAULT => '[ ' });
    $m->{konfig}->define('doksrow_finis|finis=s', { DEFAULT => ' ]' });
    $m->konfayl('makedoksrow.konf');
    $m->dpop('A2E::Makedoksrow::defvars', 1);
    return 1;
};

=head2 Special Functions

=head3 Function makedoksrow

=cut

sub makedoksrow {
    my $m = shift;
    $m->dpuc('makedoksrow', @_);
    my $dok = shift;
    my $lang = shift;
    $m->set_dok($dok) or die 'no document found here';
    $m->{lang} = $lang if length $lang;
    die 'failed to set language' unless length $m->{lang};
    $m->init_navdoks or die 'failed to initiate navigation-related documents';
    my @entries = $m->dokmenu_entries_sub;
    my $fmt = $m->{konfig}->lifmt;
    my $str = join $m->{konfig}->doksrow_separ, map { sprintf $fmt, $_ } @entries;
    print sprintf '<!-- %s -->'."\n", $m->emacs_timestamp if $m->{konfig}->htmlesc;
    print $m->{konfig}->doksrow_start . $str . $m->{konfig}->doksrow_finis . "\n" if @entries;
    $m->sav2out;
    $m->dpop('makedoksrow', 1);
    return 1;
};

=head1 The END

=cut

return 1;

# Local Variables:
# coding: utf-8
# srcfile: /adv/perl/A2E/Makedoksrow.pm.tmpl
# url: http://a2e.de/adv/perl/A2E/Makedoksrow.pm.tmpl 
# mode: perl
# End:
