=head1 NAME

A2E::MakefileCache

=head1 VERSION

0.0.1

=cut

our $VERSION = 0.0.1; 

=head1 IMPLEMENTATION

=head2 Header

=cut

package A2E::MakefileCache;
use base 'A2E::MLDK';
use File::Spec::Functions;
=head2 Overload Functions

=cut

sub makefile_cache_defvars {
    my $m = shift;
    $m->dpuc('makefile_cache_defvars', @_);
    $m->SUPER::defvars;
    $m->tmplfil_defvars;
    $m->kopi_define('conv','s','deplate');
    $m->cache_define('convs', 's@', [ qw(deplate txt2html xsltproc) ]);
    $m->cache_define('dok_vars', 's@', [ qw(dok dir url langs lang0 mldok mlnav hta_org hta_sig) ]);
    $m->cache_define('rest_vars', 's@', [ qw(topdok topdir topdirdok mbox mdir tit conv subdirs doksfiles) ]);
    $m->cache_define('xsltproc_vars', 's@', [ qw(styledir) ]);
    $m->cache_define('conv_vars', 's%'); # related to conversion from tml to outfmt
    $m->cache_define('lang_vars', 's@', [ qw(lng pag lng_faylz lng_faylz_src) ]);
    $m->cache_define('tml_vars', 's@', [ qw(mlht_targets mlht_opts main_targets pre_targets post_targets varlangs) ]); # common to all tmls
    $m->cache_define('multemp_vars', 's@', [ qw(sitemap_tmpl sitemap_index_tmpl css_tmpl doksrow_html_tmpl news_html_tmpl news_rss_tmpl news_opts html_opts plain_opts latex_opts htmsnip_opts conv_opts dba_tmpl tex_tmpl xhtml_tmpl)]); # specific to deplate
    $m->cache_define('mldbweb_vars', 's@', [ qw(styledir) ]); # specific to docbook-website
    $m->cache_define('html_vars', 's@', [ qw(sitemap_tmpl sitemap_index_tmpl doksrow_html_tmpl news_html_tmpl news_rss_tmpl news_opts css_tmpl)]); # specific to html
    $m->cache_define('makefile_cache_name', 's', 'Makefile.cache');
    $m->cache_define('makefile_lang_cache_name', 's', 'Makefile.%l.cache');
    $m->cache_define('makefile_tml_cache_name', 's', 'Makefile.%t.cache');
    $m->cache_define('lang_variants', 's%');
    $m->cache_define('tmplang_ldoknom', 's%', { multemp => '%ptop-%d.%l.txt', html => 'index.%l.html', latex => '%d.%l.tex', mulpod => '%d.%l.pod' });
    $m->konfayl('makefile_cache.konf');
    $m->dpop('makefile_cache_defvars', 1);
    return 1;
};

sub defvars {
    my $m = shift;
    $m->dpuc('A2E::MakefileCache::defvars', @_);
    $m->SUPER::defvars;
    $m->makefile_cache_defvars;
    $m->dpop('A2E::MakefileCache::defvars', 1);
    return 1;
};

sub makefile_cache_postkonfig {
    my $m = shift;
    $m->dpuc('makefile_cache_postkonfig', @_);
    $m->mlht_postkonfig;
    # moved to set_main_vars
    # my $fayl = $m->antefayl($m->pvar('dok'));$m->riid_tmplvars_fayl($fayl);
    $m->set_srcfpfx;
    $m->dpop('makefile_cache_postkonfig', 1);
    return 1;
}

sub postkonfig {
    my $m = shift;
    $m->dpuc('A2E::MakefileCache::postkonfig', @_);
    $m->SUPER::postkonfig;
    $m->tmplfil_postkonfig;
    $m->makefile_cache_postkonfig;
    $m->dpop('A2E::MakefileCache::postkonfig', 1);
    return 1;
};

=head2 Second Level Functions

=head3 Function read_cache

=cut

sub read_cache {
    my $m = shift;
    $m->dpuc('read_cache', @_);
    my $fi = shift;
    $m->open_fayl($fi, 'i', 'r');
    local $_;
    while ($m->getlini('i')) {
	chomp;
	if (m(\A(\w+)\s*[\?\:]\=\s*(.*)\s*\Z)) {
	    $key = lc $1;
	    $m->{$key} = $2;
	    $m->dgot($key => $2);
	} else {
	    warn sprintf 'useless in context of cache: %s', $_;
	};
    };
    $m->kloz_fayl('i');
    $m->dpop('read_cache', 1);
    return 1;
};

=head3 Function set_lang_vars

Set variables for use in language-dependent Makefile part such as Makefile.de.cache.
 
Given Arguments:

    $ll : 'de' : language for which to set variables

Returned Arguments:

    $ok : 1 : success

=cut

sub set_lang_vars {
    my $m = shift;
    $m->dpuc('set_lang_vars', @_);
    my $ll = shift;
    $m->{tmplvars}->{__lang} = $ll;
    my $fayl = $m->hsprintf($m->{konfig}->lexfnom, d => $m->pvar('dok'), l => $ll, p => $m->{srcfpfx});
    $m->ustop('can not set lang vars for %l: file %f not found', l => $ll, f => $fayl) unless -e $fayl;
    $m->{lng} = $fayl;
    $m->{pag} = join '.', $fayl, 'pag';
    my $tmpl = $m->pvar('tmpl');
    @{$m->{lng_faylz}} = $m->lng_faylz($ll, $m->pvar('dok'), readable => 1, typ_min => 2, par => 1, tmpl => $tmpl);
    @{$m->{lng_faylz_src}} = $m->lng_faylz_src(@{$m->{lng_faylz}});
    $m->dpop('set_lang_vars', 1);
    return 1;
};

=head3 Function main8pre_targets

=cut

sub main8pre_targets {
    my $m = shift;
    $m->dpuc('main8pre_targets', @_);
    my $tml = shift || $m->{tmplang};
    my @targets = ();
    my $dok = $m->pvar('dok');
    my $s_outfmts = $ENV{OUTFMTS} || 'html'; 
    $m->{outfmts} = [ split m(\s+), $s_outfmts ];
    $m->{tmplang} = $ENV{TMPLANG} || 'multemp';
    my @mlht_targets = ();
    foreach my $ll (@{$m->{langs}}) {
	$m->{cache}->{lang_variants}->{$ll} ||= '';
	my @lvars = split m(\s+), $m->{cache}->{lang_variants}->{$ll};
	unshift @lvars, $ll;
	my $mlht_target = $m->hsprintf($m->{cache}->{tmplang_ldoknom}->{$tml}, d => $dok, l => $ll, p => $m->{srcfpfx});
	push @mlht_targets, $mlht_target;
	for my $fmt (@{$m->{outfmts}}) { 
	    next if $fmt eq 'html';
	    next if $fmt eq 'xml' and $tml eq 'mldbweb'; 
	    push @targets, "$dok.$ll.$fmt";
	}; 
      HTML: {
	  last unless grep { $_ eq 'html' } @{$m->{outfmts}};
	  foreach my $lv (@lvars) {
	      push @targets, "index.$lv.html";
	      next if $m->{tmplvars}->{nonav};
	      push @targets, "$dok.nav.$lv.html";
	  };
	MULTEMP: {
	    last unless $tml eq 'multemp';
	    my @lv;
	  DOKROWS: {
	      last if $m->{tmplvars}->{nodokrows};
	      foreach $lv (@lvars) {
		push @targets, "$dok.doksrow.$lv.html";
	      } };
	  NEWS: {
	      # 2015-09-14 make $dok.news.* files in topdir only
	      last unless $dok eq $m->{topdok};
	      last if $m->{tmplvars}->{nonews};
	      last unless $m->{news_opts};
	      foreach $lv (@lvars) {
		  push @targets, "$dok.news.$lv.rss", "$dok.news.$lv.html";
		  1;
	      }
	    };
	  };
	MLDBWEB: {
	    last unless $tml eq 'mldbweb';
	    push @targets, "$dok.article.$ll.xml";
	  };
	};
    };
    my @pre_targets = ();
    my @post_targets = ();
  PRE_TARGETS: {
      last unless $m->pvar('dok') eq $m->{topdok};
      foreach my $tmpl (qw(css dba tex xhtml)) {
	  my $target = join '.', $m->pvar('dok'), $tmpl;
	  next unless $m->svn_status_ekso($target);
	  unshift @pre_targets, $target;
	  my $nom = join '_', $tmpl, 'tmpl'; 
	  my $source = join '_', 'a2e', $nom;
	  $source = join '.', $source, 'txt';
	  $source = catfile '/opt/a2e/share/tmpl', $tml, $source;
	  $m->{$nom} = $source;
      } };
  POST_TARGETS: {
      last unless $m->pvar('dok') eq $m->{topdok};
      push @post_targets, 'sitemap.xml.gz';
      $m->{sitemap_tmpl} = catfile '/opt/a2e/share/tmpl', 'a2e_sitemap_tmpl.txt';

    };
  SITEMAP_INDEX: {
      last unless $m->{inddoks}->{$m->pvar('dok')};
      push @post_targets, 'sitemap_index.xml.gz';
      $m->{sitemap_index_tmpl} = catfile '/opt/a2e/share/tmpl', 'a2e_sitemap_index_tmpl.txt';
    };
  HTACCESS: {
      last unless $m->{tmplvars}->{privat} or $m->{tmplvars}->{hta_prv};
      last if $m->{tmplvars}->{hta_nil};
      last unless $m->{tmplvars}->{hta_sig} and $m->{tmplvars}->{hta_org};
      $m->uwarn('Private cuz tmplvars privat is %p, tmplvars hta_prv is %h, tmplvars hta_nil is %n', p => $m->{tmplvars}->{privat}, h => $m->{tmplvars}->{hta_prv}, n => $m->{tmplvars}->{hta_nil});
      push @pre_targets, '.htaccess';
    };
    $m->dpop('main8pre_targets', \@targets, \@pre_targets, \@post_targets, \@mlht_targets);
    return \@targets, \@pre_targets, \@post_targets, \@mlht_targets;
};

=head3 Function set_main_vars

=cut

sub set_main_vars {
    my $m = shift;
    $m->dpuc('set_main_vars', @_);
    my $dok = $m->set_dok;
    $m->{hsprintf_repl}->{d} = $dok;
    $m->mldk('dbms');
    # $m->riid_tmplvars_fayl($fayl);
    # $m->tmplvars_kk_progvars('be');
    $m->ustop('no tmpl variable found in %f', f => $fayl) unless $m->{tmplvars}->{tmpl};
    $m->dok_set_tmpl or $m->ustop('cant set tmpl of dok %d', f => $m->pvar('dok'));
    # $m->{cache}->{tmpl} = $m->pvar('tmpl');
    $m->techo('hta_nil is %v', v => $m->tmplvars_getstr('hta_nil'));
    my @langs = $m->set_langs;
    $m->set_template_include_dirs;
    foreach my $ll (@langs) {
	my $lvars = $m->{cache}->{lang_variants}->{$ll};
	next unless $lvars;
	push @{$m->{varlangs}}, $ll;
	my $varnom = join '_', $ll, 'variants';
	$m->{cache}->{conv_vars}->{$varnom} = $lvars;
    };
    $m->setpvar(lang0 => $langs[0]);
    $m->setpvar(lang => $langs[0]);
    $m->uwarn('unable to set template variables: %s', $@) if $@;
    $m->set_topdoks;
    $m->set_topdok;
    $m->{topdir} = $m->dokdir($m->{topdok});
    $m->{topdirdok} = catfile $m->{topdir}, $m->{topdok};
    my $css = join '.', $m->{topdirdok}, 'css';
    warn sprintf 'cant find css file %s' unless -s $css;
    $m->{hsprintf_repl}->{d} = $m->{cache}->{dir} = $m->{tmplvars}->{dir} = $m->{dir} = $m->dokdir;
    $m->{cache}->{url} = $m->{tmplvars}->{url} = $m->{url} = $m->dir2url($m->{dir}) or die sprintf 'cant find url for dok %s', $m->pvar('dok');
  NEWS_OPTS: {
      my @opts = ();
      my $kanals = $m->{konfig}->getkanals;
      last unless $kanals;
      $kanals =~ s(\s+)(,)g;
      push @opts, '--getkanals', $kanals;
      foreach my $nom (qw(news_start news_stop news_limit news_level)) {
	  my $val = $m->{konfig}->get($nom);
	  next unless $val;
	  push @opts, '--'.$nom, $val;
      };
      $m->{news_opts} = join ' ', @opts;
      $m->dgot(m_news_opts => $m->{news_opts});
    };
    $m->put('mlhtdok', { dok => $m->pvar('dok') }, doktmp=> 'now()', dokurl => $m->pvar('url')) or die sprintf 'cant enter url %2$s for dok %1$s into database', $m->pvar('dok'), $m->pvar('url');
    my $mdir = join '/', '..', 'mbox', $dok;
  MDIR: {
      last unless -d $mdir; 
      $m->{mdir} = $mdir;
      my $mbox = join '.', $m->pvar('dok'), 'mbox';
      $m->{mbox} = join '/', '..', 'mbox', $mbox;
    };
    $m->setplst(subdirs => $m->listref($m->subdirs)); # obsolete? file:/phm/18/09/24/sig/_lng.phm_pub_sig180924.txt::doksfiles
    $m->setplst(doksfiles => $m->listref($m->doksfiles));
  DOKSROW_HTML_TMPL: {
      my $fi = $m->nav_fayl();
      my $old_nav = 0;
      $m->open_fayl($fi, 'i', 'r');
      while ($m->getlini('i')) {
	  next unless m(\A\@);
	  $old_nav = 1;
	  $m->uwarn('still using outdated menu style in file %s, convert with nav_tmpl_repl?', $fi);
	  last;
      };
      $m->kloz_fayl('i');
      last if $old_nav;
      $m->{doksrow_html_tmpl} = join '_', $m->{topdok}, 'doksrow_html_tmpl', 'txt';
      $m->{doksrow_html_tmpl} = catfile $m->{topdir}, $m->{doksrow_html_tmpl};
      last if -e $m->{doksrow_html_tmpl};
      my $last = $m->{doksrow_html_tmpl};
      $m->{doksrow_html_tmpl} = catfile '/opt/a2e/share/tmpl', 'a2e_doksrow_html_tmpl.txt';
      $m->decho('template %s not found, using default %s instead', $last, $m->{doksrow_html_tmpl});
    };
  NEWS_HTML_TMPL: {
      $m->{news_html_tmpl} = join '_', $m->{topdok}, 'news_html_tmpl', 'txt';
      $m->{news_html_tmpl} = catfile $m->{topdir}, $m->{news_html_tmpl};
      last if -e $m->{news_html_tmpl};
      my $last = $m->{news_html_tmpl};
      $m->{news_html_tmpl} = catfile '/opt/a2e/share/tmpl', 'a2e_news_html_tmpl.txt';
      $m->decho('template %s not found, using default %s instead', $last, $m->{news_html_tmpl});
    };
  NEWS_RSS_TMPL: {
      $m->{news_rss_tmpl} = join '_', $m->{topdok}, 'news_rss_tmpl', 'txt';
      $m->{news_rss_tmpl} = catfile $m->{topdir}, $m->{news_rss_tmpl};
      last if -e $m->{news_rss_tmpl};
      $last = $m->{news_rss_tmpl};
      $m->{news_rss_tmpl} = catfile '/opt/a2e/share/tmpl', 'a2e_news_rss_tmpl.txt';
      $m->decho('template %s not found, using default %s instead', $last, $m->{news_rss_tmpl});
    };
  MLDOK: {
      $f = $m->dok_fayl();
      last unless -s $f;
      $m->{mldok} = $f;
    };
  MLNAV: {
      last if $m->{tmplvars}->{nonav};
      $f = $m->nav_fayl();
      last unless -s $f;
      $m->{mlnav} = $f;
    };
    $m->{mlht_opts} = '';
  MLHT_OPTS: {
      $m->dpuc('mlht_opts');
      my @opts = ();
      my @paks = @{$m->{konfig}->pakfnoms};
      $m->dgot(paks => \@paks);
      do { $m->dpop;last } unless @paks;
      foreach my $pak (@paks) {
	  $m->dpuc('pak', $pak);
	  do { $m->dpop;next } unless -e $pak;
	  push @opts, '--tf', $pak;
	  $m->dpop('pak', @opts);
      };
      $m->{mlht_opts} = join ' ', @opts;
      $m->dpop('mlht_opts', $m->{mlht_opts});
    };
    ($m->{main_targets}, $m->{pre_targets}, $m->{post_targets}, $m->{mlht_targets}) = $m->main8pre_targets;
    push @{$m->{main_targets}}, 'MH/threads.html' if $m->{mdir};
    $m->dpop('set_main_vars', 1);
    return 1;
};

=head3 Function make_defs

=cut

sub make_defs {
    my $m = shift;
    $m->dpuc('make_defs', @_);
    my $f = shift;
    my $fmt = shift || '%s := %s';
    my ($key, $val) = ();
    foreach my $key (@_) {
	$val = do { if (exists $m->{$key}) { $m->{$key} } elsif (exists $m->{tmplvars}->{$key}) { $m->{tmplvars}->{$key} } elsif (exists $m->{cache}->{$key}) { $m->{cache}->{$key} } };
	$m->techo('makedefs has set key %k to val %v, tmplvars of key was %t, progvars %c, object %o', k => $key, v => $val, t => $m->{tmplvars}->{$key}, c => $m->{cache}->{$key}, o => $m->{$key}); 
	$m->dpuc($key, $val, { key => $key, cache_val => $m->{cache}->{$key}, object_val => $m->{$key} });
      KONFIG: {
	  last if defined $val;
	  last unless $m->{konfig}->_exists($key);
	  $val = $m->{konfig}->get($key);
	  $m->dgot(konfig_val => $val); 
	};
	do { $m->decho('no value given for key %s', $key);$m->dpop;next } if 0 == length $val;
	$val = join ' ', @$val if ref $val;
	print STDERR '.' if $m->{konfig}->verbose;
	my $uckey = uc $key;
	my $str = sprintf $fmt, $uckey, $val;
	$m->print_tu_fayl($f, $str . "\n");
	$m->dpop($key);
    };
    $m->dpop('make_defs', 1);
    return 1;
};

=head2 Top level functions

=head3 Function make_makefile_cache_lang_sub

=cut

sub make_makefile_cache_lang_sub {
    my $m = shift;
    $m->dpuc('make_makefile_cache_lang_sub', @_);
    my $ll = shift;
    $m->set_lang_vars($ll);
    my $fo = $m->hsprintf($m->{konfig}->makefile_lang_cache_name, l => $ll);
    $m->open_fayl($fo, 'o', 'w');
    $m->vecho('writing %s', $fo);
    $m->print_tu_fayl('o', join "\n", '# '.$m->emacs_top_local_variables({ mode => 'makefile', coding => 'utf-8' }), '# '.$m->emacs_timestamp, '', '');
    $m->make_defs('o', '%s := %s', @{$m->{konfig}->lang_vars});
    $m->kloz_fayl('o');$m->vecho('OK');
    $m->dpop('make_makefile_cache_lang_sub', 1);return 1;
};

=head3 Function make_makefile_cache_lang

=cut

sub make_makefile_cache_lang {
    my $m = shift; 
    $m->dpuc('make_makefile_cache_lang', @_);
    $m->read_cache($m->{konfig}->makefile_cache_name);
    $m->make_makefile_cache_lang_sub(@_);
    $m->dpop('make_makefile_cache_lang', 1);return 1;
};

=head3 Function make_makefile_cache_tml

=cut

sub make_makefile_cache_tml {
    my $m = shift;
    $m->dpuc('make_makefile_cache_tml', @_);
    my $tml = shift;
    unless ($tml eq $m->{tmplang}) { # file:/phm/18/09/25/sig/_lng.phm_pub_sig180925.txt::maketml
	($m->{main_targets}, $m->{pre_targets}, $m->{post_targets}, $m->{mlht_targets}) = $m->main8pre_targets($tml);
    };
    my $fo = $m->hsprintf($m->{konfig}->makefile_tml_cache_name, t => $tml);
    $m->open_fayl($fo, 'o', 'w');
    $m->vecho('writing %s', $fo);
    $m->print_tu_fayl('o', join "\n", '# '.$m->emacs_top_local_variables({ mode => 'makefile', coding => 'utf-8' }), '# '.$m->emacs_timestamp, '', '');
    $m->make_defs('o', '%s := %s', @{$m->{konfig}->tml_vars});
    $m->make_defs('o', '%s := %s', @{$m->{konfig}->get($tml . '_vars')});
    $m->kloz_fayl('o');$m->vecho('OK');
    $m->dpop('make_makefile_cache_tml', 1);return 1;
};


=head3 Function make_makefile_cache

=cut

sub make_makefile_cache {
    my $m = shift;
    $m->dpuc('make_makefile_cache', @_);
    $m->set_main_vars;
    my $varsfayl = $m->hsprintf($m->{konfig}->varsfnom, p => $m->{srcfpfx}, d => $m->pvar('dok'));
    $m->open_fayl($varsfayl, 'o', 'w');
    $m->print_tu_fayl('o', join "\n", $m->emacs_top_local_variables({ mode => 'makefile', coding => 'utf-8'}), '','');
    $m->make_defs('o', '%s ?= %s', @{$m->{konfig}->dok_vars});
    $m->kloz_fayl('o');$m->vecho('OK');
    my $fo = $m->{konfig}->makefile_cache_name;
    printf STDERR 'writing %s', $fo if $m->{verbose};
    $m->open_fayl($fo, 'o', 'w');
    $m->print_tu_fayl('o', join "\n", $m->emacs_top_local_variables({ mode => 'makefile', coding => 'utf-8' }), '', '');
    $m->print_tu_fayl('o', 'include %s'."\n", catfile 'a2e', 'hostsyskonfig.mk');
    $m->print_tu_fayl('o', 'include %s'."\n", $varsfayl);
    $m->make_defs('o', '%s ?= %s', @{$m->{konfig}->rest_vars});
    my $conv = $m->{konfig}->conv;
    $m->pstop('invalid converter %s', $conv) unless grep { $_ = $conv } @{$m->{konfig}->convs};
    $m->{konfig}->conv_vars->{html_opts} = join ' ', '-A', 'X,W', '-f', 'xhtml10t', '-t', "$m->{topdirdok}.xhtml";
    $m->{konfig}->conv_vars->{latex_opts} = join ' ', '-X', '-f', 'latex', '-t', join '.', grep { $_ } $m->{topdirdok}, $m->{tmplvars}->{latex_tmpl}, 'tex';
    $m->{cache}->{$_} = $m->{cache}->{conv_vars}->{$_} for keys %{$m->{cache}->{conv_vars}};
    $m->make_defs('o', '%s ?= %s', keys %{$m->{cache}->{conv_vars}});
    $m->kloz_fayl('o');
  TMLS: {
      $m->dpuc('tmls');
      my @tmls = $m->uniq_plst('tmls');
      $m->make_makefile_cache_tml($_) for @tmls;
      $m->dpop('tmls', @tmls);
    }
  LANGS: {
      $m->dpuc('langs');
      my @langs = $m->uniq_plst('langs');
      $m->make_makefile_cache_lang($_) for @langs;
      $m->dpop('langs', @langs);
    }
    $m->vecho('OK');
    $m->dpop('make_makefile_cache', 1);
    return 1;
};

=head1 TODO

=over

=item Create a version that does not need the database, so that make_site_makefile_cache can be invoked from an RPM script!

=item Have lng_faylz_src check all files for C<include> statements so that the --lng_faylz_src_extra option will rarely or never be needed.

=back

=cut

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile:/adv/perl/A2E/MakefileCache.pm.tmpl
# End:
 
