=head1 NAME A2E::Mbox2dok

=cut

package A2E::Mbox2dok;
use strict;

=head1 VERSION 0.1.0

=cut

our $VERSION = '0.1.0';

=head1 DESCRIPTION

mix-in functionality for mbox2dok, see A2E::Lokvars(3) regarding the mix-in concept.

=head1 PREREQUISITES

=cut

use base 'A2E::Dokfs';
use File::Spec::Functions qw(splitdir catdir catfile);
use Cwd;

=head1 APPLICATIONS

mbox2dok(1)
A2E::Mktdir(3)

=head1 IMPLEMENTATION

=head2 Overwrite Functions

=head3 Function mbox2dok_defvars

the non overwritable part for mixin use

=cut

sub mbox2dok_defvars {
    my $m = shift;
    $m->dpuc('mbox2dok_defvars', @_);
    $m->{konfig}->define('sig_iar_mon_rest_pats=s@', { DEFAULT => [ '\A\/sig\/(\w+)\/(\d{2})\/(\d{2})\/?(.*)\Z', '\A\/(\w+)\/sig\/(\d{2})\/(\d{2})\/?(.*)\Z' ] });
    $m->{konfig}->define('mbox_sig2fmt=s%', { DEFAULT => { ffiinnc => catdir('', 'ffiinnc', 'sig', '%02d', '%02d', 'mbox') } });
    $m->{konfig}->define('mbox_fmt=s', { DEFAULT => catdir('', 'sig', '%s', '%02d', '%02d', 'mbox') });
    $m->{konfig}->define('srcsrcdir_sig2fmt=s%', { DEFAULT => { ffiinnc => catdir('', 'mail', 'i2p', '%02d', '%02d', 'f') } });
    $m->{konfig}->define('srcsrcdir_fmt=s', { DEFAULT => catdir('', 'mail', '%s', '%02d', '%02d') });
    $m->konfayl('mbox2dok.konf');
    $m->dpop('mbox2dok_defvars', 1);
    return 1;
};

=head3 Function A2E::Mbox2dok::defvars

=cut

sub defvars {
    my $m = shift;
    $m->dpuc('A2E::Mbox2dok::defvars', @_);
    $m->SUPER::defvars;
    $m->mbox2dok_defvars;
    $m->dpop('A2E::Mbox2dok::defvars', 1);
    return 1;
};

=head2 Configuration Functions (overwritable)

=head3 Function rest_dok2bas

Deriving the mbox name from the rest (relative filename starting after current month mbox directory) and the dok, as in

  rest:adv, dok:oas_adv1003 --> adv
  rest:ocr, dok:od_ocr1003 --> d_ocr 
  rest:ocr, dok:d_ocr1003 --> ocr 
  rest:ocr, dok:adv_ocr1003 --> ocr 
  rest:adv/maus07, dok: od_maus100307 --> d_maus07
  rest:adv/maus07, dok: oas_maus100307 --> adv_maus07


=cut

sub rest_dok2bas ($$$) {
    my $m = shift;
    $m->dpuc('rest_dok2bas', @_);
    my $rest = shift;
    my $dok = shift;
    do { $m->dpop('rest_dok2bas', 'mbox');return 'mbox' } unless $rest;
    my $bas = '';
  BAS: {
      my @dirs = grep {$_} splitdir $rest;
    PREF: { # dok:od_ocr1003 --> pref:d, bas:d_ocr
	my ($pref) = $dok =~ m(\A\w(\w)\_$rest\d+\Z);
	last unless $pref;
	$bas = $pref.'_'.$dirs[$#dirs];
	last BAS;
      };
      $bas = join '_', @dirs;
    };
    $m->dpop('rest_dok2bas', $bas);
    return $bas;
};

=head2 Special Functions

=head3 Function mbox2dok

basic variant for use in A2E::Mktdir, usable in directories that are not yet committed to svn

=cut

sub mbox2dok {
    my $m = shift;
    $m->dpuc('mbox2dok', @_);
    my $target = shift;
    $m->set_pwd;
    my $dok = $m->get_dok;
    my ($sig, $iar, $mon, $rest, $pat);
    foreach $pat (@{$m->{konfig}->sig_iar_mon_rest_pats}) {
	$pat = qr{$pat};
	# pwd:/sig/oas/10/03/ocr --> sig:oas, iar:10, mon:03, rest:ocr
	($sig, $iar, $mon, $rest) = $m->{pwd} =~ m($pat);
	last if $mon;
    };
    my $src = shift || '';
    $m->dgot(sig => $sig, iar => $iar, mon => $mon, pat => $pat, dok => $dok, src => $src);
    do { $m->dpop;$m->vecho('we do not expect directories like %s to have their own mbox', $m->{pwd});return } unless $mon;
    my $ok = 0;
  NOSRC: {
      $ok = length $src ? 0 : 1;
      $m->dpuc('nosrc', $ok);
      do { $m->dpop;last } unless $ok;
      my $srcdir = do { my $fmt = $m->{konfig}->mbox_sig2fmt->{$sig};if ($fmt) { sprintf $fmt, $iar, $mon } else { $fmt = $m->{konfig}->mbox_fmt;die 'i have no idea where place the mbox directory symlink' unless $fmt;sprintf $fmt, $sig, $iar, $mon; } };
    NOSRCDIR: {
	$ok = -e $srcdir ? 0 : 1;
	$m->dpuc('nosrcdir', $ok, $srcdir);
	do { $m->dpop;last } unless $ok;
	my $srcsrcdir = do { my $fmt = $m->{konfig}->srcsrcdir_sig2fmt->{$sig};if ($fmt) { sprintf $fmt, $iar, $mon } else { $fmt = $m->{konfig}->srcsrcdir_fmt;$m->ustop('no way to build a source directory name') unless $fmt;sprintf $fmt, $sig, $iar, $mon } };
	die sprintf 'source directory %s does not exist', $srcsrcdir unless -d $srcsrcdir;
	symlink $srcsrcdir, $srcdir or die sprintf 'failed to link from mail source directory %s to intermediate source directory %s', $srcsrcdir, $srcdir;
	$m->svn_add($srcdir);
	$m->svn_commit($srcdir) or die sprintf 'couldnt commit %s', $srcdir;
	$m->dpop;
      };
      $srcdir = $m->reldir($m->{pwd}, $srcdir);
      # rest:ocr, dok:od_ocr1003 --> d_ocr 
      $src = catfile $srcdir, $m->rest_dok2bas($rest, $dok);
      $m->dpop;
    };
    die sprintf 'link source %s does not exist for cwd %s', $src, getcwd, $m->{cwd} unless -e $src;
    my $obj = join '.', $dok, 'mbox';
  OBSTAKL: {
      last unless -e $obj;
      unlink $obj;
      $ok = -e $obj;
      $m->dpuc('obstakl', $ok, $obj);
      do { $m->dpop;last } unless $ok;
      die sprintf 'file system object %s is in the way', $obj;
    };
    $m->dgot(src => $src, obj => $obj);
    symlink $src, $obj or die sprintf 'can not symlink from %s to %s', $src, $obj;
    die 'link object %s does not exist' unless -l $obj;
    $m->svn_add($obj);
    $m->dpop('mbox2dok', $obj);return $obj;
};

=head3 Function mbox2dok_commit

commmandline_variant: assuming that we are working in a directory that is already in svn

=cut

sub mbox2dok_commit {
    my $m = shift;
    $m->dpuc('mbox2dok_commit', @_);
    my $mbox = $m->mbox2dok(@_);
    if (length $mbox && -l $mbox) {
	$m->svn_commit($mbox) or die sprintf 'can not commit link object %s', $mbox;
    } else {
	die sprintf 'failed to %s', $m->koniug_et('mbox2dok', @_);
    };
    $m->exek('make', 'newmbox');
    $m->dpop('mbox2dok_commit', 1);return 1;
};

=head1 THE END

=cut

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: /adv/perl/A2E/Mbox2dok.pm.tmpl
# End:
