#!/usr/bin/perl
# Time-stamp: <2018-10-30 12:10:20 phm>

# BEGIN { $ENV{PATH} = '/opt/a2e/bin:/bin:/usr/local/bin:/usr/bin' };

=head1 NAME

    A2E::Mktop

=head1 VERSION

    This describes version B<0.1.0> of A2E::Mktop

=head1 SYNOPSIS

    my $p = new_ready A2E::Mktop;
    $p->mktop('tra');

etc

=head1 DESCRIPTION

See to it that the menus do not point to a higher level and are not pointed to from there.  Place initial stylesheets, templates and graphic files here so that they can be developped independently for this microsite.

The one and only parameter is the name of a style.  Currently 'a2e' and 'tra' exist.

=head1 OPTIONS

=over

=item --help | -h

Display short help and exit.

=item --manpage | -m

Display full documentation and exit.

=back

=head2 General options:

=over

=item --quiet

Let me keep silent while I run.

=item --verbose | -v

Let me explain what I am doing while I run.

=item --nosvn

Do not use Subversion

=item --daba <daba>

Use this database instead of the 'a2e' database.

=back

=head1 FILES

=head1 BUGS, LIMITATIONS

=head1 PREREQUISITES

    A2E::Mktdir(3)

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

Copyright (c) 2007-8 Hartmut Pilch (phm)

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=cut

package A2E::Mktop;
use base 'A2E::Mktdir';
use strict;
use File::Spec::Functions qw(catfile);

sub defvars {
    my $m = shift;
    $m->dpuc('A2E::Mktop::defvars', @_);
    $m->{progver} ||= '0.1.0';
    $m->SUPER::defvars;
    $m->konfayl('mktop.konf');
    $m->{svn_faylz} = {};
    $m->dpop('A2E::Mktop::defvars', 1);return 1;
};

sub postkonfig {
    my $m= shift;
    $m->dpuc('A2E::Mktop::postkonfig', @_);
    $m->SUPER::postkonfig;
    @{$m->{kopi_komand}} = grep {$_} ('cp', $m->{verbose} && '-v', '-f');
    $m->dpop('A2E::Mktop::postkonfig', 1);return 1;
};

sub topdok_kopi {
    my $m = shift;
    $m->dpuc('topdok_kopi', @_);
    my $sfx = shift;
    my $repl = shift;
    my $fo = join '.', $m->pvar('dok'), $sfx;
  FO_EXISTS: {
      my $ok = -e $fo;
      $m->dpuc('fo_exists', $ok, $fo);
      do { $m->dpop;last } unless $ok;
      $m->dpop('topdok_kopi');return;
    };
    my $fi_bas = join '.', $m->{konfig}->topsty, $sfx;
    my $fi = catfile $m->{konfig}->tmplsupdir, $m->{konfig}->tmplang, $fi_bas;
    unless (-e $fi) { warn sprintf 'no file %s, falling back', $fi;$fi = catfile $m->{konfig}->tmplsupdir, $fi_bas };
    die sprintf 'no file %s', $fi unless -e $fi;
    if ($repl) {
	$m->transkopi($fi, $fo, $repl);
    } else {
	$m->exek(@{$m->{kopi_komand}}, $fi, $fo);
    };
    my $ok = -s $fo;
    $m->{svn_faylz}->{$fo} = 1 if $ok;
    $m->dpop('topdok_kopi', $ok);return $ok;
};

sub mktop {
    my $m = shift;
    $m->dpuc('mktop', @_);
    my $topsty = shift;
    $m->{konfig}->topsty($topsty) if $topsty;
    $m->setpvar(dok => shift || $m->get_dok)
    die 'cant set DOK' unless length $m->pvar('dok');
    my $pwd = $m->set_pwd;
    $m->put('dokflags', { dok => $m->pvar('dok') }, varnom => 'topdok', varval => 'true');
    my $url = $m->pvar('url') || $m->dok2relurl($m->pvar('dok'));
    my $dok = $m->pvar('dok');
    $m->setpvar(url => $url);
    my $repl = { topdir => $pwd, topdok => $dok, topurl => $url };
    $m->svn_cleanup;
    $m->topdok_kopi($_, $repl) for qw(css xhtml);
    $m->topdok_kopi($_) for qw(png tex dba);
    my @faylz = keys %{$m->{svn_faylz}};
    $m->svn_add(@faylz) if @faylz;
    $m->dpop('mktop', @faylz);return @faylz;
};

=head1 The END

=cut

return 1;

# Local Variables:
# coding: utf-8
# URL: http://a2e.de/adv/perl/A2E/Mktop.pm.tmpl
# srcfile: /adv/perl/A2E/Mktop.pm.tmpl
# mode: perl
# End:
