=head1 NAME A2E::Nav2html

=cut
 
package A2E::Nav2html;
use base A2E::Dokfs;
use strict;

=head1 DESCRIPTION

Generate a left-hand menu - deprecated old way of doing it, use at a2e.de in files created before 2009-03-09.

Basis of deprecated program nav2html(1).  Nowadays we perform the same functionality by means of templates that use certain vmethods such as C<predokreks>, which we have defined in A2E::MLHT(3).

=head1 IMPLEMENTATION

=head2 Overload Functions

=head3 Function A2E::Nav2html::defvars

=cut

sub defvars {
    my $m = shift;
    $m->dpuc('A2E::Nav2html::defvars', @_);
    $m->SUPER::defvars;
    $m->{konfig}->define('urllab|u!');
    $m->{konfig}->htmlesc(1);
    $m->{konfig}->lifmt('<li>%s</li>');
    $m->konfayl('nav2html.konf');
    $m->dpop('A2E::Nav2html::defvars', 1);
    return 1;
};

=head3 Function A2E::Nav2html::postkonfig

=cut

sub postkonfig {
    my $m = shift;
    $m->dpuc('A2E::Nav2html::postkonfig', @_);
    $m->SUPER::postkonfig;
    $m->{ahref_item} = $m->make_ahref_item;
    $m->dpop('A2E::Nav2html::postkonfig', 1);
    return 1;
};

=head2 Special Functions

=head3 Functions open_sidebar etc

=cut

sub open_sidebar {
    my $m = shift;
    $m->dpuc('open_sidebar', @_);
    ++$m->{sidebar};
    my $res = "$m->{topline}\n<div id=\"sidebar\">\n";
    $m->dpop('open_sidebar', $res);
    return $res;
};

sub close_sidebar {
    my $m = shift;
    $m->dpuc('close_sidebar', @_);
    --$m->{sidebar};
    my $res =  "</div>\n";
    $m->dpop('close_sidebar', $res);
    return $res;
}

sub open_sidepanel {
    my $m = shift;
    $m->dpuc('open_sidepanel', @_);
    my $tit = shift;
    ++$m->{sidepanel};
    my $str = '<div class="sidepanel">'."\n";
    $str .= sprintf '<p class="sideBarTitle">%s</p>'."\n", $tit if length $tit;
    $m->dpop('open_sidepanel', $str);
    return $str;
}

sub close_sidepanel {
    my $m = shift;
    --$m->{sidepanel};
    return "</div>\n";
}

sub open_ulist {
    my $m = shift;
    ++$m->{ulist};
    return "<ul>\n";
};

sub close_ulist {
    my $m = shift;
    --$m->{ulist};
    return "</ul>\n";
};

sub hreftextitem {
    my $m = shift; 
    my ($href, $text) = @_;
    return sprintf "<li>%s</li>\n", $m->hreftext($href, $text);
};

sub navmenu_entries {
    my $m = shift;
    $m->dpuc('navmenu_entries', @_);
    my $tit = shift;
    do { $m->dpop;return } unless length $tit;
    my @entries = $m->dokmenu_entries($m->{konfig}->dok);
    do { $m->dpop;return } unless @entries;
    $m->dpop('navmenu_entries', $tit, @entries);return $tit, @entries;
};

sub print_dokmenu {
    my $m = shift;
    $m->dpuc('print_dokmenu', @_);
    my $tit = shift;
    my @entries = ();
    print $m->close_ulist while $m->{ulist};
    print $m->close_sidepanel while $m->{sidepanel};
    if (m(\A\@(.+)\s*\Z)) { ($tit, @entries) = $m->navmenu_entries($1);return unless $tit };
    print $m->open_sidepanel($m->htmlesc($tit));
    print $m->open_ulist;
    print join "\n", @entries;
    $m->dpop('print_dokmenu', @entries);return @entries;
};

sub make_ahref_item {
    my $m = shift;
    $m->dpuc('make_ahref_item', @_);
    my $sub = shift;
  URLLAB: {
      my $ok = $m->{konfig}->urllab;
      $m->dpuc('urllab', $ok);
      do { $m->dpop;last } unless $ok;
      $sub = sub {
	  my $url = shift;
	  my $lab = shift;
	  return $m->ahref_item($url, $lab);
      };
      $m->dpop('make_ahref_item', $sub);
      return $sub;
    };
    $sub = sub {
	my $lab = shift;
	my $url = shift;
	return $m->ahref_item($url, $lab);
    };
    $m->dpop('make_ahref_item', $sub);
    return $sub;    
};
=head3 Function nav2html1

=cut

sub nav2html1 {
    my $m = shift;
    $m->dpuc('nav2html1', @_);
    my $fi = shift;
    $m->open_fayl($fi, 'i', 'r');
    my ($url, $lab, $tit, $nl, @matches, @entries) = ();
    $nl = 0;
    print sprintf '<!-- %s -->'."\n", $m->emacs_timestamp if $m->{konfig}->htmlesc;
  LINI: while ($m->getlini('i')) {
      $nl++; 
      next if /\A\#/;
      ($url, $lab, $tit) = ();
      chomp;
      $m->dpuc('lini'.$nl, $_);
      s(\s+\Z)();
      if (m(\A(\S.*)\Z)) { $m->print_dokmenu($1);$m->dpop('lini'.$nl);next LINI };
      /\A\s+href\:(\S+)\s\t\s*(\S.*\S)\Z/ && do { print &{$m->{ahref_item}}($1, $2), "\n";$m->dpop('lini'.$nl);next LINI };
      /\A\s+dok\:(\S+)\s*\t\s*(\S.*\S)\Z/ && do { print &{$m->{ahref_item}}($1, $2), "\n";$m->dpop('lini'.$nl);next LINI };
      /\A\s+(\S+)\s*\t\s*(\S.*\S)\Z/ && do { print &{$m->{ahref_item}}($1, $2), "\n";$m->dpop('lini'.$nl);next LINI; };
      /\A\s+dok\:([\w\_]+)\Z/ && do { print $m->ahref_item($1), "\n";$m->dpop('lini'.$nl);next LINI };
      /\A\s+href\:(\S+)\Z/ && do { print $m->ahref_item($1), "\n";$m->dpop('lini'.$nl);next LINI };
      /\A\s+([\w\_]+)\Z/ && do { print $m->ahref_item($1), "\n";$m->dpop('lini'.$nl);next LINI };
      /\A\s+([^\t\n]+)\Z/ && do { print $m->ahref_item( -r $1 && $1, $1), "\n";$m->dpop('lini'.$nl);next LINI };
      $m->dpop;
  };
    $m->kloz_fayl('i');
    $m->dpop('nav2html1', 1);
    return 1;
};

=head3 Function nav2html

=cut

sub nav2html {
    my $m = shift;
    $m->dpuc('nav2html', @_);
    my $fi = shift;
    $m->{konfig}->infaylz($fi);
    my ($dok, $lang) = $m->{konfig}->infaylz->[0] =~ m(\A(\w+)\.nav\.(\w{2})\.txt) unless $m->pvar('dok'); 
    die 'failed to identify document' unless $dok;
    die 'failed to determine language' unless $lang;
    $m->setpvar(dok => $dok);
    $m->setpvar(lang => $lang);
    $m->init_navdoks;
    $m->{sidebar} = 0;
    $m->{sidepanel} = 0;
    $m->{ulist} = 0;
    $m->{topline} = sprintf '<!-- -*- coding: utf-8; mode: fundamental; srcfile: %s; -*- -->', $m->koniug_et(@{$m->{konfig}->infaylz});
    $m->nav2html1($_) for @{$m->{konfig}->infaylz};
    print $m->close_ulist while $m->{ulist};
    print $m->close_sidepanel while $m->{sidepanel};
    $m->sav2out;
    $m->dpop('nav2html', 1);return 1;
};

=head1 The End

=cut

return 1;

# Local Variables:
# coding: utf-8
# srcfile: /adv/perl/A2E/Nav2html.pm.tmpl
# url: http://a2e.de/adv/perl/A2E
# mode: perl
# End:
