=head1 NAME

A2E::Open

=cut

package A2E::Open;

=head1 DESCRIPTION

Command redirection, applications of IPC::Open3

=head1 PREREQUISITES
IPC::Open3(3)
=cut

use base A2E::Prog;
use IPC::Open3;
use IO::File;

=head2 Log

[[file:/phm/21/02/07/sig/_lng.phm_pub_sig210207.txt::adv][Y27]] 遷自 file:/adv/perl/A2E/Prog.pm.tmpl::exek_open3

=head3 Function make_pipe_sig

    Return a reference to routine that can be given to C<SIG{PIPE}> for invocation when a pipe is broken.

    Used by exek_open3, but it does not seem to be achieving anything there.

    See perldoc perlipc to learn better.

=cut

sub make_pipe_sig {
    my $db = shift;
    return sub { warn 'pipe broken 9';$db->{err} = 9 };
};

=head3 Function exek_open3

    exek_open3: like exek, use open3, store stdin/stdout/stderr for later use

    Execute a system command, return error, stdout and stderr

    Uses IPC::Open3.  Capturing of STDERR doesn't seem to be working.
    The author may not have fully understood IPC::Open3 and the perlipc manual but is trying to encapsulate what he understood into this command.

    arg is a reference to a list consisting of the program and arguments.
    The remaining arguments are input lines.  The output of the system command is captured into @col, its error messages into @cel.  Much like exek, exek_open3 returns 1 if the system command was ended successfully (with 0 exit value), 0 otherwise.

    This is achieved by collecting the exit value resulting from the waitpid call, rather than by the closure function handed to the SIG{PIPE} interrupt.  The latter doesn't seem to be working.
   	
    TODO: learn from file:/sig/oas/18/11/gentVB/tra/Daikufile::Open3, file:/sig/oas/18/12/nokjVC/tra/jp/Daikufile::Open3
=cut

use IPC::Open3;
sub exek_open3 {
    my ($m,$kmd,@inlins) = @_;
    my @komand = eval { @$kmd };
    $m->pstop('open3_nokmd') unless @komand;
    my ($ci, $co, $ce) = ();
    local $SIG{PIPE} = $m->make_pipe_sig;
    my $err = 0;
    my $pid = open3($ci, $co, $ce, @komand);
    $m->pstop(snedef => s => 'ci') unless $ci;
    $m->pstop(snedef => s => 'co') unless $co;
    $m->pstop(snedef => s => 'ce') unless $ce;
    return unless $pid;
    my @col = ();
    my @cel = ();
  PRINT_CI: while ($_ = shift @inlins) {
      print $ci $_, "\n";
  };
    if (defined $ce) {
      PUSH_CE: while (<$ce>) {
	  chomp $_;
	  push @cel, $_;
      };
	close $ce;
    } else { $m->uwarn('open3perr') };
  PUSH_CO: while (<$co>) {
      chomp $_;
      push @col, $_;
  };
    close $co or $m->pwarn('conoferm');
    close $ci or $m->pwarn('cinoferm');
    waitpid $pid, 0;
    $err ||= $? >> 8;
    # my $ok = $err == 0;
    return $err, \@col, \@cel;
}

sub fo_kmd { # use file $fo to collect output of command @kmd, e.g. $m->fo_kmd('doku.txt', 'lynx', '-dump', 'doku.html')
    my ($m, $fo, @kmd) = @_;
    # file:/sig/oas/18/11/gentVB/tra/Daikufile::open3
    my $pid = open3 $rdr, $wtr, $err, @kmd;
    my $ho = new IO::File $fo, 'w';
    my $so = <$wtr>;
    print STDERR "first slurp: $so";
    while ($so = readline $wtr) { $ho->print($so) };
    $ho->close;
    waitpid $pid, 0;
    my $exit_status = $? >> 8;
    printf "exit status: %d\n", $exit_status;
    my $fsz = -s $fo;
    printf "size of file %s is %d\n", $fo, $fsz;
    $fsz;
}
sub lynx_dump { # app file:/sig/adv/prg/lynx_dump.pl tested file:/phm/21/02/07/sig/_lng.phm_pub_sig210207.txt::adv
    my ($m, $fi, $fo) = @_;
    $m->fo_kmd($fo, 'lynx', '-dump', '-width', '9999', '-nomargins', '-image_links', '-display_charset=utf-8', '-assume_local_charset=utf-8', 'assume_unrec_charset=utf-8', $fi);
};
1;

# Local Variables:
# coding: utf-8
# mode: perl
# End:
