=head1 NAME

A2E::SVN

=head1 VERSION

This describes version B<0.0.1>

=cut

$VERSION = '0.0.1';

=head1 DESCRIPTION

wrapper around SVN::Client

=head1 BUGS

=head1 PREREQUISITES

         SVN::Client(4)
         A2E::Prog(3)
         Sys::Hostname(3)

=head1 APPLICATIONS

        A2E::Dokfs(3)

=head1 SEE ALSO

       A2E::Lokvars(3)

on mixin mechanism 

=head2 Alternatives on multiple inheritance

    mixin(3)
    Class::C3(3)

=head1 AUTHOR

    Hartmut Pilch

=cut

package A2E::SVN;
use base 'A2E::Prog';
use Cwd qw(getcwd abs_path);
use File::Basename;
use SVN::Client;
use POSIX qw(strftime);
use strict;
use File::Spec::Functions qw(catdir catfile splitpath);
use File::Path qw(rmtree);
use Sys::Hostname;

=head1 OPTIONS

=cut
our %MSGS = (
    failfayl => 'failed to add file %f to subversion repository: %s',
    faylekzistp => 'file %f is $(SW|no longer|still) there',
    nodir => 'no such directory %d',
    svn_off => 'svn inactive in procedure %p',
    noinfourl => 'cant obtain svn info url for %p aka %d',
    noinfofrom => 'cant obtain svn info from %p aka %d: %m' );

sub spuc { # quit if ready
    my ($m,$prog,@args) = @_;
    return if $m->{svn_repos}->{offline};
    $m->dpuc($prog,@args);
    ($m,$prog,@args);
};
sub svn_defvars {
    my ($m,$prog) = shift->qpuc('svn_defvars',@_) or return;
=over
=item --svn_off
dont use SVN at all in this application
=cut
    $m->konfig_define('svn_off', '!', $ENV{SVN_OFF} || '0');
    $m->prog_define('svn_add_ok', '!', $ENV{SVN_ADD_OK} || '0');
=item --svn_offline
=cut
    $m->{konfig}->define('svn_offline!', { DEFAULT => $ENV{SVN_OFFLINE} });
=item --svnhostnames
=cut
    $m->{konfig}->define('svnhostnames=s@');
=item --svnhost
=cut
    $m->cache_define([ 'svnhost', 'H' ]);
=item --svn_info_dir
=cut
    $m->konfig_define('svn_info_dir', 's', { DEFAULT => $ENV{SVN_INFO_DIR} || '/svn/a2e' });
=back
=head1 FILES

	/etc/opt/a2e/svn.konf
	~/.svn_konf
	./_svn

=cut

    $m->konfayl('svn.konf');
    $m->prog_msgdefs('svn',%MSGS);
    $m->dpop1($prog,1);
};

sub defvars {
    my ($m,$prog) = shift->qpuc('A2E::SVN::defvars',@_) or return;
    $m->SUPER::defvars;
    $m->svn_defvars;
    $m->dpop1($prog,1);
};

=head3 Function svn_info

encapsulate {svn}->info to avoid hassle.
e.g. the directory must exist and be a svn directory.
$rcv is a receiver function.
optional arg2 is a directory.
return the svn info url if svn info is available.

=cut

sub svn_info { 
    my ($m,$prog,$dir,$rev,$rcv,$rek) = shift->ppuc('svn_info', @_);
    $m->{svn_repos}->{offline} = $m->{konfig}->svn_offline;
    $m->{svn_repos}->{off} = $m->{konfig}->svn_off;
    return $m->Wpop1($prog, u => svn_off => p => $prog) if $m->{svn_repos}->{off};
    $dir ||= $m->{konfig}->svn_info_dir || $m->pstop('no svn_info_dir specified');
    my $paf = &abs_path($dir);
    return $m->Wpop0($prog, u => nodir => d => $dir) unless -d $paf;
    $rcv ||= $m->{svn_info_receiver};
    $rek //= 0;
    $rev = undef if !$rev or $m->{svn_repos}->{offline};
    my $url = eval { $m->{svn_repos}->{url} = undef;$m->{svn}->info($paf, undef, $rev, $rcv, $rek);$m->{svn_repos}->{url} };
    return $m->Wpop0($prog, p => noinfofrom => p => $paf, d => $dir, m => $@) if $@;
    return $m->Wpop0($prog, p => noinfourl => p => $paf, d => $dir) unless $url;
    $m->dpop1($prog,$url);
};

sub svn_lastmod {
    my ($m,$prog,@args) = shift->ppuc('svn_lastmod', @_);
    my $lastmod = $m->{hodie};
    my $receiver = sub {
	my ($path, $info, $pool) = @args;
	my $sekund = $info->last_changed_date / 1000000;
	my @tempus = localtime $sekund;
	$lastmod = strftime '%Y-%m-%d', @tempus;
    };
    $m->svn_info(&getcwd);
    $m->dpop1($prog,$lastmod);
};

sub svn_log { 
    my ($m,$prog,$log_receiver,$fyl) = shift->ppuc('svn_log', @_);
    $fyl ||= &getcwd();
    $m->{svn}->log($fyl, 'HEAD', 0, 0, 1, $log_receiver);
    $m->dpop($prog, 1);
};

=head3 Function svn_postkonfig

=cut

sub svn_postkonfig {
    my ($m,$prog) = shift->qpuc('svn_postkonfig',@_) or return;
    $m->{hostname} = hostname;
    my ($localhost) = grep { $_ eq $m->{hostname} } @{$m->{konfig}->svnhostnames}; 
    $m->{konfig}->svnhost('localhost') if $localhost;
    $m->dgot(localhost => $localhost, hostname => $m->{hostname}, svnhost => $m->{konfig}->svnhost);
    $m->{svn} = new SVN::Client or die 'failed to construct a svn client';
    my $user = $m->{konfig}->user or $m->pstop("cant get user in $prog");
    my $fpas = catfile '', 'home', $user, '.svn_pass';
    my $pass = $m->fayl_get_str($fpas) or $m->ustop("cant read $fpas in $prog");chomp $pass;
    my $kfgdir = $m->{konfig}->svn_info_dir or $m->pstop("failed to obtain kfgdir in $prog");
    my $kfg = SVN::Core::config_get_config($kfgdir) or $m->pstop("failed to obtain kfg in $prog");
    my $kfgnom = $SVN::Core::CONFIG_CATEGORY_CONFIG or $m->pstop("failed to obtain kgfnom in $prog");
    my $kfgktg = $kfg->{$kfgnom} or $m->pstop("failed to obtain config category");
    my $auth_baton = SVN::Core::cmdline_create_auth_baton(1,$user,$pass,$kfgdir,1,1,$kfgktg,undef,undef) or $m->pstop("failed to obtain auth baton");
    $m->{svn}->auth($auth_baton) or $m->pstop("failed to auth with baton");
    $m->{svn_info_receiver} = sub {
	my ($path, $info, $pool) = @_;
	my $repos_url = $info->repos_root_URL;
	my ($prot, $host, $hdir) = $repos_url =~ m(\A(svn|file|https?)\:\/\/([a-z][^\/]*)\/(.*)\Z);
	my $url = $info->URL;
	return unless $url;
	my ($stem) = $url =~ m(\A$repos_url\b(.*)\Z);
	my (@fols) = split '/', $stem;shift @fols unless $fols[0];
	my $cwd = getcwd;
	my ($root) = $cwd =~ m(\A(.*)$stem\Z);
	my %info = (prot => $prot, host => $host, hdir => $hdir, url => $url, stem => $stem, fols => \@fols, root => $root, off => $m->{konfig}->svn_off, offline => $m->{konfig}->svn_offline);
	foreach my $k (keys %info) { $m->{svn_repos}->{$k} = $info{$k} };
	$url;
    };
    my $url = '';
  OFFLINE: {
      $m->{svn_repos}->{offline} ||= $ENV{SVN_OFFLINE};
      $m->{svn_repos}->{off} ||= $ENV{SVN_OFF};
      last if $m->{svn_repos}->{off};
      $url = eval { $m->svn_info($kfgdir) };
      do { $m->{svn_repos}->{offline} = 1;$m->{svn_repos}->{off} = 1;$m->uwarn('unable to fetch svn info: %m, working without svn: %r', m => $@, r => $m->{svn_repos}) } if $@;
      last if $m->{svn_repos}->{offline};
      do { $m->{svn_repos}->{offline} = 1;$m->{svn_repos}->{off} = 1;$m->pwarn('svn_info was unable to obtain a svn url, switching svn off');last } unless $url;
      last if $m->{svn_repos}->{offline};
      $m->{svn_repos}->{offline} = 1 unless $m->netp($m->{konfig}->svnhost);; 
    };
    my $uri = $url ? SVN::Core::uri_canonicalize($url) : SVN::Core::dirent_canonicalize($kfgdir);
    $m->dpop1($prog,$uri);
};

sub postkonfig {
    my ($m,$prog) = shift->qpuc('A2E::SVN::postkonfig', @_) or return;
    $m->SUPER::postkonfig;
    $m->svn_postkonfig;
    $m->dpop1($prog,1);
};

=head3 Function svn_faylz

select those files that are already in svn

argument 1 is a hash of selection options which can be cumulated:

  nsf: non subversion files: files that are not yet in svn
  sf: subversion files: those already in sf
  nf: non files: objects that do not exist in the file system  

used by svn_faylz(1) and ::svn_add.

arguments 2ff are files.  if none are given, all files of the current directory are traversed.

=cut

our %text_status_ekso = ( 
# $SVN::Wc::Status::ignored => 1,
# $SVN::Wc::Status::none => 1,
 $SVN::Wc::Status::unversioned => 1
    );

sub svn_faylz {
    my ($m,$prog,$opts,@args) = shift->ppuc('svn_faylz', @_);
    return if $m->{konfig}->svn_off;
    my @faylz = ();
    my $fayl = '';
    my $rekurs = 0;
    unless (@args) { my $cwd = getcwd;@args = ($cwd);$rekurs = 1 };
    $m->decho('args: %A', A => \@args);
    my $non_svn = 0;
    $m->decho('constants: unversioned %u, conflicting %c, none %N, normal %n, added %a, missing %m, ignored %i, obstructed %o, external %e, incomplete %p, modified %f'."\n", u => $SVN::Wc::Status::unversioned, c => $SVN::Wc::Status::conflicted, n => $SVN::Wc::Status::normal, a => $SVN::Wc::Status::added, m => $SVN::Wc::Status::missing, N => $SVN::Wc::Status::none, i => $SVN::Wc::Status::ignored, o => $SVN::Wc::Status::obstructed, e => $SVN::Wc::Status::external, p => $SVN::Wc::Status::incomplete, f => $SVN::Wc::Status::modified);
    my $find_non_svn = sub {
	$fayl = shift;
	my $wcstat = shift;
	$m->decho('file %f: text status %w', f => $fayl, w => $wcstat);
	my $text_status = $wcstat->text_status;
	$non_svn = $text_status_ekso{$text_status};
	unshift @faylz, $fayl if ($opts->{nsf} and $non_svn) or ($opts->{sf} and !$non_svn);
	$m->decho('non_svn %n faylz %F', n => $non_svn, F => \@faylz);
    };
    foreach my $fi (@args) {
	do { push @faylz, $fi if $opts->{nf};next } unless -e $fi;
	$m->decho('existing file %f', f => $fi);
        # $m->{svn}->status($fi, 'HEAD', $find_non_svn, $rekurs, 1, 0, 1);
	my $rev = $m->{svn_repos}->{offline} ? 'BASE' : 'HEAD';
	my $rekurs = 0;my $get_all = 1;my $update = 0;my $no_ignore = 1;
	eval { $m->{svn}->status($fi, $rev, $find_non_svn, $rekurs, $get_all, $update, $no_ignore) };
    };
    shift @faylz if $rekurs;
    $m->dpop($prog, @faylz);
};

=head3 Function svn_status

return text status value of the file entry $fi

=cut
sub svn_status {
    my ($m,$prog,$fi) = shift->ppuc('svn_status',@_);
    my $text_status; 
    my $get_status = sub {
	my ($fayl,$wcstat) = @_;
	$text_status = $wcstat->text_status; # side effect in callback function
	$m->decho('file %f: wcstat %w, text_status %t', f => $fayl, w => $wcstat, t => $text_status);
    };
    my $rev = $m->{svn_repos}->{offline} ? 'BASE' : 'HEAD';
    my $rekurs = 0;my $get_all = 1;my $update = 0;my $no_ignore = 1;
    $m->{svn}->status($fi, $rev, $get_status, $rekurs, $get_all, $update, $no_ignore);
    $m->dpop1($prog,$text_status);
};

=head3 Function svn_status_ekso

true if file entry arg1 exists and is not in SVN repository

=cut

sub svn_status_ekso {
    my ($m,$prog,$fi) = shift->ppuc('svn_status_ekso',@_);
    return $m->dpop($prog) unless -e $fi;
    my $text_status = eval { $m->svn_status($fi) };
    return $m->dpop1($prog,1) if $@;
    my $ekso = $text_status_ekso{$text_status};
    $m->dpop1($prog,$ekso);
};

=head3 Function svn_cleanup

=cut

sub svn_cleanup {
    my ($m,$prog,$dir) = shift->spuc('svn_cleanup', @_) or return;
    $dir ||= &getcwd();
    eval { $m->{svn}->cleanup($dir) };
    $m->ustop('failed to clean up directory %d: %m', d => $dir, m => $@) if $@;
    $m->dpop1($prog,1);
};

=head3 Function svn_mkdir

Make a directory, and, if appropriate, for SVN, ensuring that one can change to this directory and use any needed further svn commands there.

=cut

sub svn_mkdir {
    my ($m,$prog,@dirs) = shift->ppuc('svn_mkdir', @_);
    my $mkdir = $m->{svn_repos}->{off} ? sub { my ($dir) = @_;mkdir $dir } : sub { my ($dir) = @_;$m->{svn}->mkdir($dir);$m->svn_commit('making directory', $dir) }; 
    foreach my $dir (@dirs) {
	eval { &$mkdir($dir) };
	$m->ustop('failed to make directory %d: %m', d => $dir, m => $@) if $@;	
	$m->pstop('ran mkdir on %d without error but in vain', d => $dir) unless -d $dir;
    };
    $m->dpop1($prog,1);
};

=head3 Function svn_add: add files/dirs to repository

add to svn as far as needed, return the names of the added files.
arg 1 may be a filename or listref of filenames.
arg 2 whether to recursively add children
arg 3 whether to recursively add unversioned parent directories
return the files that were successfully added

=cut

sub svn_add {
    my ($m,$prog,$paths,$rekurs,$forc,$noign,$parents) = shift->ppuc('svn_add',@_);
    $rekurs ||= 0;$forc ||= 0;$noign ||= 0;$parents ||= 0;
    return $m->dpop($prog,1) if $m->{svn_repos}->{off};
    my @faylz = 'ARRAY' eq ref $paths ? @$paths : ($paths);
    # @faylz = $m->svn_faylz({ nsf => 1 }, @faylz);
    my @faylz_ok = ();
  FAYL: foreach my $fayl (@faylz) { 
      # complex old version file:/adv/perl/A2E/OLD/SVN.pm::PARENTS no longer needed with add4
      next unless -e $fayl;
    ENDO: {
	last unless $m->svn_status_ekso($fayl);
	eval { $m->{svn}->add4($fayl,$rekurs,$forc,$noign,$parents) };
	next FAYL if $@ or $m->svn_status_ekso($fayl);
      };
      push @faylz_ok, $fayl;
      # use case file:/phm/21/10/12/sig/_lng.phm_pub_sig211012.txt::apml
  }; # FAYL
    $m->dpop($prog,@faylz_ok);
};

sub svn_add_if {
    my ($m,$prog,@faylz) = shift->ppuc('svn_add_if',@_);
    my $ok = $m->pvar(svn_add_ok => 1);
    return $m->dpop($prog) unless $ok;
    @faylz = $m->svn_add(\@faylz);
    $m->dpop($prog,@faylz);
};
sub svn_symlink_add {
    my ($m,$prog,$fol,$lfol) = shift->ppuc('svn_symlink_add',@_);
    symlink $fol, $lfol;
    my $ok = $m->{svn}->add($lfol);
    $m->dpop1($prog,$ok);
} 

=head3 Function svn_delete($pafs, $forc)

delete from svn repository as far as needed, return the names of the deleted files.
$pafs may be a file or a ref to list of files
$forc $forc:
  may set force on so as to delete even files that are in not in svn
  try non-svn removal if == 2
  not well tested 

=cut

sub svn_delete {
    my ($m,$prog,$pafs,$forc) = shift->ppuc('svn_delete',@_);
    my @faylz = ();push @faylz, 'ARRAY' eq ref $pafs ? @$pafs : $pafs;
  OFF: { 
      last unless $m->{svn_repos}->{off};
    FORC: {
	last unless $forc;
	eval { rmtree(@faylz) };
	last unless $@;
	$m->ustop('failed to rmtree non-svn paths %F: %m', F => \@faylz, m => $@);
      };
      return $m->dpop1($prog,1);
    }; 
    # @faylz = $m->svn_faylz({ sf => 1 }, @faylz) unless $forc;
    my @faylz_ok = (); 
    foreach my $fayl (@faylz) { 
	($fayl) = $fayl =~ m(\A(.*\w)\/*\Z);
	$m->dpuc($fayl,$forc);
	eval { $m->{svn}->delete($fayl, $forc) };
	if ($@) {
	  FORCKO: {
	      $m->uwarn('failed to delete file %f from subversion repository $(SW|without|with) force: %m', f => $fayl, SW => $forc, m => $@);
	      $m->uwarn(faylekzistp => f => $fayl, SW => -e $fayl);
	      last unless $forc > 1;
	      eval { unlink $fayl };
	      last unless $@;
	      $m->uwarn('failed to unlink file %f: %m', f => $fayl, m => $@);
	      $m->uwarn(faylekzistp => f => $fayl, SW => -e $fayl);
	    };
	} else {
	    push @faylz_ok, $fayl;
	    $m->vecho("D\t$fayl");
	};
	$m->dpop($fayl);
    };
    $m->dpop($prog,@faylz_ok);
};

=head4 Function svn_move

Rename file $fi to file $fo.

    args := fi, fo, force

=cut

sub svn_move {
    my ($m,$prog,$fi,$fo,$forc) = shift->ppuc('svn_move',@_);
    return $m->dpop1(1) if $m->{svn_repos}->{off};
    $forc ||= 0;
    my $cinfo = eval { $m->{svn}->move($fi, 'HEAD', $fo, $forc) };
    $m->ustop('failed to svn_move %i to %o: %m', i => $fi, o => $fo, m => $@) if $@;
    $m->dpop1($prog,$cinfo);
};

=head4 Function svn_update

Try to update files supplied in fileargs.  Die unless fully successful.
If no fileargs are given, update current path.

=cut

sub svn_update {
    my ($m,$prog,@pafs) = shift->spuc('svn_update',@_) or return;
    my $rekurs = 1;
    if (!@pafs or $pafs[0] eq '.') { $pafs[0] = &getcwd();$rekurs = 1 };
    my $conflicted = 0;
    my $obstructed = 0;
    my $obstructed_last = 0;
    my $find_error = sub {
	my $fi = shift;
	my $wcstat = shift;
	my $status = $wcstat->text_status;
	$obstructed = $status == $SVN::Wc::Status::obstructed;
	return if $obstructed;
	$conflicted = $status == $SVN::Wc::Status::conflicted;
	return if $conflicted;
    };
  PAF: while (@pafs) { 
      my $paf = $pafs[0];
      # my $rekursp = -d $paf ? $rekurs : 0;
      eval { $m->{svn}->update($paf, 'HEAD', $rekurs) };
      $m->pstop('svn client failed to update file %f: %m', f => $paf, m => $@) if $@;
      eval { $m->{svn}->status($paf, 'HEAD', $find_error, 1, 1, 0, 1) };
      $m->ustop('failed to query status of file %f: %m', f => $paf, m => $@) if $@;
      $m->ustop('version conflict found in file %f', f => $paf) if $conflicted;
    OBSTRUCTED: {
	last unless $obstructed;
	$m->ustop('file %f is still obstructed in spite of renaming', f => $paf) if $obstructed_last;
	eval { $m->{svn}->revert($paf, 1) };
	$m->pwarn('failed to revert obstructed file %f: %m', f=> $paf, m => $@) if $@;
	my $paf_bak = join $paf, $m->timestamp;
	$m->uwarn('renaming obstructing local copy %f to %b', f => $paf, b => $paf_bak); 
	eval { rename $paf, $paf_bak };
	$m->pstop('failed to rename %f to %b: %m', f => $paf, b => $paf_bak, m => $@) if $@;
	$obstructed_last = 1;
	next PAF;
      };
      $obstructed_last = 0;
      shift @pafs;
  };
    $m->dpop1($prog,1);
};
=head3 svn_update_topdir

Update the whole tree in which we are

=cut
sub svn_topdir {
    my ($m,$prog,$dir) = shift->ppuc('svn_topdir',@_) or return;
    $m->svn_info($dir);
    my $topdir = catdir '', map { $_ } $m->{svn_repos}->{hdir}, $m->{svn_repos}->{fols}->[0];
    $m->dpop1($prog,$topdir);
};
sub svn_update_topdir {
    my ($m,$prog) = shift->spuc('svn_update_topdir') or return;
    my $topdir = $m->svn_topdir;
    my $ok = $m->svn_update($topdir);
    $m->dpop1($prog,$ok);
};

=head3 Function svn_commit

reset message argument to information about internal state, then invoke svn_msg_commit with that.
The remaining arguments are files to be committed.
If none are given, commit current path.

=cut

sub svn_commit {
    my ($m,$prog,@pafs) = shift->spuc('svn_commit', @_) or return;
    my $msg = join '.', $m->{prognom}, @{$m->{funz}};
    $m->svn_msg_commit($msg, @pafs);
    $m->dpop1($prog,1);
};

=head3 Function svn_msg_commit

Commit fileargs with message in arg1.
If a filearg is an absolute path, go to its parent and commit it from there.
If no fileargs are given, commit current path.

=cut

sub svn_msg_commit {
    my ($m,$prog,$msg,@pafs) = shift->spuc('svn_msg_commit', @_) or return;
    my $nonrekurs = 0;
    my $info;
    my $lwd = &getcwd();
    push @pafs, $lwd unless @pafs;
    my ($paf, $vol, $dir, $fol, @pafinfs) = ();
    foreach $paf (@pafs) {
	my $fyl = $paf;
	my $dyr = $paf;
	unless (-d $paf) {
	    ($vol, $dir, $fol) = eval { splitpath $paf };
	    $m->pwarn('cant splitpath %f: %,', f => $paf, m => $@) if $@;
	    $m->chwd($dir) if $dir;
	    my $dyr = $dir || $lwd;
	    my $fyl = $fol || $paf; 
	}
	$m->vecho('in directory %d committing %f', d => $dyr, f => $fyl);
	$info = eval { $m->{svn}->commit($fyl, $nonrekurs) };
	if ($@) { $m->vecho('failed to commit %s: %s', $fyl, $@) } else { push @pafinfs, $paf, $info };
	$m->chwd($lwd) if $dir;
    };
    $m->dpop($prog,@pafinfs);
};

=head3 Function svn_add_commit

Do whatever needs to be done to submit the file to SVN and return 1 unless we are in trouble.

=cut

sub svn_add_commit {
    my ($m,$prog,@pafs) = shift->ppuc('svn_add_commit',@_);
    my @adds = $m->svn_add(\@pafs);
    my $info = $m->svn_msg_commit($prog,@adds);
    $m->dpop1($prog,$info);
};

=head3 svn_get_keyword, svn_put_keyword, svn_get_deps, svn_put_deps
Keyword 'deps' is used by &A2E::MLDK::dkf file:/adv/perl/A2E/MLDK.pm.tmpl::/svn_get_deps
df file:/phm/18/ont/tdir/phm_pub_sig181005/_log.phm_pub_sig181005.c15.txt::1198
=cut

sub svn_get_keyword {
    my ($m,$prog,$fi,$nom) = shift->ppuc('svn_get_keyword',@_);
    my $res = eval { $m->{svn}->propget('svn:keywords', $fi,'BASE',0) };
    return $m->dpop($prog) if $@;
    my $keyws = $res->{$fi};
    return $m->dpop($prog) unless $keyws;
    $m->dpop1($prog,$keyws->{$nom});
};
sub svn_get_deps { shift->svn_get_keyword(shift, 'deps') };
sub svn_put_keyword {
    my ($m,$prog,$fi,$nom,$val) = shift->ppuc('svn_put_keyword',@_);
    $val = join ' = ', $nom, $val;
    eval { $m->{svn}->propset('svn:keywords', $nom, $val, $fi, 0) };
    my $OK = $@ ? 0 : 1;
    $m->dpop1($prog,$OK);
}
sub svn_put_deps { shift->svn_put_keyword(shift,'deps', shift) };
    
=head1 COPYRIGHT

    Copyright (c) 2009 Hartmut Pilch (phm)

    This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=cut

1;

# Local Variables:
# coding: utf-8
# srcfile: /adv/perl/A2E/SVN.pm.tmpl
# mode: perl
# End:
