#!/usr/bin/perl

=head1 NAME

A2E::SiteMakefileCache

=head1 VERSION

0.0.1

=cut

our $VERSION = 0.0.1; 

=head1 IMPLEMENTATION

=head2 Header

=cut

package A2E::SiteMakefileCache;
use base A2E::MakefileCache, A2E::Osver;
use File::Spec::Functions;
use Sys::Hostname;
use English -no_match_vars;

=head2 Functions

=head3 Function perlver

=head3 Function defvars

=head3 Function perl_dir

=cut

sub perl_dir {
    my $arg = shift;
    die 'perl_dir invoked without parameter' unless $arg;
    my $part = quotemeta $arg;
    die 'no include path known to perl_dir' unless @INC;
    foreach my $dir (@INC) { 
	next unless $dir =~ m(\/$part\/[\d\.]+\Z);
	return $dir;
    };
    die sprintf 'perl_dir %s ending without result', $arg;
};

sub perl_ver {
    my $m = shift;
    my $perl_ver = $PERL_VERSION;
    $m->dpuc('perl_ver', $perl_ver);
    my $ver = '';
  NOV: { # WEIRD: in perl 5.8.8 $PERL_VERSION is empty but not null
      $m->dpuc('nov', $perl_ver);
      ($ver) = $perl_ver =~ m(\Av*(\d[\d\.]*\d)\Z);
      $m->dpop('nov', $ver);
  };
  OLD: {
      $m->dpuc('old', $ver);
      do { $m->dpop;last } if $ver;
      my $cdr = $];
      $m->uwarn('old version %s without valid PERL_VERSION variable', $cdr);
      my $car = int $cdr;
      my @cars = ();
      push @cars, $car;
      $cdr = 1000 * ($cdr - $car);
      $car = int $cdr;
      push @cars, $car; 
      $cdr = 1000 * ($cdr - $car);
      $car = int $cdr;
      push @cars, $car;
      $ver = join '.', @cars;
      $m->dpop('old', $ver);
  };
    $m->ustop('no perl version found') unless $ver;
    $m->dpop('perl_ver', $ver);
    return $ver;
};


=head3 defvars

=head4 TODO

vendor_perl does not exist under Debian, change invocation so that in this case /usr/lib/perl5 is used instead, test on genba.a2e.de

=cut

sub defvars {
    my $m = shift;
    $m->dpuc('A2E::SiteMakefileCache::defvars', @_);
    $m->SUPER::defvars;
    $m->cache_define('site_vars', 's@', [ qw(adming gnupghome webgrp mailto pgp admins host osver osvernum macklas deplate_ver latex_preproc users perl_ver vendor_perl_dir site_perl_dir docbook_website_dir) ]);
    $m->cache_define('site_makefile_cache_name', 's', 'site_makefile_cache.mk');
    $m->cache_define('admingrp', 's', 'adv'); 
    $m->cache_define('mailto', 's', 'adv@a2e.de');
    $m->cache_define('perl_ver', 's');
# TODO vendor_perl does not exist under Debian, change invocation so that in this case /usr/lib/perl5 is used instead, test on genba.a2e.de
    $m->cache_define('vendor_perl_dir', 's', &perl_dir('vendor_perl')); 
    $m->cache_define('site_perl_dir', 's', &perl_dir('site_perl')); 
    $m->cache_define('gnupghome', 's', '/homdisk/.gnupg'); 
    $m->cache_define('server_name', 's', 'bartu');
    $m->cache_define('latex_preproc', 's', 'u2tex -p 1');
    $m->konfayl('site_makefile_cache.konf');
    $m->dpop('A2E::SiteMakefileCache::defvars', 1);return 1;
};

=head2 Second Level Functions

=head3 Function set_site_vars

=cut

sub set_site_vars {
    my $m = shift;
  ADMING: foreach my $grp (qw(adv wheel admin root)) {
      $m->open_fayl('/etc/group', 'i', 'r');
      while ($m->getlini('i')) {
	  chomp;
	  next unless m(\A$grp\:.*:.*:(.+)\Z);
	  $m->{adming} = $grp;
	  @{$m->{admins}} = split ',', $1;
	  $m->kloz_fayl('i');
	  last ADMING;
      };
      $m->kloz_fayl('i');
  };
    $m->{webgrp} = 'www';
  WEBGRP: foreach my $grp (qw(apache www wwwrun)) {
      $m->open_fayl('/etc/group', 'i', 'r');
      while ($m->getlini('i')) {
	  chomp;
	  next unless m(\b$grp\b);
	  $m->{webgrp} = $grp;
	  $m->kloz_fayl('i');
	  last WEBGRP;
      };
      $m->kloz_fayl('i');
  };
    $m->{pgp} = -s catfile($m->{gnupghome}, 'secring.gpg') && 'OK';
    $m->{host} ||= &hostname;
    my @matches = $m->{host} =~ m(\A([^\.]+)\.);
    $m->{host} = $matches[0] if @matches;
    $m->{osver} = $m->osver;
    $m->{osvernum} = $m->osvernum;
    $m->{perl_ver} ||= $m->perl_ver;
    $m->{docbook_website_dir} = glob('/usr/share/xml/docbook/custom/website/?.?.?');
    $m->{macklas} = $m->{host} eq $m->{konfig}->server_name ? 'susrv' : 'sukli';
    local $_ = `deplate --version`;chomp;$m->{deplate_ver} = $_;
    $m->{mailto} = $m->{konfig}->mailto;
    $_ = `a2e_users`;chomp;@{$m->{users}} = split /\s+/;
};

=head3 Function make_site_makefile_cache

=cut

sub make_site_makefile_cache {
    my $m = shift;
    $m->dpuc('make_site_makefile_cache', @_);
    $m->set_site_vars;
    my $fo = $m->{konfig}->site_makefile_cache_name;
    printf STDERR 'writing %s', $fo if $m->{konfig}->verbose; 
    $m->open_fayl($fo, 'o', 'w');
    $m->print_tu_fayl('o', join "\n", '# '.$m->emacs_top_local_variables({ mode => 'makefile', coding => 'utf-8' }), '# '.$m->emacs_timestamp, '', '');
    $m->make_defs('o', 'export %s = %s', @{$m->{konfig}->site_vars});
    $m->kloz_fayl('o');$m->vecho('OK');
    $m->dpop('make_site_makefile_cache', 1);return 1;
};

=head1 The End

=cut

return 1;
 
# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: /adv/perl/A2E/SiteMakefileCache.pm.tmpl
# End:
