# -*- mode: perl; coding: utf-8 -*-
# Time-stamp: <2019-05-29 15:04:36 phm>

=head1 NAME

A2E::SitemapTmplfil

=cut

package A2E::SitemapTmplfil;

=head1 DESCRIPTION

Generate a Sitemap file for a given document, using a given output template,
crawling down the subdocument tree, based on document metadata found in the
backend database.

=head1 BUGS

Report to adv@a2e.de

=head1 VERSION

our $VERSION = '0.0.2';

=head1 PREREQUISITES

=over

=item Base: A2E::MLHT(3)

=cut

use base 'A2E::MLHT';
use strict;
use File::Spec::Functions;
use POSIX qw(strftime);

=item Template(3)

=cut

use Template;
# use A2E::Template::Provider;
# use Template::Config;
# $Template::Config::Provider = 'A2E::Template::Provider';
# =item A2E::SArb::Make(3)

# The workhorse of make-like macro expansion in lexical source files.

=cut

# use A2E::SArb::Make;
use Encode qw(decode encode);

=back

=head1 OPTIONS

The options of the base library A2E::MLHT(3) and its ancestors are also available here.

=head3 Function sitemap_tmplfil_postkonfig

Here the work of reading makefiles into template variables is done.  Any application that inherits this library will have the template variables read in, based on configuration and commandline options, before the main program starts.

=cut

sub sitemap_tmplfil_postkonfig {
    my $m = shift;
    $m->dpuc('sitemap_tmplfil_postkonfig', @_);
    my $tmplopts = $m->pvar('tmplopts',1);
    $m->{template} = new Template $tmplopts;
    $m->set_dok;$m->set_langs;$m->pvar('lang') = $m->{langs}->[0] unless grep { $_ eq $m->pvar('lang') } @{$m->{langs}}; 
    # $m->mlht_set_langvars;
    $m->mlht_set_restvars;
    $m->{tmplvars}->{url} = $m->pvar('url') || $m->dok2absurl($m->pvar('dok'));
    $m->{tmplvars}->{dirs} = [];
    ($m->{tmplvars}->{prot}, $m->{tmplvars}->{dom}, @{$m->{tmplvars}->{dirs}}) = $m->url_parts($m->{tmplvars}->{url});
    $m->dpop('sitemap_tmplfil_postkonfig', 1);
    return 1;
}

=head3 Function postkonfig

Wrapper around sitemap_tmplfil_postkonfig for automatic use in A2E::Prog::new_ready constructor mechanism.

=cut

sub postkonfig {
    my $m = shift;
    $m->dpuc('A2E::SitemapTmplfil::postkonfig', @_);
    $m->SUPER::postkonfig;
    $m->sitemap_tmplfil_postkonfig;
    $m->dpop('A2E::SitemapTmplfil::postkonfig', 1);
    return 1;
};

=head2 Top level functions

=head3 Function tmplfil1

Process one template.

=cut

sub tmplfil1 {
    my $prog = 'tmplfil1';my $m = shift;$m->dpuc($prog,@_);
    my $fi = shift;
    $m->open_fayl($fi, 'i', 'r');
    $m->{template}->include($m->fh('i'), $m->{tmplvars}) or die $m->{template}->error();
    $m->kloz_fayl('i');
    $m->sav2out;
    $m->dpop($prog,1);return 1;
};

=head3 Function sitemap_tmplfil

Process all templates supplied through the infaylz option i.e. usually the commandline arguments.

=cut

sub sitemap_tmplfil {
    my $m = shift;
    $m->dpuc('sitemap_tmplfil', @_);
    $m->{konfig}->infaylz($_) for @_;
    $m->mlht_set_dokvars;
    $m->mlht_set_tmplvars;
    my @faylz = @{$m->{konfig}->infaylz};
    $m->tmplfil1($_) for @faylz;
    $m->put('topdok', { topdok => $m->pvar('dok') }, lastmod => $m->pvar('hodie'));
    $m->dpop('sitemap_tmplfil', 1);return 1;
};

=head1 TODO

 - debug

=cut

return 1;

# Local Variables:
# coding: utf-8
# srcfile: /adv/perl/A2E/SitemapTmplfil.pm
# mode: perl
# End:
