=head1 NAME

A2E::Subdoks

=cut

package A2E::Subdoks;
use strict;
use File::Spec::Functions qw(catfile);

=head1 VERSION

    This describes version B<0.1.0> of A2E::Subdoks.

=cut

    our $VERSION = '0.1.0';

=head1 SYNPOSIS

    use base 'A2E::MLHT', 'A2E::Subdoks', 'A2E::Mbox2dok';
    my $p = new_ready A2E::MLHT;
    $p->add_subdoks($dok, $pwd);

=head1 DESCRIPTION

Write metadata that come with presence of subordinate documents, such as a
'subdoks' or 'grpdoks' file, and make corresponding entries in the database.

mixin library, does not invoke the constructor of its base library.

=head1 Prerequisites

    A2E::Dokfs(3)

=cut

use base 'A2E::Dokfs';

=head1 SEE ALSO

    A2E::Mktdir(3)

=head1 Implementation

=head2 Overwritable Functions

=head3 defvars

=cut
our %MSGS = (
    srcfpfxr => 'failed to obtain srcfpfx in routine %r',
    psetdok => 'cant set dok in directory %p',
    );

sub subdoks_defvars {
    my ($m,$prog) = shift->qpuc('subdoks_defvars',@_) or return;
    $m->cache_define('supdok');
    $m->cache_define('dokstyp');
    $m->konfig_define('pushdok', '!');
    $m->cache_define('new_dok');
    $m->cache_define('doksfayl');
    $m->prog_msgdefs('subdoks', %MSGS);
    $m->qpop1($prog, 1);
};

=head2 Subroutines

=head3 ar_doks_add

add extra documents to an array reference such as $m->plst('subdoks') or $m->plst('grpdoks')

=cut

sub ar_doks_add {
    my ($m,$prog) = shift->ppuc('ar_doks_add',@_);
    my $dok = $m->pvar(dok => 1);
    my $dokstyp = $m->pvar(dokstyp => 1);
    if ($m->{konfig}->pushdok) {
	push @{$m->{progvars}->{$dokstyp}}, $dok;
    } else {
	unshift @{$m->{progvars}->{$dokstyp}}, $dok;
    };
    $m->dpop1($prog,1);
};

=head3 db_doks_push

add extra documents to the end of present group/directory by writing database records according to the predetermined syntax.

=cut

sub db_doks_push {
    my ($m,$prog) = shift->ppuc('db_doks_push',@_);
    my $dokstyp = $m->pvar(dokstyp => 1);
    my @doks = $m->plst($dokstyp);
    my $predok = $doks[$#doks];
    my $dok = shift || $m->pvar(dok => 1);
    $m->dgot(dokstyp => $dokstyp, doks => \@doks, predok => $predok, dok => $dok);
    $m->sql_query(qq{update dokprop set predok = null, pretyp = 0 where predok = '$dok' and pretyp = 1});
    $m->put(mlhtdok => { dok => $dok }, doktmp => 'now()');
    $m->put(dokprop => { dok => $dok }, pretyp => 1, predok => $predok);
    $m->{progvars}->{pretyp_hac}->{$dok} = 1;
    $m->{progvars}->{predok_hac}->{$dok} = $predok;
    $m->dpop1($prog,1);
}; 

=head3 db_doks_unshift

Like db_doks_push but add the documents to the beginning.

=cut

sub db_doks_unshift {
    my ($m,$prog) = shift->ppuc('db_doks_unshift',@_);
    my $dokstyp = $m->pvar(dokstyp => 1);
    my @doks = $m->plst($dokstyp);
    my $dok0 = $doks[0];
    my $supdok = $m->pvar(supdok => 1);
    my $dok = $m->pvar(dok => 1);
    my $num = 0;
    my $pretyp = $m->pvar(pretyp => 1);
    $m->dgot(dokstyp => $dokstyp, doks => \@doks, dok0 => $dok0, dok => $dok, supdok => $supdok, pretyp => $pretyp);
    $num = $m->sql_query(qq{update dokprop set predok = null, pretyp = 0 where predok = '$supdok' and pretyp > 1});
    $m->decho("UPDATE $num");
    foreach my $d (keys %{$m->{progvars}->{predok_hac}}) { next unless $m->{progvars}->{predok_hac}->{$d} eq $supdok and $pretyp > 1;delete $m->{progvars}->{predok_hac}->{$d};delete $m->{progvars}->{pretyp_hac}->{$d} };
    $num = $m->sql_query(qq{update dokprop set predok = null, pretyp = 0 where predok = '$dok' and pretyp = 1});
    $m->decho("UPDATE $num");
    foreach my $d (keys %{$m->{progvars}->{predok_hac}}) { next unless $m->{progvars}->{predok_hac}->{$d} eq $dok and $pretyp eq 1;delete $m->{progvars}->{predok_hac}->{$d};delete $m->{progvars}->{pretyp_hac}->{$d} };
    $m->put(mlhtdok => { dok => $dok0 }, doktmp => 'now()');
    $m->put(dokprop => { dok => $dok0 }, pretyp => 1, predok => $dok);
    $m->{progvars}->{pretyp_hac}->{$dok0} = 1;
    $m->{progvars}->{predok_hac}->{$dok0} = $dok;
    $m->put(mlhtdok => { dok => $dok }, doktmp => 'now()');
    $m->put(dokprop => { dok => $dok }, pretyp => $pretyp, predok => $supdok);
    $m->{progvars}->{pretyp_hac}->{$dok} = $pretyp;
    $m->{progvars}->{predok_hac}->{$dok} = $supdok; 
    $m->dpop1($prog,1);
};

=head3 db_doks_add

=cut

sub db_doks_add {
    my ($m,$prog) = shift->ppuc(@_);
    my $dok = $m->pvar(dok => 1);
    my $dokstyp = $m->pvar(dokstyp => 1);
  OLD_DOK: {
      my @doks = $m->plst($dokstyp);
      last if !@doks;
      my $dok0 = grep { $_ eq $dok } @doks;
      return $m->dpop1($prog,$dok0) if $dok0;
    };
    $m->setpvar(new_dok => $dok);
    $m->put(mlhtdok => { dok => $dok }, doktmp => 'now()');
    if ($m->{konfig}->pushdok) { $m->db_doks_push } else { $m->db_doks_unshift };
    $m->ar_doks_add;
    $m->dpop1($prog,1);
};

=head3 Function extra_elements

args 1 and 2 are array references.

return the elements that are contained only in array 2 but not array 1.

=cut

sub extra_elements {
    my ($m,$prog,$lr0,$lr1) = shift->ppuc('extra_elements',@_);
    my @l0 = @$lr0;
    my @l1 = @$lr1;
    my @extra = ();
    foreach my $e1 (@l1) {
	# print STDERR "e1 $e1 compared to @l0\n";
	local $_;next if grep { $e1 eq $_ } @l0;
	push @extra, $e1;
    };
    # print STDERR "extra_elements return extra @extra\n";
    $m->dpop($prog,@extra);
};

=head3 Function db_doks_new 

=cut

sub db_doks_new ($) {
    my ($m,$prog) = shift->ppuc('db_doks_new',@_);
    my $dok = $m->pvar(dok => 1);
    $m->setpvar(new_dok => $dok);
    my $supdok = $m->pvar(supdok => 1);
    my $pretyp = $m->pvar(pretyp => 1);
    $m->put(mlhtdok => { dok => $dok }, doktmp => 'now()');
    my $num = $m->sql_query(qq{update dokprop set predok = null, pretyp = 0 where predok = '$supdok' and pretyp > 1});
    $m->decho(qq{UPDATE $num});
    $num = $m->sql_query(qq{update dokprop set predok = null, pretyp = 0 where predok = '$dok' and pretyp = 1});
    $m->decho(qq{UPDATE $num});
    $m->put(dokprop => { dok => $dok }, pretyp => $pretyp, predok=> $supdok);
    $m->{progvars}->{pretyp_hac}->{$dok} = $pretyp;
    $m->{progvars}->{predok_hac}->{$dok} = $supdok;
    $m->ar_doks_add;
    $m->dpop1($prog,1);
};

=head3 Function find_splice_pos

Find position in list of lines in a text file where to insert the new document id.

If we push ($m->{konfig}->pushdok set), we must insert at the end, otherwise at the beginning.

In either case we must skip over possible comment lines.

=cut

sub find_splice_pos {
    my ($m,$prog,@lines) = shift->ppuc('find_splice_pos',@_);
    my ($p, $l) = ();
    local $_;
  TRY: {
      $m->dpuc('try', @lines);
    PUC: {
	my $ok = $m->{konfig}->pushdok;
	$m->dpuc('puc', $ok);
	do { $m->dpop;last } unless $ok;
	$p = $#lines;
	while ($p >= 0) {
	    $l = $_ = $lines[$p];
	    $m->dpuc($p, $l);
	    do { $m->dpop;next } if m(\A\#);
	    do { $m->dpop('puc');last };
	} continue { $p-- };
	$p++;
	$m->dpop('try', $p);
	last TRY;
      };
    POP: {
	$p = 0;
	$m->dpuc('pop', $p);
	while ($p <= $#lines) {
	    $l = $_ = $lines[$p];
	    $m->dpuc($p, $l);
	    do { $m->dpop;next } if m(\A\#);
	    $p++ if m(\A\s*\Z);
	    $m->dpop('pop', $p);last;
	} continue { $p++ };
      };
      $m->dpop('try', $p);
    };
    $m->dpop1($prog,$p);
};

=head3 Function db_add

=cut

sub db_add {
    my ($m,$prog,@doks) = shift->ppuc('db_add',@_);
    my $dokstyp = $m->pvar(dokstyp => 1);
  TRY: {
      $m->dpuc(try => \@doks);
    DOKS: {
	$m->dpuc(doks => \@doks);
	do { $m->dpop;last } unless @doks;
	$m->db_doks_add;
	$m->dpop(try => 1);last TRY;
      };
      $m->db_doks_new;
      $m->dpop(try => 0);
    };
    @doks = $m->plst($dokstyp);
    $m->dpop($prog,@doks);
};

=head3 Function fayl_lins_doks

parse file, return a reference to the lines and one to the doks found therein

=cut

sub fayl_lins_doks {
    my ($m,$prog,$fi) = shift->ppuc('fayl_lins_doks',@_);
    $m->open_fayl($fi, 'i', 'r');
    my ($n, $lini, @fayl_lins, @fayl_doks, @lini_doks) = ();
    $n = 0;
    while ($m->getlini('i')) {
	chomp;
	$n++;
	$m->dpuc('l'.$n, $_);
	push @fayl_lins, $_;
	do { $m->dpop;next } if m(\A\#) or m(\A\s*\Z);
	($lini) = m(\A([^\#]+));
	@lini_doks = grep { m(\S) } split m(\s+), $lini;
	push @fayl_doks, @lini_doks;
	$m->dpop('l'.$n, @lini_doks);
    };
    $m->kloz_fayl('i');
    $m->dpop($prog,\@fayl_lins,\@fayl_doks);
};

=head3 Function fayl_nov

=cut

sub fayl_nov {
    my ($m,$prog,$fi,@doks) = shift->ppuc('fayl_nov',@_);
    my $DokDoks = $m->CamelCase($m->pvar(supdok => 1)) . ucfirst($m->pvar(dokstyp => 1));
    $m->fayl_put_str($fi, join("\n", "# -*- dokl: $DokDoks; mode: conf; coding: utf-8 -*-", '# '.$m->emacs_timestamp, '', @doks, '', '#EOF'));
    $m->svn_add_if($fi);
    $m->dpop($prog,@doks);
};

=head3 Function fayl_ins

=cut

sub fayl_ins {
    my ($m,$prog,$fi,@db_doks) = shift->ppuc('fayl_ins',@_);
    $m->svn_update($fi);
    my ($fayl_lins, $fayl_doks) = $m->fayl_lins_doks($fi);
    my @extra_doks = $m->extra_elements($fayl_doks, \@db_doks);
  NO_EXTRA: {
      last if @extra_doks;
      return $m->dpop($prog,[v => 'found only known documents, nothing to add to the %f file', f => $fi]);
    };
  SWAO: { # file:/phm/19/06/07/sig/_lng.phm_pub_sig190607.txt::hsprintf
      $m->vecho('adding %n documents to %(SW|beginning|end) of file %f', n => scalar(@extra_doks), SW => $m->{konfig}->pushdok, f => $fi);
    };
    my $p = $m->find_splice_pos(@$fayl_lins);
    $m->dgot(fayl_lins => $fayl_lins, extra_doks => \@extra_doks, p => $p);
    splice @$fayl_lins, $p, 0, @extra_doks;
    my $str = join "\n", @$fayl_lins;
    $m->fayl_put_str($fi, $str);
    $m->dpop($prog,@$fayl_lins);
};

=head3 Function fayl_add

register subdocument in 'grpdoks' or 'subdoks' file $fi.

=cut

sub fayl_add {
    my ($m,$prog,$fayl,@doks) = shift->ppuc('fayl_add',@_);
    my @lins = do { if (-r $fayl) { $m->fayl_ins($fayl, @doks) } else { $m->fayl_nov($fayl, @doks) } };
    $m->svn_commit($fayl) if @lins;
    $m->dpop($prog,@lins);
}

=head3 Function doksfayl_add

Add entries to _subdoks/_grpdoks file and database.

=cut

sub doksfayl_add ($) {
    my ($m,$prog,$doksfayl,$dokstyp) = shift->ppuc('doksfayl_add',@_);
    $doksfayl ||= $m->pvar(doksfayl => 1);
    $dokstyp ||= $m->pvar(dokstyp => 1);
    my @doks = $m->plst($dokstyp);
    @doks = $m->db_add(@doks);
    my @lins = $m->fayl_add($doksfayl, @doks);
    $m->dpop($prog,@lins);
};

=head2 Added Interface Functions

Used in A2E::Mktdir

add node to menu-wise parent, both in the database and in the bookkeeping files of the local filesystem aka doksfayl ie 'grpdoks' or 'subdoks' and in the corresponding internal variables ie pvar supdir et al.
file:/adv/perl/A2E/MLMK.pm.tmpl::/\badd_subdoks\b/

=head3 Function add_subdoks

=cut

sub add_subdoks ($) {
    my ($m,$prog,$dok,$supdir,$dokstyp,%opts) = shift->ppuc('add_subdoks',@_);
    $m->setpvar(dok => $dok);
    $m->setpvar(supdir => $supdir);
    $m->setpvar(dokstyp => $dokstyp);
    # my ($cwd,$lwd) = $m->chwd($supdir);
    $m->setpvar(supdir => $supdir);
    my $supdok = $m->pvar(supdok => 1);
    my $pretyp = $dokstyp eq 'grpdoks' ? 2 : 3;
    $m->setpvar(pretyp => $pretyp);
    my $pfx = $m->pvar(srcfpfx => 1);
    my $doksfayl = $m->absfayl($pfx.$dokstyp,$supdir);
    $m->setpvar(doksfayl => $doksfayl);
    $m->set_doks($dokstyp, FDoks => $doksfayl);
    $m->putdoks($dokstyp, $supdir, $supdok);
    my @lins = $m->doksfayl_add($doksfayl,$dokstyp);
    # $m->chwd($lwd) if $lwd;
    $m->dpop($prog,@lins);
};

=head1 Changelog

=over
=item 2008-11-29 fixed bugs in fayl_lins_doks, mktjar now works more robustly
=back

=head1 The End

=cut

1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/Subdoks.pm.tmpl"
# End:
