#!/usr/bin/perl
# Time-stamp: <2023-02-02 00:12:32 phm>
# BEGIN { $ENV{PATH} = '/opt/a2e/bin:/bin:/usr/local/bin:/usr/bin' };

=head1 NAME

       A2E::SvnCommit(3)

=cut

package A2E::SvnCommit;
use base A2E::SVN;
use strict;

=head1 SYNOPSIS

    svn_update -v

=head1 DESCRIPTION

Force update from subversion archive, handling some common errors.

If Subversion is not in use or the repository is not accessible, still exit successfully, but warn about this when in verbose mode.

E.g. when a local object is found to have been added although an oject of the same name already existed in the repository, the repository object is renamed and a warning is issued.

=head1 APPLICATIONS

=over
=item svn_update(1)
=back

=head1 CAVEAT

Use with care in cases of emergency, only interactively, never as a matter of routine in automatic maintenance scripts.

=head1 OPTIONS

Inherited from the prerequisites, partially with special meanings

=over

=item --svn

we are using svn

=item --svn_repo

we have access to the svn repository

=item --force

Ruthlessly remove anything that is in the way.

=back

=head1 FILES

=head1 BUGS, LIMITATIONS

=head1 PREREQUISITES

    A2E::Prog(3)

=cut

use base 'A2E::Prog';

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

Copyright (c) 2007-8 Hartmut Pilch (phm)

This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

=cut

=head2 Construction

=head3 svn_update_defvars

=cut

sub svn_commit_defvars {
    my $m = shift;
    $m->{progver} ||= '0.0.1';
    $m->{konfig}->define('message|m=s');
    $m->konfayl('svn_commit.konf');
    1;
}

=head3 defvars

=cut

sub defvars {
    my $m = shift;
    $m->SUPER::defvars;
    $m->svn_commit_defvars;
    1;
};

=head2 Execution

=head3 Function commit

=cut

sub commit {
    my $m = shift;
    $m->dpuc('commit', @_);
    my $msg = $m->{konfig}->message;
    $m->svn_msg_commit($msg, @_);
    $m->dpop('commit');
    1;
}

=head1 The End

=cut

1;

# Local Variables: ***
# coding: utf-8 ***
# mode: perl ***
# srcfile:/adv/perl/A2E/SvnCommit.pm ***
# End: ***
