# -*- coding: utf-8; mode: perl; srcfile: /adv/perl/A2E/Template/Context.pm -*-[% PROCESS 'perl_prog_tmpl.txt' %]
# Time-stamp: <2019-07-20 09:26:50 phm>

=head1 NAME

A2E::Template::Context

=cut

package A2E::Template::Context;

=head1 DESCRIPTION

Does by means of virtual methods what has so far been done by means of the filter A2E::Template::Plugin::Textbau.

Rather than writing

    [% USE Textbau 'tb' %]
    [% pi = '3.14' %]
    [% 'value of pi: ${1}' | tb 'call' pi %]
    [% 'value of pi is ${pi}' | tb %]
    [% 'ahurl = {ah url="${1}":${2}}' %]
    [% ahurl | tb 'call' 'http://www.a2e.de' 'A2E' %]
    [% 'This is $(ae:our company)' | tb 'fill' 'ae=:call:ahurl' %]

we would write

In textchunk file:

    pi := '3.14'
    ahurl = {ah url="${1}":${2}}
    hello = 'This is $(ae:our company)'

In document:

    [% var = 'value of pi: ${1}';var.call(pi) %]
    [% var = 'value of pi is ${pi}';var.expand %]
    [% ahurl.call('http://www.a2e.de', 'A2E') %]
    [% hello.fill('ae=:call:ahurl') %]

More info can be found in dok:dynavars090307.

=head1 PREREQUISITES

    Template::Context
    A2E::SArb::Make

=cut

use base 'Template::Context';
use A2E::SArb::Make;
use strict;

=head1 Wrappers around external functionality

=head2 dpop, dpuc, dgot

Wrappers around some basic functionality from elsewhere, especially $m->{sarb}.

=cut

sub dpop { return shift->{sarb}->dpop(@_) };
sub dpuc { return shift->{sarb}->dpuc(@_) };
sub dgot { return shift->{sarb}->dgot(@_) };

=head2 defin_bas_tmplfuns

Define some low level vmethods and other template language extension elements used for text building.
In &A2E::MLHT::defin_mlht_tmplfuns we define some higher-level, i.e. application specific, language elements.

=cut

sub defin_bas_tmplfuns {
    my $m = shift;
    # dok:oas_mlht090408
    $m->{sarb}->add_renfun(litsproc => sub { my $sarb = shift;$m->dpuc('proc', @_);my $vars = shift;my ($block, @litvals) = map { $sarb->render({}, $_) } @_;die 'missing block name in proc function' unless $block;my $params = { litvals => \@litvals };my @rest = ();($block, @rest) = @{$sarb->textsep2listref($block)};while (@rest) { my $k = shift @rest;my $v = shift @rest;$params->{$k} = $v };$m->dgot(block => $block, params => $params);my $body = eval { $m->process($block, $params) };warn sprintf 'failed to process template %s: %s', $block, $@ if $@;$m->dpop('proc', $body);return $body });
    $m->define_vmethod('ARRAY', 'litsproc', sub { 
=pod

apply a block as usually defined for litscat, involving special variables litvals and litlvl, on a list variable, so as to create nested lists within the template language, in a somewhat clumsy way, as seen in the usage example from dok:munnL2:

    [% items = [] -%]
    [% items.push('{ah url=munnL2_pass.asc:Kennsatz}') -%]
    [% items.push('{ah dok=a2e_cryptfs:Erfahrungen und Konventionen im Bereich der Datenträgerverschlüssselung bei A2E}') -%]
    [% subitems = [] -%]
    [% subitems.push('{ah url=munnL2_cdrom.tar.bz2:Dokumentation auf der mitgelieferten CDROM}') -%]
    [% subitems.push('Teil A') -%]
    [% subitems.push('Teil B') -%]
    [% subitems.push('Teil C') -%]
    [% items.push(subitems.litsproc(2, 'tititemlist')) -%]
    [% items.litsproc(1, 'enumlist') -%]

This yields:

  #. {ah url=munnL2_pass.asc:Kennsatz}
  #. {ah dok=a2e_cryptfs:Erfahrungen und Konventionen im Bereich der Datenträgerverschlüssselung bei A2E}
  #. {ah url=munnL2_cdrom.tar.bz2:Dokumentation auf der mitgelieferten CDROM}
    - Teil A
    - Teil B
    - Teil C

It is better to do this using the text chunk language.  Alternatively, we may need to subclass the Template::Directive module to allow something like

  [% LIST enumlist -%]
  [% ITEM %]{ah url=munnL2_pass.asc:Kennsatz}[% END %]
  [% ITEM %]{ah dok=a2e_cryptfs:Erfahrungen und Konventionen im Bereich der Datenträgerverschlüssselung bei A2E}[% END -%]
  [% ITEM %]{ah url=munnL2_cdrom.tar.bz2:Dokumentation auf der mitgelieferten CDROM}
  [% LIST tititemlist %] 
  [% ITEM %]Teil A[% END -%]
  [% ITEM %]Teil B[% END -%]
  [% ITEM %]Teil C[% END -%]
  [% END %][%# enumlist.item.tititemlist -%]
  [% END %][%# enumlist.item -%] 
  [% END %][%# enumlist -%]

While this may still look inelegantly verbose, it provides some counter-value for the verbosity.  It is the 'Template'ish way to solve the problem, and by allowing nested structures this way it provides a powerful abstraction by which also tables and other more-dimensional lists could be represented.

=cut
	$m->dpuc('litsproc', @_);
	my $params = { litvals => shift, litlvl => shift };
	my $block = shift;
	while (@_) { my $k = shift;my $v = shift;$params->{$k} = $v };
	$m->dgot(params => $params, block => $block);
	my $body = eval { $m->process($block, $params) };
	die sprintf 'failed to process template %s: %s', $block, $@ if $@;
	$m->dpop('litsproc', $body);
	return $body; });
    $m->define_block('vset', sub { my $context = shift;my $stash = $context->stash;my $output = '';my $error;$stash->set($stash->get('key'), $stash->get($stash->get('val')));$output = ''; });
    $m->define_vmethod($_, 'value', sub { my $var = shift;return $m->stash->get($var) }) for qw(SCALAR LIST HASH);
    $m->define_filter('value', sub { my $text = shift;return $m->stash->get($text) });
    $m->define_vmethod('SCALAR', 'rtrim', sub { local $_ = shift;s(\A(.*?)[\s\n\r]*\Z)($1);return $_ });
    $m->define_vmethod('SCALAR', 'strip', sub { local $_ = shift;s(\A[\s\n\r]*(.*?)[\s\n\r]*\Z)($1);return $_ });
    $m->define_vmethod('SCALAR', 'fill', sub { my $text = shift;return $m->fill($text, @_) });
    $m->define_vmethod('SCALAR', 'cfil', sub { my $text = shift;return $m->cfil($text, @_) });
    $m->define_filter('fill', sub { my $context = shift;my @args = @_;return sub { my $text = shift;$m->fill($text, @args) } }, 1);
    $m->define_vmethod('SCALAR', 'call', sub { my $text = shift;my $res = $m->call($text, @_);return $res });
    $m->define_filter('call', sub { my $context = shift;my @args = @_;return sub { my $text = shift;$m->call($text, @args) } }, 1);
    $m->define_vmethod('SCALAR', 'funcall', sub { my $text = shift;my $verb = shift;my $fun = $m->{sarb}->getfun($verb) or die sprintf 'no such function %s', $verb;$text = &$fun($m->{sarb}, {}, @_, $text);return $text });
    $m->define_filter('funcall', sub { my $context = shift;my $verb = shift;my $fun = $m->{sarb}->getfun($verb) or die sprintf 'no such function %s', $verb;my @args = @_;return sub { my $text = shift;return &$fun($m->{sarb}, {}, @args, $text) } }, 1);
    return 1;
}


=head2 new_ready

construct the object with the a2e language elements

=cut

sub new_ready {
    my $class = shift;
    my $tmplopts = shift;
    my $m = new $class $tmplopts;
    $m->{sarb} = shift;
    $m->defin_bas_tmplfuns;
    return $m;
};

=head2 Function vmethod_arglist

read arguments, turning any array references (that might come from template variables) into lists of scalar arguments.

internal function, not an object method

=cut

sub vmethod_arglist {
    my @args = ();
    while (my $arg = shift) {
	if (ref($arg) eq 'ARRAY') {
	    push @args, @{$arg};
	} else {
	    push @args, $arg;
	};
    };
    return @args;
}

=head2 Function fill

Part of implementation of calls like

Given

    [% kom = 'This is $(ae:our company)' %]

we can say

    [% kom.fill('ae=:call:ahurl:${url}') %]

or, more safely, with a list variable

    [% kom.fill([ 'ae', 'call', 'ahurl', url ]) %]


=cut

sub fill {
    my $m = shift;
    $m->dpuc('fill', @_);
    my $text = shift;
    my @args = vmethod_arglist @_;
    my $arb = $m->{sarb}->parse($text);
    my $pos = 0;
    my $key = '';
    my $vars = {};
    my $val;
    my @funs = ();
    foreach my $arg (@args) {
	($pos, $key, $val) = $m->{sarb}->fillsetvar_from_tmplvar($pos, $arg);
	$vars->{$key} = $val;
    };
    $m->dgot(text => $text, vars => $vars, clue => 794);
    $text = $m->{sarb}->render($vars, $arb);
    $m->dpop('fill', $text);
    return $text;
}

=head2 Function call

Part of implementation of calls like

    [% 'value of pi: ${1}' | tb 'call' '3.14' %]

or

    [% 'value of pi is ${pi}' | tb %]

=cut

sub call {
    my $m = shift;
    $m->dpuc('call', @_);
    my $text = shift;
    my @args = vmethod_arglist @_;
    $text = $m->{sarb}->val_args_call($text, undef, @args);
    $m->dpop('call', $text);
    return $text;
}

=head1 The END

=cut

return 1;
