#!/usr/bin/perl -w
# Time-stamp: <2019-01-03 02:11:16 phm>

=head1 NAME

    A2E::TestUnits(3)

=cut

package A2E::TestUnits;
use strict;

=head1 SYNOPSIS

	my $p = new_ready A2E::TestUnits;
	$p->testunits(@ARGV) or die sprintf 'failed to %s', $p->koniug_et('testunits', @ARGV);

=head1 DESCRIPTION

Run a series of tests

=head1 APPLICATIONS

=over
=item testunits(1)
=back

=head1 PREREQUISITES

=over

=item A2E::Prog(3)

Further options inherited from here

=item Test::Unit(3)

with Test::Unit::Debug(3), Test::Unit::Runner(3)

=back    

=cut

use base 'A2E::Prog';
use warnings;
use diagnostics;
use Test::Unit::Debug qw(debug_pkgs);
use Test::Unit::TestRunner;

=head1 IMPLEMENTATION

=head2 Constructor Functions

=head3 Function defvars

=cut

sub defvars {
    my $m = shift;
    $m->SUPER::defvars;
    $m->testunits_defvars;
    return 1;
};

=head3 Function testunit_defvars

=cut

sub testunits_defvars {
    my $prog = 'testunits_defvars';my $m = shift;$m->dpuc($prog,@_);
    $m->{konfig}->define('incdirs|inc=s@');
    $m->konfayl('testunits.konf');
    $m->dpop($prog,1);return 1;
};

=head2 Executor Functions

=head3 Function lib2fe

Generate name for logfile

=cut

sub lib2fe {
    my $m = shift;
    local $_ = shift;
    s(\:+)(_)g;
    $_ = lc $_;
    $_ .= '.log';
    return $_;
};

=head3 Function testunits

=cut

sub redirect_stderr {
    my $m = shift;
    my $lib = shift;
    my $test = shift;
    my $fe = $m->lib2fe($lib);
    my $fe_tmp = join '.', $fe, 'tmp';
    my $sav;
    open $sav, '>&STDERR' or die 'failed to save stderr';
    open STDERR, '>:utf8', $fe_tmp or die sprintf 'failed to redirect STDERR to %s', $fe_tmp;
    my $ok = eval { $test->start($lib) };
    open STDERR, '>:utf8&=', $sav or die 'failed to restore stderr';
    if ($@ or not $ok) { print STDERR $m->fayl_get_str($fe_tmp);die "failed $0: $@: $!" };
    rename $fe_tmp, $fe; 
    $m->svn_add_commit($fe);
    return 1;
}

sub testunits {
    my $m = shift;
    $m->dpuc('testunits', @_);
    my @libs = @_;
    my @incdirs = @{$m->{konfig}->incdirs};
    @INC = (@incdirs, @INC);
    my $test = new Test::Unit::TestRunner;
    $test->print_stream->binmode(':utf8');
    debug_pkgs qw(Test::Unit::TestCase);
    my $verbose = $m->pvar('verbose');
    foreach my $lib (@libs) {
	print "Testing $lib\n" if $verbose;
	do { $m->redirect_stderr($lib, $test);next } if $m->{_redirektu_mian_eroron_};
	$test->start($lib, '-v') or die sprintf 'can not start test on lib %s', $lib;
	print "\n" if $verbose;
    };
    $m->dpop('testunits', 1);
    return 1;
}

=head1 The END

=cut

return 1;

# Local Variables:
# coding: utf-8
# srcfile: /adv/perl/A2E/TestUnits.pm
# mode: perl
# End:
