=head1 NAME

A2E::Texchars

=cut

package A2E::Texchars;
use strict;
use utf8;
# use Encode qw(_utf8_on);

=head1 VERSION

    This describes version B<0.1.0> of A2E::Texchars.

=cut

    our $VERSION = '0.1.0';

=head1 SYNPOSIS

    use base 'A2E::Dokfs', 'A2E::Valid', 'A2E::Texchars';
    my $p = new_ready A2E::MLHT;
    ...
    $c->define_filter('u2tex', sub { my $context = shift;return sub { my $text = shift;return $m->str_u2tex($text) } }, 1);
    ...

=head1 DESCRIPTION

Replace 'übermäßig' with '\"uberm\"a{\ss}ig'.

Support standalone program u2tex and Template filter u2tex to do this.

Allow user to add new characters via --texchars € '{\euro}'.

This is a mixin library.  It does not invoke the constructor of its base library.

=head1 Prerequisites

    A2E::Prog(3)

=cut

use base 'A2E::Dokfs';

=head1 SEE ALSO

    A2E::Mktdir(3)

=head1 Implementation

=head2 Overwritable Functions

=head3 defvars

=cut

our $texchars = { ā => q{\=a}, ō => q{\=o}, ū => q{\=u}, ē => q{\=e}, ī => q{\=i}, ä => q{\"a}, ö => q{\"o}, ü => q{\"u}, Ä => q{\"A}, Ö => q{\"O}, Ü => q{\"U}, ß => '{\ss}', ê => q{\^e}, É => '\^E', é => q{\'e}, É => q{\'E}, è => q{\`e}, È => '\`E', á => '\'a', Á => q{\'A}, à => q{\`a}, â => q{\^a}, À => q{\`A}, ô => '\^o', û => '\^u', ĉ => '^c', Ĉ => '\^C', ĝ => '\^g', ç => '\c{c}', Ĝ => '\^G', ĵ => '\^j', Ĵ => '\^J', ĥ => '\^h', '„' => '``', "\"" => q{''}, '“' => q{``}, '”' => q{''}, '«' => '``', '»' => q{''}, '%' => '\%', '&' => '\&' , '_' => '＿', '"' => "''", '#' => '\#' };
our $htmchars = { ä => q{&auml;}, ö => q{&ouml;}, ü => q{&uuml;}, Ä => q{&Auml;}, Ö => q{&Ouml;}, Ü => q{&Uuml;}, ß => q{&szlig;}, '&' => q{&amp;} };

sub texchars_defvars {
    my $m = shift;
    $m->dpuc('texchars_defvars');
    $m->cache_define('texchars', 's%', $texchars);
    $m->cache_define('htmchars', 's%', $htmchars);
    $m->konfayl('texchars.konf');
    $m->dpop('texchars_defvars', 1);
    return 1;
};

=head2 Subroutines

=head3 str_u2tex

replace string 'äöüß' with '\"a\"o\"u{\ss}'

=cut

sub str_u2tex {
    my $prog = 'str_u2tex';my $m = shift;$m->dpuc($prog,@_);
    my $s = shift;
    $s = join '', map { my $chr = pack 'U', $_;my $tex = $m->{cache}->{texchars}->{$chr};$tex || $chr } unpack 'U*', $s;
    $m->dpop($prog,$s);return $s;
}
sub str_u2htm {
    my $prog = 'str_u2htm';my $m = shift;$m->dpuc($prog,@_);
    my $s = shift;
    $s = join '', map { my $chr = pack 'U', $_;my $htm = $m->{cache}->{htmchars}->{$chr};$htm || $chr } unpack 'U*', $s;
    $m->dpop('str_u2htm',$s);
    return $s;
}

=head1 Changelog

=over
=item 2017-06-09 moved out from A2E::U2tex
=back

=head1 The End

=cut

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# End: 
