package A2E::TmplArb;
use base 'A2E::Prog';
our %MSGS = (
    putundef => 'putting var %n to undefined value' );

=head4 $lihs = $m->lihs_new: generate a new lihs object

In A2E::Tmplfil this is enriched so that each new $lihs object has basic variables like __dabarels__ set.

=cut

sub lihs_initvars {
    my ($m,$prog,$lihs) = shift->lpuc('lihs_initvars',@_);
    $m->dpop1($prog,1);
};

=head4 ($lihs,$oldlihs,$dir) = $m->lihs_new($m->{lihs}): Generate new current lihs object and return old one into a variable for later restoring

Set up a new A2E::Lihs object for use in A2E::Tmplfil and above, equipped with basic variables.
$lihs becomes the current ($m->{lihs}) lihs object so that certain functions (template vmethods, sarb functions) can also be connected to it.
CAVEAT: the object does not yet have its directory, which is set up with $lihs->setdir($dir) later (e.g. when $dir is ready for use).
Support file:/adv/perl/A2E/MLMK.pm.tmpl::lihs_new

=cut

sub lihs_new {
    my ($m,$oldlihs,$dir) = @_;
    my $lihs = new A2E::Lihs;
    $lihs->setdir($dir) if $dir;
    $m->lihs_initvars($lihs);
    $m->{lihs} = $lihs;
    $m->{sarb}->{lihs} = $lihs;
    ($lihs,$oldlihs);
};

=head4 lihs_old;

Support file:/adv/perl/A2E/MLMK.pm.tmpl::lihs_old
=cut
sub lihs_old {
    my ($m,$prog,$oldlihs) = shift->ppuc('lihs_old',@_);
    $m->{lihs} = $oldlihs;
    $m->{sarb}->{lihs} = $oldlihs;
    $m->dpop1($prog,1);
};


=head4 $m->get_lihs($lihs, %opts)

Obtain a lihs object, either from argument $lihs or default from $m->{lihs}.
With option insinstp => 1, insist that the argument must be a valid lihs object. 

=cut
sub get_lihs {
    my ($m,$lihs,%opts) = @_;
  LIHS: {
      last if $lihs or $opts{insistp};
      $lihs = $m->{lihs};
      last if $lihs;
      ($lihs) = $m->lihs_new;
    };
    $m->pstop('I need reference to a A2E::Lihs object, not to %r', r => ref($lihs)) if 'A2E::Lihs' ne ref $lihs;
    $m->{lihs} = $lihs;
    $lihs;
};
=head4 ($m,$prog,$lihs,@_) = shift->lpuc($prog,$lihs,@_): Simplify beginning of functions that extend $m with $lihs
=cut

sub lpuc {
    my ($m,$prog,$lihs,@args) = @_;
    $lihs = $m->get_lihs($lihs);
    $m->dpuc($prog,@args);
    ($m,$prog,$lihs,@args);
};
=head4 $k = $m->vahrflags($k,$vahr,%opts{nov}): set modalities of validation

$k => dok_get
$k :: in this case switch 'get' flag on

The nov option says that the flags are newly set and will not be deleted by the next valid3fun8nom but only thereafter.  This is done via an internal __nov__ flag.
valid3fun8nom offers this facility but valid_args no longer needs it.

=cut
sub vahrflags {
    my ($m,$k,$vahr,%opts) = @_;
    ($k,my @flags) = split '_', $k;
    $vahr->{__flagp__}->{$_} = 1 for @flags;
    $vahr->{__flagp__}->{__nov__} = 1 if $opts{nov};
    $k;
};
=head4 ($validp,$nom) = $m->valid3fun8nom($validp,$vahr): resolve the validator symbol/reference to the ultimate coderef/name pair

When $validp is a coderef, it is returned without $nom, which is suboptimal because we want to store the final validated argument in $vahr->{$nom} later.
Optimally $validp is a symbolword which is returned as $nom and whose $m->{var2validp}->{$nom} is a coderef to a returnable validator function.
The $vahr hashref is not returned but by side-effect enhanced with __flagp__ flags that are available to the function and are currently 
	let :: allow null values
	get :: obtain a default, based on info from $vahr et al, when value is null
Old flags (from parent functions) are discarded in the VAHR section. 

=cut
our %vah0 = ( __flagp__ => {} );
sub valid3fun8nom {
    my ($m,$validp,$vahr) = @_;
  VAHR: { # reset $vahr
    NOV: { # new $vahr 
	last if $vahr;
	$vahr = { %vah0 };
	last VAHR;
      };
      $m->pstop('invalid vahr %v', v => $vahr) if 'HASH' ne ref $vahr;
    FLAGP: {
	last if $vahr->{__flagp__};
	$m->pwarn('__flagp__ should always be present in a vahr such as %v, reparing', v => $vahr);
	$vahr->{__flagp__} = {};
	last VAHR;
      };
      if ($vahr->{__flagp__}->{__nov__}) {
	  $vahr->{__flagp__}->{__nov__} = 0;
      } else {
	  %{$vahr->{__flagp__}} = ();
      };
    };
    my $ref = ref $validp;
    return $validp if 'CODE' eq $ref;
    $m->pstop('bad validator %v, must be either code or string', v => $validp) if $ref;
    my $nom = $m->vahrflags($validp,$vahr);
    $validp = $m->{var2validp}->{$nom};
    return $validp, $nom if 'CODE' eq ref $validp;
    $m->pstop('bad validator %v, should have been a coderef', v => $validp);
};

=head4 $arg = $m->validigu($lihs,$valid,$arg): Validate argument with validator

Apply the validator $valid to return a potentially modified argument.
Return the argument as is if the validator is null.
Stop with error if the validator is inapplicable.
The validator can be a code reference or a key of $m->{var2validp} whose value is a code reference.
The same validators are used in file:/adv/perl/A2E/MLHT.pm.tmpl::var2validp for setpvar.
Cf also file:/adv/perl/A2E/MLMK.pm.tmpl::var2validp for a trg i.e. target validator.

=cut
sub validigu {
    my ($m,$lihs,$valid,$arg,$vahr) = @_;
    return $arg if !$valid;
    ($valid,$nom) = $m->valid3fun8nom($valid,$vahr);
    $arg = &$valid($m,$lihs,$arg,$vahr);
    if ($vahr->{__nom__}) { $nom = $vahr->{__nom__};$vahr->{__nom__} = '' };
    $vahr->{$nom} = $arg if $nom and $arg;
    $arg;
};

=head4 ($m,$prog,$lihs,@_) = shift->vlpuc($prog,$lihs,$validr,@_)

$validr => [[\&valid_arg1, ... \&valid_arg9], \&valid_map_rest ]

Like lpuc but with validators as in opuc and with additional possibilities of handling options if the $validr argumet is so set, e.g.

$validr => [[\&valid_arg1, ... \&valid_arg9],[ \&valid_dok, { prelang => \&valid_lang } ]]

Hereby opuc and lpuc have converged, the main difference being that vlpuc takes an obligatory $lpuc positional argument whereas opuc takes this from the options and therefore should be used with an options-validator and not a rest-validator, whereas vlpuc could have either of the two, although for the options-validator variant the use of opuc is the way to go at least in the frequent case where $lihs is not among the positional arguments.

file:/adv/perl/A2E/Dokdata2db.pm::vlpuc

=cut
sub vlpuc {
    my ($m,$prog,$validr,$lihs,$vahr,@args) = @_;
    $lihs = $m->get_lihs($lihs);
    my ($vahr,$argvr,%opts) = $m->valid_args($lihs,$validr,$vahr,@args);
    my @argv = @$argvr;
    $m->dpuc($prog,@argv);
    ($m,$prog,$lihs,$vahr,@argv);
};
    
=head4 my ($m,$prog,$lihs,\%opts,@_) = shift->tpuc($prog,\%opts,@_): Simplify beginning of functions that extend $m with a $lihs contained in \%opts

e.g.
	($ok, $nom2) = $m->grupopen(\%opts,$_);
  . =>    my ($m,$prog,$lihs,$optsr,$s) = shift->ppuc('grupopen',@_);


Compared to %opts the optsr is makes it easier to passing $lihs on down the tree and to deal with an unknown number of positional arguments.
The \%opts is used in A2E::Tmplfil like a local extension of $m/$lihs object.

=cut
sub tpuc {
    my $m = shift;
    my $prog = shift;
    my $optsr = shift;
    my $lihs = $m->get_lihs($optsr->{lihs});$optsr->{lihs} = $lihs;
    my $gvars = $optsr->{gvars} || $m->{grupvars};$optsr->{gvars} ||= $gvars;
    $m->dpuc($prog,@_);
    ($m,$prog,$lihs,$optsr,@_);
};
=head3 ($vahr,\@argv,%opts) = $m->valid_args($lihs,$validr,$vahr,@args): validate complete argument line

Handles the validator spec parameter in opuc/vlpuc/vopuc.
This can consist of 1,2 or 3 parts.
 1 :: positional arguments
 2a :: option names transferred to positional arguments
 2b :: single name of rest argument e.g. 'lang' for @langs
 3 :: mapping from unique argument/option name to validator

Examples

    my ($m,$prog,$lihs,$tml,$tmpl,$prez,$tmkz) = shift->opuc('tml8tmpl7pre8tmk3dirs',[[qw(tml tmpl prez tmkz)],{ prez => lstr, tmkz => lstr }],@_);
    my ($m,$prog,$lihs,$dok,$lang,$data,%opts) = shift->opuc('ml_dok2absurl',[[qw(dok_get)],['lang_let', 'data'],{ data => 'hacr_let' }],@_);
    my ($m,$prog,$lihs,$dok,$lang,%opts) = shift->opuc('ml_dok2absdir',[[qw(dok)],[qw(lang_let)]],@_);
    my ($m,$prog,$lihs,$dok,%opts) = shift->opuc('db_dok2absurl',[qw(dok)],@_);
    my ($m,$prog,$lihs,$vahr,$dok,$tmpl,@langs) = shift->vlpuc('dok_langs_put_littups',[[qw(dok tmpl)],'lang'],@_);

=cut

sub valid_args {
    my ($m,$lihs,$validr,$vahr,@args) = @_;$vahr ||= { %vah0 };
    my @kvalids = ();
    my %kvalidp = ();
    my %opts = ();
    my @argv = ();
    my @valids = @$validr;$validr = undef;
    if ('ARRAY' eq ref $valids[0]) {
	my $valids0 = shift @valids;
      KVALIDP: {
	  last if !@valids;
	  my $larg = pop @valids;
	  my $lref = ref $larg;
	  my $karg = shift @valids;
	  my $lhacp = 'HASH' eq $lref;
	  if ($lhacp) {
	      %kvalidp = %$larg;
	  } elsif ($karg) {
	      $m->pstop('bad last validator argument %l, must be hashref when second argument %k is present', k => $karg, l => $larg) ;
	  };
	VALIDR: {
	    $karg ||= $larg if !$lhacp;
	    last if !$karg;
	    my $kref = ref $karg;
	    if ($kref) {
		if ('ARRAY' eq $kref) {
		    @kvalids = @$karg;
		} elsif ('CODE' ne $kref) {
		    $m->pstop('validator has invalid middle argument %k', k => $karg);
		} 
	    } else {
		$validr = $karg;		
	    };
	  };
	};
	@valids = @$valids0;
    };
    my $kvalidp7valid = sub {
	my ($valid) = @_;
	my $ref = ref $valid;
	return $valid if 'CODE' eq $ref;
	$m->pstop('bad validator %v', v => $valid) if $ref;
	(my $kvalid,%kvalidp) = $m->shifthash($valid,%kvalidp);
	return $valid if !$kvalid;
	$vahr->{__nom__} = $valid;
	$kvalid;
    };
    foreach my $valid (@valids) {
	my $arg = shift @args;
	$valid = &$kvalidp7valid($valid);
	push @argv, $m->validigu($lihs,$valid,$arg,$vahr);
    };
    if ($validr) { # process rest of arguments, e.g. @langs with a single validator 'lang'
	$validr = &$kvalidp7valid($validr);
	while (@args) {
	    my $arg = shift @args;
	    push @argv, $m->validigu($lihs,$validr,$arg,$vahr);
	};
    } else { # process options
	%opts = $m->valid_opts(@args);
	while (@kvalids) { # first turn some options into positional parameters
	    my $k = shift @kvalids;
	    # $k = $m->vahrflags($k,$vahr,nov => 1); # no longer needed since we do not use flags in cases where %kvalidp contains the final validator name
	    my ($k0) = split '_', $k;
	    (my $arg,%opts) = $m->shifthash($k0,%opts);
	    $valid = &$kvalidp7valid($k);
	    push @argv, $m->validigu($lihs,$valid,$arg,$vahr);
	};
	foreach my $k (keys %opts) { # then check the rest in place
	    last if !%kvalidp;
	    (my $valid,%kvalidp) = $m->shifthash($k,%kvalidp);
	    $opts{$k} = $m->validigu($lihs,$valid,$opts{$k},$vahr);
	};
    };
    ($vahr,\@argv,%opts);
};
=head4 ($val,@args) = $m->popfintups($nom,@args): find value in final %opts part of @args

similar to $m->shifthash($nom,%opts) but search in an unknown argument list from end backward as far as it looks likes %opts.

=cut
sub popfintups {
    my ($m,$nom,@args) = @_;
    my ($arg,@argv) = ();
    while (($#args > 0) && ($args[$#args-1] =~ m(\A\w+\Z))) {
	my $v = pop @args;
	my $k = pop @args;
	do { $arg = $v;last } if $k eq $nom;
	unshift @argv, $k, $v;
    };
    unshift @argv, @args;
    ($arg,@argv);
}

=head4 ($m,$prog,$lihs,$arg*,%opts) = shift->opuc($prog,$npa,@_,_keys_ = \@optskeys): Simplify argument processing of functions that transport $lihs in %opts

$prog :: function name
$npa :: number of positional arguments
$arg* :: arguments in number given by $npa

Store positional arguments and named arguments as indicated and get $lihs from options, validate/generate it and store the valid version so it can be passed down the function chain.

Example: 

	my ($m,$prog,$lihs,$nom,%opts) = shift->opuc('grup_add_grupvars',1,@_);

the invocation of which

	$m->grup_add_grupvars($sup,%opts)

assigns $sup to $nom, %opts to %opts and $opts{lihs} or a new object to $lihs and $opts{lihs}.

Or:
        my ($m,$prog,$lihs,$nom,$val,%opts) = shift->opuc('dabarel',2,@_);
invoked as
	$m->dabarel($nom,$val,%opts)

An alternative form supplies validators by a listref-type $npa (\@valids).
If $valids[0] is a listref it becomes what otherwise would be @valids, and $valids[1] receives additional meaning.
If $valids[1] is a listref, it becomes an instruction for handling of %opts, consisting of option-validator tuples that check/rectify those options and transfer them to additional positional arguments, with an optional final hashref of option-validator tuples that are used to check/rectify further %opts in place without transfer.
If $valids[1] is a single element, it is a validator with which to check all remaining arguments, but this is used only in vlpuc.
Any further %opts, especially $opts{lihs} can still be captured (by a final %opts argument) and handed down (by the same).  We advise against using such unconstrained %opts except as a quick and dirty way of programming to be corrected later on.

Example:

    sub valid_prelang { my ($m,$arg,$lihs) = @_;return '' if !$arg;return $arg if $arg =~ m(\A\w{2}\Z);$m->pstop('invalid prelang argument %a', a => $arg) };
    my ($m,$prog,$lihs,$dok,$tmpl,$lang,$prelang,%opts) = shift->opuc('dok_lang_getput_littups',[[\&valid_dok,\&valid_tmpl,\&valid_lang],[prelang => \&valid_prelang, { litstag => \&valid_litstag }]],@_);

invoked as

	$m->prog($dok,$tmpl,$lang, prelang => $prelang, lihs => $lihs, litstag => 'l')

cf file:/adv/perl/A2E/Dokdata2db.pm::valid_prelang

It seems desirable to fork the initial validation procedures of many functions out into such shared validator functions.

$validr changes meaning in the process.  Initially it is a reference to the validators list, finally it becomes the validator for the rest arguments, or void in order to signify that such do not exist.
This may seem kludgy but is a matter of convenience and economy.  Why not reuse an unneeded variable?

$vahr is a hashref that is passed to validators as optional last argument.  
$vahr contains variables that are useful to validators and may need to be retrieved in a more time-consuming way if not passed downward via $vahr.
Validators usually set a tuple in $vahr after validating.  $vahr is passed down and back up so as to collect current state information and speed up validation.
The $vahr argument that appears in vlpuc and vopuc is hidden in opuc0/opuc as an option.


The same validation system is used in vlpuc, the better version of lpuc. 

=cut

sub opuc0 {
    my ($m,$prog,$npa,@args) = @_;
    my $argvr = [];
    my %opts = ();
    my ($lihs,$vahr);
    if ($npa =~ m(\A\d\Z)) {
	do { my $arg = shift @args;push @$argvr, $arg } for 1..$npa;
	%opts = $m->valid_opts(@args);
	($vahr,%opts) = $m->shifthash('vahr',%opts);$vahr ||= { %vah0 };
	($lihs,%opts) = $m->shifthash('lihs',%opts);$lihs = $m->get_lihs($lihs);	
    } elsif ('ARRAY' eq ref $npa) {
	($vahr,@args) = $m->popfintups('vahr',@args);$vahr ||= { %vah0 };
	($lihs,@args) = $m->popfintups('lihs',@args);$lihs = $m->get_lihs($lihs);
	($vahr,$argvr,%opts) = eval { $m->valid_args($lihs,$npa,$vahr,@args) };
	$m->pstop('failed to validate %p arguments %A: %m', p => $prog, A => \@args, m => $@) if $@;
    };
    $opts{lihs} = $lihs;
    $opts{vahr} = $vahr;
    $m->dpuc($prog,$argvr,%opts);
    ($m,$prog,$lihs,$argvr,%opts);
};
sub opuc {
    my ($m,$prog,$lihs,$argsr,%opts) = shift->opuc0(@_);
    ($m,$prog,$lihs,@$argsr,%opts);
};
sub vopuc {
    my ($m,$prog,$lihs,$argsr,%opts) = shift->opuc0(@_);
    ($m,$prog,$lihs,$opts{vahr},@$argsr,%opts);
}
sub qlpuc { # quit if ready
    my ($m,$prog,$lihs,@args) = @_;
    return if $m->{ready}->{$prog};
    $lihs = $m->get_lihs($lihs);
    $m->dpuc($prog,@args);
    ($m,$prog,$lihs,@args);
};


=head4 Constant @arbseps

Listref of separator characters, used for coding structured expressions in strings used by C<arb2str>, defined by configuration option arbseps in tmplfil.konf, e.g.

	arbseps = | + : - ! & / * ? . ^ , ~ 61 35

such that 61 and 35 will be stored as the characters whose numbers they are, namely C<=> and C<#> respectively.
=cut

our @arbseps = ('+', '|', ':', '!', '*', '/', '.', '&', '-', '?', '^', '~', '#', '=');
our %arbsepp = ();$arbsepp{$_} = 1 for @arbseps;
sub arbsepp {
    my ($m,$prog,$sep) = shift->ppuc('arbsepp',@_);
    $m->dpop1($prog,$arbsepp{$sep});
};

=head3 $arg = $m->valid_\w+($lihs,$arg,$vahr): validators
for use with A2E::TmplArb vlpuc, opuc, vopuc via $m->{var2validp}
=cut
sub valid_text {
    my ($m,$lihs,$text,$vahr) = @_;
    $m->pstop('empty text') if !$text and !$vahr->{__flagp__}->{let};
    $m->pstop('need string, not reference') if ref $text;
    $text;
};
sub valid_nom {
    my ($m,$lihs,$nom,$vahr) = @_;
    return '' if !$nom and $vahr->{__flagp__}->{let};
    $m->pstop('need symbol name, not %n', n => $nom) if $nom !~ m(\A\w+\Z);
    $nom;
};
sub valid_nom3 { # document name with trailing stuff e.g. oas_pub2001/index.de.html, oas_pub2001.ont; m(\A\w{3}\S*\w{3}\Z)
    my ($m,$lihs,$nom,$vahr) = @_;
    return '' if !$nom and $vahr->{__flagp__}->{let};
    $m->pstop('need symbol name, optionally with trailing stuff, but not %n', n => $nom) if $nom !~ m(\A\w+\Z) and $nom !~ m(\A\w{3}\S*\w{3}\Z);
    $nom;
};
sub valid_url {
    my ($m,$lihs,$url,$vahr) = @_;
  NOURL: {
      last if $url;
      return '' if $vahr->{__flagp__}->{let};
      last if !$vahr->{__flagp__}->{get};
      $url = $lihs->{vars}->{__url__} || $lihs->{vars}->{url} || $m->pvar('url') || $m->dir2url($vahr->{pwd} || $m->get_cwd);
      $m->setpvar(url => $url, lihs => $lihs);
    };
  URLMATCH: { 
      last if $url =~ m(\A\w+\:\/\/\w);
      $m->pstop('need URL, not %u', u => $url);
    };
    $url;
};
sub valid_fun {
    my ($m,$lihs,$fun,$vahr) = @_;
    return sub { @_ } if !$fun and $vahr->{__flagp__}->{let};
    $m->pstop('need coderef, not %f', f => $fun) if 'CODE' ne ref $fun;
    $fun;
};
sub valid_hacr {
    my ($m,$lihs,$hacr,$vahr) = @_;
    return {} if !$hacr and $vahr->{__flagp__}->{let};
    $m->pstop('need hashref, not %h', h => $hacr) if 'HASH' ne ref $hacr;
    $hacr;
};
sub valid_lstr {
    my ($m,$lihs,$lstr,$vahr) = @_;
    return [] if !$lstr and $vahr->{__flagp__}->{let};
    $m->pstop('need listref, not %l', l => $lstr) if 'ARRAY' ne ref $lstr;
    $lstr;
};

=head4 %opts = $m->valid_opts(@args): turn arguments into hash if suitable or fail
=cut
sub valid_opts {
    my ($m,%opts) = @_;
    do { next if m(\A\w+\Z);
	 $m->pstop('bad option key %k', k => $_);
    } for keys %opts;
    %opts;
};
sub tmplarb_defvars {
    my ($m,$prog,$lihs) = shift->qlpuc('tmplarb_defvars',@_) or return;
    $m->prog_msgdefs('tmplarb',%MSGS);
  SArb: {
      last if 'A2E::SArb::Make' eq ref $m;
      $m->{sarb} = new_ready A2E::SArb::Make noexec => 1, lihs => $m->{lihs}; # new_sarb: gvars => $m->{grupvars}
      do { $m->pwarn('lihs is not the same between sarb and m, resetting');$m->{sarb}->{lihs} = $m->{lihs} } unless $m->{lihs} eq $m->{sarb}->{lihs};
      $m->pstop('why is lihs not a proper object?') unless 'A2E::Lihs' eq ref $m->{lihs};
      $m->pstop('why is sarb_lihs still not prog_lihs') unless $m->{sarb}->{lihs}->{vars} eq $m->{lihs}->{vars};
      sub textsep2listref { return shift->{sarb}->textsep2listref(@_) }
      sub textsep2seplist { return shift->{sarb}->textsep2seplist(@_) }
      sub textsep2list { return shift->{sarb}->textsep2list(@_) }
      sub text2seplist { return shift->{sarb}->text2seplist(@_) }
      sub text2list { return shift->{sarb}->text2list(@_) }
    };
    $m->{var2validp}->{text} = \&valid_text;
    $m->{var2validp}->{url} = \&valid_url;
    $m->{var2validp}->{nom} = \&valid_nom;
    $m->{var2validp}->{nom3} = \&valid_nom3;
    $m->{var2validp}->{fun} = \&valid_fun;
    $m->{var2validp}->{hacr} = \&valid_hacr;
    $m->{var2validp}->{lstr} = \&valid_lstr;
    $m->qpop1($prog,1);
};
sub defvars {
    my ($m,$prog) = shift->qpuc('A2E::TmplArb::defvars', @_) or return;
    $m->SUPER::defvars;
    $m->tmplarb_defvars;
    $m->qpop1($prog,1);
};

sub x2list { my ($m,$x) = @_;return ref $x ? $m->ref2list($x) : $m->text2list($x) };

=head3 Function str2arb and strl2arb

    args := str, lvlmax?
    rets := arb
    str := string
    arb := tree
    tree := listref || hashref || scalar
    str => '|a:5|b:/2a,2b/3/|'
    arb => [ [ 'a', '5' ], [ 'b', [ [ '2a', '2b' ], '3' ] ] ]
    str => '||a|1||b|2a,2b|'
    arb => { a => '1', b => [ '2a', '2b' ] }
    str => 'abc'
    arb => 'abc'
    lvlmax := integer
    lvlmax :: maximal nesting depth; zero or negative means unlimited nesting.

Assuming that elements consist of \w+ only and anything else, in particular arbseps, are separators, and that the first encountered separator belongs to the next level, parse a string into a structure.

I.o.w. return a structure where one is found and to the extent that this is allowed by lvlmax.

strl2arb requires the second argument whereas str2arb reads it from str and optionally accepts it if str didnt yield.

See also reverse functions arb2str and arb2strl

=cut

sub str2str8lvlmax {
    my ($m,$prog,$s) = shift->ppuc('str2str8lvlmax',@_);
    return $m->dpop($prog,'',$s) if $s =~ m(\A\d*\Z); 
    my ($lvlmax,$str,$sep) = $s =~ m(\A(\d*)((\W).*\3)\s*\Z);$lvlmax ||= 0;
    $m->pstop('expression %e lacks the separator that marks structured variables', e => $s) if !$str;
    $m->dpop($prog,$str,$lvlmax,$sep);
};
sub strl2arb {
    my ($m,$prog,$str,$lvlmax,$sep) = shift->ppuc('strl2arb',@_);$lvlmax ||= 0;
    $lvlmax--; # breakpoint is zero, must not stay zero if parsing is to proceed without fixed limits
    ($sep, my @elms) = $m->textsep2seplist($str, $sep, noblanksepp => 1); # file:/phm/19/06/10/sig/_lng.phm_pub_sig190610.txt::noblanksepp
    my $arb = undef;
  TRY: {
      last unless @elms;
      do { $arb = $elms[0];last TRY } if !$sep and !$#elms;
    LIST: {
	my $elm0 = $elms[0];
	last unless length $elm0; # initial empty field indicates a HASH: ||key|val||nom|val||... but |0|key|... is an array [2018-12-25 Di 11:47]
      REKURS: {
	  ($sep) = $m->text2seplist($elm0);$sep = '' unless $sep eq substr($elm0,0,1) or grep { $_ eq $sep } '+', '/', '|'; # allow + in middle e.g. /proc+linioi/, be more generous when elm starts with sep
	  do { $arb = \@elms;last } if !$sep or !$lvlmax; # interrupt recursion when lvlmax descends to 0 or first element does not contain sep
	  $arb = [ map { $m->strl2arb($_, $lvlmax, $sep) } @elms ];
	};
	last TRY;
      };
      $arb = {};
      my $noms = [];
      while (@elms) {
	  shift @elms;
	  my $nom = shift @elms;
	  unless ($nom =~ m(\A\w+\Z)) {
	      $m->decho('non-symbol hash key %n', n => $nom);
	  };
	  my $val = shift @elms;
	  $val = $m->strl2arb($val, $lvlmax) unless $lvlmax == 0;
	  $arb->{$nom} = $val;
	  push @$noms, $nom;
      };
      $arb->{'_keys_'} = $noms;
    };
    $m->dpop1($prog,$arb);
};
sub str2arb {
    my ($m,$prog,$str,$lvlmax0) = shift->ppuc('str2arb',@_);
    ($str,my $lvlmax) = $m->str2str8lvlmax($str);
    $lvlmax ||= $lvlmax0;
    my $arb = $m->strl2arb($str,$lvlmax);
    $m->dpop1($prog,$arb);
};

=head3 Functions arb2strl and arb2str

Recursively generate string representation from tree structure.
Inversion of C<str2arb>. 

    args := arb, lvl?
    arb :: a tree structure
    lvl :: number of the hierarchy level, default 0.
    rets := str, lvlmax
    str :: a string representation of it, using separators C<arbseps> for the layers of the hierarchy.

The arb2strl version returns the string and lvlmax, whereas arb2str only returns a string (? perhaps soon with lvlmax prepended)

The __arb2slu version additionally takes an list of unused separators @arbseps and returns a list of used ones $uslr.

__arb2slu is the base function and it uses subroutine ssu2sep8seps to make sure that if the separator is found in some terminal string of the tree then we try to use another separator which is searched from the end of the list of separators.
The original list is found in @arbseps, where it comes via configuration variable of same name from tmplfil.konf, and should contain some good separators at the end.

The C<arbseps> separators are not reused at lower levels because this would derail parsing.
If the arbseps separators list is exhausted, an error occurs.
The list is already much to long for this to normally occur, but Unicode provides a vast reservoir of wide characters for extending it.  Even implausible characters can be specified, also by means of their number.  This system might break if we tried to abuse arb2str for encoding large Perl objects but hopefully not for anything that came from text via str2arb.

=cut

sub __ssus2sep {
    my ($m,$prog,$sep,$sepsr,$ushr,@strl) = shift->ppuc('__ssus2sep',@_);
    my @seps = @$sepsr;
    my %ush = %$ushr;
    @seps = grep { !$ush{$_} } @seps;
    push @seps, $sep;
    my @strlu = map { unpack 'U*', $_ } @strl; # cant use regexp here since m($sep) resolves to m(|) which is interpreted as matching everything
    while (@seps) {
	$sep = pop @seps;
        my $sepu = unpack 'U', $sep;
	next if grep { $sepu == $_ } @strlu;
	return $m->dpop1($prog,$sep);
    };
    $m->pstop('none of the remaining separators from arbseps %A can encode these items: %S', A => $sepsr, S => \@strl);    
};

sub __arbls2strlu { 
    my ($m,$prog,$arb,$lvl,$seplr) = shift->ppuc('__arbls2strlu',@_);
    return $m->dpop($prog,$arb,$lvl) unless ref $arb;
    my $str = '';
    $lvl++;
    my $lvlmax = $lvl;
    my @sepl = @$seplr;
    my $sep = shift @sepl;
    my %ush = (); # used separators
  TRY: {
      my @strl = ();
      local $_;
      my $subarb7strlu = sub {
	  my ($subarb) = @_;
	  my ($str,$lvlsub,$uslr) = $m->__arbls2strlu($subarb,$lvl,\@sepl);
	  $lvlmax = $lvlsub if $lvlsub > $lvlmax;
	  $ush{$_} = 1 for @$uslr;
	  push @strl, $str;
      };
    ARRAY: {
	last unless 'ARRAY' eq ref $arb;
	foreach my $subarb (@$arb) { &$subarb7strlu($subarb) };
	$sep = $m->__ssus2sep($sep,\@sepl,\%ush,@strl);
	do { $str = '';$lvlmax--;last TRY } unless @strl;
	$str = join $sep, '', @strl, '';
	last TRY;
      };
      $arb->{_keys_} = [ keys %$arb ] unless 'ARRAY' eq ref $arb->{_keys_};
      my $nomlr = $arb->{_keys_};
      my @keys = grep { $_ ne '_keys_' } keys %$arb; 
      if ('ARRAY' eq ref $nomlr) {
	  $m->uniqpush($nomlr,\@keys);
      } else {
	  $nomlr = \@keys;
      }
      $arb->{_keys_} = $nomlr;
      foreach my $nom (@$nomlr) { 
	  push @strl, '', $nom;
	  &$subarb7strlu($arb->{$nom});
      };
      $sep = $m->__ssus2sep($sep,\@sepl,\%ush,@strl);
      do { $str = '';$lvlmax--;last } unless @strl;
      $str = join $sep, '', @strl, '';
    };
    my @usl = keys %ush;
    push @usl, $sep;
    # $str = $lvlmax . $str if $lvlmax;
    $m->dpop($prog,$str,$lvlmax,\@usl);
};

sub arb2strl {
    my ($m,$prog,$arb,$seplr) = shift->ppuc('arb2strl',@_);
    $seplr ||= [ @arbseps ];
    my ($str,$lvlmax) = $m->__arbls2strlu($arb,0,$seplr);
    $m->dpop($prog,$str,$lvlmax);
}
sub arb2str {
    my ($m,$prog,$arb,$seplr) = shift->ppuc('arb2str',@_);
    my ($str,$lvlmax) = $m->arb2strl($arb,$seplr);
    # $str = $lvlmax . $str if $lvlmax;
    # 最終目標但易出錯, file:/phm/18/10/14/sig/_log.j.phm_pub_sig181014.txt::21878
    $m->dpop1($prog,$str);
}
sub klavs {
    my ($m,$prog,$hac) = shift->ppuc('klavs',@_);
    my $klavs = $hac->{_keys_};
    return @$klavs if 'ARRAY' eq ref $klavs;
    keys %$hac;
}

=head3 Function non_dynavar_p

Whether a name is unacceptable for dynamic variables.
Regexp used to decide based on the textchunk name whether this structured variable may contain references.
By default such a variable should contain '__' somewhere in its name.
Being considered dynamic does not do harm to a static structure, apart from a slight reduction of processing speed due to the need of executing arb2xarb in nom2hac/nom2lst.
    args := nom
    nom :: variable name
    rets := ok
    ok := boolean

=cut

sub non_dynavar_p {
    my $m = shift;
    local $_ = shift;
    m(\_\_) ? 0 : 1;
};

=head4 __defarb__

Establish a tree variable of some kind.
Internal function of this library.

Parameters

    fun
    reftyp
    tmpltyps
    dyn
    redef
    lvlmax :: maximum allowed number of nexting levels
    nom :: variable name
    val :: value

The latter parameters are usually passed on from application functions, the former are used by functions of this library.

dbmvars_put is an underlying function that lets the user put to a dbm variables hash of his choice in the same sane way whose purpose is ultimately suitability to dbm storage.

=cut

sub __defarb__ {
    my ($m,$fun,$typ,$dyn,$redef,$lvlmax,$nom,$val,$lihs,$seplr,%putopts) = @_;$lihs = $m->get_lihs($lihs);
    my $tref = $lihs->tref($typ) or $m->pstop('no reftyp for typ %t', t => $typ); # (?:ARRAY|HASH)
    my ($treg,$typp) = $lihs->tyrp($typ) or $m->pstop('no typs for typ %t', t => $typ); # 'lsts', 'lstp'
    my $iniarb = $lihs->iniarb($typ);$m->pstop('no iniarb for typ %t', t => $typ) unless defined $iniarb; # []
    $m->ustop('invalid name %n for dynamic variable, not conforming to value specified with dynavar option', n => $nom) if $dyn and $m->non_dynavar_p($nom);
    do { $m->pwarn('%r variable %n is already known as %v', r => $tref, n => $nom, v => $val);return $val } if $lihs->{vars}->{$val} and !$redef;
    my ($arb,$str);
  VAL: {
      do { $arb = $iniarb;last } unless $val;
      if (ref $val) {
	  $arb = $val;
	  ($str,$lvlmax) = $m->arb2strl($arb,$seplr);
      } else {
	  $str = $val;
	  $arb = $m->str2arb($str,$lvlmax);
      };
      last unless $arb; # NULLARB file:/phm/18/10/26/sig/_lng.phm_pub_sig181026.txt::narb06
      last if $tref eq ref $arb;
      $m->ustop('value %v of %n is not of type %r', n => $nom, r => $tref, v => $val);
    };
    $lihs->{$treg}->{$nom} = $arb;
    $lihs->{$typp}->{$nom} = 1; # OLD
    $str = $lvlmax . $str if $lvlmax;
    $lihs->putstr($nom => $str, lvlmax => $lvlmax, %putopts);
    delete $m->{tmplxarbs}->{$nom}; # oldp?
    $arb;
};
=head4 $m->defarb($lihs,$typ,$nom,$val,$lvlmax,%opts) => $arb: define complex variable
=cut
sub defarb {
    my ($m,$prog,$lihs,$typ,$nom,$val,$lvlmax,%putopts) = shift->lpuc('defarb',@_); # typ => '@', nom => 'tml', val => [qw(html latex)]
    my $seplr = $opts{arbseps};
    my $arb = $m->__defarb__($prog,$typ,0,1,$lvlmax,$nom,$val,$lihs,$seplr,%putopts);
    $m->dpop1($prog,$arb);
};		

=head3 Functions defhac, deflst, defxhac, defxlst

Declare a list/hash variable and initialise it, either with a value or with an empty list, and store it in $m->{lihs}->{hacs} and $m->{lihs}->{lsts}.
The 'x' versions can contain references like %hash, @list and $scalar that are not expanded before storage but only before use.

    args := nom, val?
    nom :: name of hash/list variable
    nom => 'colors'
    val :: value of hash/list variable, in notation understood by str2arb
    val => '|red|green|blue|'
    rets := arb
    arb :: structured representation initial value: listref or hashref
    arb => [ 'red', 'green', 'blue' ]

=cut

sub defhac {
    my ($m,$lihs,$nom,$val,$seplr) = @_;    
    $m->__defarb__('defhac', '%', 0, 0, 0, $nom, $val, $lihs, $seplr);
};
sub deflst {
    my ($m,$lihs,$nom,$val,$seplr) = @_;    
    $m->__defarb__('deflst', '@', 0, 0, 0, $nom,$val,$lihs,$seplr);
};
sub sethac {
    my ($m,$nom,$val,$lihs,$seplr) = @_;    
    $m->__defarb__('defhac', '%', 0, 1, 0, $nom,$val,$lihs,$seplr);
};
sub setlst {
    my ($m,$nom,$val,$lihs,$seplr) = @_;
    $m->__defarb__('setlst', '@', 0, 1, 0, $nom, $val, $lihs, $seplr);
};
sub defxhac {
    my ($m,$lihs,$nom,$val,$seplr) = @_;
    $m->__defarb__('defxhac', '%', 1, 0, 0, $nom, $val, $lihs, $seplr);
};
sub defxlst {
    my ($m,$lihs,$nom,$val,$seplr) = @_;
    $m->__defarb__('defxlst', '@', 1, 0, 0, $nom, $val, $lihs, $seplr);
};
sub setxhac {
    my ($m,$nom,$val,$lihs,$seplr) = @_;
    $m->__defarb__('defxhac', '%', 1, 1, 0, $nom, $val, $lihs, $seplr);
};
sub setxlst {
    my ($m,$nom,$val,$lihs,$seplr) = @_;
    $m->__defarb__('defxlst', '@', 1, 1, 0, $nom, $val, $lihs, $seplr);
};

=head3 Function lihs_define

    like konfig_define, but additionally store the variable to %{$m->{lihs}, in the same way that C<cache_define> stores it to %{$m->{progvars}}, but additionally store '__'.$nom.  The transfer of the values is done in new_ready right before invocation of postkonfig.

=cut

sub lihs_define ($$$$$;$) {
    my ($m,$prog,$lihs,$nom,$typ,$val,$lvlmax) = shift->ppuc('lihs_define',@_);$lihs = $m->get_lihs($lihs);
    $m->kopivar_define('lihs',$nom,$typ,$val);
    my $nonstorp = $lihs->{vars}->{__nonstorp__} ? 1 : 0;
    $lihs->{vars}->{__nonstorp__} = 1;
    if ($lihs->arbtypp($typ)) {
	$m->defarb($lihs,$typ,$nom,$val,$lvlmax);
    } else {
	$lihs->putstr($nom,$val,lihs => $lihs);
    };
    $lihs->{vars}->{__nonstorp__} = $nonstorp;
    $m->dpop1($prog,$nom);
};

sub lihs_put {
    my ($m,$prog,$lihs,$nom,$val,%opts) = shift->ppuc('lihs_put',@_);
    return $m->wpop($prog,[p => 'putundef', n => $nom]) unless length $val;
    my $typ = $opts{typ};$lihs = $m->get_lihs($lihs);
    my $str = $val;
    my $ref = ref $val;
  REF: {
      last unless $ref;
      $str = $m->arb2str($val,$opts{arbseps});
      if ('@' eq $typ or 'ARRAY' eq $ref) {
	  $m->pstop('variable %n registered as %t but ref %r', t => $typ, r => $ref) unless 'ARRAY' eq $ref;
	  $lihs->{lsts}->{$nom} = $val;
	  $lihs->{lstp}->{$nom} = 1;
      } elsif ('%' eq $typ or 'HASH' eq $ref) {
	  $m->pstop('variable %n registered as %t but ref %r', t => $typ, r => $ref) unless 'HASH' eq $ref;
	  $lihs->{hacs}->{$nom} = $val;
	  $lihs->{hacp}->{$nom} = 1;
      } };
    $lihs->putstr($nom,$str,%opts);
    $m->dpop($prog,$val,$ref);
};

# file:/phm/18/10/14/sig/_lng.phm_pub_sig181014.txt::rm_arb2lvlmax

1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/TmplArb.pm"
# End:
