=head1 NAME

A2E::Tmplvars

=head1 VERSION

This describes version B<0.0.1>

=cut

$VERSION = '0.0.1';

=head1 DESCRIPTION

Writing local variables and other meta information that can be read by editing
programs like Emacs.

This class needs to be mixed into A2E::Tmplfil and A2E::Dokfs.  It is their intersection set.

This method is of a mix-in type, i.e. it contains no fields and its methods have names which try not to duplicate any names of methods used in any other classes.  Other mix-in libraries include 

    A2E::Osver
    A2E::Sqltab

Since it contains only methods, no fields, and none of these is intended to
overwrite any other method of another class, invocation with the base pragma
in the sense of

	package MyApp;
	use base 'A2E::Daba', 'A2E::Tmplkonf';
        sub defvars {
	    my $m = shift;
	    $m->SUPER::defvars;
	    $m->tmplkonf_defvars();
            $m->{konfig}->define('my_app_var|a=s');
	    [...]
	    $m->konfayl('my_app', 'my_app.konf');
	    return 1;		
	}
	sub process {
	    my $m = shift;	
	    [...]
	}
	package Main;
	my $app = new_ready MyApp;
	$app->process;

will do the job.

=head1 BUGS

Seems to work fine.

A2E::Mktdir::skrib_fnom now supplements contents of $m->{cache}->{tmplkonf} with those of $m->{cache}->{tmplopts} in order to get missing INCLUDE_PATH:
    foreach my $opt (keys %{$m->{cache}->{tmplopts}}) { $tmplkonf{$opt} = $m->{cache}->{tmplopts}->{$opt} unless defined $tmplkonf{$opt} };
Looks kludgy.

=head1 PREREQUISITES

       A2E::Prog(3)

=head1 APPLICATIONS

        db2auth(1)
 	mlht(1)
	A2E::Mktdir::Mktdir(3)

=head1 SEE ALSO

    mktdir5(1)
    db2auth(1)
    perl(1)

=head2 Alternatives on multiple inheritance

    mixin(3)
    Class::C3(3)

=head1 AUTHOR

    Hartmut Pilch

=head1 COPYRIGHT

    Copyright (c) 2007 Hartmut Pilch (phm)

    This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.

=head1 IMPLEMENTATION

=head2 Header

=cut

package A2E::Tmplkonf;
use base 'A2E::TmplArb';
use strict;use warnings 'all';no warnings 'uninitialized';
use File::Spec::Functions;

=head2 Construction and extension of underlying object

We inherit the constructor 'new' from the A2E::Prog package, but supply 'tmplkonf_defvars' to be used within the 'setup' routine of applications. 

=head3 Function tmplkonf_defvars

=cut

sub tmplkonf_defvars ($$) {
    my ($m,$prog,$lihs) = shift->qpuc('tmplkonf_defvars',@_) or return;$lihs = $m->get_lihs($lihs);

=over

=item --inkluddirs I<dir>

directories in which to search for include files

=cut

    $m->cache_define('inkluddirs','@');

=item --tmplsupdir I<dir>

directory under which to search for templates.
The templates at this first level are related to the process flow (e.g. makefile templates) and output formatter rather than document types or vocabulary

=cut

    $m->cache_define('tmplsupdir', 's', '/opt/a2e/share/tmpl');

=item --tmpl oas_akt_aammdd

This is needed in order to make tmpl available as progvar right from the start, without need to fire up libraries and read _pre.*.txt.pag.

=cut
    $m->cache_define('tmpl','s');
    
=item --tmplopts|--to key=val

set options of the Template object, as defined in Template(3).  Meaningful options include INTERPOLATE, ENCODING, DEBUG, and PLUGIN_BASE.

=cut

    $m->cache_define('tmplopts', 's%', { INTERPOLATE => 0, PRIVATE => 0, EVAL_PERL => 0, ENCODING => 'utf8', DEBUG => $m->pvar(debug => 0), PLUGIN_BASE => 'A2E::Template::Plugin', PRE_PROCESS => 'html_tmpl.txt' });

=item --tmplang I<text_markup_language>

locally defined set of output formatting conventions used in this case.  This may lead to a choice of different plugins and configuration settings as well as a different choice of document template, but not a different choice of vocabulary files (except those in the tmplangdir which interfaces between both, see below)
At a2e we currently have html and latex as tmls, deplate-based 'multemp' was given up.
The value can be read from the environment variable MLTML, cf MLANG for language.  These two help reduce the load and make operations cleaner, e.g.

	export MLANG=zh MLTML=latex
	mldk dok

instead of

	mldk -l zh -t latex dok

This will do no harm, because limited operations are cleaner.
file:/phm/19/07/16/sig/_lng.phm_pub_sig190716.txt::mlang

=cut

    $m->prog_define(['tmplang','tml','mltml','t'], '$', 'html', $ENV{MLTML});


=item --doktmpldir /opt/a2e/share/tmpl/mlht

directory under which multilingual document templates with vocabulary building files are found

=cut
    my $test_modus = $m->pvar(test_modus => 0);
    $m->cache_define('doktmpldir', 's', $test_modus ? '/adv/tmpl' : '/opt/a2e/share/tmpl/mlht');
    $m->cache_define('sysinkluddir', 's', $test_modus ? '/adv/mktdir/include' : '/usr/include/a2e');

=item TODO: sth like --supdir_tmpl_r %d.news_tmpl.txt

Look for dok template at


	$SUPDIR/$SUPDOK.news_dok_tmpl.txt

or upper directories up to

	$TOPDIR/$TOPDOK.news_dok_tmpl.txt

or

	/opt/a2e/share/tmpl/news_dok_tmpl.txt

or, as last fallback, the default template

	/opt/a2e/share/tmpl/dok_tmpl.txt

Look for nav template at


	$SUPDIR/$SUPDOK.news_nav_tmpl.txt

or upper directories up to

	$TOPDIR/$TOPDOK.news_nav_tmpl.txt

or

	/opt/a2e/share/tmpl/news_nav_tmpl.txt

or, as last fallback, the default template

	/opt/a2e/share/tmpl/nav_tmpl.txt

and likewise for any other subordinate parts of the document.

Collect variables in a hierarchy to be implemented in lng_faylz

	/opt/a2e/share/tmpl/news_lang.txt
	/opt/a2e/share/tmpl/news_lang.en.txt
	$TOPDIR/@lng.$TOPDOK.news.txt
	$TOPDIR/@lng.$TOPDOK.news.en.txt

etc down to the current directory level, as specified in --tmplnglangfnom

See also mktdir0(1)

=cut

=item --tmls html --tmls latex

Text Markup Languages are also called TMPLANGs and the current one is stored in $m->pvar('tmplang') and used by &Tmplkonf::set_template_include_dirs.
file:/phm/18/09/19/sig/_lng.phm_pub_sig180919.txt::tml

=cut
    $m->cache_define(['tmls', 'L' ], 's@', ['html']);

=item --tmplapp a2e

template application, i.e. a symbol denoting a set of websites/documentations to which this set of multilingual document templates belongs and under which it can be retrieved.  We assume that application-specific templates live in /opt/a2e/share/tmpl/mlht/$tmplapp/$tmpl, the application's default template lives in /opt/a2e/share/tmpl/mlht/$tmplapp, and the software's default template lives in /opt/a2e/share/tmpl/mlht.  The vocabulary files live in these directories, whereas the language-neutral template files live in subdirectories with the name given by the tmplang option.

=cut

    $m->cache_define('tmplapp', 's', 'a2e');

=item --suptmpls apr_pub_aamm=pub_aamm,apr_aamm --suptmpls apr_aamm=sig_aamm,apr

parent template which is to be evaluated before the actual template and to which the actual template applies modifications.
the parent can again have a parent.  all are in parallel directories, named somethin lke

	/opt/a2e/share/tmpl/mlht/a2e/apr_pub_aamm
	/opt/a2e/share/tmpl/mlht/a2e/pub_aamm

used by A2E::Dokfs::lng_tmpl_faylz::suptmpls to help generate the list of vocabulary files to be evaluated.
Inheritance from multiple parents is supported, just separate with commas or spaces or a combination thereof.

=cut

    $m->cache_define('suptmpls','s%');

=item --conv deplate

=cut

    $m->cache_define('conv', 's', 'deplate');
    
=item --tmplapp_tmpls a2e=jpat,konf,log_aammdd,oas_aa,oas_aamm,sig,sig_aa,sig_aamm,sig_aammdd,log_aammdd

comma-separated list of document types available witin one tmplapp (e.g. C<a2e>), for which again specific document part templates like C<a2e_sig_aamm_dokfs_konf_tmpl.txt> might be copied to the top directory (topdir) when creating a microsite.

=cut

    $m->cache_define('tmplapp_tmpls','s%');

=item --alias_tmpls oas_tra = oas_akt_aammdd

Define template alias so that the one on the left hand can be used as an alias for the real name on the right hand side

=cut

#    $m->{konfig}->define('alias_tmpls=s%');

=item --alias_tmls deplate = multemp --alias_tmls tex => latex    

Define template alias so that the favorite name on the left hand can be used as an alias for the system name on the right hand side

=cut
#    $m->konfig_define('alias_tmls','s%', { deplate => 'multemp', 'xhtml' => 'html', 'tex' => 'latex' });    
=item --tmpl_vars sig_aa:privat = 1

Set variables of templates

=cut
    $m->cache_define('tmpl_vars','%');
    $m->cache_define('tmplkonf_fnom');
    $m->lihs_define($lihs,'include_extra', '@', ['..'], 1);
=back

=cut
    $m->konfayl('tmplkonf.konf');
    $m->qpop1($prog,1);
};

sub defvars ($) {
    my ($m,$prog,$lihs) = shift->qlpuc('A2E::Tmplvars::defvars',@_) or return;
    $m->SUPER::defvars($lihs);
    $m->tmplkonf_defvars($lihs);
    $m->qpop1($prog,1);
};

=head3 Function tmplkonf_postkonfig

=cut

sub tmplkonf_postkonfig ($) {
    my ($m,$prog) = shift->qpuc('tmplkonf_postkonfig',@_) or return;
    my $tmpl = $m->pvar('tmpl');
    do { $m->uwarn('no tmpl known even at end of configuration');$m->dpop;return } unless $tmpl;
  FNOM: {
      my $fnom = $m->{cache}->{tmplkonf_fnom};
      do { $m->pwarn('empty tmplkonf_fnom option');last } unless $fnom;
      my $fayl = $m->hsprintf($fnom, t => $tmpl);
      if (-r $fayl) { $m->{konfig}->file($fayl) } else { $m->decho('why not prepare a doctype-specific configuration file %f?', f => $fayl) };
    };
    $m->qpop1($prog,1);
}

=head3 Function dok_set_tmpl

Set document type (tmpl) based on configuration values read from _pre.*.txt or _dokfs
Set whatever was found to $m->pvar('tmpl') and return it.


=cut

sub dok_set_tmpl {
    my ($m,$prog,$lihs) = shift->ppuc('dok_set_tmpl',@_);
    my $tmpl = $m->pvar('tmpl') || $m->{konfig}->tmpl;
    $m->setpvar(tmpl => $tmpl, lihs => $lihs) or $m->ustop('failed to set tmpl');
    $m->dpop1($prog,$tmpl);
};

=head3 Function set_template_include_dirs

Set variables that point to template directories (potential members of include path), based on current configuration.
This needs to be invocable outside the startup routine, e.g. in A2E::Privat::Lnsig
Inside the startup routine it makes little sense since the tmpl is usually not known yet.

=cut
sub tmpldir {
    my ($m,$tmpl) = @_;
    catdir $m->pvar(tmplappdir => 1), $tmpl;
};
sub tmpl_set_tmpldir ($$$) {
    my ($m,$prog,$lihs,$tmpl) = shift->ppuc('tmpl_set_tmpldir',@_);$lihs = $m->get_lihs($lihs);
    my $tmpldir = $m->tmpldir($tmpl);
    $m->setpvar(tmpldir => $tmpldir, lihs => $lihs);
    $m->dpop1($prog,$tmpldir);
};

sub set_template_include_dirs ($$) {
    my ($m,$prog,$lihs) = shift->ppuc('set_template_include_dirs',@_);$lihs = $m->get_lihs($lihs);
    my $setcat = sub { 
	my $nov = shift;
	my $dirnom = shift;
	my $folnom = shift;
	my $dir = $m->pvar($dirnom);
	return unless $dir;
	my $fol = $m->pvar($folnom);
	return unless $fol;
	$dir = catfile $dir, $fol;
	$m->setpvar($nov => $dir, lihs => $lihs);
    };
=head4 directories
=over
=item tmplangsupdir /opt/a2e/share/tmpl/multemp
=cut
    &$setcat(qw(tmplangsupdir tmplsupdir tmplang));
=item tmplangdir  /opt/a2e/share/tmpl/mlht/multemp
=cut
    &$setcat(qw(tmplangdir doktmpldir tmplang));
=item tmplappdir /opt/a2e/share/tmpl/mlht/a2e
=cut
    &$setcat(qw(tmplappdir doktmpldir tmplapp));
=item tmplapptmplangdir /opt/a2e/share/tmpl/mlht/a2e/multemp
=cut
    &$setcat(qw(tmplapptmplangdir tmplappdir tmplang));
=item tmpldir /opt/a2e/share/tmpl/mlht/a2e/sig_int_aa
=cut
    &$setcat(qw(tmpldir tmplappdir tmpl));
=item tmpltmplangdir /opt/a2e/share/tmpl/mlht/a2e/sig_int_aa/multemp
=cut
    &$setcat(qw(tmpltmplangdir tmpldir tmplang));
=back
=cut
    $m->dpop1($prog,1);
};

=head2 Function alias_get_tmpl

Resolve tmpl aliases as specified using the C<alias_tmpls> config option.

Take a tmpl and return its real name.

=cut

sub alias_get_tmpl { my ($m,$tmpl) = @_;return $m->ali2val(tmpl => $tmpl) };
sub alias_get_tml { my ($m,$tml) = @_;return $m->ali2val(tmplang => $tml) };


=head3 Function set_template_include_path

Wrapper around set_template_inlcude_dirs, additionally set $m->{tmplopts}->{INCLUDE_PATH}

=cut

sub set_template_include_path ($%) {
    my ($m,$prog,$lihs,%opts) = shift->opuc('set_template_include_path',0,@_);
    my @dirs = @{$m->{konfig}->inkluddirs};
    my $tmplsupdir = $m->{konfig}->tmplsupdir;
    my $tmplopts = $m->{cache}->{tmplopts};
    unshift @dirs, $tmplsupdir; 
  TMPLDIRS: {
      if ($opts{tmplang}) { $m->setpvar(tmplang => $opts{tmplang}, lihs => $lihs) };
      $m->set_template_include_dirs($lihs);
      my $tmpltmplangdir = $m->pvar('tmpltmplangdir');
      my $tmpldir = $m->pvar('tmpldir');
      my $tmplapptmplangdir = $m->pvar('tmplapptmplangdir');
      my $tmplappdir = $m->pvar('tmplappdir');
      my $tmplangdir = $m->pvar('tmplangdir');
      my $tmplangsupdir = $m->pvar('tmplangsupdir');
      my $sysinkluddir = $m->pvar('sysinkluddir');
      my @tmpldirs = grep {$_} $tmpltmplangdir, $tmpldir, $tmplapptmplangdir, $tmplappdir, $tmplangdir, $tmplangsupdir;
      unshift @dirs, @tmpldirs;
      my @extras = $m->ref2list($lihs->{lsts}->{include_extra});
      unshift @dirs, @extras;
    };
    unshift @dirs, '.';
    do { $m->pwarn('dangerous: /usr/include was missing here!');push @dirs, '/usr/include' } unless grep { $_ eq '/usr/include' } @dirs; 
    my $path = \@dirs;
    $tmplopts->{INCLUDE_PATH} = $path;
    $m->dpop1($prog,$path);
};

=head1 END

=cut

1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/Tmplkonf.pm.tmpl"
# End:
