
=head1 NAME

A2E::Tmplvars

=cut

package A2E::Tmplvars;

=head1 DESCRIPTION

Read string variables to and from $m->{lihs}->{vars}, maintain list and hash variables, transport them in strings by encoding and decoding, so that DBM files can be used for more complex structures.
Provide a reduced basic interface to this storage.
A2E::Dokfs should be fine with read-only access to complex variables and write-access to simple ones only.
We may fork out A2E::Tmplarbs for the rest soon.

=head1 BUGS

Report to adv@a2e.de

=head1 VERSION

=over

=item 2018-10-22 version 0.0.1

Provide only most basic things like tmplvars_get*, leave rest in Tmplfil at least for the time being.
A2E::Dokfs should perhaps be fine with read-only access to tmplvars.

=back

=cut

our $VERSION = '0.0.2';
our $novitas = $ENV{NOVITAS} || '0';
use File::Spec::Functions qw(catfile);

=head1 LIMITATIONS

=head1 TODO

Move the tmplvars related stuff over.

=head1 PREREQUISITES

=over

=item A2E::Prog(3)
=item Encode

=cut

use base 'A2E::TmplArb';
use A2E::Lihs;
use strict;use warnings 'all';no warnings 'uninitialized';
use Encode qw(_utf8_on);
use File::Basename;
our %MSGS = (
    tmplvarsdetour => 'why take a detour to A2E::TmplArb::%o via A2E::Tmplvars::tmplvars_%o?' );
=back

=cut

=head3 Startup Configuration Functions

=cut
    
sub tmplvars_defvars {
    my ($m,$prog,$lihs) = shift->qlpuc('tmplvars_defvars',@_) or return;
    $m->tmplarb_defvars($lihs);
    foreach my $typ ($lihs->typl) { # $ @ %
	my $treg = $lihs->treg($typ); # 'hacs'
	my $ttyr = $lihs->ttyr($typ); # 'tmplhacs'
	next if 'HASH' eq ref $m->{$ttyr};
	$m->{$ttyr} = $lihs->{$treg};
    };
    $m->{kopivars_hac2reg}->{tmplvars} = 'tmplvarp'; 
    $m->{kopivars_hac2fun}->{tmplvars} = sub { my ($hac,$nom,$val,$typ) = @_;$m->lihs_put($lihs,$nom,$val, typ => $typ, nonstorp => 1);$lihs->{vars}->{$nom} = $val unless ref $val }; # old kludge, cf file:/adv/perl/A2E/Prog.pm.tmpl::/\bkopivars_hac2fun\b/
    my $cwd = $m->get_cwd;
    my $supdir = dirname $cwd;
    $m->lihs_define($lihs,'progver', '$', '4.5');
    $m->lihs_define($lihs,'progvernum', 'i', '405');
    $m->lihs_define($lihs,'dokprogver', '$');
    my $user = $ENV{USER};
    my $name = $m->{progvars}->{progbas} || $0;
    my $time = $m->timestamp;
    $m->pstop('m lihs is not lihs') unless $lihs eq $m->{lihs};
    $m->{tmplvars} = $m->{lihs}->{vars};
    $m->{tmpllsts} = $m->{lihs}->{lsts};
    $m->{tmplhacs} = $m->{lihs}->{hacs};
    $m->{lvlmax} = $m->{lihs}->{lvlmax};
    $m->setptvar(prog_user => $user);
    $m->setptvar(prog_time => $time);
    $m->setptvar(prog_name => $name);

=item --dokprogver

Version of software that the document was created with.
Based on this info backward compatibility is enabled or switched off, e.g. with the dokprogverpre function as in

	$m->{lihs}->{vars}->{$nom.'_jungils'} = $val if $m->dokprogverpre(400);

bolean routine.
The dokprogver must be preset to a value that represents the lowest allowable progver of the site, so that when a local directory does not have its own dokprogver value this one can be safely used.

=cut
    $m->konfayl('tmplvars.konf');
    $m->{tmplvarp} = {};
    $m->prog_msgdefs('tmplvars',%MSGS);
    $m->qpop1($prog,1);
};

sub defvars {
    my ($m,$prog) = shift->qpuc('A2E::Tmplvars::defvars',@_) or return;
    $m->SUPER::defvars;
    $m->tmplvars_defvars;
    $m->qpop1($prog,1);
};

sub tmplvars_postkonfig {
    my ($m,$prog,$lihs) = shift->qlpuc('tmplvars_postkonfig',@_) or return;
  TMPLVARS_KOPIVAR: { # now subsumed under file:/adv/perl/A2E/Prog.pm.tmpl::kopi_konfigvars
      last if $novitas;
      my $set_tmplvars = sub {
	  my @noms = ();
	  while (my ($nom,$typ) = each %{$m->{tmplvarp}}) {
	      next unless defined $typ;
	      push @noms, $nom;
	      my $val = $m->{konfig}->get($nom);
	      next unless $val;
	      $m->lihs_put($lihs,$nom,$val);
	      # $m->lihs_put($lihs,'__'.$nom,$val);
	  };
	  return @noms;
      };
      &$set_tmplvars;
    };
    $m->qpop1($prog,1);
};

=head3 Function dokprogverpre

current program version $dpver is earlier than the required minimum $minver

=cut

sub dokprogverpre {
    my ($m,$prog,$minver,$lihs) = shift->ppuc('dokprogverpre',@_);$lihs = $m->get_lihs($lihs);
    local $_ = $lihs->getstr('dokprogver') || $m->{konfig}->dokprogver || $m->ustop('dokprogver not set');
    my @dpfs = reverse split '\.';
    my $dpver = 0;my $mul = 1;foreach my $dpf (@dpfs) { $dpver += $dpf * $mul } continue { $mul *= 100 };
    my $ok = $dpver < $minver;
    $m->dpop1($prog,$ok);
}

sub setptvar {
    my ($m,$prog,$nom,$val,%opts) = shift->ppuc('setptvar',@_);
    (my $lihs,%opts) = $m->shifthash('lihs',%opts);$lihs = $m->get_lihs($lihs);
    $m->setpvar($nom,$val,lihs => $lihs);
    if (ref $val) {
	$m->lihs_put($lihs,$nom,$val,%opts);
    } else {
	$lihs->putstr($nom,$val,%opts);
    };
    $m->dpop1($prog,$val);
};

sub lihs_putarb {
    my ($m,$prog,$lihs,$nom,$val,%opts) = shift->ppuc('lihs_putarb',@_);$lihs = $m->get_lihs($lihs);
    return $m->wpop($prog,$m->pwarn('refusing to put undefined arb value for key %n', n => $nom)) unless defined $val;
    my $ref = ref $val;
    return $m->dpop1($prog,$lihs->putstr($nom => $val, %opts)) unless $ref;
    return $m->dpop1($prog,$m->defarb($lihs,'@',$nom,$val)) if 'ARRAY' eq $ref;
    return $m->dpop1($prog,$m->defarb($lihs,'%',$nom,$val)) if 'HASH' eq $ref;
};

=head1 The END

=cut

return 1;
# (write-file (let ((dok "perl_a2e_pub") (fol "Tmplvars.pm.tmpl")) (format "~/tdir/%s/%s" dok fol)))
# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/Tmplvars.pm.tmpl"
# End:
