
=head1 NAME 

A2E::Txt2dbm

=cut

package A2E::Txt2dbm;

=head1 SYNOPSIS

       txt2dbm _lng.a2e_template_pub.en.txt.pag /opt/a2e/include/tmpl/mldok_lang.en.txt /adv/_lng.adv_pub.en.txt _lng.a2e_template_pub.en.txt

=head1 DESCRIPTION

Create a DBM vocabulary database from a series of text file sources.

This is usually done automatically by Makefile rules.

=head1 FILES

=head2 Configuration

    /etc/opt/a2e/txt2dbm.konf
    ~/txt2dbm.konf
    _txt2dbm

=head2 Makefile component

    /opt/a2e/include/dok.mk

=head2 Prerequisites

    SDBM_File(3)
    A2E::Prog(3)
    perl(1)

=cut

use base 'A2E::MLHT';
use strict;
use File::Spec::Functions qw(catfile);
use Fcntl;
use SDBM_File;
use DBM_Filter;
# use Data::Dumper;

=head2 Other 

    txt2dbm(1)
    dbmvals(1)
    A2E::MLHT(3)

=head1 IMPLEMENTATION

=cut

# use Encode qw(_utf8_on);

=head2 Overwrite Functions

=head3 Function A2E::Txt2dbm::defvars

=cut

sub txt2dbm_defvars ($) {
    my $m = shift;
    $m->dpuc('txt2dbm_defvars', @_);
    do { $m->dpop;return } if $m->{ready}->{txt2dbm_defvars};
    $m->{konfig}->define('dbmvars|V=s%', { DEFAULT => {} });
    $m->qre_define('txt2dbm_setnomval', 's', '\A(\w+)[\:\=]\s*(.*)\s*\Z');
    $m->{konfig}->define('pur!');
    $m->konfayl('txt2dbm.konf');
    $m->{ready}->{txt2dbm_defvars} = 1;
    $m->dpop('txt2dbm_defvars', 1);
    return 1;
};

sub defvars ($) {
    my $m = shift;
    $m->dpuc('A2E::Txt2dbm::defvars', @_);
    $m->SUPER::defvars;
    $m->txt2dbm_defvars;
    $m->dpop('A2E::Txt2dbm::defvars', 1);
    return 1;
}

=head3 Function A2E::Txt2dbm::postkonfig

Not applied yet, needs further testing

=cut

sub txt2dbm_postkonfig {
    my $m = shift;
    $m->dpuc('txt2dbm_postkonfig', @_);
    $m->mlht_postkonfig;
    $m->upon_setvar('sems', 'set_sems');
    $m->{ready}->{txt2dbm_postkonfig} = 1;
    $m->dpop('txt2dbm_postkonfig', 1);
    return 1;

}
sub postkonfig ($) {
    my $m = shift;
    $m->dpuc('A2E::Txt2dbm::postkonfig', @_);
    $m->SUPER::postkonfig;
    $m->txt2dbm_postkonfig;
    $m->dpop('A2E::Txt2dbm::postkonfig', 1);
    return 1;
};

=head2 Special Functions

=head3 Function txt2dbm

=cut

sub dbm_fayl_arg {
    local $_ = shift;
    my ($fo) = m(\A(.*)\.pag\Z);
    $fo || $_;
};
sub txt2dbm ($$) {
    my $m = shift;
    my $fo = dbm_fayl_arg shift;
    my @fis = @_;
    my %opts = (msglini => 1, setnomval_pur => $m->{konfig}->pur, badnom_ok => 1);
    $opts{setnomval} = $m->{qre}->{txt2dbm_setnomval} if $opts{setnomval_pur};
    my $lang = $m->{konfig}->lang;
    my @langs = $m->set_langs;
    $lang ||= $langs[0];
    $m->setpvar(lang => $lang);
    my $lihs = new A2E::Lihs;
    $opts{lihs} = $lihs;
    $m->riid_tmplvars_faylz($m->{progvars}, \%opts, @fis) if @fis;
    # $m->tmplvars_kk_vars(bd => $m->{progvars});
    # $m->tmplvars_arb2var; # file:/phm/18/09/sig/_org.phm_pub_sig1809.txt::var2arb
    my %vars = ();
    my $db = tie %vars, 'SDBM_File', $fo, O_RDWR|O_CREAT, 0666 or die $m->hsprintf('can not tie file %f to hash: %m', f => $fo, m => $!);
    $db->Filter_Push('DBM_Filter::utf8');
    # |將來可以嘗試以${1}替代|$m->dbmvars_kk_tmplvars(\%vars, 'bd')|
    foreach my $nom (keys %{$lihs->{vars}}) {
	my $val = $lihs->{vars}->{$nom};
	$vars{$nom} = $val || $nom;
    }
    1;
};

=head1 THE END

=cut

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# End:
