#!/usr/bin/perl
# a2e_users: print names of valid registered users on the current systems

package A2E::Users;
use Getopt::Long;Getopt::Long::Configure('gnu_getopt');
use AppConfig qw(:expand);
use File::Spec::Functions;
use Cwd;
use IO::File;
use strict;

sub new {
    my $class = shift;
    my $m = {};
    bless $m, $class;
    $m->{konfig} = new AppConfig { CASE => 1, GLOBAL => { EXPAND => EXPAND_ALL | EXPAND_WARN } };
    $m->{konfig}->define('version!');
    $m->{konfig}->define('verbose|v!');
    $m->{konfig}->define('manpage|man!');
    $m->{konfig}->define('help|usage|h!');
    my $etcdir = catfile $INC[0], 'etc';
    $m->{test_modus} = $INC[0] =~ m(\A\/adv\/perl\b) and -d $etcdir;
    if ($m->{test_modus}) { warn sprintf '%s test run with %s as etcdir in %s with args %s', $0, $etcdir, &getcwd(), join(' ', @ARGV) } else { $etcdir = '/etc/opt/a2e' };
    $m->{konfig}->define('etcdir=s', { DEFAULT => $etcdir });

    $m->{konfig}->define('group|G!');
    $m->{konfig}->define('non_users|nu=s@');
    $m->{users} = {};
    my $konfayl = catfile $m->{konfig}->etcdir, 'a2e_users.konf';
    $m->{konfig}->file($konfayl);
    return $m;
};

sub get_users {
    my $m = shift;
    my $fi = shift;
    my $hi = new IO::File $fi, 'r' or die sprintf 'failed to open file %s for reading', $fi;
    local $_;
    while ($_ = readline $hi) {
	chomp $_;
	next unless m(\A(\w+)\b);
	next unless -d catdir '', 'home', $1;
	next if grep { $_ eq $1 } @{$m->{konfig}->non_users};
	$m->{users}->{$1} = 1;
    };
    $hi->close;
    1;
};

sub list_users {
    my $m = shift;
    my @files = ();
    push @files, catfile '/etc/opt/a2e/uidgids', 'uids.txt';
    push @files, catfile '/etc/opt/a2e/uidgids', 'uidgids.txt' unless $m->{konfig}->group;
    $m->get_users($_) for @files;
    my @users = ();
    foreach my $user (keys %{$m->{users}}) { next unless $m->{users}->{$user};push @users, $user }; 
    return @users;
}

return 1;

# Local Variables:
# coding: utf-8
# mode: perl
# srcfile: "/adv/perl/A2E/Users.pm.tmpl"
# End:
